package com.qcg.comp.sdk.rsync;

import java.io.IOException;
import java.io.InputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;

import org.metastatic.rsync.Checksum32;
import org.metastatic.rsync.ChecksumPair;
import org.metastatic.rsync.Configuration;
import org.metastatic.rsync.DataBlock;
import org.metastatic.rsync.Delta;
import org.metastatic.rsync.ListenerException;
import org.metastatic.rsync.MatcherEvent;
import org.metastatic.rsync.MatcherListener;
import org.metastatic.rsync.MatcherStream;
import org.metastatic.rsync.Offsets;
import org.metastatic.rsync.Rdiff;

public class DeltaInputStream extends InputStream {

    private Configuration c;
    private MatcherStream match;
    private InputStream in;
    private byte[] buffer = new byte[1024*1024];
    private int bufferPos = 0;
    private int bufferAvail = 0;

    private void write(byte[] bytes) {	
	for (byte b : bytes)
	    writeByte(b);	    
    }

    private void writeByte(byte b) {
	if(bufferAvail == buffer.length) {
	    byte []newBuffer = new byte[buffer.length*2];
	    
	    for(int i = 0; i < buffer.length; ++i)
		newBuffer[i] = buffer[i];
	    
	    buffer = newBuffer;
	}
	buffer[bufferAvail++] = b; 
    }

    private void writeInt(int i) {
	writeByte((byte) ((i >>> 24) & 0xff));
	writeByte((byte) ((i >>> 16) & 0xff));
	writeByte((byte) ((i >>> 8) & 0xff));
	writeByte((byte) (i & 0xff));
    }

    private void writeInt(long l, int len) {
	for (int i = len - 1; i >= 0; i--) {
	    writeByte((byte) ((int) (l >>> i * 8) & 0xff));
	}
    }

    private void writeCopy(Offsets off) {
	writeByte(Rdiff.OP_COPY_N4_N4);
	writeInt(off.getOldOffset(), 4);
	writeInt(off.getBlockLength());
    }

    private static int integerLength(long l) {
	if ((l & ~0xffL) == 0) {
	    return 1;
	} else if ((l & ~0xffffL) == 0) {
	    return 2;
	} else if ((l & ~0xffffffffL) == 0) {
	    return 4;
	}
	return 8;
    }

    private void writeLiteral(DataBlock d) {
	byte cmd = 0;
	int param_len;

	switch (param_len = integerLength(d.getBlockLength())) {
	case 1:
	    cmd = Rdiff.OP_LITERAL_N1;
	    break;
	case 2:
	    cmd = Rdiff.OP_LITERAL_N2;
	    break;
	case 4:
	    cmd = Rdiff.OP_LITERAL_N4;
	    break;
	}

	writeByte(cmd);
	writeInt(d.getBlockLength(), param_len);
	write(d.getData());
    }

    public DeltaInputStream(List<ChecksumPair> sums, InputStream in) throws NoSuchAlgorithmException, IOException {
	c = new Configuration();
	c.strongSum = MessageDigest.getInstance("MD4");
	c.weakSum = new Checksum32(Params.CHAR_OFFSET);
	c.blockLength = Params.BLOCK_LENGTH;
	c.strongSumLength = Params.STRONG_SUM_LENGTH;
	match = new MatcherStream(c);
	match.setChecksums(sums);
	writeInt(Rdiff.DELTA_MAGIC);
	match.addListener(new MatcherListener() {
	    public void update(MatcherEvent me) throws ListenerException {		
		Delta d = me.getDelta();
		if (d instanceof Offsets) {
		    writeCopy((Offsets) d);
		} else if (d instanceof DataBlock) {
		    writeLiteral((DataBlock) d);
		}		
	    }
	});
	this.in = in;	
    }

    int state = 0;
    byte[] buf = new byte[10240];

    private boolean step() throws IOException {
	int len;
	
	switch (state) {
	case 0:
	    if ((len = in.read(buf)) != -1) {		
		try {
		    match.update(buf, 0, len);
		} catch (ListenerException le) {
		    throw (IOException) le.getCause();
		}
	    } else
		state++;
	    break;
	case 1:
	    try {
		match.doFinal();
	    } catch (ListenerException le) {
		throw (IOException) le.getCause();
	    }
	    writeByte((byte) 0);
	    state++;
	    break;
	default:
	    return false;
	}
	return true;
    }

    @Override
    public int read() throws IOException {
	int val;
	
	//we reached the end of the buffer, buffer must be refilled 
	if(bufferPos == bufferAvail) {
	    bufferAvail = 0;
	    bufferPos = 0;
	}

	while (bufferAvail == 0 && step())
	    ;
		
	if (bufferAvail == 0)
	    val = -1;
	else {
	    val = buffer[bufferPos++];	    
	}	
	
	return val;
    }

}