/* $Id: atz_ardl_filter.c 2 2009-10-05 10:09:52Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>

#include <qcg-core/sm.h>
#include <qcg-core/compat.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/signal.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/atz_module.h>

#include <include/ardl_filter.h>
#include <xmlstructs/xmlstructs.h>

__SM_ID("$Id: atz_ardl_filter.c 2 2009-10-05 10:09:52Z mamonski $");


static int setup(sm_module_t);
static void cleanup(sm_module_t);
static sm_value_t *handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params);

static sm_atz_module_t mapfile_module = NULL;
static struct sm_transport_s dummy_transport;


SM_GENERAL_MODULE = {

	.header = {
		.entry_size = SM_GENERAL_MODULE_ENTRY_SIZE,
		.api_type = SM_API_GENERAL,

		.name = "atz_ardl_filter",
		.descr = "User DN to local username ARDL mapper",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	},
};

static int
setup(sm_module_t this sm_unused)
{

	SM_DEBUG_ENTER();

	SM_TRY_FAULT(sm_signal_connect("smc_ardl_filter", handler));

	memset(&dummy_transport, 0, sizeof(dummy_transport));

	mapfile_module = (sm_atz_module_t)this->module_manager->find_by_name_first(this->module_manager, "atz_mapfile");

	return SM_OK;

fault:

	return SM_FAULT;
}

static void
cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();
}

static sm_value_t *
handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params)
{
	sm_alloc_tracked_t mh;
	_ardl__ReservationDefinition ardl;
	const char *input_ardl;
	char *output_ardl;
	sm_value_t *result = NULL;
	SM_BOOL modified = SM_FALSE;

	SM_DEBUG_ENTER();

	if (sm_previous_signal_failed(chained_params))
		return sm_value_clone(chained_params);
	else if (sm_signal_first(chained_params)) {
		input_ardl = params[0]->vstring;
	} else
		input_ardl = chained_params[1]->vstring;

	SM_TRY_NOT(mh = sm_gsoap_buffer_get(input_ardl, &ardl, ardl__ARDLReservationDefinition_USCOREType, "ardl:ReservationDefinition", xmlstructs_namespaces));

	if (ardl.ardl__ReservationDescription
		&& ardl.ardl__ReservationDescription->ardl__Resources
		&& ardl.ardl__ReservationDescription->ardl__Resources->UserName
		&& ardl.ardl__ReservationDescription->ardl__Resources->UserName[0] == '/') {
		sm_atc_token_t token;

		token = sm_atc_token_create(&dummy_transport);
		token->add(token, SM_ATC_TOKEN_X509_SUBJECT, ardl.ardl__ReservationDescription->ardl__Resources->UserName);

		if (!mapfile_module) {
			sm_emerg("No atz_mapfile module loaded");
			result = sm_value_create(2, SM_VALUE_RET_CODE, ARDL_FILTER_INTERNAL_FAULT, "No atz_mapfile module loaded", SM_VALUE_NONE);
			sm_atc_token_free(&token);
			goto fault;
		}

		if (mapfile_module->map_user((sm_module_t)mapfile_module, token) != SM_OK) {
			result = sm_value_create(2, SM_VALUE_RET_CODE, ARDL_FILTER_INVALID_REQUEST_MESSAGE_FAULT, "Invalid user DN", SM_VALUE_NONE);
			sm_atc_token_free(&token);
			goto fault;
		}

		sm_info("ARDL User DN = %s mapped to %s", ardl.ardl__ReservationDescription->ardl__Resources->UserName, ((sm_user_t )token->mapped_users->data)->username);

		SM_TRY_NOT(ardl.ardl__ReservationDescription->ardl__Resources->UserName = sm_alloc_tracked_strdup(mh, ((sm_user_t )token->mapped_users->data)->username));
		modified = SM_TRUE;

		sm_atc_token_free(&token);
	}

	if (modified) {
		SM_TRY_NOT(output_ardl = sm_gsoap_buffer_put(&ardl, ardl__ARDLReservationDefinition_USCOREType, "ardl:ReservationDefinition", xmlstructs_namespaces));

		result = sm_value_create(2, SM_VALUE_RET_CODE, ARDL_FILTER_OK, SM_VALUE_STRING, output_ardl);
		SM_FREE(output_ardl);
		SM_TRY_NOT(result);
	} else {
		SM_TRY_NOT(result = sm_value_create(2, SM_VALUE_RET_CODE, ARDL_FILTER_OK, SM_VALUE_STRING, input_ardl));
	}

	sm_alloc_tracked_free(&mh);
	return result;

fault:
	sm_alloc_tracked_free(&mh);

	signal->stop_emission(signal);

	if (result)
		return result;
	else
		return sm_value_create(2, SM_VALUE_RET_CODE, ARDL_FILTER_INTERNAL_FAULT, "Internal fault in application_mapper", SM_VALUE_NONE);
}
