/* $Id: atz_sandbox.c 24 2009-10-30 17:29:00Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <regex.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <qcg-core/sm.h>
#include <qcg-core/odbc.h>
#include <qcg-core/logger.h>
#include <qcg-core/signal.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/atz_module.h>

#include <db_sandbox.h>

__SM_ID("$Id: atz_sandbox.c 24 2009-10-30 17:29:00Z mamonski $");


static int atz_sandbox_setup(sm_module_t);
static void atz_sandbox_cleanup(sm_module_t);
static int atz_sandbox_authorize(sm_module_t, sm_atc_token_t, const char *);
static int atz_sandbox_map_user(sm_module_t, sm_atc_token_t);

static sm_value_t *handle_failed(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_cancelled(sm_signal_t, sm_value_t *, const sm_value_t *);
static sm_value_t *handle_finished(sm_signal_t, sm_value_t *, const sm_value_t *);
static int handle_purge(const char *username, const char *homedir, int uid);


SM_ATZ_MODULE = {

	.header = {
		.entry_size = SM_ATZ_MODULE_ENTRY_SIZE,
		.api_type = SM_API_ATZ,

		.name = "atz_sandbox",
		.descr = "Sandbox authorization and user mapping",

		.version_major = 0,
		.version_minor = 1,

		.setup = atz_sandbox_setup,
		.cleanup = atz_sandbox_cleanup,

		.multiple_instances_supported = SM_FALSE
	},

	.authorize = atz_sandbox_authorize,
	.map_user = atz_sandbox_map_user
};


typedef enum { smc_release_on_job_completion, smc_release_on_stage_out} smc_release_on_t;

static smc_release_on_t release_on = smc_release_on_job_completion;
static const char *purge_script = NULL;


static int
parse_accounts_pool_file(const char *account_pool_file)
{
	FILE *f = NULL;
	char account_name[256];
	int count = 0;

	SM_DEBUG_ENTER_STR(account_pool_file);

	if ((f = fopen(account_pool_file, "r")) == NULL) {
		sm_error("Failed to open file: %s", account_pool_file);
		return SM_FAULT;
	}

	SM_TRY_FAULT(smc_db_sandbox_clear_accounts_pool_relation());

	while (fscanf(f,"%255s", account_name) == 1) {
		struct passwd *pwd = NULL;

		if ((pwd = getpwnam(account_name)) == NULL) {
			sm_perror("getpwnam failed");
			sm_error("Unable to resolve account: %s", account_name);
			goto fault;
		}

		sm_info("account[%d] = %s (home=%s,uid=%d)",++count, account_name, pwd->pw_dir, pwd->pw_uid);
		SM_TRY_FAULT(smc_db_sandbox_add_account_entry(account_name, pwd->pw_dir, pwd->pw_uid));
	}

	sm_info("Total number of accounts in pool: %d", count);
	fclose(f);
	return SM_OK;
fault:
	fclose(f);
	return SM_FAULT;
}

static int
atz_sandbox_setup(sm_module_t this)
{
	int ret = SM_OK;
	static const char *account_pool_file = NULL;
	static const char *release_on_str = NULL;

	SM_DEBUG_ENTER();

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:AccountsPoolFile", &account_pool_file) != SM_OK)
		goto fault;

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:ReleaseOn", &release_on_str) == SM_FAULT)
		goto fault;

	if (release_on_str && strcmp(release_on_str, "JobCompletion") == 0) {
		release_on = smc_release_on_job_completion;
	} else if (release_on_str && strcmp(release_on_str, "StageOut") == 0) {
		release_on = smc_release_on_stage_out;
	} else {
		assert(release_on_str == NULL);
	}

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:PurgeScript", &purge_script) == SM_FAULT)
		goto fault;

	sm_info("AccountsPoolFile=%s", account_pool_file);
	sm_info("ReleaseOn=%s", release_on_str);
	sm_info("PurgeScript=%s", purge_script);

	/*database setup*/
	if (smc_db_sandbox_setup(this->conf_ctxt) != SM_OK) {
		sm_error("Failed to set up database module");
		goto fault;
	}

	if (parse_accounts_pool_file(account_pool_file)!= SM_OK) {
		sm_error("parsing %s accounts pool file failed", account_pool_file);
		goto fault;
	}

	sm_info("connecting to signals...");
	ret |= sm_signal_connect("smc_job_finished", handle_finished);
	ret |= sm_signal_connect("smc_job_failed", handle_failed);
	ret |= sm_signal_connect("smc_job_cancelled", handle_cancelled);

	if (ret & SM_FAULT) {
		sm_error("Failed to connect to signals");
		goto fault;
	}

	return SM_OK;

fault:
	smc_db_sandbox_cleanup();

	return SM_FAULT;
}

static void
atz_sandbox_cleanup(sm_module_t this)
{
	SM_DEBUG_ENTER();

	smc_db_sandbox_cleanup();
}

/*
static int
atz_sandbox_readline(const regex_t *regex, FILE *sandbox_fd, char **remote_subject, char **mappings, char **routines)
{
	char buf[sandbox_LINE_MAXLEN + 1];
	regmatch_t pmatch[4];

nextline:
	if (fgets(buf, sizeof(buf), sandbox_fd)) {
		char *ptr;

		if ((ptr = strchr(buf, '\r')))
			*ptr = '\0';
		if ((ptr = strchr(buf, '\n')))
			*ptr = '\0';

		if (buf[0] == '#' || buf[0] == '\0')
			goto nextline;

		if (regexec(regex, buf, 4, pmatch, 0)) {
			sm_notice("Ignoring line: \"%s\"", buf);
			goto nextline;
		}

		*remote_subject = sm_malloc(pmatch[1].rm_eo - pmatch[1].rm_so + 2);
		*mappings = sm_malloc(pmatch[2].rm_eo - pmatch[2].rm_so + 2);

		if (pmatch[3].rm_so == -1) {
			*routines = sm_strdup("*");
		} else {
			*routines = sm_malloc(pmatch[3].rm_eo - pmatch[3].rm_so + 2);
		}

		if (NULL == *remote_subject || NULL == *mappings || NULL == *routines) {
			SM_FREE(*remote_subject);
			SM_FREE(*mappings);
			SM_FREE(*routines);
			return SM_FAULT;
		}

		buf[pmatch[1].rm_eo] = '\0';
		strcpy(*remote_subject, buf + pmatch[1].rm_so);
		buf[pmatch[2].rm_eo] = '\0';
		strcpy(*mappings, buf + pmatch[2].rm_so);

		if (pmatch[3].rm_so != -1) {
			buf[pmatch[3].rm_eo] = '\0';
			strcpy(*routines, buf + pmatch[3].rm_so);
		}

		sm_debug("Read: |%s|%s|%s|", *remote_subject, *mappings, *routines);
	} else {
		if (ferror(sandbox_fd)) {
			sm_error("fgets failed");
			clearerr(sandbox_fd);
		}
		return SM_FAULT;
	}

	return SM_OK;
}

static int
atz_sandbox_get_data(const char *path, const regex_t *regex, const char *subject, char **mapping, char **routines)
{
	char *sandbox_subject = NULL;
	char *sandbox_mappings = NULL;
	char *sandbox_routines = NULL;
	FILE *sandbox_fd = NULL;

	SM_MANDATORY_PARAM(path, SM_FAULT);
	SM_MANDATORY_PARAM(regex, SM_FAULT);
	SM_MANDATORY_PARAM(subject, SM_FAULT);

	SM_DEBUG_ENTER_STR(subject);

	if (!(sandbox_fd = fopen(path, "r"))) {
		sm_perror("fopen");
		return SM_FAULT;
	}

	sm_debug("Opened sandbox: %s", path);

	while (SM_OK == atz_sandbox_readline(regex, sandbox_fd, &sandbox_subject, &sandbox_mappings, &sandbox_routines)) {

		if (!strcmp(subject, sandbox_subject)) {

			fclose(sandbox_fd);
			sandbox_fd = NULL;

			SM_FREE(sandbox_subject);

			if (mapping)
				*mapping = sandbox_mappings;
			else
				SM_FREE(sandbox_mappings);

			if (routines)
				*routines = sandbox_routines;
			else
				SM_FREE(sandbox_routines);

			return SM_OK;
		} else {
			SM_FREE(sandbox_subject);
			SM_FREE(sandbox_mappings);
			SM_FREE(sandbox_routines);
		}
	}

	SM_FREE(sandbox_subject);
	SM_FREE(sandbox_mappings);
	SM_FREE(sandbox_routines);
	fclose(sandbox_fd);

	sm_notice("Unknown user: %s", subject);
	return SM_NEG;
}
*/

static int
atz_sandbox_authorize(sm_module_t this, sm_atc_token_t atc_token, const char *routine)
{
	char *mapped_user = NULL;
	const char *subject = NULL;
	SM_BOOL unix_user = SM_FALSE;
	int ret = SM_FAULT;
	SM_DEBUG_ENTER();

	subject = atc_token->get_last_oneline(atc_token);

	sm_info("subject = %s", subject);

	if (!subject || !strcmp(subject,SM_ACCESS_ANONYMOUS_SUBJECT)) {
		/*1.a NO session context given */
		if (strcmp(routine, "CreateActivity") == 0) {
			/*2. initial call - need to create context */
			ret = smc_db_sandbox_acquire_account(&mapped_user);
		} else {
			/*2. let the the service deduce the user using context available in message (e.g.: activity id) */
			mapped_user = sm_strdup(SM_ACCESS_ANY);
		}
	} else {
		/*1.b have session context*/
		/* TODO */
		assert(0);
	}

	if (mapped_user) {
		sm_info("mapped to  %s", mapped_user);

		if (!strcmp(mapped_user, SM_ACCESS_ANY))
			unix_user = SM_FALSE;
		else
			unix_user = SM_TRUE;

		ret = atc_token->map_user(atc_token, subject, mapped_user, unix_user);
		SM_FREE(mapped_user);
	}

	if (ret == SM_NEG) {
		sm_error("Account pool exhausted");
	}

	return ret;
}

static int
atz_sandbox_map_user(sm_module_t this, sm_atc_token_t atc_token)
{
	int ret = SM_OK;



/*	atz_sandbox_ctxt_t ctxt;

	SM_DEBUG_ENTER();

	ctxt = this->ctxt;

	subject = atc_token->get_last_oneline(atc_token);

	authorize must have succeeded as we're here *
	assert(subject);

	ret = atz_sandbox_get_data(ctxt->sandbox_path, &ctxt->sandbox_regex, subject, &mappings, NULL);

	if (ret == SM_OK) {
		char *current;
		char *ptr;

		current = mappings;
		ptr = strchr(current, ',');

		for (;;) {
			SM_BOOL unix_user;
			int has_more = 0;

			if (ptr) {
				*ptr = '\0';
				has_more = 1;
			}

			sm_debug("Current mapping: %s", current);

			if (!strcmp(current, "*"))
				unix_user = SM_FALSE;
			else
				unix_user = SM_TRUE;

			if (strlen(current) > 0) {
				ret = atc_token->map_user(atc_token, subject, current, unix_user);

				if (ret != SM_OK)
					break;
			}

			if (has_more) {
				current = ptr + 1;
				ptr = strchr(current, ',');
			} else
				break;
		}
	}

	SM_FREE(mappings);
*/
	return ret;
}

sm_value_t *
handle_finished(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *mapped_user = params[6]->vstring;
	SM_DEBUG_ENTER_STR(mapped_user);

	smc_db_sandbox_release_account(mapped_user, handle_purge);

	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_failed(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *mapped_user = params[6]->vstring;
	SM_DEBUG_ENTER_STR(mapped_user);

	smc_db_sandbox_release_account(mapped_user, handle_purge);

	return SM_VALUE_NONE_SINGLETON;
}

sm_value_t *
handle_cancelled(sm_signal_t sig sm_unused, sm_value_t *params, const sm_value_t *previous sm_unused)
{
	char *mapped_user = params[6]->vstring;
	SM_DEBUG_ENTER_STR(mapped_user);

	smc_db_sandbox_release_account(mapped_user, handle_purge);

	return SM_VALUE_NONE_SINGLETON;
}
int
handle_purge(const char *username, const char *homedir, int uid)
{
	SM_DEBUG_ENTER_STR(username);
	sm_info("purging directory: %s", homedir);
	return SM_OK;
}
