/* $Id: pbs_jsdl_filter.c 313 2010-10-13 15:49:44Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>

#include <qcg-core/sm.h>
#include <qcg-core/compat.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/signal.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <include/jsdl_filter.h>
#include <xmlstructs/xmlstructs.h>

__SM_ID("$Id: pbs_jsdl_filter.c 313 2010-10-13 15:49:44Z mamonski $");

static int setup(sm_module_t);
static void cleanup(sm_module_t);
static sm_value_t *handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params);

SM_GENERAL_MODULE = {

	.header = {
		.entry_size = SM_GENERAL_MODULE_ENTRY_SIZE,
		.api_type = SM_API_GENERAL,

		.name = "pbs_jsdl_filter",
		.descr = "Torque/PBS Pro specific module for mapping non-DRMAA JSDL elements to native specification",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	},
};

static int
setup(sm_module_t this sm_unused)
{

	SM_DEBUG_ENTER();

	SM_TRY_FAULT(sm_signal_connect("smc_jsdl_filter", handler));

	return SM_OK;

fault:

	return SM_FAULT;
}

static void
cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();
}
static int
parse_range(struct jsdl__RangeValue_USCOREType *range, long long *minv, long long *maxv)
{
	double min_d = 0;
	double max_d = 1099511627776.0;
	SM_MANDATORY_P(range);
	SM_MANDATORY_P(minv);
	SM_MANDATORY_P(maxv);

	SM_DEBUG_ENTER_PTR(range);

	if (range->__sizeExact) {
		min_d = max_d = range->Exact[0].__item;

		if (range->Exact[0].epsilon) {
			min_d -= *range->Exact[0].epsilon;
			max_d += *range->Exact[0].epsilon;
		}
	}

	if (range->LowerBoundedRange) {
		min_d = range->LowerBoundedRange->__item;
		if (range->LowerBoundedRange->exclusiveBound && *range->LowerBoundedRange->exclusiveBound) {
			min_d += 0.000001;
		}
	}

	 if (range->UpperBoundedRange) {
		max_d = range->UpperBoundedRange->__item;
		if (range->UpperBoundedRange->exclusiveBound && *range->UpperBoundedRange->exclusiveBound) {
			max_d += 0.000001;
		}
	}

	*minv = (long long)(min_d);
	*maxv = (long long)(max_d);

	if (*minv > *maxv)
		return SM_FAULT;
	else
		return SM_OK;

}

static sm_value_t *
handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params)
{
	sm_alloc_tracked_t mh;
	_jsdl__JobDefinition jsdl;
	const char *input_jsdl;
	char *output_jsdl;
	char native[1024] = "";
	sm_value_t *result = NULL;
	SM_BOOL modified = SM_FALSE;

	SM_DEBUG_ENTER();

	if (sm_previous_signal_failed(chained_params))
		return sm_value_clone(chained_params);
	else if (sm_signal_first(chained_params))
		input_jsdl = params[0]->vstring;
	else
		input_jsdl = chained_params[1]->vstring;

	SM_TRY_NOT(mh = sm_gsoap_buffer_get(input_jsdl, &jsdl, jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", xmlstructs_namespaces));


	if (jsdl.jsdl__JobDescription->jsdl__Resources) {
		struct jsdl__Resources_USCOREType *r = jsdl.jsdl__JobDescription->jsdl__Resources;
		char resources[256] = "";

		if (r->jsdl_qcg_comp_factory__NativeSpecification)
			strlcat(native, r->jsdl_qcg_comp_factory__NativeSpecification, sizeof(native));

		if (r->jsdl__CandidateHosts && r->jsdl__CandidateHosts->__sizeHostName == 1) {

			strlcat(native, " -l host=", sizeof(native));
			strlcat(native, r->jsdl__CandidateHosts->jsdl__HostName[0], sizeof(native));

			r->jsdl__CandidateHosts = NULL;
			modified = SM_TRUE;
		}

		if (r->jsdl__TotalCPUCount && !r->jsdl__IndividualCPUCount && !r->jsdl__TotalResourceCount) {
			long long min_cpu = 1;
			long long max_cpu = 1;

			if (parse_range(r->jsdl__TotalCPUCount, &min_cpu, &max_cpu) == SM_OK && (min_cpu == max_cpu)) {
				char tmp[64];

				snprintf(tmp, 64, " -l nodes=%lld", min_cpu);
				strlcat(native, tmp, sizeof(native));
				r->jsdl__TotalCPUCount = NULL;
				modified = SM_TRUE;
			}
		} else if (!r->jsdl__TotalCPUCount && r->jsdl__IndividualCPUCount) {
			long long min_ppn = 1;
			long long max_ppn = 1;
			long long min_nodes = 1;
			long long max_nodes = 1;

			if (parse_range(r->jsdl__IndividualCPUCount, &min_ppn, &max_ppn) != SM_OK || (min_ppn != max_ppn)) {
				sm_notice("Invalid IndividualCPUCount value");
			} else if (r->jsdl__TotalResourceCount && (parse_range(r->jsdl__TotalResourceCount, &min_nodes, &max_nodes) != SM_OK || (min_nodes != max_nodes))) {
				sm_notice("Invalid TotalResourceCount value");
			} else {
				char tmp[64];

				snprintf(tmp, 64, " -l nodes=%lld:ppn=%lld", min_nodes, min_ppn);
				strlcat(native, tmp, sizeof(native));
				r->jsdl__IndividualCPUCount = NULL;
				r->jsdl__TotalResourceCount = NULL;
				modified = SM_TRUE;
			}
		}

		if (r->jsdl__IndividualCPUTime) {
			long long min_time = 0;
			long long max_time = 0;

			if (parse_range(r->jsdl__IndividualCPUTime, &min_time, &max_time) != SM_OK || (min_time != max_time) || (min_time <= 0)) {
				sm_notice("Invalid IndividualCPUTime value");
			} else {
				char tmp[64];

				snprintf(tmp, 64, " -l walltime=%lld", min_time);
				strlcat(native, tmp, sizeof(native));
				r->jsdl__IndividualCPUTime = NULL;
				modified = SM_TRUE;
			}
		}

		if (r->jsdl__IndividualPhysicalMemory) {
			long long min_mem = 0;
			long long max_mem = 0;

			if (parse_range(r->jsdl__IndividualPhysicalMemory, &min_mem, &max_mem) != SM_OK || (min_mem != max_mem) || (min_mem <= 0)) {
				sm_notice("Invalid IndividualPhysicalMemory value");
			} else {
				char tmp[64];

				snprintf(tmp, 64, " -l pmem=%lldb", min_mem);
				strlcat(native, tmp, sizeof(native));
				r->jsdl__IndividualPhysicalMemory = NULL;
				modified = SM_TRUE;
			}
		}

		if (r->jsdl__TotalPhysicalMemory) {
			long long min_mem = 0;
			long long max_mem = 0;

			if (parse_range(r->jsdl__TotalPhysicalMemory, &min_mem, &max_mem) != SM_OK || (min_mem != max_mem) || (min_mem <= 0)) {
				sm_notice("Invalid TotalPhysicalMemory value");
			} else {
				char tmp[64];

				snprintf(tmp, 64, " -l mem=%lldb", min_mem);
				strlcat(native, tmp, sizeof(native));
				r->jsdl__TotalPhysicalMemory = NULL;
				modified = SM_TRUE;
			}
		}

		if (r->jsdl_qcg_comp_factory__Queue) {
			strlcat(native, " -q ", sizeof(native));
			strlcat(native, r->jsdl_qcg_comp_factory__Queue , sizeof(native));

			r->jsdl_qcg_comp_factory__Queue = NULL;
			modified = SM_TRUE;
		}

		if (r->jsdl__CPUArchitecture && r->jsdl__CPUArchitecture->jsdl__CPUArchitectureName != jsdl__ProcessorArchitectureEnumeration__other) {
			const char *arch = soap_jsdl__ProcessorArchitectureEnumeration2s(NULL, r->jsdl__CPUArchitecture->jsdl__CPUArchitectureName);

			if (strlen(resources)) {
				strlcat(resources,",", sizeof(resources));
			}

			strlcat(resources, "arch=", sizeof(resources));
			strlcat(resources, arch, sizeof(resources));

			r->jsdl__CPUArchitecture = NULL;
			modified = SM_TRUE;
		}

		if (r->jsdl__OperatingSystem
				&& r->jsdl__OperatingSystem->jsdl__OperatingSystemType
				&& r->jsdl__OperatingSystem->jsdl__OperatingSystemType->jsdl__OperatingSystemName != jsdl__OperatingSystemTypeEnumeration__other) {
			const char *osname = soap_jsdl__OperatingSystemTypeEnumeration2s(NULL, r->jsdl__OperatingSystem->jsdl__OperatingSystemType->jsdl__OperatingSystemName);

			if (strlen(resources)) {
				strlcat(resources,",", sizeof(resources));
			}

			strlcat(resources, "opsys=", sizeof(resources));
			strlcat(resources, osname, sizeof(resources));

			if (!r->jsdl__OperatingSystem->jsdl__OperatingSystemVersion) {
				r->jsdl__OperatingSystem = NULL;
			}

			modified = SM_TRUE;
		}

		if (strlen(resources)) {
			strlcat(native, " -l ", sizeof(native));
			strlcat(native, resources, sizeof(native));
		}
	}

	if (jsdl.jsdl__JobDescription->jsdl__JobIdentification && jsdl.jsdl__JobDescription->jsdl__JobIdentification->__sizeJobProject) {
		strlcat(native, " -A ", sizeof(native));
		strlcat(native, jsdl.jsdl__JobDescription->jsdl__JobIdentification->jsdl__JobProject[0], sizeof(native));

		jsdl.jsdl__JobDescription->jsdl__JobIdentification->__sizeJobProject = 0;
		modified = SM_TRUE;
	}


	if (modified) {
		struct jsdl__Resources_USCOREType rst;
		sm_info("final native specification = %s", native);

		if (!jsdl.jsdl__JobDescription->jsdl__Resources) {
			memset(&rst, 0, sizeof(rst));
			jsdl.jsdl__JobDescription->jsdl__Resources = &rst;
		}

		jsdl.jsdl__JobDescription->jsdl__Resources->jsdl_qcg_comp_factory__NativeSpecification = native;

		SM_TRY_NOT(output_jsdl = sm_gsoap_buffer_put(&jsdl, jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", xmlstructs_namespaces));

		result = sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_OK, SM_VALUE_STRING, output_jsdl);
		SM_FREE(output_jsdl);
		SM_TRY_NOT(result);
	} else {
		SM_TRY_NOT(result = sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_OK, SM_VALUE_STRING, input_jsdl));
	}

	sm_alloc_tracked_free(&mh);
	return result;

fault:
	sm_alloc_tracked_free(&mh);

	signal->stop_emission(signal);

	if (result)
		return result;
	else
		return sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_INTERNAL_FAULT, "Internal fault in lsf JSDL module", SM_VALUE_NONE);
}
