/* $Id: reservation_python.c 2 2009-10-05 10:09:52Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>
#include <dlfcn.h>
#include <time.h>

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/python.h>
#include <qcg-core/utils.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <include/reservation_module.h>

__SM_ID("$Id: reservation_python.c 2 2009-10-05 10:09:52Z mamonski $");

typedef struct reservation_python_ctxt_s {
	char *module_name;
	sm_python_object_t module;
} *reservation_python_ctxt_t;

static int setup(sm_module_t);
static void cleanup(sm_module_t);
static sm_value_t *reservation_python_create(sm_module_t, sm_value_t *);
static sm_value_t *reservation_python_terminate(sm_module_t, sm_value_t *);
static sm_value_t *reservation_python_get_status(sm_module_t, sm_value_t *);
static sm_value_t *reservation_python_get_active(sm_module_t);

RESERVATION_MODULE = {

	.header = {
		.entry_size = RESERVATION_MODULE_ENTRY_SIZE,
		.api_type = RESERVATION_MODULE_API,

		.name = "reservation_python",
		.descr = "Python reservation module",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_TRUE
	},

	.create = reservation_python_create,
	.terminate = reservation_python_terminate,
	.get_status = reservation_python_get_status,
	.get_active = reservation_python_get_active
};

static int
setup(sm_module_t this)
{
	char *base = NULL, *p = NULL, *config_xml = NULL;
	const char *path = NULL;
	reservation_python_ctxt_t ctxt = NULL;
	char buf[4096];

	SM_DEBUG_ENTER();

        snprintf(buf, sizeof(buf),
	"RESERVATION_OK = %d\n"
        "RESERVATION_INTERNAL_FAULT = %d\n"
        "RESERVATION_NOT_AUTHORIZED_FAULT = %d\n"
        "RESERVATION_UNSUPPORTED_FEATURE_FAULT = %d\n"
        "RESERVATION_INVALID_REQUEST_MESSAGE_FAULT = %d\n"
        "RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT = %d\n"
        "RESERVATION_NO_SLOTS_AVAILABLE_FAULT = %d\n"
	"RESERVATION_NO_MATCHING_RESOURCES_FAULT = %d\n",
	RESERVATION_OK,
	RESERVATION_INTERNAL_FAULT,
	RESERVATION_NOT_AUTHORIZED_FAULT,
	RESERVATION_UNSUPPORTED_FEATURE_FAULT,
	RESERVATION_INVALID_REQUEST_MESSAGE_FAULT,
	RESERVATION_UNKNOWN_RESERVATION_IDENTIFIER_FAULT,
	RESERVATION_NO_SLOTS_AVAILABLE_FAULT,
	RESERVATION_NO_MATCHING_RESOURCES_FAULT);

	if (sm_python_load_module_code_from_string("qcg.comp", buf) != SM_OK) {
		sm_error("Failed to extend Python module");
		goto fault;
	}

	if (!(ctxt = sm_malloc(sizeof(*ctxt))))
		goto fault;

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/@path", &path) != SM_OK)
		goto fault;

	if (sm_conf_node_raw(this->conf_ctxt, SM_TRUE, "/", &config_xml) != SM_OK)
		goto fault;

	if (!(base = sm_basename(path)))
		goto fault;

	if ((p = strchr(base, '.')))
		*p = '\0';

	ctxt->module_name = base;

	sm_info("Setting up: %s", path);

	if (!(ctxt->module = sm_python_load_module(path, config_xml,
					sm_conf_ctxt_get_default_prefix(this->conf_ctxt),
					sm_conf_ctxt_get_service_prefix(this->conf_ctxt),
					sm_conf_ctxt_get_service_namespace(this->conf_ctxt)))) {
		sm_error("Failed to load module: %s", path);
		goto fault;
	}

	if (!sm_python_object_has_attr(ctxt->module, "create", SM_TRUE)) {
		sm_error("The module doesn't define create function");
		goto fault;
	}

	if (!sm_python_object_has_attr(ctxt->module, "terminate", SM_TRUE)) {
		sm_error("The module doesn't define terminate function");
		goto fault;
	}

	if (!sm_python_object_has_attr(ctxt->module, "get_status", SM_TRUE)) {
		sm_error("The module doesn't define get_status function");
		goto fault;
	}

	if (!sm_python_object_has_attr(ctxt->module, "get_active", SM_TRUE)) {
		sm_error("The module doesn't define get_active function");
		goto fault;
	}

	SM_FREE(config_xml);

	this->ctxt = ctxt;

	return SM_OK;

fault:
	if (ctxt->module)
		sm_python_unload_module(base);
	SM_FREE(base);
	SM_FREE(config_xml);
	SM_FREE(ctxt);
	return SM_FAULT;
}

static void
cleanup(sm_module_t this sm_unused)
{
	reservation_python_ctxt_t ctxt;
	SM_DEBUG_ENTER();

	if (!this->ctxt)
		return;

	ctxt = this->ctxt;

	sm_python_unload_module(ctxt->module_name);
	SM_FREE(this->ctxt);
}

static sm_value_t *
reservation_python_create(sm_module_t this, sm_value_t *definition)
{
	reservation_python_ctxt_t ctxt;
	sm_python_object_t result;
	sm_value_t *value;

	SM_MANDATORY_PARAM(this, NULL);
	SM_MANDATORY_PARAM(definition, NULL);

	assert(sm_value_count(definition) == 1);
	assert(SM_VALUE_IS_STRING(definition[0]));

	SM_DEBUG_ENTER_STR(definition[0]->vstring);

	ctxt = this->ctxt;

	result = sm_python_call_method(ctxt->module, "create", "(s)", definition[0]->vstring);
	if (!result)
		return sm_value_create(2, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, SM_VALUE_STRING, "Failed to call Python script \"create\" method");

	value = sm_python_object_to_value(result);
	sm_python_object_free(&result);

	if (!value)
		return sm_value_create(2, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, SM_VALUE_STRING, "Failed to parse Python script return value");

	SM_VALUE_DEBUG(value);
	return value;
}

static sm_value_t *
reservation_python_terminate(sm_module_t this, sm_value_t *identifier)
{
	reservation_python_ctxt_t ctxt;
	sm_python_object_t result;
	sm_value_t *value;

	SM_MANDATORY_PARAM(this, NULL);
	SM_MANDATORY_PARAM(identifier, NULL);

	assert(sm_value_count(identifier) == 1);
	assert(SM_VALUE_IS_STRING(identifier[0]));

	SM_DEBUG_ENTER_STR(identifier[0]->vstring);

	ctxt = this->ctxt;

	result = sm_python_call_method(ctxt->module, "terminate", "(s)", identifier[0]->vstring);
	if (!result)
		return sm_value_create(2, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, SM_VALUE_STRING, "Failed to call Python script \"terminate\" method");

	value = sm_python_object_to_value(result);
	sm_python_object_free(&result);

	if (!value)
		return sm_value_create(2, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, SM_VALUE_STRING, "Failed to parse Python script return value");

	SM_VALUE_DEBUG(value);
	return value;
}

static sm_value_t *
reservation_python_get_status(sm_module_t this, sm_value_t *identifier)
{
	reservation_python_ctxt_t ctxt;
	sm_python_object_t result;
	sm_value_t *value;

	SM_MANDATORY_PARAM(this, NULL);
	SM_MANDATORY_PARAM(identifier, NULL);

	assert(sm_value_count(identifier) == 1);
	assert(SM_VALUE_IS_STRING(identifier[0]));

	SM_DEBUG_ENTER_STR(identifier[0]->vstring);

	ctxt = this->ctxt;

	result = sm_python_call_method(ctxt->module, "get_status", "(s)", identifier[0]->vstring);
	if (!result)
		return sm_value_create(2, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, SM_VALUE_STRING, "Failed to call Python script \"get_status\" method");

	value = sm_python_object_to_value(result);
	sm_python_object_free(&result);

	if (!value)
		return sm_value_create(2, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, SM_VALUE_STRING, "Failed to parse Python script return value");

	SM_VALUE_DEBUG(value);
	return value;
}

static sm_value_t *
reservation_python_get_active(sm_module_t this)
{
	reservation_python_ctxt_t ctxt;
	sm_python_object_t result;
	sm_value_t *value;

	SM_MANDATORY_PARAM(this, NULL);

	SM_DEBUG_ENTER();

	ctxt = this->ctxt;

	result = sm_python_call_method(ctxt->module, "get_active", "()");
	if (!result)
		return sm_value_create(2, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, SM_VALUE_STRING, "Failed to call Python script \"get_active\" method");

	value = sm_python_object_to_value(result);
	sm_python_object_free(&result);

	if (!value)
		return sm_value_create(2, SM_VALUE_INT32, RESERVATION_INTERNAL_FAULT, SM_VALUE_STRING, "Failed to parse Python script return value");

	SM_VALUE_DEBUG(value);
	return value;
}
