/* $Id: ardl.c 2 2009-10-05 10:09:52Z mamonski $ */

/*
 *  Copyright 2006-2009 QCG Project Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

#include <stdbool.h>

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/signal.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>

#include <common.h>
#include <ardl.h>
#include <include/ardl_filter.h>
#include <stub/qcgcompStub.h>
#include <stub/qcgcompH.h>

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "ardl"
	}
};

extern struct Namespace qcgcomp_namespaces[];

sm_status_ret_code_t
smc_ardl_setup(void)
{
	sm_status_ret_code_t ret;

	SM_DEBUG_ENTER();

	ret = sm_signal_register("smc_ardl_filter", SM_SIGNAL_DEFAULT,
		(sm_value_type_t []) { SM_VALUE_STRING }, 1,
		(sm_value_type_t []) { SM_VALUE_RET_CODE, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG }, 2);

	return ret;
}

void
smc_ardl_cleanup(void)
{
	SM_DEBUG_ENTER();

	sm_signal_unregister("smc_ardl_filter");
}

const char *
smc_ardl_get_username(_ardl__ReservationDefinition *ardl)
{
	SM_MANDATORY_P(ardl);

	SM_DEBUG_ENTER();

	if (ardl->ardl__ReservationDescription->ardl__Resources)
		return SM_DEBUG_LEAVE_STR(ardl->ardl__ReservationDescription->ardl__Resources->UserName);

	return SM_DEBUG_LEAVE_STR(NULL);
}

sm_status_ret_code_t
smc_ardl_set_username(_ardl__ReservationDefinition *ardl, const char *username, smc_malloc_function malloc_func, smc_strdup_function strdup_func, void *ctxt)
{
	if (!ardl->ardl__ReservationDescription->ardl__Resources) {
		if (!(ardl->ardl__ReservationDescription->ardl__Resources = malloc_func(ctxt, sizeof(ardl->ardl__ReservationDescription->ardl__Resources))))
			return SM_FAULT;

		soap_default__ardl__Resources(ctxt, ardl->ardl__ReservationDescription->ardl__Resources);
	}

	if (!(ardl->ardl__ReservationDescription->ardl__Resources->UserName = strdup_func(ctxt, username)))
		return SM_FAULT;

	return SM_OK;
}

char *
smc_ardl_str(_ardl__ReservationDefinition *ardl, smc_strdup_function func, void *ctxt)
{
	char *tmp = NULL;
	char *result = NULL;

	SM_MANDATORY_P(ardl);

	SM_DEBUG_ENTER();

	tmp = sm_gsoap_buffer_put(ardl, ardl__ARDLReservationDefinition_USCOREType, "ardl:ReservationDefinition", qcgcomp_namespaces);

	if (tmp) {
		if (func) {
			SM_MANDATORY_P(ctxt);

			result = func(ctxt, tmp);
			SM_FREE(tmp);
		} else
			result = tmp;
	} else
		sm_error("Failed to get string representation of ARDL");

	return result;
}

sm_alloc_tracked_t
smc_ardl_from_str(const char *ardl_str, _ardl__ReservationDefinition *ardl)
{
	sm_alloc_tracked_t handle;

	SM_MANDATORY_P(ardl_str);
	SM_MANDATORY_P(ardl);

	SM_DEBUG_ENTER();

	handle = sm_gsoap_buffer_get(ardl_str, ardl, ardl__ARDLReservationDefinition_USCOREType, NULL, qcgcomp_namespaces);

	if (!handle)
		sm_error("Failed to parse ARDL from string");

	return handle;
}

sm_ret_code_t
smc_ardl_handle_filters(const char *ardl_str, const char **out, smc_strdup_function func, void *ctxt)
{
	sm_list_t return_values = NULL;
	sm_list_t l = NULL;
	sm_value_t *value = NULL;

	SM_DEBUG_ENTER();

	SM_MANDATORY_P(func);
	SM_MANDATORY_P(ctxt);

	SM_TRY_FAULT(sm_signal_emit("smc_ardl_filter", ardl_str, &return_values));

	/* We're interested in the output of the last one. */
	for (l = return_values; l; l = l->next) {
		if (!l->next)
			break;
	}

	if (l && (value = l->data)) {
		if (SM_VALUE_RET_CODE_C(value[0]))
			return sm_ret_code_create(SM_VALUE_RET_CODE_C(value[0]), SM_VALUE_RET_CODE_STRING(value[0]));

		sm_debug("After going through a chain of %zu filters, the final ARDL is: %s", sm_list_count(return_values), value[1]->vstring);
		SM_TRY_NOT(*out = func(ctxt, value[1]->vstring));
	} else
		SM_TRY_NOT(*out = func(ctxt, ardl_str));

	sm_list_free_user(return_values, (sm_list_free_user_function) sm_value_free);
	return sm_ret_code_create(ARDL_FILTER_OK, NULL);

fault:
	sm_list_free_user(return_values, (sm_list_free_user_function) sm_value_free);
	return sm_ret_code_create(ARDL_FILTER_INTERNAL_FAULT, "Internal fault in ARDL filter");
}
