/* $Id: bes_factory.c 312 2010-10-13 14:55:42Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <qcg-core/sm.h>
#include <qcg-core/access.h>
#include <qcg-core/logger.h>
#include <qcg-core/signal.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/atz_module.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>

#include <common.h>
#include <faults.h>
#include <icomm.h>
#include <identifiers.h>
#include <job.h>
#include <jsdl.h>
#include <user.h>
#include <stub/qcgcompStub.h>
#include <stub/qcgcompH.h>
#include <include/extensions.h>
#include <include/jsdl_filter.h>
#include <include/namespaces.h>
#include <include/submission_module.h>


__SM_ID("$Id: bes_factory.c 312 2010-10-13 14:55:42Z mamonski $");

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "bes_factory"
	}
};

extern struct Namespace qcgcomp_namespaces[];

static sm_status_ret_code_t
get_job_from_epr(sm_transport_t transport, sm_transport_fault_t fault, struct wsa5__EndpointReferenceType *ActivityIdentifier, bool readonly, sm_list_t *result)
{
	const char *id = NULL;
	sm_ret_code_t ret;
	bool return_neg;

	SM_MANDATORY_P(transport);
	SM_MANDATORY_P(ActivityIdentifier);
	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER();

	return_neg = false;
	*result = NULL;

	if (smc_parse_activity_epr(ActivityIdentifier, &id) != SM_OK)
		THROW_INTERNAL_FAULT("Cannot parse activity EPR");

	if (readonly)
		ret = smc_job_get_readonly(id, result);
	else
		ret = smc_job_get(id, result);

	if (ret.code == SMC_JOB_NOT_FOUND) {
		return_neg = true;
		THROW_APP_FAULT(bes_factory_UnknownActivityIdentifierFault, id);
	} else if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	return SM_OK;

fault:
	if (return_neg)
		return SM_NEG;
	else
		return SM_FAULT;
}

/*
 * Possible faults: NotAuthorizedFault, NotAcceptingNewActivitiesFault,
 * UnsupportedFeatureFault, InvalidRequestMessageFault
 */
int
__bes_factory__CreateActivity(
	struct soap *soap,
	struct bes_factory__CreateActivityType *bes_factory__CreateActivity,
	struct bes_factory__CreateActivityResponseType *bes_factory__CreateActivityResponse
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	sm_ret_code_t ret;
	_jsdl__JobDefinition *jsdl = NULL;
	const char *jsdl_str = NULL;
	const char *jsdl_str_after_filters = NULL;
	sm_alloc_tracked_t tracked = NULL;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	const char *subject = NULL;
	char *username = NULL;
	sm_value_t *atc_token = NULL;
	uid_t uid;

	extern SM_BOOL smc_accepting_new_activities;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "CreateActivity", fault) != SM_OK)
		goto fault;

	if (transport->access_check(transport, SM_TRUE, "CreateActivity", fault) != SM_ACCESS_GRANTED)
		goto fault;

	if (!smc_accepting_new_activities)
		THROW_APP_FAULT(bes_factory_NotAcceptingNewActivitiesFault, NULL);

	jsdl = bes_factory__CreateActivity->bes_factory__ActivityDocument->jsdl__JobDefinition;
	SM_TRY_NOT(jsdl_str = smc_jsdl_str(jsdl, (smc_strdup_function) soap_strdup, soap));

	if (strlen(jsdl_str) > SMC_MAX_JSDL_LENGTH)
		THROW_INTERNAL_FAULT("JSDL size exceeds predefined limit");

	ret = smc_jsdl_handle_filters(jsdl_str, &jsdl_str_after_filters, (smc_strdup_function) soap_strdup, soap);

	switch (ret.code) {
		case JSDL_FILTER_OK:
			tracked = smc_jsdl_from_str(jsdl_str_after_filters, jsdl);
			break;
		default:
			THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));
			break;
	}

	ret = smc_user_handle_submission(smc_jsdl_get_username(jsdl), transport->atc_token, &username, &uid);
	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));


	subject = transport->atc_token->get_last_oneline(transport->atc_token);
	atc_token = transport->atc_token->serialize(transport->atc_token, transport->atc_token->get_last(transport->atc_token));

	if (atc_token) {
		SM_TRY_NOT(params = sm_value_create(5, SM_VALUE_STRING, subject,
				SM_VALUE_STRING, username, SM_VALUE_STRING, jsdl_str,
				SM_VALUE_STRING, jsdl_str_after_filters, SM_VALUE_ARRAY,
				atc_token));
	} else { /*anonymous*/
		SM_TRY_NOT(params = sm_value_create(4, SM_VALUE_STRING, subject,
				SM_VALUE_STRING, username, SM_VALUE_STRING, jsdl_str,
				SM_VALUE_STRING, jsdl_str_after_filters));
	}

	if (smc_icomm_request_CreateActivity(uid, params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if (!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT_FROM_VAL(result);

	sm_info("Created activity: id = %s", result[2]->vstring);

	if (smc_create_activity_epr(transport, &bes_factory__CreateActivityResponse->ActivityIdentifier, result[2]->vstring) != SM_OK)
		THROW_INTERNAL_FAULT("Activity created but failed to create EPR");

	sm_alloc_tracked_free(&tracked);
	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->exit(transport, NULL);

fault:
	sm_alloc_tracked_free(&tracked);
	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->fault(transport, fault);
}

static sm_status_ret_code_t
get_drms_state(smc_job_t job, smc_job_state_t *state, char **extra_info)
{
	uid_t uid;
	sm_ret_code_t ret;
	sm_status_ret_code_t res;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_MANDATORY_P(job);
	SM_MANDATORY_P(state);

	*state = SMC_UNSPECIFIED_STATE;
	res = SM_OK;

	ret = smc_user_get_uid(job->get_mapped_user(job), &uid);
	if (ret.code) {
		sm_error("Failed to get user UID: %d:%s", ret.code, ret.string);
		goto fault;
	}

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, job->get_drms_id(job)));

	if (smc_icomm_request_GetActivityStatuses(uid, params, &result) != SM_OK) {
		sm_error("Failed to send request to session process");
		goto fault;
	}

	if (!SM_VALUE_RET_CODE_OK(result[0])) {
		if (SM_VALUE_RET_CODE_C(result[0]) == bes_factory_UnknownActivityIdentifierFault) {
			res = SM_NEG;
		} else {
			sm_error("Failed to get job status from DRMS: %d:%s", SM_VALUE_RET_CODE_C(result[0]), SM_VALUE_RET_CODE_STRING(result[0]));
			goto fault;
		}

	} else {
		assert(SM_VALUE_IS_INT(result[1]));
		*state = result[1]->vint32;
		if (sm_value_count(result) == 3 && extra_info) {
			assert(SM_VALUE_IS_ARRAY(result[2]));
			assert(SM_VALUE_IS_STRING(result[2]->varray[0]));
			*extra_info = sm_strdup(result[2]->varray[0]->vstring);
		} else if (extra_info) {
			*extra_info = NULL;
		}
	}

	sm_value_free(&params);
	sm_value_free(&result);
	return res;

fault:
	sm_value_free(&params);
	sm_value_free(&result);
	return SM_FAULT;
}

/*
 * Possible faults: UnknownActivityIdentifierFault, NotAuthorizedFault
 */
int
__bes_factory__GetActivityStatuses(
	struct soap *soap,
	struct bes_factory__GetActivityStatusesType *bes_factory__GetActivityStatuses,
	struct bes_factory__GetActivityStatusesResponseType *bes_factory__GetActivityStatusesResponse
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	sm_list_t jobs = NULL;
	smc_job_t job = NULL;
	struct bes_factory__GetActivityStatusResponseType *response;
	struct bes_factory__ActivityStatusType *st;
	smc_job_state_t state;
	smc_job_state_t drms_state;
	sm_ret_code_t ret;
	sm_status_ret_code_t status;
	char *extra_info = NULL;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetActivityStatuses", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "GetActivityStatuses", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	if (bes_factory__GetActivityStatuses->__sizeActivityIdentifier != 1)
		THROW_APP_FAULT(bes_factory_UnsupportedFeatureFault, "Processing more than one EPR not supported");

	SM_TRY_NOT_OK(get_job_from_epr(transport, fault, bes_factory__GetActivityStatuses->ActivityIdentifier, true, &jobs));
	SM_ASSERT(sm_list_count(jobs) <= 1, "Parameter sweep not implemented in %s", __FUNCTION__);
	job = jobs->data;
	state = job->get_state(job);

	sm_debug("Database state: %s", smc_job_state_str(state));

	switch (state) {
		case SMC_STAGE_IN:
		case SMC_STAGE_OUT:
		case SMC_FINISHED:
		case SMC_FAILED:
		case SMC_CANCELLED:
			sm_debug("Using database state");
			goto out;
		default:
			break;
	}

	/* If state is not final and is one of those that are monitored
	 * periodically by the poller, ask DRMS for the most recent state. */
	if ((status = get_drms_state(job, &drms_state, &extra_info)) == SM_FAULT)
		THROW_INTERNAL_FAULT("Couldn't contact DRMS to get the most recent state of an activity");
	else if (status == SM_NEG)
		goto out;

	sm_debug("DRMS state: %s", smc_job_state_str(drms_state));

	if (state == drms_state) {
		sm_debug("Database state == DRMS state. That's OK.");
		goto out;
	}


	/* Now get a RW job object that is locked and update its state
	 * according to DRMS state. */
	SM_LIST_FREE_USER(jobs, smc_job_free);
	SM_TRY_NOT_OK(get_job_from_epr(transport, fault, bes_factory__GetActivityStatuses->ActivityIdentifier, false, &jobs));
	job = jobs->data;
	state = job->get_state(job);

	if ((status = get_drms_state(job, &drms_state, NULL)) == SM_FAULT)
		THROW_INTERNAL_FAULT("Couldn't contact DRMS to get the most recent state of an activity");
	else if (status == SM_NEG)
		goto out;

	sm_debug("Updating database state with DRMS state (%s)", smc_job_state_str(drms_state));

	state = drms_state;

	job->set_state(job, state);
	ret = job->save(job);
	if (ret.code)
		THROW_INTERNAL_FAULT("Couldn't update object in database");

out:
	SM_TRY_FAULT(status = smc_user_authorize(SMC_USER_AUTH_GET_STATUS, job->get_subject(job), transport->atc_token->get_last_oneline(transport->atc_token)));
	if (status == SM_NEG)
		THROW_APP_FAULT(bes_factory_NotAuthorizedFault, "You are not authorized to get status of this activity");

	SM_TRY_NOT(bes_factory__GetActivityStatusesResponse->Response = transport->transport_malloc(transport, sizeof(*response)));
	response = bes_factory__GetActivityStatusesResponse->Response;
	soap_default_bes_factory__GetActivityStatusResponseType(soap, response);

	bes_factory__GetActivityStatusesResponse->__sizeResponse = 1;

	if (smc_create_activity_epr(transport, &response->ActivityIdentifier, job->get_id(job)) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to create activity EPR");

	SM_TRY_NOT(response->bes_factory__ActivityStatus = soap_malloc(soap, sizeof(struct bes_factory__ActivityStatusType)));
	st = response->bes_factory__ActivityStatus;
	soap_default_bes_factory__ActivityStatusType(soap, st);

	switch (state) {
		case SMC_UNSPECIFIED_STATE:
			SM_ASSERT(0, "Job %s in unspecified state", job->get_id(job));
			break;
		case SMC_QUEUED:
			st->state = bes_factory__ActivityStateEnumeration__Pending;
			st->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Queued;
			SM_TRY_NOT(st->union_ActivityStatusType.qcg_comp_factory__Queued = soap_malloc(soap, sizeof(struct _qcg_comp_factory__Queued)));
			break;
		case SMC_STAGE_IN:
			st->state = bes_factory__ActivityStateEnumeration__Pending;
			st->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Stage_in;
			SM_TRY_NOT(st->union_ActivityStatusType.qcg_comp_factory__Stage_in = soap_malloc(soap, sizeof(struct _qcg_comp_factory__Stage_in)));
			break;
		case SMC_HELD:
			st->state = bes_factory__ActivityStateEnumeration__Pending;
			st->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Held;
			SM_TRY_NOT(st->union_ActivityStatusType.qcg_comp_factory__Held = soap_malloc(soap, sizeof(struct _qcg_comp_factory__Held)));
			break;
		case SMC_EXECUTING:
			st->state = bes_factory__ActivityStateEnumeration__Running;
			st->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Executing;
			SM_TRY_NOT(st->union_ActivityStatusType.qcg_comp_factory__Executing = soap_malloc(soap, sizeof(struct _qcg_comp_factory__Executing)));
			break;
		case SMC_STAGE_OUT:
			st->state = bes_factory__ActivityStateEnumeration__Running;
			st->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Stage_out;
			SM_TRY_NOT(st->union_ActivityStatusType.qcg_comp_factory__Stage_out = soap_malloc(soap, sizeof(struct _qcg_comp_factory__Stage_out)));
			break;
		case SMC_SUSPENDED:
			st->state = bes_factory__ActivityStateEnumeration__Running;
			st->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Suspended;
			SM_TRY_NOT(st->union_ActivityStatusType.qcg_comp_factory__Suspended = soap_malloc(soap, sizeof(struct _qcg_comp_factory__Suspended)));
			break;
		case SMC_CANCELLED:
			st->state = bes_factory__ActivityStateEnumeration__Cancelled;
			break;
		case SMC_FINISHED:
			st->state = bes_factory__ActivityStateEnumeration__Finished;

			SM_TRY_NOT(st->qcg_comp_factory__ActivityEndStatus = soap_malloc(soap, sizeof(struct qcg_comp_factory__ActivityEndStatus_USCOREType)));
			soap_default_qcg_comp_factory__ActivityEndStatus_USCOREType(soap, st->qcg_comp_factory__ActivityEndStatus);

			if (job->has_exit_status(job)) {
				SM_TRY_NOT(st->qcg_comp_factory__ActivityEndStatus->ExitStatus = soap_malloc(soap, sizeof(ULONG64)));
				*(st->qcg_comp_factory__ActivityEndStatus->ExitStatus) = job->get_exit_status(job);
			}
			break;
		case SMC_FAILED:
		{
			const char *staging_fault;

			st->state = bes_factory__ActivityStateEnumeration__Failed;

			SM_TRY_NOT(st->qcg_comp_factory__ActivityEndStatus = soap_malloc(soap, sizeof(struct qcg_comp_factory__ActivityEndStatus_USCOREType)));
			soap_default_qcg_comp_factory__ActivityEndStatus_USCOREType(soap, st->qcg_comp_factory__ActivityEndStatus);

			if (job->get_posix_signal_name(job))
				SM_TRY_NOT(st->qcg_comp_factory__ActivityEndStatus->POSIXSignalName = soap_strdup(soap, job->get_posix_signal_name(job)));
			if (job->has_core_dumped(job)) {
				SM_TRY_NOT(st->qcg_comp_factory__ActivityEndStatus->CoreDumped = soap_malloc(soap, sizeof(enum xsd__boolean)));
				*(st->qcg_comp_factory__ActivityEndStatus->CoreDumped) = job->get_core_dumped(job);
			}

			if ((staging_fault = job->get_staging_fault(job))) {
				struct soap_dom_element *f;

				/*
				 * XXX: There's a bug in gSOAP -- if we use response->Fault then it
				 * is sent out as bes-factory:Fault. That's why we put this element
				 * using DOM. We still leave Fault element in the schema as this
				 * should make life easier for clients.
				 */

				SM_TRY_NOT(f = soap_malloc(soap, sizeof(*f)));
				soap_default_xsd__anyType(soap, f);
				SM_TRY_NOT_OK(sm_gsoap_buffer_get_alloc_soap(staging_fault, f, xsd__anyType, NULL, qcgcomp_namespaces, soap));

				response->__size = 1;
				response->__any = f;
			}

			break;
		}
	}

	if (extra_info){
		struct soap_dom_element *job_monitoring_dom = NULL;

		sm_info("Returning additional job information: %s", extra_info);

		SM_TRY_NOT(job_monitoring_dom = transport->transport_malloc(transport, sizeof(*job_monitoring_dom)));

		if (sm_gsoap_buffer_get_alloc_soap(extra_info, job_monitoring_dom, xsd__anyType, NULL, qcgcomp_namespaces, soap) != SM_OK)
			THROW_INTERNAL_FAULT("Failed to parse job monitoring element");

		st->__size = 1;
		st->__any = job_monitoring_dom;

	}

	SM_FREE(extra_info);
	SM_LIST_FREE_USER(jobs, smc_job_free);
	return transport->exit(transport, NULL);

fault:
	SM_FREE(extra_info);
	SM_LIST_FREE_USER(jobs, smc_job_free);
	return transport->fault(transport, fault);
}

/*
 * Possible faults: UnknownActivityIdentifierFault, NotAuthorizedFault,
 * CantApplyOperationToCurrentStateFault
 */
int
__bes_factory__TerminateActivities(
	struct soap *soap,
	struct bes_factory__TerminateActivitiesType *bes_factory__TerminateActivities,
	struct bes_factory__TerminateActivitiesResponseType *bes_factory__TerminateActivitiesResponse
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	sm_list_t jobs = NULL;
	smc_job_t job = NULL;
	sm_ret_code_t ret;
	sm_status_ret_code_t status;
	uid_t uid;
	sm_value_t *result = NULL;
	sm_value_t *params = NULL;
	struct bes_factory__TerminateActivityResponseType *response;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "TerminateActivities", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "TerminateActivities", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	if (bes_factory__TerminateActivities->__sizeActivityIdentifier != 1)
		THROW_APP_FAULT(bes_factory_UnsupportedFeatureFault, "Processing more than one EPR not supported");

	SM_TRY_NOT_OK(status = get_job_from_epr(transport, fault, bes_factory__TerminateActivities->ActivityIdentifier, true, &jobs));
	SM_ASSERT(sm_list_count(jobs) <= 1, "Parameter sweep not implemented in %s", __FUNCTION__);

	job = jobs->data;

	SM_TRY_FAULT(status = smc_user_authorize(SMC_USER_AUTH_TERMINATE, job->get_subject(job), transport->atc_token->get_last_oneline(transport->atc_token)));
	if (status == SM_NEG)
		THROW_APP_FAULT(bes_factory_NotAuthorizedFault, "You are not authorized to terminate this activity");

	if (job->get_state(job) == SMC_STAGE_IN || job->get_state(job) == SMC_STAGE_OUT)
		THROW_APP_FAULT2INT(bes_factory_CantApplyOperationToCurrentStateFault, "Terminating an activity in stage in/out phase not supported", job->get_state(job));

	if (job->get_state(job) == SMC_FAILED || job->get_state(job) == SMC_FINISHED || job->get_state(job) == SMC_CANCELLED)
		THROW_APP_FAULT2INT(bes_factory_CantApplyOperationToCurrentStateFault, "Activity already in final state", job->get_state(job));

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, job->get_drms_id(job)));

	ret = smc_user_get_uid(job->get_mapped_user(job), &uid);
	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));


	if (smc_icomm_request_TerminateActivities(uid, params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if (SM_VALUE_RET_CODE_C(result[0]) == bes_factory_UnknownActivityIdentifierFault) {
		/* Race condition. The job is not known in DRMS but it still
		 * exists in the active jobs table -- the waiting thread hasn't
		 * updated the job state yet. */

		/* XXX: We can't tell now whether it is Finished or Failed
		 * state. */
		THROW_APP_FAULT2INT(bes_factory_CantApplyOperationToCurrentStateFault, "Activity is already in final, but not yet known, state: Finished or Failed.", SMC_FINISHED);
	}

	if (SM_VALUE_RET_CODE_C(result[0]))
		THROW_APP_FAULT_FROM_VAL(result);

	sm_info("Terminated activity: id = %s (drms_id = %s)", job->get_id(job), SM_SAFE_STR(job->get_drms_id(job)));

	SM_TRY_NOT(bes_factory__TerminateActivitiesResponse->Response = transport->transport_malloc(transport, sizeof(*response)));
	response = bes_factory__TerminateActivitiesResponse->Response;
	soap_default_bes_factory__TerminateActivityResponseType(soap, response);

	bes_factory__TerminateActivitiesResponse->__sizeResponse = 1;

	/* drmaa_control() is not synchronous so we cannot guarantee that the
	 * job is in fact terminated right now -- it will be eventually. */
	response->Terminated = 0;

	if (smc_create_activity_epr(transport, &response->ActivityIdentifier, job->get_id(job)) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to create activity EPR");

	sm_value_free(&params);
	sm_value_free(&result);
	SM_LIST_FREE_USER(jobs, smc_job_free);
	return transport->exit(transport, NULL);

fault:
	sm_value_free(&params);
	sm_value_free(&result);
	SM_LIST_FREE_USER(jobs, smc_job_free);
	return transport->fault(transport, fault);
}

/*
 * Possible faults: UnknownActivityIdentifierFault, NotAuthorizedFault
 */
int
__bes_factory__GetActivityDocuments(
	struct soap *soap,
	struct bes_factory__GetActivityDocumentsType *bes_factory__GetActivityDocuments,
	struct bes_factory__GetActivityDocumentsResponseType *bes_factory__GetActivityDocumentsResponse
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	struct bes_factory__GetActivityDocumentResponseType *response;
	sm_list_t jobs = NULL;
	smc_job_t job = NULL;
	sm_status_ret_code_t status;
	bool parameter_sweep;
	const char *jsdl;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetActivityDocuments", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_FALSE, "GetActivityDocuments", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	if (bes_factory__GetActivityDocuments->__sizeActivityIdentifier != 1)
		THROW_APP_FAULT(bes_factory_UnsupportedFeatureFault, "Processing more than one EPR not supported");

	SM_TRY_NOT_OK(get_job_from_epr(transport, fault, bes_factory__GetActivityDocuments->ActivityIdentifier, true, &jobs));
	parameter_sweep = (sm_list_count(jobs) > 1);
	job = jobs->data;

	SM_TRY_FAULT(status = smc_user_authorize(SMC_USER_AUTH_GET_DOCUMENT, job->get_subject(job), transport->atc_token->get_last_oneline(transport->atc_token)));
	if (status == SM_NEG)
		THROW_APP_FAULT(bes_factory_NotAuthorizedFault, "You are not authorized to get document of this activity");

	SM_TRY_NOT((bes_factory__GetActivityDocumentsResponse->Response = transport->transport_malloc(transport, sizeof(*response))));
	response = bes_factory__GetActivityDocumentsResponse->Response;
	soap_default_bes_factory__GetActivityDocumentResponseType(soap, response);
	bes_factory__GetActivityDocumentsResponse->__sizeResponse = 1;

	response->ActivityIdentifier = *bes_factory__GetActivityDocuments->ActivityIdentifier;
	SM_TRY_NOT(response->JobDefinition = transport->transport_malloc(transport, sizeof(*response->JobDefinition)));

	/* In case of parameter sweep we can only return one JSDL so it only
	 * makes sense to return the input JSDL. */
	if (parameter_sweep)
		jsdl = job->get_submitted_jsdl(job);
	else
		jsdl = job->get_jsdl(job);

	if (sm_gsoap_buffer_get_alloc_soap(jsdl, response->JobDefinition,
		jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", qcgcomp_namespaces, soap) != SM_OK) {

		THROW_INTERNAL_FAULT("Unable to parse JSDL document from database");
	}

	/* During staging the job doesn't have DRMS id yet. */
	if (!parameter_sweep && job->get_drms_id(job))
		SM_TRY_NOT(response->qcg_comp_factory__LocalActivityId = transport->transport_strdup(transport, job->get_drms_id(job)));

	SM_LIST_FREE_USER(jobs, smc_job_free);
	return transport->exit(transport, NULL);

fault:
	SM_LIST_FREE_USER(jobs, smc_job_free);
	return transport->fault(transport, fault);
}

/*
 * Possible faults: none
 */
int
__bes_factory__GetFactoryAttributesDocument(
	struct soap *soap,
	struct bes_factory__GetFactoryAttributesDocumentType *bes_factory__GetFactoryAttributesDocument,
	struct bes_factory__GetFactoryAttributesDocumentResponseType *bes_factory__GetFactoryAttributesDocumentResponse
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	_bes_factory__FactoryResourceAttributesDocument *doc = NULL;
	sm_list_t jobs = NULL;
	sm_list_t factory_attr_list = NULL;
	sm_list_t contained_resources_list = NULL;
	sm_list_t subjects_list = NULL;

	sm_ret_code_t ret;
	bool return_activites;
	bool return_resources;

	extern SM_BOOL smc_accepting_new_activities;
	extern bool smc_return_list_of_activities;
	extern bool smc_return_list_of_contained_resources;
	extern const char *smc_attributes_common_name;
	extern const char *smc_attributes_long_description;
	extern const char *smc_attributes_local_resource_manager_type;
	extern char *smc_attributes_contained_resources;
	extern char *smc_factory_attributes;
	extern submission_module_t smc_sumbission_module;
	extern sm_atz_module_t smc_mapfile_module;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetFactoryAttributesDocument", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_FALSE, "GetFactoryAttributesDocument", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	SM_TRY_NOT(bes_factory__GetFactoryAttributesDocumentResponse->bes_factory__FactoryResourceAttributesDocument = transport->transport_malloc(transport, sizeof(*doc)));
	doc = bes_factory__GetFactoryAttributesDocumentResponse->bes_factory__FactoryResourceAttributesDocument;
	soap_default__bes_factory__FactoryResourceAttributesDocument(NULL, doc);

	doc->IsAcceptingNewActivities = smc_accepting_new_activities;
	doc->CommonName = (char *) smc_attributes_common_name;
	doc->LongDescription = (char *) smc_attributes_long_description;

	if (bes_factory__GetFactoryAttributesDocument->hpcp_bp__BasicFilter) {
		return_activites = bes_factory__GetFactoryAttributesDocument->hpcp_bp__BasicFilter->ActivityReferences;
		return_resources = bes_factory__GetFactoryAttributesDocument->hpcp_bp__BasicFilter->ContainedResources;
	} else {
		return_activites = smc_return_list_of_activities;
		return_resources = smc_return_list_of_contained_resources;
	}

	if (return_activites) {
		sm_list_t l;
		size_t count;
		size_t i;

		ret = smc_job_get_active_readonly(&jobs);

		if (ret.code)
			THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

		count = sm_list_count(jobs);
		doc->TotalNumberOfActivities = count;
		doc->__sizeActivityReference = count;

		if (count)
			SM_TRY_NOT(doc->ActivityReference = transport->transport_malloc(transport, count * sizeof(struct wsa5__EndpointReferenceType)));

		for (l = jobs, i = 0; l; l = l->next, i++) {
			smc_job_t job = l->data;

			SM_TRY_NOT_OK(smc_create_activity_epr(transport, &doc->ActivityReference[i], job->get_id(job)));
		}
	} else
		doc->TotalNumberOfActivities = smc_job_count();

	if (!smc_attributes_local_resource_manager_type) { /*TODO race condition!!!*/
		sm_value_t *value = smc_sumbission_module->get_drms((sm_module_t)smc_sumbission_module);
		if (SM_VALUE_RET_CODE_OK(value[0])) {
			int drms_type = value[1]->vint32;
			char *drms_namespace = value[2]->vstring;
			sm_debug("drms type = %d (ns=%s)" , drms_type, drms_namespace);

			smc_attributes_local_resource_manager_type = sm_strdup(drms_namespace);
		} else {
			sm_error("get_drms() failed");
		}
		sm_value_free(&value);
	}

	SM_TRY_NOT_OK(sm_signal_emit("smc_contained_resources", smc_attributes_local_resource_manager_type, &contained_resources_list));

	SM_ASSERT((sm_list_count(contained_resources_list) <= 1), "Only one handler can be connected to the smc_smc_contained_resources signal");

	if (sm_list_count(contained_resources_list) == 1) {
		SM_FREE(smc_attributes_contained_resources); /*TODO race condition !!!*/
		smc_attributes_contained_resources = sm_strdup(((sm_value_t *)(contained_resources_list->data))[0]->vstring);
	}

	if (smc_attributes_contained_resources && strlen(smc_attributes_contained_resources)) {
		struct soap_dom_element dom;
		struct soap_dom_element *ptr;
		size_t count;
		size_t i;

		memset(&dom, 0, sizeof(dom));

		sm_debug("smc_attributes_contained_resources = \"%s\"", smc_attributes_contained_resources);

		if (sm_gsoap_buffer_get_alloc_soap(smc_attributes_contained_resources, &dom, xsd__anyType, NULL, qcgcomp_namespaces, soap) != SM_OK)
			THROW_INTERNAL_FAULT("Failed to parse ContainedResources configuration element");

		for (count = 0, ptr = dom.elts; ptr; count++)
			ptr = ptr->next;

		doc->TotalNumberOfContainedResources = count;

		if (return_resources) {
			doc->__sizeContainedResource = count;
			SM_TRY_NOT(doc->ContainedResource = transport->transport_malloc(transport, count * sizeof(struct soap_dom_element)));

			for (ptr = dom.elts, i = 0; ptr; ptr = ptr->next, i++) {
				struct soap_dom_attribute *attr;
				struct soap_dom_element *p;

				memcpy(&doc->ContainedResource[i], ptr, sizeof(struct soap_dom_element));
				doc->ContainedResource[i].nstr = SMC_NAMESPACE_BES_FACTORY;

				for (p = doc->ContainedResource[i].elts; p; p = p->next) {
					if (p->nstr && !strcmp(p->nstr, SMC_NAMESPACE_BES_FACTORY))
						continue;
					if (p->nstr && !strcmp(p->nstr, SMC_NAMESPACE_JSDL))
						continue;
					if (p->nstr && !strcmp(p->nstr, SMC_NAMESPACE_MONITORING))
						continue;

					sm_error("All the elements in ContainedResource must be from bes-factory, qcg-monitoring or jsdl namespace.");
					sm_error("Attach bes-factory and jsdl namespace definitions to ContainedResources or elements below.");
					THROW_INTERNAL_FAULT("Configuration error");
				}

				SM_TRY_NOT(attr = transport->transport_malloc(transport, sizeof(struct soap_dom_attribute)));
				attr->name = "xsi:type";
				attr->data = "bes-factory:BasicResourceAttributesDocumentType";
				attr->next = doc->ContainedResource[i].atts;
				doc->ContainedResource[i].atts = attr;
			}
		} else {
			doc->__sizeContainedResource = 0;
		}
	}

	SM_TRY_NOT_OK(sm_signal_emit("smc_factory_attributes", smc_attributes_local_resource_manager_type, &factory_attr_list));

	if (factory_attr_list) {
		doc->__size = sm_list_count(factory_attr_list);
		sm_debug("smc_factory_attributes signal handled %d times", doc->__size);
	} else
		doc->__size = 0;

	{
		struct soap_dom_element *factory_attributes;
		struct soap_dom_element *ptr;
		size_t count;
		size_t i;

		SM_TRY_NOT(factory_attributes = transport->transport_malloc(transport, sizeof(*factory_attributes)));

		if (sm_gsoap_buffer_get_alloc_soap(smc_factory_attributes, factory_attributes, xsd__anyType, NULL, qcgcomp_namespaces, soap) != SM_OK)
			THROW_INTERNAL_FAULT("Failed to parse FactoryAttributes configuration element");

		for (count = 0, ptr = factory_attributes->elts; ptr; ptr = ptr->next) {
			if (ptr->nstr && strcmp(ptr->nstr, SMC_NAMESPACE_CONFIG))
				count++;
		}

		doc->__size += count;

		SM_TRY_NOT(doc->__any = transport->transport_malloc(transport, sizeof(*doc->__any) * doc->__size));

		for (i = 0, ptr = factory_attributes->elts; ptr; ptr = ptr->next) {

			/* Add everything that is not from our namespace. */
			if (ptr->nstr && strcmp(ptr->nstr, SMC_NAMESPACE_CONFIG)) {
				doc->__any[i] = *ptr;
				i++;
			}
		}
	}

	if (!doc->__any)
		SM_TRY_NOT(doc->__any = transport->transport_malloc(transport, sizeof(*doc->__any) * doc->__size));

	{
		sm_list_t iter;
		int i = doc->__size - 1;

		for (iter = factory_attr_list; iter; iter = iter->next, i--) {
			sm_value_t *value = iter->data;

			/* XXX: We assume the signal handler returns only one
			 * XML element. */
			if (sm_gsoap_buffer_get_alloc_soap(value[0]->vstring, &(doc->__any[i]), xsd__anyType, NULL, qcgcomp_namespaces, soap) != SM_OK)
				THROW_INTERNAL_FAULT("Failed to parse FactoryAttributes retrieved from signal handler");
		}
	}

	SM_TRY_NOT(doc->NamingProfile = transport->transport_malloc(transport, sizeof(char **)));
	doc->NamingProfile[0] = SMC_NAMESPACE_NAMING_PROFILE;
	doc->__sizeNamingProfile = 1;

	{
		sm_value_t *extensions = NULL;
		size_t extensions_count;
		size_t i;

		extensions = smc_extensions_get();

		if ((extensions_count = sm_value_count(extensions)) > 0) {
			SM_TRY_NOT(doc->BESExtension = transport->transport_malloc(transport, extensions_count * sizeof(char **)));

			for (i = 0; i < extensions_count; i++)
				SM_TRY_NOT(doc->BESExtension[i] = transport->transport_strdup(transport, extensions[i]->vstring));

			doc->__sizeBESExtension = extensions_count;
		}
	}

	doc->LocalResourceManagerType = (char *) smc_attributes_local_resource_manager_type;

	if (smc_mapfile_module) {
		sm_list_t list_iter = NULL;
		size_t subjects_count = 0;
		int i = 0;
		sm_debug("mapfile module available");

		smc_mapfile_module->list_authorized_subjects((sm_module_t)smc_mapfile_module, &subjects_list);

		if ((subjects_count = sm_list_count(subjects_list)) != 0) {
			SM_TRY_NOT(doc->qcg_comp_factory__AuthorizedUsers = transport->transport_malloc(transport, sizeof(struct qcg_comp_factory__AuthorizedUsers_USCOREType)));
			SM_TRY_NOT(doc->qcg_comp_factory__AuthorizedUsers->Subject = transport->transport_malloc(transport, sizeof(char *) * subjects_count));
			doc->qcg_comp_factory__AuthorizedUsers->__sizeSubject = subjects_count;
		}

		for (list_iter = subjects_list; list_iter; list_iter = list_iter->next) {
			sm_debug("subject = %s", (char*)list_iter->data);
			SM_TRY_NOT(doc->qcg_comp_factory__AuthorizedUsers->Subject[i++] = transport->transport_strdup(transport, (const char*)list_iter->data));
		}

	}

	SM_LIST_FREE_USER(jobs, smc_job_free);
	SM_LIST_FREE_USER(factory_attr_list, sm_value_free);
	SM_LIST_FREE_USER(contained_resources_list, sm_value_free);
	sm_list_free(subjects_list, SM_TRUE);
	return transport->exit(transport, NULL);

fault:
	SM_LIST_FREE_USER(jobs, smc_job_free);
	SM_LIST_FREE_USER(factory_attr_list, sm_value_free);
	SM_LIST_FREE_USER(contained_resources_list, sm_value_free);

	return transport->fault(transport, fault);
}
