/* $Id: faults.c 252 2010-08-12 13:58:43Z smaug $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/transport.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>

#include <faults.h>
#include <job.h>
#include <stub/qcgcompH.h>
#include <stub/qcgcompStub.h>

__SM_ID("$Id: faults.c 252 2010-08-12 13:58:43Z smaug $");

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "faults"
	}
};

static const char *
strfault(int value)
{
	switch (value) {
		case 0: return "Success";
		case 1: return "InternalFault";

		case bes_factory_NotAuthorizedFault: return "NotAuthorizedFault";
		case bes_factory_NotAcceptingNewActivitiesFault: return "NotAcceptingNewActivitiesFault";
		case bes_factory_UnsupportedFeatureFault: return "UnsupportedFeatureFault";
		case bes_factory_CantApplyOperationToCurrentStateFault: return "CantApplyOperationToCurrentStateFault";
		case bes_factory_UnknownActivityIdentifierFault: return "UnknownActivityIdentifierFault";
		case bes_factory_InvalidRequestMessageFault: return "InvalidRequestMessageFault";

		case bes_management_NotAuthorizedFault: return "NotAuthorizedFault";

		case qcg_comp_ares_NotAuthorizedFault: return "NotAuthorizedFault";
		case qcg_comp_ares_UnsupportedFeatureFault: return "UnsupportedFeatureFault";
		case qcg_comp_ares_InvalidRequestMessageFault: return "InvalidRequestMessageFault";
		case qcg_comp_ares_UnknownReservationIdentifierFault: return "UnknownReservationIdentifierFault";
		case qcg_comp_ares_NoSlotsAvailableFault: return "NoSlotsAvailableFault";
		case qcg_comp_ares_NoMatchingResourcesFault: return "NoMatchingResourcesFault";

		case qcg_comp_staging_DeleteOnTerminationFault: return "DeleteOnTerminationFault";
		case qcg_comp_staging_FileNotFoundFault: return "FileNotFoundFault";
		case qcg_comp_staging_StagingFault: return "StagingFault";
		case qcg_comp_staging_NotAuthorizedFault: return "NotAuthorizedFault";

		case RSyncIOError: return "RSyncIOError";
		case RSyncInvalidPathError: return "RSyncInvalidPathError";
		case RSyncNoMemoryError: return "RSyncNoMemoryError";
		case RSyncAttachmentNotFound: return "RSyncAttachmentNotFound";

		default:
			SM_ASSERT(0, "Unknown error code (%d) in %s", value, __FUNCTION__);
	}
}

const char *
smc_get_fault_msg(sm_transport_t transport sm_unused, sm_transport_fault_t fault)
{
	SM_DEBUG_ENTER();
	assert(fault->type == SM_APPLICATION_FAULT);
	return strfault(fault->application_fault.code);
}

sm_transport_fault_class_t
smc_get_fault_class(sm_transport_t transport sm_unused, sm_transport_fault_t fault)
{
	SM_DEBUG_ENTER();

	assert(fault->type == SM_APPLICATION_FAULT);

	switch (fault->application_fault.code) {
		case 1: /* all internal faults */
		case bes_factory_NotAcceptingNewActivitiesFault:
		case bes_factory_CantApplyOperationToCurrentStateFault:
		case qcg_comp_staging_DeleteOnTerminationFault:
		case qcg_comp_staging_StagingFault:
		case qcg_comp_ares_NoSlotsAvailableFault:
			return SM_RECEIVER_FAULT;
	}

	/* bad request, no use resending it */
	return SM_SENDER_FAULT;
}

void *
smc_create_fault_detail(sm_transport_t super_transport, sm_transport_fault_t fault)
{
	struct ecm_gsoap_transport_s *transport = (struct ecm_gsoap_transport_s *)super_transport;
	struct SOAP_ENV__Detail *detail = NULL;
	sm_value_t *data = NULL;

	assert(fault->type == SM_APPLICATION_FAULT);

	SM_DEBUG_ENTER();

	data = fault->application_fault.data;
	if (data)
		SM_VALUE_DEBUG(data);

	SM_TRY_NOT(detail = super_transport->transport_malloc(super_transport, sizeof(*detail)));

	soap_default_SOAP_ENV__Detail(transport->soap, detail);

	sm_debug("application_fault_code = %d", fault->application_fault.code);

	switch (fault->application_fault.code) {
		/* all internal faults */
		case 1:
		{
			if (sm_value_count(data) > 0) {
				assert(SM_VALUE_IS_STRING(data[0]));
				SM_TRY_NOT(fault->application_fault.msg = super_transport->transport_strdup(super_transport, data[0]->vstring));
			}
			detail = NULL;
			break;
		}
		case bes_management_NotAuthorizedFault:
		case bes_factory_NotAuthorizedFault:
		{
			struct bes_factory__NotAuthorizedFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_bes_factory__NotAuthorizedFaultType(transport->soap, d);
			detail->bes_factory__NotAuthorizedFault = d;
			break;
		}
		case bes_factory_NotAcceptingNewActivitiesFault:
		{
			struct bes_factory__NotAcceptingNewActivitiesFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_bes_factory__NotAcceptingNewActivitiesFaultType(transport->soap, d);
			detail->bes_factory__NotAcceptingNewActivitiesFault = d;
			break;
		}
		case bes_factory_UnsupportedFeatureFault:
		{
			struct bes_factory__UnsupportedFeatureFaultType *d;
			size_t count, i;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_bes_factory__UnsupportedFeatureFaultType(transport->soap, d);

			count = sm_value_count(data);
			if (count) {
				SM_TRY_NOT(d->Feature = super_transport->transport_malloc(super_transport, count * sizeof(char *)));

				for (i = 0; i < count; i++) {
					assert(SM_VALUE_IS_STRING(data[i]));

					SM_TRY_NOT(d->Feature[i] = super_transport->transport_strdup(super_transport, data[i]->vstring));
				}

				d->__sizeFeature = count;
			}

			detail->bes_factory__UnsupportedFeatureFault = d;
			break;
		}
		case bes_factory_CantApplyOperationToCurrentStateFault:
		{
			struct bes_factory__CantApplyOperationToCurrentStateFaultType *d;
			struct bes_factory__ActivityStatusType *state;
			union _bes_factory__union_ActivityStatusType state_union;

			if (!data) {
				sm_error("CantApplyOperationToCurrentStateFault with no detail");
				goto fault;
			}

			assert(sm_value_count(data) == 2);
			assert(SM_VALUE_IS_STRING(data[0]));
			assert(SM_VALUE_IS_INT32(data[1]));

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_bes_factory__CantApplyOperationToCurrentStateFaultType(transport->soap, d);

			SM_TRY_NOT(state = super_transport->transport_malloc(super_transport, sizeof(*state)));
			soap_default_bes_factory__ActivityStatusType(transport->soap, state);
			memset(&state_union, 0, sizeof(state_union));

			d->Message = data[0]->vstring;

			switch (data[1]->vint32) {
				case SMC_CANCELLED:
					state->state = bes_factory__ActivityStateEnumeration__Cancelled;
					break;
				case SMC_FINISHED:
					state->state = bes_factory__ActivityStateEnumeration__Finished;
					break;
				case SMC_FAILED:
					state->state = bes_factory__ActivityStateEnumeration__Failed;
					break;
				case SMC_STAGE_IN:
					state->state = bes_factory__ActivityStateEnumeration__Pending;
					state->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Stage_in;
					if (!(state_union.qcg_comp_factory__Stage_in = super_transport->transport_malloc(super_transport, sizeof(struct _qcg_comp_factory__Stage_in))))
						goto fault;
					state->union_ActivityStatusType = state_union;
					break;
				case SMC_HELD:
					state->state = bes_factory__ActivityStateEnumeration__Pending;
					state->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Held;
					if (!(state_union.qcg_comp_factory__Held = super_transport->transport_malloc(super_transport, sizeof(struct _qcg_comp_factory__Held))))
						goto fault;
					state->union_ActivityStatusType = state_union;
					break;
				case SMC_QUEUED:
					state->state = bes_factory__ActivityStateEnumeration__Pending;
					state->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Queued;
					if (!(state_union.qcg_comp_factory__Queued = super_transport->transport_malloc(super_transport, sizeof(struct _qcg_comp_factory__Queued))))
						goto fault;
					state->union_ActivityStatusType = state_union;
					break;
				case SMC_EXECUTING:
					state->state = bes_factory__ActivityStateEnumeration__Running;
					state->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Executing;
					if (!(state_union.qcg_comp_factory__Executing = super_transport->transport_malloc(super_transport, sizeof(struct _qcg_comp_factory__Executing))))
						goto fault;
					state->union_ActivityStatusType = state_union;
					break;
				case SMC_SUSPENDED:
					state->state = bes_factory__ActivityStateEnumeration__Running;
					state->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Suspended;
					if (!(state_union.qcg_comp_factory__Suspended = super_transport->transport_malloc(super_transport, sizeof(struct _qcg_comp_factory__Suspended))))
						goto fault;
					state->union_ActivityStatusType = state_union;
					break;
				case SMC_STAGE_OUT:
					state->state = bes_factory__ActivityStateEnumeration__Running;
					state->__union_ActivityStatusType = SOAP_UNION__bes_factory__union_ActivityStatusType_qcg_comp_factory__Stage_out;
					if (!(state_union.qcg_comp_factory__Stage_out = super_transport->transport_malloc(super_transport, sizeof(struct _qcg_comp_factory__Stage_out))))
						goto fault;
					state->union_ActivityStatusType = state_union;
					break;
				default:
					sm_error("Wrong detail for CantApplyOperationToCurrentStateFault");
					goto fault;
			}

			d->ActivityStatus = state;

			detail->bes_factory__CantApplyOperationToCurrentStateFault = d;
			break;
		}
		case bes_factory_UnknownActivityIdentifierFault:
		{
			struct bes_factory__UnknownActivityIdentifierFaultType *d;

			if (!data) {
				sm_error("UnknownActivityIdentifierFault with no detail");
				goto fault;
			}

			assert(sm_value_count(data) == 1);
			assert(SM_VALUE_IS_STRING(data[0]));

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_bes_factory__UnknownActivityIdentifierFaultType(transport->soap, d);

			SM_TRY_NOT(d->Message = super_transport->transport_strdup(super_transport, data[0]->vstring));

			detail->bes_factory__UnknownActivityIdentifierFault = d;
			break;
		}
		case bes_factory_InvalidRequestMessageFault:
		{
			struct bes_factory__InvalidRequestMessageFaultType *d;
			size_t count, i;

			count = sm_value_count(data);
			if (!count) {
				sm_error("InvalidRequestMessageFault with no detail");
				goto fault;
			}

			assert(SM_VALUE_IS_STRING(data[0]));

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_bes_factory__InvalidRequestMessageFaultType(transport->soap, d);

			SM_TRY_NOT(d->Message = super_transport->transport_strdup(super_transport, data[0]->vstring));

			if (count > 1) {
				count--;

				SM_TRY_NOT(d->InvalidElement = super_transport->transport_malloc(super_transport, count * sizeof(char *)));

				for (i = 0; i < count; i++) {
					assert(SM_VALUE_IS_STRING(data[i + 1]));

					SM_TRY_NOT(d->InvalidElement[i] = super_transport->transport_strdup(super_transport, data[i + 1]->vstring));
				}

				d->__sizeInvalidElement = count;
			}

			detail->bes_factory__InvalidRequestMessageFault = d;
			break;
		}
		case qcg_comp_ares_NotAuthorizedFault:
		{
			struct qcg_comp_ares__NotAuthorizedFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_qcg_comp_ares__NotAuthorizedFaultType(transport->soap, d);

			detail->qcg_comp_ares__NotAuthorizedFault = d;

			assert(SM_VALUE_IS_STRING(data[0]));
			SM_TRY_NOT(fault->application_fault.msg = super_transport->transport_strdup(super_transport, data[0]->vstring));

			break;
		}
		case qcg_comp_ares_NoSlotsAvailableFault:
		{
			struct qcg_comp_ares__NoSlotsAvailableFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_qcg_comp_ares__NoSlotsAvailableFaultType(transport->soap, d);

			detail->qcg_comp_ares__NoSlotsAvailableFault = d;

			assert(SM_VALUE_IS_STRING(data[0]));
			SM_TRY_NOT(fault->application_fault.msg = super_transport->transport_strdup(super_transport, data[0]->vstring));

			break;
		}
		case qcg_comp_ares_NoMatchingResourcesFault:
		{
			struct qcg_comp_ares__NoMatchingResourcesFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_qcg_comp_ares__NoMatchingResourcesFaultType(transport->soap, d);

			detail->qcg_comp_ares__NoMatchingResourcesFault = d;

			assert(SM_VALUE_IS_STRING(data[0]));
			SM_TRY_NOT(fault->application_fault.msg = super_transport->transport_strdup(super_transport, data[0]->vstring));

			break;
		}
		case qcg_comp_ares_UnknownReservationIdentifierFault:
		{
			struct qcg_comp_ares__UnknownReservationIdentifierFaultType *d;

			if (!data) {
				sm_error("UnknownReservationIdentifierFault with no detail");
				goto fault;
			}

			assert(sm_value_count(data) == 1);
			assert(SM_VALUE_IS_STRING(data[0]));

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_qcg_comp_ares__UnknownReservationIdentifierFaultType(transport->soap, d);

			SM_TRY_NOT(d->Message = super_transport->transport_strdup(super_transport, data[0]->vstring));

			detail->qcg_comp_ares__UnknownReservationIdentifierFault = d;
			break;
		}
		case qcg_comp_ares_UnsupportedFeatureFault:
		{
			struct qcg_comp_ares__UnsupportedFeatureFaultType *d;
			size_t count, i;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_qcg_comp_ares__UnsupportedFeatureFaultType(transport->soap, d);

			count = sm_value_count(data);
			if (count) {
				SM_TRY_NOT(d->Feature = super_transport->transport_malloc(super_transport, count * sizeof(char *)));

				for (i = 0; i < count; i++) {
					assert(SM_VALUE_IS_STRING(data[i]));

					SM_TRY_NOT(d->Feature[i] = super_transport->transport_strdup(super_transport, data[i]->vstring));
				}

				d->__sizeFeature = count;
			}

			detail->qcg_comp_ares__UnsupportedFeatureFault = d;
			break;
		}
		case qcg_comp_ares_InvalidRequestMessageFault:
		{
			struct qcg_comp_ares__InvalidRequestMessageFaultType *d;
			size_t count, i;

			count = sm_value_count(data);
			if (!count) {
				sm_error("InvalidRequestMessageFault with no detail");
				goto fault;
			}

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_qcg_comp_ares__InvalidRequestMessageFaultType(transport->soap, d);

			assert(SM_VALUE_IS_STRING(data[0]));

			SM_TRY_NOT(d->Message = super_transport->transport_strdup(super_transport, data[0]->vstring));

			if (count > 1) {
				count--;

				SM_TRY_NOT(d->InvalidElement = super_transport->transport_malloc(super_transport, count * sizeof(char *)));

				for (i = 0; i < count; i++) {
					assert(SM_VALUE_IS_STRING(data[i + 1]));

					SM_TRY_NOT(d->InvalidElement[i] = super_transport->transport_strdup(super_transport, data[i + 1]->vstring));
				}

				d->__sizeInvalidElement = count;
			}

			detail->qcg_comp_ares__InvalidRequestMessageFault = d;
			break;
		}
		case qcg_comp_staging_DeleteOnTerminationFault:
		{
			size_t count = sm_value_count(data);
			if (!count) {
				sm_error("DeleteOnTerminationFault with no detail");
				goto fault;
			}
			assert(SM_VALUE_IS_STRING(data[0]));

			SM_TRY_NOT(detail->qcg_comp_staging__DeleteOnTerminationFault = super_transport->transport_malloc(super_transport, sizeof(*detail->qcg_comp_staging__DeleteOnTerminationFault)));
			SM_TRY_NOT(detail->qcg_comp_staging__DeleteOnTerminationFault->File = super_transport->transport_strdup(super_transport, data[0]->vstring));
			break;
		}
		case qcg_comp_staging_FileNotFoundFault:
		{
			size_t count = sm_value_count(data);
			if (!count) {
				sm_error("FileNotFoundFault with no detail");
				goto fault;
			}
			assert(SM_VALUE_IS_STRING(data[0]));
			SM_TRY_NOT(detail->qcg_comp_staging__FileNotFoundFault = super_transport->transport_malloc(super_transport, sizeof(*detail->qcg_comp_staging__FileNotFoundFault)));
			SM_TRY_NOT(detail->qcg_comp_staging__FileNotFoundFault->__item = super_transport->transport_strdup(super_transport, data[0]->vstring));
			break;
		}
		case qcg_comp_staging_StagingFault:
		{
			size_t count = sm_value_count(data);
			if (!count) {
				sm_error("StagingFault with no detail");
				goto fault;
			}
			assert(SM_VALUE_IS_STRING(data[0]));

			SM_TRY_NOT(detail->qcg_comp_staging__StagingFault = super_transport->transport_malloc(super_transport, sizeof(*detail->qcg_comp_staging__StagingFault)));
			SM_TRY_NOT(detail->qcg_comp_staging__StagingFault->__item = super_transport->transport_strdup(super_transport, data[0]->vstring));
			break;
		}
		case qcg_comp_staging_NotAuthorizedFault:
		{
			struct qcg_comp_staging__NotAuthorizedFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));
			soap_default_qcg_comp_staging__NotAuthorizedFaultType(transport->soap, d);

			detail->qcg_comp_staging__NotAuthorizedFault = d;
			break;
		}
		case RSyncIOError:
		{
			struct qcg_comp_rsync__IOFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));

			detail->qcg_comp_rsync__IOFault = d;
			break;
		}
		case RSyncInvalidPathError:
		{
			struct qcg_comp_rsync__InvalidPathFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));

			detail->qcg_comp_rsync__InvalidPathFault = d;
			break;
		}
		case RSyncNoMemoryError:
		case RSyncAttachmentNotFound:
		{
			struct qcg_comp_rsync__InternalFaultType *d;

			SM_TRY_NOT(d = super_transport->transport_malloc(super_transport, sizeof(*d)));

			detail->qcg_comp_rsync__InternalFault = d;
			break;
		}
		default:
			SM_ASSERT(0, "Unknown error code (%d) in %s", fault->application_fault.code, __FUNCTION__);
	}

	return detail;

fault:
	return NULL;
}

const char *
smc_fault_code2s(sm_transport_t super_transport, sm_transport_fault_t fault)
{
	char buf[1024] = "";
	char *detail_debug;
	sm_value_t *detail;

	SM_DEBUG_ENTER();

	strlcpy(buf, strfault(fault->application_fault.code), sizeof(buf));
	strlcat(buf, " ", sizeof(buf));

	detail = (sm_value_t *) fault->application_fault.data;

	if (detail) {
		if ((detail_debug = sm_value_debug(detail)))
			strlcat(buf, detail_debug, sizeof(buf));

		SM_FREE(detail_debug);
	}

	return super_transport->transport_strdup(super_transport, buf);
}

void
smc_free_fault_data(sm_transport_t transport sm_unused, void **data)
{
	sm_value_t *value;

	SM_MANDATORY_P(data);

	SM_DEBUG_ENTER();

	if (*data) {
		value = *data;
		sm_value_free(&value);
		*data = NULL;
	}
}
