/* $Id: job.c 220 2010-08-06 10:14:15Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <time.h>

#include <qcg-core/sm.h>
#include <qcg-core/access.h>
#include <qcg-core/logger.h>
#include <qcg-core/odbc.h>
#include <qcg-core/uuid.h>
#include <qcg-core/signal.h>
#include <qcg-core/string.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>

#include <identifiers.h>
#include <job.h>
#include <jsdl.h>
#include <rsvn.h>
#include <user.h>
#include <stub/qcgcompH.h>

#define SMC_JOB_INTERNAL_FAULT_STR	"Internal fault in job module"
#define SMC_JOB_DATABASE_FAULT_STR	"Database fault in job module"
#define SMC_JOB_NOT_FOUND_STR		"Job not found"

#ifdef TEST
#include <stub/qcgcomp.nsmap>
#else
extern struct Namespace qcgcomp_namespaces[];
#endif

__SM_ID("$Id: job.c 220 2010-08-06 10:14:15Z mamonski $");

SM_INTERNAL_GENERAL_MODULE = {
        .header = {
                .name = "job"
        }
};

typedef struct smc_concrete_job_s {
	struct smc_job_s super;

	int serial;

	char *id;
	char *drms_id;
	char *reservation_id;
	char *submitted_jsdl;
	char *jsdl;
	_jsdl__JobDefinition *jsdl_parsed;
	char *subject;
	char *mapped_user;
	smc_job_state_t state;
	char *staging_fault;
	struct tm submission_time;
	struct tm start_time;
	struct tm end_time;
	unsigned char exit_status;
	bool core_dumped;
	char *posix_signal_name;
	double cpu_usage;
	double mem_usage;
	double vmem_usage;
	char *queue;
	char *project;
	int walltime;
	char *nodes;
	sm_value_t *atc_token;

	smc_job_state_t previous_state;
	bool state_set;
	bool submission_time_set;
	bool start_time_set;
	bool end_time_set;
	bool exit_status_set;
	bool core_dumped_set;
	bool cpu_usage_set;
	bool mem_usage_set;
	bool vmem_usage_set;
	bool walltime_set;


	sm_alloc_tracked_t tracked;
	sm_ret_code_t error;

	bool in_transaction;
	bool new_entry;
	bool modified;
	bool readonly;
} *smc_concrete_job_t;

static const char *find_txt = "SELECT * FROM jobs WHERE id = ? FOR UPDATE";
static const char *find_by_drms_id_txt = "SELECT * FROM jobs WHERE drms_id = ? FOR UPDATE";

static const char *find_readonly_txt = "SELECT * FROM jobs WHERE id = ?";
static const char *find_by_drms_id_readonly_txt = "SELECT * FROM jobs WHERE drms_id = ?";

static const char *find_acc_readonly_txt = "SELECT * FROM jobs_acc WHERE id = ?";
static const char *find_acc_by_drms_id_readonly_txt = "SELECT * FROM jobs_acc WHERE drms_id = ? ORDER BY submission_time DESC LIMIT 1";

static const char *insert_txt = "INSERT INTO jobs (id, drms_id, reservation_id,"
	"submitted_jsdl, jsdl, subject, mapped_user, state, staging_fault, submission_time, start_time,"
	"end_time, exit_status, core_dumped, posix_signal_name, cpu_usage, mem_usage, vmem_usage, queue, project, walltime, nodes)"
	"VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

static const char *update_txt = "UPDATE jobs SET id = ?, drms_id = ?,"
	"reservation_id = ?, submitted_jsdl = ?, jsdl = ?, subject = ?, mapped_user = ?,"
	"state = ?, staging_fault = ?, submission_time = ?, start_time = ?, end_time = ?, exit_status = ?,"
	"core_dumped = ?, posix_signal_name = ?, cpu_usage = ?, mem_usage = ?, vmem_usage = ?, queue = ?, project = ?, walltime = ?, nodes = ? WHERE serial = ?";

static const char *delete_txt = "DELETE FROM jobs WHERE id = ?";

static const char *insert_acc_txt = "INSERT INTO jobs_acc (id, drms_id, reservation_id,"
	"submitted_jsdl, jsdl, subject, mapped_user, state, staging_fault, submission_time, start_time,"
	"end_time, exit_status, core_dumped, posix_signal_name, cpu_usage, mem_usage, vmem_usage, queue, project, walltime, nodes)"
	"VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

static const char *count_txt = "SELECT count(*) FROM jobs";
static const char *find_all_readonly_txt = "SELECT * from jobs";

static sm_odbc_handle_t odbc;

/*
 * Statements are global as we assume that odbc->begin_transaction locks the
 * ODBC handle and thus only one statement can be used at a time.
 */
static sm_odbc_stmt_t find_stmt;
static sm_odbc_stmt_t find_by_drms_id_stmt;
static sm_odbc_stmt_t find_readonly_stmt;
static sm_odbc_stmt_t find_by_drms_id_readonly_stmt;
static sm_odbc_stmt_t find_acc_readonly_stmt;
static sm_odbc_stmt_t find_acc_by_drms_id_readonly_stmt;
static sm_odbc_stmt_t insert_stmt;
static sm_odbc_stmt_t insert_stmt;
static sm_odbc_stmt_t update_stmt;
static sm_odbc_stmt_t delete_stmt;
static sm_odbc_stmt_t insert_acc_stmt;
static sm_odbc_stmt_t count_stmt;
static sm_odbc_stmt_t find_all_readonly_stmt;

static const char *smc_job_get_id(smc_concrete_job_t this);
static const char *smc_job_get_drms_id(smc_concrete_job_t this);
static const char *smc_job_get_reservation_id(smc_concrete_job_t this);
static smc_job_res_info_t smc_job_get_reservation_info(smc_concrete_job_t this);
static const char *smc_job_get_submitted_jsdl(smc_concrete_job_t this);
static const char *smc_job_get_jsdl(smc_concrete_job_t this);
static _jsdl__JobDefinition *smc_job_get_jsdl_parsed(smc_concrete_job_t this);
static const char *smc_job_get_subject(smc_concrete_job_t this);
static const char *smc_job_get_mapped_user(smc_concrete_job_t this);
static smc_job_state_t smc_job_get_state(smc_concrete_job_t this);
static const char *smc_job_get_staging_fault(smc_concrete_job_t this);
static struct tm smc_job_get_submission_time(smc_concrete_job_t this);
static struct tm smc_job_get_start_time(smc_concrete_job_t this);
static struct tm smc_job_get_end_time(smc_concrete_job_t this);
static unsigned char smc_job_get_exit_status(smc_concrete_job_t this);
static bool smc_job_has_exit_status(smc_concrete_job_t this);
static bool smc_job_get_core_dumped(smc_concrete_job_t this);
static bool smc_job_has_core_dumped(smc_concrete_job_t this);
static const char *smc_job_get_posix_signal_name(smc_concrete_job_t this);
static double smc_job_get_cpu_usage(smc_concrete_job_t this);
static double smc_job_get_mem_usage(smc_concrete_job_t this);
static double smc_job_get_vmem_usage(smc_concrete_job_t this);
static const char *smc_job_get_queue(smc_concrete_job_t this);
static const char *smc_job_get_project(smc_concrete_job_t this);
static int smc_job_get_walltime(smc_concrete_job_t this);
static const char* smc_job_get_nodes(smc_concrete_job_t this);
static sm_value_t *smc_job_get_atc_token(smc_concrete_job_t this);

static const char *smc_job_set_id(smc_concrete_job_t this, const char *value);
static const char *smc_job_set_drms_id(smc_concrete_job_t this, const char *value);
static const char *smc_job_set_submitted_jsdl(smc_concrete_job_t this, const char *value);
static const char *smc_job_set_jsdl(smc_concrete_job_t this, const char *value);
static const char *smc_job_set_jsdl_parsed(smc_concrete_job_t this, _jsdl__JobDefinition *value);
static const char *smc_job_set_subject(smc_concrete_job_t this, const char *value);
static const char *smc_job_set_mapped_user(smc_concrete_job_t this, const char *value);
static smc_job_state_t smc_job_set_state(smc_concrete_job_t this, smc_job_state_t value);
static const char *smc_job_set_staging_fault(smc_concrete_job_t this, const char *value);
static struct tm smc_job_set_start_time(smc_concrete_job_t this, struct tm value);
static unsigned char smc_job_set_exit_status(smc_concrete_job_t this, unsigned char value);
static bool smc_job_set_core_dumped(smc_concrete_job_t this, bool value);
static const char *smc_job_set_posix_signal_name(smc_concrete_job_t this, const char *value);
static double smc_job_set_cpu_usage(smc_concrete_job_t this, double value);
static double smc_job_set_mem_usage(smc_concrete_job_t this, double value);
static double smc_job_set_vmem_usage(smc_concrete_job_t this, double value);
static const char *smc_job_set_queue(smc_concrete_job_t this, const char *value);
static const char *smc_job_set_project(smc_concrete_job_t this, const char *value);
static int smc_job_set_walltime(smc_concrete_job_t this, int value);
static const char *smc_job_set_nodes(smc_concrete_job_t this, const char *value);
static sm_value_t *smc_job_set_atc_token(smc_concrete_job_t this, sm_value_t *value);

static sm_ret_code_t smc_job_save(smc_concrete_job_t this);

static smc_job_t smc_job_create_internal(bool readonly, bool new, bool transaction);
static sm_status_ret_code_t smc_job_insert_acc(smc_concrete_job_t this);
static sm_status_ret_code_t smc_job_insert(smc_concrete_job_t this);
static sm_status_ret_code_t smc_job_update(smc_concrete_job_t this);
static sm_status_ret_code_t smc_job_move_to_acc(smc_concrete_job_t this);
static sm_ret_code_t smc_job_process_jsdl(smc_concrete_job_t this);
static sm_ret_code_t smc_job_get_internal(const char *id, sm_list_t *result, bool readonly);
static sm_ret_code_t smc_job_get_by_drms_id_internal(const char *drms_id, smc_job_t *result, bool readonly);
static void smc_job_add_estimated_time(smc_concrete_job_t this);
static void smc_job_process_state_change(smc_concrete_job_t this);

static bool smc_job_is_readonly(smc_concrete_job_t this);
static struct tm timetm(void);

sm_status_ret_code_t
smc_job_setup(sm_odbc_handle_t odbc_handle, bool session_process)
{
	int ret;

	SM_MANDATORY_P(odbc_handle);

	SM_DEBUG_ENTER();

	odbc = odbc_handle;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	ret = sm_signal_register("smc_job_stage_in", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* reservation_id */,
		SM_VALUE_STRING /* submitted_jsdl */, SM_VALUE_STRING /* jsdl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */ },
		7,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);
	ret |= sm_signal_register("smc_job_queued", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* drms_id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* reservation_id */,
		SM_VALUE_STRING /* submitted_jsdl */, SM_VALUE_STRING /* jsdl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */ },
		8,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);
	ret |= sm_signal_register("smc_job_held", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* drms_id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* reservation_id */,
		SM_VALUE_STRING /* submitted_jsdl */, SM_VALUE_STRING /* jsdl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */ },
		8,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);
	ret |= sm_signal_register("smc_job_executing", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* drms_id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* reservation_id */,
		SM_VALUE_STRING /* submitted_jsdl */, SM_VALUE_STRING /* jsdl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */, SM_VALUE_INT32 /* start_time */ },
		9,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);
	ret |= sm_signal_register("smc_job_suspended", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* drms_id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* reservation_id */,
		SM_VALUE_STRING /* submitted_jsdl */, SM_VALUE_STRING /* jsdl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */, SM_VALUE_INT32 /* start_time */ },
		9,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);
	ret |= sm_signal_register("smc_job_stage_out", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* drms_id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* reservation_id */,
		SM_VALUE_STRING /* submitted_jsdl */, SM_VALUE_STRING /* jsdl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */, SM_VALUE_INT32 /* start_time */,
		SM_VALUE_INT32 /* end_time */ },
		10,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);
	ret |= sm_signal_register("smc_job_cancelled", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* drms_id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* reservation_id */,
		SM_VALUE_STRING /* submitted_jsdl */, SM_VALUE_STRING /* jsdl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_INT32 /* submission_time */, SM_VALUE_INT32 /* start_time */,
		SM_VALUE_INT32 /* end_time */ },
		10,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);
	ret |= sm_signal_register("smc_job_finished", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* 0. id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* 1. drms_id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* 2. reservation_id */,
		SM_VALUE_STRING /* 3. submitted_jsdl */, SM_VALUE_STRING /* 4. jsdl */, SM_VALUE_STRING /* 5. subject */,
		SM_VALUE_STRING /* 6. mapped_user */, SM_VALUE_INT32 /* 7. submission_time */, SM_VALUE_INT32 /* 8. start_time */,
		SM_VALUE_INT32 /* 9. end_time */, SM_VALUE_INT32 /* 10. exit_status */, SM_VALUE_DOUBLE /* 11. cpu_usage */, SM_VALUE_DOUBLE /* 12. mem_usage */, 	SM_VALUE_DOUBLE /* 13. vmem_usage */,
		SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* 14. queue */, SM_VALUE_STRING /* 15. project */, SM_VALUE_INT32 /* 16. walltime */, SM_VALUE_STRING /* 17. nodes */ },
		18,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);
	ret |= sm_signal_register("smc_job_failed", 0,
		(sm_value_type_t []) { SM_VALUE_STRING /* id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* drms_id */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* reservation_id */,
		SM_VALUE_STRING /* submitted_jsdl */, SM_VALUE_STRING /* jsdl */, SM_VALUE_STRING /* subject */,
		SM_VALUE_STRING /* mapped_user */, SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* staging_fault */, SM_VALUE_INT32 /* submission_time */,
		SM_VALUE_INT32 /* start_time */, SM_VALUE_INT32 /* end_time */, SM_VALUE_BOOL /* core_dumped */,
		SM_VALUE_STRING | SM_VALUE_MAY_BE_NONE_FLAG /* posix_signal_name */ },
		13,
		(sm_value_type_t []) { SM_VALUE_NONE }, 1);

	if (ret & SM_FAULT)
		goto fault;

	SM_TRY_NOT(find_stmt = sm_odbc_stmt_create(odbc, find_txt));
	SM_TRY_NOT(find_by_drms_id_stmt = sm_odbc_stmt_create(odbc, find_by_drms_id_txt));
	SM_TRY_NOT(find_readonly_stmt = sm_odbc_stmt_create(odbc, find_readonly_txt));
	SM_TRY_NOT(find_by_drms_id_readonly_stmt = sm_odbc_stmt_create(odbc, find_by_drms_id_readonly_txt));
	SM_TRY_NOT(find_acc_readonly_stmt = sm_odbc_stmt_create(odbc, find_acc_readonly_txt));
	SM_TRY_NOT(find_acc_by_drms_id_readonly_stmt = sm_odbc_stmt_create(odbc, find_acc_by_drms_id_readonly_txt));
	SM_TRY_NOT(insert_stmt = sm_odbc_stmt_create(odbc, insert_txt));
	SM_TRY_NOT(update_stmt = sm_odbc_stmt_create(odbc, update_txt));
	SM_TRY_NOT(delete_stmt = sm_odbc_stmt_create(odbc, delete_txt));
	SM_TRY_NOT(insert_acc_stmt = sm_odbc_stmt_create(odbc, insert_acc_txt));
	SM_TRY_NOT(count_stmt = sm_odbc_stmt_create(odbc, count_txt));
	SM_TRY_NOT(find_all_readonly_stmt = sm_odbc_stmt_create(odbc, find_all_readonly_txt));

	SM_TRY_FAULT(odbc->commit(odbc));

	if (!session_process) {
		int count = smc_job_count();

		if (count > 0)
			sm_notice("There are %d active jobs in database that seem to be orphaned. Is there any session process running? If no, then you should delete all entries from 'jobs' table and start the service again.", count);
		else if (count == -1) {
			sm_error("Failed to retrieve the number of active jobs from database. Check configuration.");
			smc_job_cleanup();
			return SM_FAULT;
		}
	}

	return SM_OK;

fault:
	smc_job_cleanup();
	odbc->rollback(odbc);
	return SM_FAULT;
}

void
smc_job_cleanup(void)
{
	SM_DEBUG_ENTER();

	sm_odbc_stmt_free(&find_stmt);
	sm_odbc_stmt_free(&find_by_drms_id_stmt);
	sm_odbc_stmt_free(&find_readonly_stmt);
	sm_odbc_stmt_free(&find_by_drms_id_readonly_stmt);
	sm_odbc_stmt_free(&find_acc_readonly_stmt);
	sm_odbc_stmt_free(&find_acc_by_drms_id_readonly_stmt);
	sm_odbc_stmt_free(&insert_stmt);
	sm_odbc_stmt_free(&update_stmt);
	sm_odbc_stmt_free(&delete_stmt);
	sm_odbc_stmt_free(&insert_acc_stmt);
	sm_odbc_stmt_free(&count_stmt);
	sm_odbc_stmt_free(&find_all_readonly_stmt);

	sm_signal_unregister("smc_job_stage_in");
	sm_signal_unregister("smc_job_queued");
	sm_signal_unregister("smc_job_held");
	sm_signal_unregister("smc_job_executing");
	sm_signal_unregister("smc_job_suspended");
	sm_signal_unregister("smc_job_stage_out");
	sm_signal_unregister("smc_job_cancelled");
	sm_signal_unregister("smc_job_finished");
	sm_signal_unregister("smc_job_failed");
}

smc_job_t
smc_job_create(void)
{
	SM_DEBUG_ENTER();
	return smc_job_create_internal(false, true, false);
}

static smc_job_t
smc_job_create_internal(bool readonly, bool new, bool transaction)
{
	smc_concrete_job_t ret;

	SM_DEBUG_ENTER();

	if (!(ret = sm_malloc(sizeof(*ret))))
		return NULL;

	memset(ret, 0, sizeof(*ret));

	ret->super.get_id = (smc_job_string_getter_function) smc_job_get_id;
	ret->super.get_drms_id = (smc_job_string_getter_function) smc_job_get_drms_id;
	ret->super.get_reservation_id = (smc_job_string_getter_function) smc_job_get_reservation_id;
	ret->super.get_reservation_info = (smc_job_reservation_info_getter_function) smc_job_get_reservation_info;
	ret->super.get_submitted_jsdl = (smc_job_string_getter_function) smc_job_get_submitted_jsdl;
	ret->super.get_jsdl = (smc_job_string_getter_function) smc_job_get_jsdl;
	ret->super.get_jsdl_parsed = (smc_job_jsdl_parsed_getter_function) smc_job_get_jsdl_parsed;
	ret->super.get_subject = (smc_job_string_getter_function) smc_job_get_subject;
	ret->super.get_mapped_user = (smc_job_string_getter_function) smc_job_get_mapped_user;
	ret->super.get_state = (smc_job_state_getter_function) smc_job_get_state;
	ret->super.get_staging_fault = (smc_job_string_getter_function) smc_job_get_staging_fault;
	ret->super.get_submission_time = (smc_job_time_getter_function) smc_job_get_submission_time;
	ret->super.get_start_time = (smc_job_time_getter_function) smc_job_get_start_time;
	ret->super.get_end_time = (smc_job_time_getter_function) smc_job_get_end_time;
	ret->super.get_exit_status = (smc_job_byte_getter_function) smc_job_get_exit_status;
	ret->super.has_exit_status = (smc_job_bool_getter_function) smc_job_has_exit_status;
	ret->super.get_core_dumped = (smc_job_bool_getter_function) smc_job_get_core_dumped;
	ret->super.has_core_dumped = (smc_job_bool_getter_function) smc_job_has_core_dumped;
	ret->super.get_posix_signal_name = (smc_job_string_getter_function) smc_job_get_posix_signal_name;
	ret->super.get_cpu_usage = (smc_job_double_getter_function) smc_job_get_cpu_usage;
	ret->super.get_mem_usage = (smc_job_double_getter_function) smc_job_get_mem_usage;
	ret->super.get_vmem_usage = (smc_job_double_getter_function) smc_job_get_vmem_usage;
	ret->super.get_queue = (smc_job_string_getter_function) smc_job_get_queue;
	ret->super.get_project = (smc_job_string_getter_function) smc_job_get_project;
	ret->super.get_walltime = (smc_job_int_getter_function) smc_job_get_walltime;
	ret->super.get_nodes = (smc_job_string_getter_function) smc_job_get_nodes;
	ret->super.get_atc_token = (smc_job_value_getter_function) smc_job_get_atc_token;


	ret->super.set_id = (smc_job_string_setter_function) smc_job_set_id;
	ret->super.set_drms_id = (smc_job_string_setter_function) smc_job_set_drms_id;
	ret->super.set_submitted_jsdl = (smc_job_string_setter_function) smc_job_set_submitted_jsdl;
	ret->super.set_jsdl = (smc_job_string_setter_function) smc_job_set_jsdl;
	ret->super.set_jsdl_parsed = (smc_job_jsdl_parsed_setter_function) smc_job_set_jsdl_parsed;
	ret->super.set_subject = (smc_job_string_setter_function) smc_job_set_subject;
	ret->super.set_mapped_user = (smc_job_string_setter_function) smc_job_set_mapped_user;
	ret->super.set_state = (smc_job_state_setter_function) smc_job_set_state;
	ret->super.set_staging_fault = (smc_job_string_setter_function) smc_job_set_staging_fault;
	ret->super.set_start_time = (smc_job_time_setter_function) smc_job_set_start_time;
	ret->super.set_exit_status = (smc_job_byte_setter_function) smc_job_set_exit_status;
	ret->super.set_core_dumped = (smc_job_bool_setter_function) smc_job_set_core_dumped;
	ret->super.set_posix_signal_name = (smc_job_string_setter_function) smc_job_set_posix_signal_name;
	ret->super.set_cpu_usage = (smc_job_double_setter_function) smc_job_set_cpu_usage;
	ret->super.set_mem_usage = (smc_job_double_setter_function) smc_job_set_mem_usage;
	ret->super.set_vmem_usage = (smc_job_double_setter_function) smc_job_set_vmem_usage;
	ret->super.set_queue = (smc_job_string_setter_function) smc_job_set_queue;
	ret->super.set_project = (smc_job_string_setter_function) smc_job_set_project;
	ret->super.set_walltime = (smc_job_int_setter_function) smc_job_set_walltime;
	ret->super.set_nodes = (smc_job_string_setter_function) smc_job_set_nodes;
	ret->super.set_atc_token = (smc_job_value_setter_function) smc_job_set_atc_token;

	ret->super.is_readonly = (smc_job_bool_getter_function) smc_job_is_readonly;
	ret->super.save = (smc_job_save_function) smc_job_save;

	ret->readonly = readonly;
	ret->new_entry = new;
	ret->in_transaction = transaction;

	return (smc_job_t) ret;
}

#define CHECK_THIS	SM_ASSERT(this, "You must supply 'this' pointer")

#define SET(field) \
	do { \
		this->modified = true; \
		this->field = value; \
		this->field##_set = true; \
		return this->field; \
	} while (0)

#define SET_STR(field) \
	do {  \
		this->modified = true; \
		SM_FREE(this->field); \
		this->field = sm_strdup(value); \
		if (!this->field) { \
			this->error = sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, "Out of memory"); \
			return NULL; \
		} \
		return this->field; \
	} while (0)

#define SET_VALUE(field) \
	do {  \
		this->modified = true; \
		sm_value_free(&this->field); \
		this->field = sm_value_clone(value); \
		if (!this->field) { \
			this->error = sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, "Out of memory"); \
			return NULL; \
		} \
		return this->field; \
	} while (0)

static const char *smc_job_get_id(smc_concrete_job_t this) { CHECK_THIS; return this->id; }
static const char *smc_job_get_drms_id(smc_concrete_job_t this) { CHECK_THIS; return this->drms_id; }
static const char *smc_job_get_reservation_id(smc_concrete_job_t this) { CHECK_THIS; smc_job_process_jsdl(this); return this->reservation_id; }
static const char *smc_job_get_submitted_jsdl(smc_concrete_job_t this) { CHECK_THIS; return this->submitted_jsdl; }
static const char *smc_job_get_jsdl(smc_concrete_job_t this) { CHECK_THIS; smc_job_process_jsdl(this); return this->jsdl; }
static const char *smc_job_get_subject(smc_concrete_job_t this) { CHECK_THIS; return this->subject; }
static const char *smc_job_get_mapped_user(smc_concrete_job_t this) { CHECK_THIS; return this->mapped_user; }
static smc_job_state_t smc_job_get_state(smc_concrete_job_t this) { CHECK_THIS; return this->state; }
static const char *smc_job_get_staging_fault(smc_concrete_job_t this) { CHECK_THIS; return this->staging_fault; }
static struct tm smc_job_get_submission_time(smc_concrete_job_t this) { CHECK_THIS; return this->submission_time; }
static struct tm smc_job_get_start_time(smc_concrete_job_t this) { CHECK_THIS; return this->start_time; }
static struct tm smc_job_get_end_time(smc_concrete_job_t this) { CHECK_THIS; return this->end_time; }
static unsigned char smc_job_get_exit_status(smc_concrete_job_t this) { CHECK_THIS; return this->exit_status; }
static bool smc_job_get_core_dumped(smc_concrete_job_t this) { CHECK_THIS; return this->core_dumped; }
static const char *smc_job_get_posix_signal_name(smc_concrete_job_t this) { CHECK_THIS; return this->posix_signal_name; }
static double smc_job_get_cpu_usage(smc_concrete_job_t this) { CHECK_THIS; return this->cpu_usage; }
static double smc_job_get_mem_usage(smc_concrete_job_t this) { CHECK_THIS; return this->mem_usage; }
static double smc_job_get_vmem_usage(smc_concrete_job_t this) { CHECK_THIS; return this->vmem_usage; }
static const char *smc_job_get_queue(smc_concrete_job_t this) { CHECK_THIS; return this->queue; }
static const char *smc_job_get_project(smc_concrete_job_t this) { CHECK_THIS; return this->project; }
static int smc_job_get_walltime(smc_concrete_job_t this) { CHECK_THIS; return this->walltime; }
static const char* smc_job_get_nodes(smc_concrete_job_t this) { CHECK_THIS; return this->nodes; }
static sm_value_t *smc_job_get_atc_token(smc_concrete_job_t this) { CHECK_THIS; return this->atc_token; }

static const char *smc_job_set_id(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(id); }
static const char *smc_job_set_drms_id(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(drms_id); }
static const char *smc_job_set_submitted_jsdl(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(submitted_jsdl); }
static const char *smc_job_set_jsdl(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(jsdl); }
static const char *smc_job_set_subject(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(subject); }
static const char *smc_job_set_staging_fault(smc_concrete_job_t this, const char *value) { CHECK_THIS; if (!this->staging_fault) SET_STR(staging_fault); else return NULL; }
static const char *smc_job_set_mapped_user(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(mapped_user); }
static struct tm smc_job_set_start_time(smc_concrete_job_t this, struct tm value) { CHECK_THIS; SET(start_time); }
static unsigned char smc_job_set_exit_status(smc_concrete_job_t this, unsigned char value) { CHECK_THIS; SET(exit_status); }
static bool smc_job_set_core_dumped(smc_concrete_job_t this, bool value) { CHECK_THIS; SET(core_dumped); }
static const char *smc_job_set_posix_signal_name(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(posix_signal_name); }
static double smc_job_set_cpu_usage(smc_concrete_job_t this, double value) { CHECK_THIS; SET(cpu_usage); }
static double smc_job_set_mem_usage(smc_concrete_job_t this, double value) { CHECK_THIS; SET(mem_usage); }
static double smc_job_set_vmem_usage(smc_concrete_job_t this, double value) { CHECK_THIS; SET(vmem_usage); }
static const char *smc_job_set_queue(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(queue); }
static const char *smc_job_set_project(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(project); }
static int smc_job_set_walltime(smc_concrete_job_t this, int value) { CHECK_THIS; SET(walltime); }
static const char *smc_job_set_nodes(smc_concrete_job_t this, const char *value) { CHECK_THIS; SET_STR(nodes); }
static sm_value_t *smc_job_set_atc_token(smc_concrete_job_t this, sm_value_t *value) { CHECK_THIS; SET_VALUE(atc_token); }


static smc_job_res_info_t
smc_job_get_reservation_info(smc_concrete_job_t this)
{
	smc_rsvn_t rsvn = NULL;
	smc_job_res_info_t result = NULL;
	sm_ret_code_t ret;
	struct tm end_time_tm, start_time_tm;
	time_t cur_time = time(NULL) + 2; /* add 2s offset */
	time_t start_time, end_time;

	CHECK_THIS;

	smc_job_process_jsdl(this);

	if (!this->reservation_id)
		return NULL;

	ret = smc_rsvn_get_readonly(this->reservation_id, &rsvn);

	if (ret.code) {
		sm_error("Failed to resolve reservation id: %d:%s", ret.code, ret.string);
		return NULL;
	}
	start_time_tm = rsvn->get_start_time(rsvn);
	end_time_tm = rsvn->get_end_time(rsvn);
	start_time = mktime(&start_time_tm);
	end_time = mktime(&end_time_tm);

	SM_TRY_NOT(result = sm_malloc(sizeof(*result)));
	sm_debug(" start_time=%u end_time=%u curr_time=%u\n", (unsigned int)start_time,(unsigned int) end_time, (unsigned int)cur_time);

	result->cancelled = SM_FALSE;
	result->expired = SM_FALSE;
	result->duration = 0;

	if (end_time - start_time == 0)
		result->cancelled = SM_TRUE;
	else if (cur_time < start_time)
		result->duration = end_time - start_time;
	else if (start_time <= cur_time && cur_time < end_time)
		result->duration = end_time - cur_time;
	else
		result->expired = SM_TRUE;

	SM_TRY_NOT(result->drms_id = sm_strdup(rsvn->get_drms_id(rsvn)));
	smc_rsvn_free(&rsvn);

	return result;
fault:
	smc_rsvn_free(&rsvn);
	return NULL;
}

static bool
smc_job_is_readonly(smc_concrete_job_t this)
{
	CHECK_THIS;
	return this->readonly;
}

static smc_job_state_t
smc_job_set_state(smc_concrete_job_t this, smc_job_state_t value)
{
	CHECK_THIS;

	if (value == this->state)
		return value;

	this->previous_state = this->state;
	this->state = value;
	this->modified = true;
	return this->state;
}

static _jsdl__JobDefinition *
smc_job_get_jsdl_parsed(smc_concrete_job_t this)
{
	_jsdl__JobDefinition j;

	CHECK_THIS;

	if (this->jsdl_parsed && !this->modified)
		return this->jsdl_parsed;

	if (!this->jsdl)
		return NULL;

	this->jsdl_parsed = NULL;
	sm_alloc_tracked_free(&this->tracked);

	memset(&j, 0, sizeof(j));
	SM_TRY_NOT(this->tracked = smc_jsdl_from_str(this->jsdl, &j));
	return sm_alloc_tracked_memdup(this->tracked, &j, sizeof(j));

fault:
	return NULL;
}

static const char *
smc_job_set_jsdl_parsed(smc_concrete_job_t this, _jsdl__JobDefinition *value)
{
	CHECK_THIS;

	this->modified = true;

	SM_FREE(this->jsdl);

	if (!(this->jsdl = smc_jsdl_str(value, NULL, NULL)))
		this->error = sm_ret_code_create_oom();

	return this->jsdl;
}

static bool
smc_job_has_exit_status(smc_concrete_job_t this)
{
	CHECK_THIS;
	SM_DEBUG_ENTER();
	return this->exit_status_set;
}

static bool
smc_job_has_core_dumped(smc_concrete_job_t this)
{
	CHECK_THIS;
	SM_DEBUG_ENTER();
	return this->core_dumped_set;
}

static sm_status_ret_code_t
smc_job_read_from_db(smc_concrete_job_t job, sm_odbc_stmt_t stmt)
{
	char *db_string = NULL;
	struct tm db_time;
	int db_int = 0;
	SM_BOOL db_bit = SM_FALSE;
	double db_double = 0.0;
	sm_status_ret_code_t status;

	SM_MANDATORY_P(job);
	SM_MANDATORY_P(stmt);

	SM_DEBUG_ENTER();

	memset(&db_time, 0, sizeof(db_time));

	SM_TRY_FAULT(stmt->get_int(stmt, 1, &job->serial));
	SM_TRY_FAULT(stmt->get_string(stmt, 2, &job->id));
	SM_TRY_FAULT(stmt->get_string(stmt, 3, &job->drms_id));
	SM_TRY_FAULT(stmt->get_string(stmt, 4, &job->reservation_id));
	SM_TRY_FAULT(stmt->get_string(stmt, 5, &job->submitted_jsdl));
	SM_TRY_FAULT(stmt->get_string(stmt, 6, &job->jsdl));
	SM_TRY_FAULT(stmt->get_string(stmt, 7, &job->subject));
	SM_TRY_FAULT(stmt->get_string(stmt, 8, &job->mapped_user));

	SM_TRY_FAULT(stmt->get_string(stmt, 9, &db_string));
	if (db_string) {
		job->state = smc_job_state_parse(db_string);
		SM_FREE(db_string);
	}

	SM_TRY_FAULT(stmt->get_string(stmt, 10, &job->staging_fault));

	SM_TRY_FAULT(status = stmt->get_timestamp(stmt, 11, &db_time));
	if (status != SM_NEG) {
		job->submission_time = db_time;
		job->submission_time_set = true;
	}

	SM_TRY_FAULT(status = stmt->get_timestamp(stmt, 12, &db_time));
	if (status != SM_NEG) {
		job->start_time = db_time;
		job->start_time_set = true;
	}

	SM_TRY_FAULT(status = stmt->get_timestamp(stmt, 13, &db_time));
	if (status != SM_NEG) {
		job->end_time = db_time;
		job->end_time_set = true;
	}

	SM_TRY_FAULT(status = stmt->get_int(stmt, 14, &db_int));
	if (status != SM_NEG) {
		job->exit_status = db_int;
		job->exit_status_set = true;
	}

	SM_TRY_FAULT(status = stmt->get_bit(stmt, 15, &db_bit));
	if (status != SM_NEG) {
		job->core_dumped = (bool) db_bit;
		job->core_dumped_set = true;
	}

	SM_TRY_FAULT(stmt->get_string(stmt, 16, &job->posix_signal_name));

	SM_TRY_FAULT(status = stmt->get_double(stmt, 17, &db_double));
	if (status != SM_NEG) {
		job->cpu_usage = db_double;
		job->cpu_usage_set = true;
	}

	SM_TRY_FAULT(status = stmt->get_double(stmt, 18, &db_double));
	if (status != SM_NEG) {
		job->mem_usage = db_double;
		job->mem_usage_set = true;
	}

	SM_TRY_FAULT(status = stmt->get_double(stmt, 19, &db_double));
	if (status != SM_NEG) {
		job->vmem_usage = db_double;
		job->vmem_usage_set = true;
	}

	SM_TRY_FAULT(stmt->get_string(stmt, 20, &job->queue));
	SM_TRY_FAULT(stmt->get_string(stmt, 21, &job->project));

	SM_TRY_FAULT(status = stmt->get_int(stmt, 22, &db_int));
	if (status != SM_NEG) {
		job->walltime = db_int;
		job->walltime_set = true;
	}

	SM_TRY_FAULT(stmt->get_string(stmt, 23, &job->nodes));

	job->atc_token = NULL; /* We do not save credentials into database as this would led to some security risk */

	return SM_OK;

fault:
	SM_FREE(db_string);
	return SM_FAULT;
}

#define SMC_DB_TRY_FAULT(value) \
	do { \
		if ((value) == SM_FAULT) \
			goto db_fault; \
	} while(0);

static sm_ret_code_t
smc_job_get_internal(const char *id, sm_list_t *result, bool readonly)
{
	sm_status_ret_code_t status;
	sm_odbc_stmt_t first_stmt = NULL;
	sm_odbc_stmt_t stmt = NULL;
	smc_job_t job = NULL;

	SM_MANDATORY_P(id);
	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER_STR(id);

	*result = NULL;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return sm_ret_code_create(SMC_JOB_DATABASE_FAULT, SMC_JOB_DATABASE_FAULT_STR);

	/* Try in jobs first. */
	if (readonly)
		stmt = find_readonly_stmt;
	else
		stmt = find_stmt;

	first_stmt = stmt;

	SMC_DB_TRY_FAULT(stmt->execute(stmt, "s", id));
	SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

	/* Now try in jobs_accounting. */
	if (status == SM_NEG) {
		if (readonly)
			stmt = find_acc_readonly_stmt;
		else
			goto not_found;

		SMC_DB_TRY_FAULT(stmt->execute(stmt, "s", id));
		SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

		if (status == SM_NEG)
			goto not_found;
	}

	/* Read all jobs. */
	for (;;) {
		SM_TRY_NOT(job = smc_job_create_internal(readonly, false, !readonly));

		SMC_DB_TRY_FAULT(smc_job_read_from_db((smc_concrete_job_t) job, stmt));

		SM_TRY_NOT(sm_list_add(result, job, 0));
		job = NULL;

		SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));
		if (status == SM_NEG)
			break;
	}

	first_stmt->close_cursor(first_stmt);
	stmt->close_cursor(stmt);

	if (readonly)
		SM_TRY_FAULT(odbc->commit(odbc));

	return sm_ret_code_create(SMC_JOB_OK, NULL);

not_found:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	return sm_ret_code_create(SMC_JOB_NOT_FOUND, SMC_JOB_NOT_FOUND_STR);

db_fault:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_job_free(&job);
	sm_list_free_user(*result, (sm_list_free_user_function) smc_job_free);
	return sm_ret_code_create(SMC_JOB_DATABASE_FAULT, SMC_JOB_DATABASE_FAULT_STR);

fault:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_job_free(&job);
	sm_list_free_user(*result, (sm_list_free_user_function) smc_job_free);
	return sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, SMC_JOB_INTERNAL_FAULT_STR);
}

sm_ret_code_t
smc_job_get(const char *id, sm_list_t *result)
{
	SM_DEBUG_ENTER_STR(id);
	return smc_job_get_internal(id, result, false);
}

sm_ret_code_t
smc_job_get_readonly(const char *id, sm_list_t *result)
{
	SM_DEBUG_ENTER_STR(id);
	return smc_job_get_internal(id, result, true);
}

static sm_ret_code_t
smc_job_get_by_drms_id_internal(const char *drms_id, smc_job_t *result, bool readonly)
{
	sm_status_ret_code_t status;
	sm_odbc_stmt_t stmt = NULL;
	sm_odbc_stmt_t first_stmt = NULL;

	SM_MANDATORY_P(drms_id);
	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER_STR(drms_id);

	*result = NULL;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return sm_ret_code_create(SMC_JOB_DATABASE_FAULT, SMC_JOB_DATABASE_FAULT_STR);

	/* Try in jobs first. */
	if (readonly)
		stmt = find_by_drms_id_readonly_stmt;
	else
		stmt = find_by_drms_id_stmt;

	first_stmt = stmt;

	SMC_DB_TRY_FAULT(stmt->execute(stmt, "s", drms_id));
	SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

	/* Now try in jobs_accounting. */
	if (status == SM_NEG) {
		if (readonly)
			stmt = find_acc_by_drms_id_readonly_stmt;
		else
			goto not_found;

		SMC_DB_TRY_FAULT(stmt->execute(stmt, "s", drms_id));
		SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

		if (status == SM_NEG)
			goto not_found;
	}

	SM_TRY_NOT(*result = smc_job_create_internal(readonly, false, !readonly));

	SMC_DB_TRY_FAULT(smc_job_read_from_db((smc_concrete_job_t) *result, stmt));

	first_stmt->close_cursor(first_stmt);
	stmt->close_cursor(stmt);

	if (readonly)
		SM_TRY_FAULT(odbc->commit(odbc));

	return sm_ret_code_create(SMC_JOB_OK, NULL);

not_found:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	return sm_ret_code_create(SMC_JOB_NOT_FOUND, SMC_JOB_NOT_FOUND_STR);

db_fault:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_job_free(result);
	return sm_ret_code_create(SMC_JOB_DATABASE_FAULT, SMC_JOB_DATABASE_FAULT_STR);

fault:
	if (first_stmt)
		first_stmt->close_cursor(first_stmt);
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_job_free(result);
	return sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, SMC_JOB_INTERNAL_FAULT_STR);
}

sm_ret_code_t
smc_job_get_by_drms_id_readonly(const char *drms_id, smc_job_t *result)
{
	SM_DEBUG_ENTER_STR(drms_id);
	return smc_job_get_by_drms_id_internal(drms_id, result, true);
}

sm_ret_code_t
smc_job_get_by_drms_id(const char *drms_id, smc_job_t *result)
{
	SM_DEBUG_ENTER_STR(drms_id);
	return smc_job_get_by_drms_id_internal(drms_id, result, false);
}

sm_ret_code_t
smc_job_get_active_readonly(sm_list_t *result)
{
	sm_status_ret_code_t status;
	sm_odbc_stmt_t stmt = NULL;
	smc_job_t job = NULL;

	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER();

	*result = NULL;

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return sm_ret_code_create(SMC_JOB_DATABASE_FAULT, SMC_JOB_DATABASE_FAULT_STR);

	stmt = find_all_readonly_stmt;

	SMC_DB_TRY_FAULT(stmt->execute(stmt, NULL));
	SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));

	if (status == SM_OK) {

		for (;;) {
			SM_TRY_NOT(job = smc_job_create_internal(true, false, false));

			SMC_DB_TRY_FAULT(smc_job_read_from_db((smc_concrete_job_t) job, stmt));

			SM_TRY_NOT(sm_list_add(result, job, 0));
			job = NULL;

			SMC_DB_TRY_FAULT(status = stmt->fetch_row(stmt));
			if (status == SM_NEG)
				break;
		}
	}

	stmt->close_cursor(stmt);

	SM_TRY_FAULT(odbc->commit(odbc));

	return sm_ret_code_create(SMC_JOB_OK, NULL);

db_fault:
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_job_free(&job);
	sm_list_free_user(*result, (sm_list_free_user_function) smc_job_free);
	return sm_ret_code_create(SMC_JOB_DATABASE_FAULT, SMC_JOB_DATABASE_FAULT_STR);

fault:
	if (stmt)
		stmt->close_cursor(stmt);
	odbc->rollback(odbc);
	smc_job_free(&job);
	sm_list_free_user(*result, (sm_list_free_user_function) smc_job_free);
	return sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, SMC_JOB_INTERNAL_FAULT_STR);
}

void
smc_job_free(smc_job_t *job)
{
	smc_concrete_job_t concrete;

	SM_MANDATORY_P(job);

	SM_DEBUG_ENTER();

	concrete = (smc_concrete_job_t) *job;

	if (!concrete)
		return;

	if (concrete->in_transaction) {
		odbc->rollback(odbc);
		concrete->in_transaction = false;
	}

	SM_FREE(concrete->id);
	SM_FREE(concrete->drms_id);
	SM_FREE(concrete->reservation_id);
	SM_FREE(concrete->submitted_jsdl);
	SM_FREE(concrete->jsdl);
	SM_FREE(concrete->staging_fault);
	SM_FREE(concrete->subject);
	SM_FREE(concrete->mapped_user);
	SM_FREE(concrete->posix_signal_name);
	SM_FREE(concrete->queue);
	SM_FREE(concrete->nodes);
	SM_FREE(concrete->project);

	sm_value_free(&concrete->atc_token);

	sm_alloc_tracked_free(&concrete->tracked);

	SM_FREE(*job);
}

const char *
smc_job_state_str(smc_job_state_t state)
{
	switch (state) {
		case SMC_UNSPECIFIED_STATE: return NULL;
		case SMC_STAGE_IN: return "Stage-in";
		case SMC_HELD: return "Held";
		case SMC_QUEUED: return "Queued";
		case SMC_EXECUTING: return "Executing";
		case SMC_SUSPENDED: return "Suspended";
		case SMC_STAGE_OUT: return "Stage-out";
		case SMC_FINISHED: return "Finished";
		case SMC_FAILED: return "Failed";
		case SMC_CANCELLED: return "Cancelled";
		default: SM_ASSERT(false, "Unknown state in %s", __FUNCTION__);
	}
}

smc_job_state_t
smc_job_state_parse(const char *state)
{
	if (!state)
		return SMC_UNSPECIFIED_STATE;
	else if (!strcmp(state, "Stage-in"))
		return SMC_STAGE_IN;
	else if (!strcmp(state, "Held"))
		return SMC_HELD;
	else if (!strcmp(state, "Queued"))
		return SMC_QUEUED;
	else if (!strcmp(state, "Executing"))
		return SMC_EXECUTING;
	else if (!strcmp(state, "Suspended"))
		return SMC_SUSPENDED;
	else if (!strcmp(state, "Stage-out"))
		return SMC_STAGE_OUT;
	else if (!strcmp(state, "Finished"))
		return SMC_FINISHED;
	else if (!strcmp(state, "Failed"))
		return SMC_FAILED;
	else if (!strcmp(state, "Cancelled"))
		return SMC_CANCELLED;
	else
		SM_ASSERT(false, "Unknown state in %s", __FUNCTION__);
}

#define FIELD2DB(field)	((field##_set) ? &(field) : NULL)

static struct tm
timetm(void)
{
	time_t now = time(NULL);
	struct tm nowtm;

	localtime_r(&now, &nowtm);
	return nowtm;
}

static sm_status_ret_code_t
smc_job_insert_acc(smc_concrete_job_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	SM_ASSERT(!this->in_transaction, "Job object inconsistent state");

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	SM_TRY_FAULT(insert_acc_stmt->execute(insert_acc_stmt, "ssssssssstttibsdd",
			this->id,
			this->drms_id,
			this->reservation_id,
			this->submitted_jsdl,
			this->jsdl,
			this->subject,
			this->mapped_user,
			smc_job_state_str(this->state),
			this->staging_fault,
			FIELD2DB(this->submission_time),
			FIELD2DB(this->start_time),
			FIELD2DB(this->end_time),
			FIELD2DB(this->exit_status),
			FIELD2DB(this->core_dumped),
			this->posix_signal_name,
			FIELD2DB(this->cpu_usage),
			FIELD2DB(this->mem_usage),
			FIELD2DB(this->vmem_usage),
			this->queue,
			this->project,
			FIELD2DB(this->walltime),
			this->nodes));

	SM_TRY_FAULT(odbc->commit(odbc));
	return SM_OK;

fault:
	odbc->rollback(odbc);
	return SM_FAULT;
}

static sm_status_ret_code_t
smc_job_insert(smc_concrete_job_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	SM_ASSERT(!this->in_transaction, "Job object inconsistent state");

	if (odbc->begin_transaction(odbc) == SM_FAULT)
		return SM_FAULT;

	SM_TRY_FAULT(insert_stmt->execute(insert_stmt, "ssssssssstttibsdddssis",
			this->id,
			this->drms_id,
			this->reservation_id,
			this->submitted_jsdl,
			this->jsdl,
			this->subject,
			this->mapped_user,
			smc_job_state_str(this->state),
			this->staging_fault,
			FIELD2DB(this->submission_time),
			FIELD2DB(this->start_time),
			FIELD2DB(this->end_time),
			FIELD2DB(this->exit_status),
			FIELD2DB(this->core_dumped),
			this->posix_signal_name,
			FIELD2DB(this->cpu_usage),
			FIELD2DB(this->mem_usage),
			FIELD2DB(this->vmem_usage),
			this->queue,
			this->project,
			FIELD2DB(this->walltime),
			this->nodes));

	SM_TRY_FAULT(odbc->commit(odbc));
	return SM_OK;

fault:
	odbc->rollback(odbc);
	return SM_FAULT;
}

static sm_status_ret_code_t
smc_job_update(smc_concrete_job_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	SM_ASSERT(this->in_transaction, "Job object inconsistent state");

	SM_TRY_FAULT(update_stmt->execute(update_stmt, "ssssssssstttibsdddssisi",
			this->id,
			this->drms_id,
			this->reservation_id,
			this->submitted_jsdl,
			this->jsdl,
			this->subject,
			this->mapped_user,
			smc_job_state_str(this->state),
			this->staging_fault,
			FIELD2DB(this->submission_time),
			FIELD2DB(this->start_time),
			FIELD2DB(this->end_time),
			FIELD2DB(this->exit_status),
			FIELD2DB(this->core_dumped),
			this->posix_signal_name,
			FIELD2DB(this->cpu_usage),
			FIELD2DB(this->mem_usage),
			FIELD2DB(this->vmem_usage),
			this->queue,
			this->project,
			FIELD2DB(this->walltime),
			this->nodes,
			&this->serial
			));

	SM_TRY_FAULT(odbc->commit(odbc));
	this->in_transaction = false;
	return SM_OK;

fault:
	odbc->rollback(odbc);
	this->in_transaction = false;
	return SM_FAULT;
}

static sm_status_ret_code_t
smc_job_move_to_acc(smc_concrete_job_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	SM_ASSERT(this->in_transaction, "Job object inconsistent state");

	SM_TRY_FAULT(delete_stmt->execute(delete_stmt, "s", this->id));

	SM_TRY_FAULT(insert_acc_stmt->execute(insert_acc_stmt, "ssssssssstttibsdddssis",
			this->id,
			this->drms_id,
			this->reservation_id,
			this->submitted_jsdl,
			this->jsdl,
			this->subject,
			this->mapped_user,
			smc_job_state_str(this->state),
			this->staging_fault,
			FIELD2DB(this->submission_time),
			FIELD2DB(this->start_time),
			FIELD2DB(this->end_time),
			FIELD2DB(this->exit_status),
			FIELD2DB(this->core_dumped),
			this->posix_signal_name,
			FIELD2DB(this->cpu_usage),
			FIELD2DB(this->mem_usage),
			FIELD2DB(this->vmem_usage),
			this->queue,
			this->project,
			FIELD2DB(this->walltime),
			this->nodes));

	SM_TRY_FAULT(odbc->commit(odbc));
	this->in_transaction = false;
	return SM_OK;

fault:
	odbc->rollback(odbc);
	this->in_transaction = false;
	return SM_FAULT;
}

static sm_ret_code_t
smc_job_process_jsdl(smc_concrete_job_t this)
{
	sm_alloc_tracked_t tracked = NULL;
	_jsdl__JobDefinition jsdl;
	_jsdl_hpcpa__HPCProfileApplication *hpcpapp = NULL;
	sm_ret_code_t ret;

	SM_DEBUG_ENTER_STR(this->id);

	memset(&ret, 0, sizeof(ret));

	if (!this->jsdl)
		return sm_ret_code_create(SMC_JOB_OK, NULL);

	if (!this->modified)
		return sm_ret_code_create(SMC_JOB_OK, NULL);

	if (!(tracked = smc_jsdl_from_str(this->jsdl, &jsdl)))
		return sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, "Failed to parse JSDL document");

	/* Put this->mapped_user into jsdl-hpcpa:UserName if not already
	 * present. */
	if (this->mapped_user && jsdl.jsdl__JobDescription->jsdl__Application) {
		hpcpapp = jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication;

		if (hpcpapp && !hpcpapp->jsdl_hpcpa__UserName) {
			SM_TRY_NOT(hpcpapp->jsdl_hpcpa__UserName = sm_alloc_tracked_malloc(tracked, sizeof(_jsdl_hpcpa__UserName)));
			SM_TRY_NOT(hpcpapp->jsdl_hpcpa__UserName->__item = sm_alloc_tracked_strdup(tracked, this->mapped_user));

			sm_debug("Set jsdl-hpcpa:UserName with value from this->mapped_user (%s)", this->mapped_user);
		}
	}

	/* Put user's home directory into jsdl-hpcpa:WorkingDirectory if not
	 * already present. */
	if (this->mapped_user && jsdl.jsdl__JobDescription->jsdl__Application) {
		hpcpapp = jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication;

		/*if working directory is relative*/
		if (hpcpapp && hpcpapp->jsdl_hpcpa__WorkingDirectory && hpcpapp->jsdl_hpcpa__WorkingDirectory->__item && hpcpapp->jsdl_hpcpa__WorkingDirectory->__item[0] != '/' ) {
			char *tmp = NULL;

			ret = smc_user_get_home(this->mapped_user, &tmp);
			if (ret.code) {
				sm_error("Failed to get user's home dir: %s", ret.string);
				goto fault;
			}

			/*prepend it with user's home directory */
			hpcpapp->jsdl_hpcpa__WorkingDirectory->__item = sm_alloc_tracked_asprintf(tracked,"%s/%s", tmp, hpcpapp->jsdl_hpcpa__WorkingDirectory->__item);

			SM_FREE(tmp);

			SM_TRY_NOT(hpcpapp->jsdl_hpcpa__WorkingDirectory->__item);
		}

		/*if working directory is not set at all*/
		if (hpcpapp && !hpcpapp->jsdl_hpcpa__WorkingDirectory) {
			char *tmp = NULL;

			ret = smc_user_get_home(this->mapped_user, &tmp);
			if (ret.code) {
				sm_error("Failed to get user's home dir: %s", ret.string);
				goto fault;
			}

			hpcpapp->jsdl_hpcpa__WorkingDirectory = sm_alloc_tracked_malloc(tracked, sizeof(*hpcpapp->jsdl_hpcpa__WorkingDirectory));
			hpcpapp->jsdl_hpcpa__WorkingDirectory->__item = sm_alloc_tracked_strdup(tracked, tmp);

			SM_FREE(tmp);

			SM_TRY_NOT(hpcpapp->jsdl_hpcpa__WorkingDirectory);
			SM_TRY_NOT(hpcpapp->jsdl_hpcpa__WorkingDirectory->__item);
		}

		/* add activity id as environment variable */
		if (this->atc_token && hpcpapp) {
			int i;

			if (hpcpapp->__sizeEnvironment) {
				struct jsdl_hpcpa__Environment_USCOREType *origEnv = hpcpapp->jsdl_hpcpa__Environment;

				SM_TRY_NOT(hpcpapp->jsdl_hpcpa__Environment = sm_alloc_tracked_calloc(this->tracked, hpcpapp->__sizeEnvironment + 2 /*make room for X509_USER_PROXY*/, sizeof(struct jsdl_hpcpa__Environment_USCOREType)));

				for (i=0; i < hpcpapp->__sizeEnvironment; i++)
					memcpy(&(hpcpapp->jsdl_hpcpa__Environment[i]), &(origEnv[i]), sizeof( struct jsdl_hpcpa__Environment_USCOREType ));

			} else {
				SM_TRY_NOT(hpcpapp->jsdl_hpcpa__Environment = sm_alloc_tracked_calloc(this->tracked, 2 /*make room for X509_USER_PROXY*/, sizeof(struct jsdl_hpcpa__Environment_USCOREType)));
			}

			for (i = hpcpapp->__sizeEnvironment; i < hpcpapp->__sizeEnvironment + 2; i++)
				memset(&(hpcpapp->jsdl_hpcpa__Environment[i]), 0, sizeof( struct jsdl_hpcpa__Environment_USCOREType ));

			hpcpapp->jsdl_hpcpa__Environment[hpcpapp->__sizeEnvironment].__item = this->id;
			hpcpapp->jsdl_hpcpa__Environment[hpcpapp->__sizeEnvironment].name = "QCG_ACTIVITY_ID";
			hpcpapp->__sizeEnvironment++;
		}

		if (this->atc_token && this->atc_token[0]->vint32 == SM_ATC_TOKEN_X509_PROXY && hpcpapp) {
			char *proxy_file_name = NULL;
			char *working_dir = (hpcpapp->jsdl_hpcpa__WorkingDirectory && hpcpapp->jsdl_hpcpa__WorkingDirectory->__item) ? hpcpapp->jsdl_hpcpa__WorkingDirectory->__item : ".";

			SM_ASSERT(SM_VALUE_IS_BYTES(this->atc_token[1]),"Invalid X509 proxy atc_token");

			proxy_file_name = sm_alloc_tracked_asprintf(this->tracked, "%s/.x509.proxy.%s", working_dir, this->id);

			sm_info("saving X509 user proxy into %s", proxy_file_name);

			if (sm_mem_to_file(proxy_file_name, this->atc_token[1]->vbytes, this->atc_token[1]->size) != SM_OK) {
				sm_error("Failed to save proxy certificate as %s", proxy_file_name);
				sm_value_free(&this->atc_token);
				ret = sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, "Failed to save user X509 proxy");
				goto fault;
			}

			/* necessary space was allocated in previous if block */
			hpcpapp->jsdl_hpcpa__Environment[hpcpapp->__sizeEnvironment].__item = proxy_file_name;
			hpcpapp->jsdl_hpcpa__Environment[hpcpapp->__sizeEnvironment].name = "X509_USER_PROXY";
			hpcpapp->__sizeEnvironment++;

		}
		
		sm_value_free(&this->atc_token);
	}

	/* Get ReservationId from JSDL and save into this->reservation_id */
	if (!this->reservation_id && jsdl.jsdl__JobDescription->jsdl__Resources && jsdl.jsdl__JobDescription->jsdl__Resources->jsdl_qcg_comp_ares__Reservation) {
		sm_status_ret_code_t status;
		const char *tmp = NULL;

		status = smc_parse_reservation_epr(&jsdl.jsdl__JobDescription->jsdl__Resources->jsdl_qcg_comp_ares__Reservation->ReservationIdentifier, &tmp);

		/* Invalid ReservationId format */
		if (status == SM_FAULT) {
			ret = sm_ret_code_create(SMC_JOB_INVALID_ARGUMENT, "JobDefinition/JobDescription/Resources/Reservation/ReservationIdentifier/ReferenceParameters/ReservationId");
			goto fault;

		} else if (status == SM_OK) {
			SM_FREE(this->reservation_id);
			SM_TRY_NOT(this->reservation_id = sm_strdup(tmp));
		}
	}

	SM_FREE(this->jsdl);
	SM_TRY_NOT(this->jsdl = smc_jsdl_str(&jsdl, NULL, NULL));

	sm_alloc_tracked_free(&tracked);
	return sm_ret_code_create(SMC_JOB_OK, NULL);

fault:
	sm_alloc_tracked_free(&tracked);
	if (ret.code)
		return ret;
	else
		return sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, SMC_JOB_INTERNAL_FAULT_STR);
}

static void
smc_job_add_estimated_time(smc_concrete_job_t this)
{
	SM_DEBUG_ENTER();

	if (this->new_entry) {
		if (!this->submission_time_set) {
			this->submission_time_set = true;
			this->submission_time = timetm();
		}
	}

	/* We can enter the Running state starting from Executing or Stage-out. */
	if (this->state == SMC_EXECUTING || this->state == SMC_STAGE_OUT)  {
		if (!this->submission_time_set) {
			this->submission_time_set = true;
			this->submission_time = timetm();
		}
		if (!this->start_time_set) {
			this->start_time_set = true;
			this->start_time = timetm();
		}
	}

	/* End time includes the time needed for staging out. */
	if (SMC_JOB_STATE_TERMINAL(this->state)) {
		if (!this->submission_time_set) {
			this->submission_time_set = true;
			this->submission_time = timetm();
		}
		if (!this->start_time_set) {
			this->start_time_set = true;
			this->start_time = timetm();
		}
		if (!this->end_time_set) {
			this->end_time_set = true;
			this->end_time = timetm();
		}
	}
}

static void
smc_job_enter_state(smc_concrete_job_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	switch (this->state) {
		case SMC_UNSPECIFIED_STATE:
			SM_ASSERT(0, "Job in unspecified state");
			break;
		case SMC_STAGE_IN:
			sm_signal_emit("smc_job_stage_in", this->id,
			(this->reservation_id), this->submitted_jsdl, this->jsdl,
			this->subject, this->mapped_user, mktime(&this->submission_time), NULL);
			break;
		case SMC_QUEUED:
			sm_signal_emit("smc_job_queued", this->id, this->drms_id,
			this->reservation_id, this->submitted_jsdl, this->jsdl,
			this->subject, this->mapped_user, mktime(&this->submission_time), NULL);
			break;
		case SMC_HELD:
			sm_signal_emit("smc_job_held", this->id, this->drms_id,
			this->reservation_id, this->submitted_jsdl, this->jsdl,
			this->subject, this->mapped_user, mktime(&this->submission_time), NULL);
			break;
		case SMC_EXECUTING:
			sm_signal_emit("smc_job_executing", this->id,
			this->drms_id, this->reservation_id,
			this->submitted_jsdl, this->jsdl, this->subject,
			this->mapped_user, mktime(&this->submission_time), mktime(&this->start_time), NULL);
			break;
		case SMC_SUSPENDED:
			sm_signal_emit("smc_job_suspended", this->id,
			this->drms_id, this->reservation_id,
			this->submitted_jsdl, this->jsdl, this->subject,
			this->mapped_user, mktime(&this->submission_time), mktime(&this->start_time), NULL);
			break;
		case SMC_STAGE_OUT:
			sm_signal_emit("smc_job_stage_out", this->id,
			this->drms_id, this->reservation_id,
			this->submitted_jsdl, this->jsdl, this->subject,
			this->mapped_user, mktime(&this->submission_time), mktime(&this->start_time), mktime(&this->end_time), NULL);
			break;
		case SMC_FINISHED:
			sm_signal_emit("smc_job_finished", this->id,
			this->drms_id, this->reservation_id,
			this->submitted_jsdl, this->jsdl, this->subject,
			this->mapped_user, mktime(&this->submission_time), mktime(&this->start_time), mktime(&this->end_time),
			this->exit_status, this->cpu_usage, this->mem_usage, this->vmem_usage, this->queue, this->project, this->walltime, this->nodes, NULL);
			break;
		case SMC_CANCELLED:
			sm_signal_emit("smc_job_cancelled", this->id,
			this->drms_id, this->reservation_id,
			this->submitted_jsdl, this->jsdl, this->subject,
			this->mapped_user, mktime(&this->submission_time), mktime(&this->start_time), mktime(&this->end_time), NULL);
			break;
		case SMC_FAILED:
			sm_signal_emit("smc_job_failed", this->id, this->drms_id,
			this->reservation_id, this->submitted_jsdl, this->jsdl,
			this->subject, this->mapped_user, this->staging_fault, mktime(&this->submission_time), mktime(&this->start_time), mktime(&this->end_time),
			this->core_dumped, this->posix_signal_name, NULL);
			break;
	}
}

#define STATE_CHANGE(prev, now)	(this->previous_state == (prev) && this->state == (now))
#define ENTER_STATE(state) \
	do { \
		this->super.set_state((smc_job_t) this, state); \
		smc_job_enter_state(this); \
	} while(0)

static void
smc_job_process_state_change(smc_concrete_job_t this)
{
	SM_DEBUG_ENTER_STR(this->id);

	assert(this->previous_state != this->state);

	sm_debug("previous_state = %s, state = %s", SM_SAFE_STR(smc_job_state_str(this->previous_state)), SM_SAFE_STR(smc_job_state_str(this->state)));

	/* Handle special cases. */
	if (STATE_CHANGE(SMC_STAGE_IN, SMC_FAILED)) {

		/* Stage-in -> Queued -> Executing -> Failed */
		ENTER_STATE(SMC_QUEUED);
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_FAILED);

	} else if (STATE_CHANGE(SMC_STAGE_IN, SMC_FINISHED)) {

		/* Stage-in -> Queued -> Executing -> Finished */
		ENTER_STATE(SMC_QUEUED);
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_FINISHED);

	} else if (STATE_CHANGE(SMC_QUEUED, SMC_FAILED)) {

		/* Queued -> Executing -> Failed */
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_FAILED);

	} else if (STATE_CHANGE(SMC_QUEUED, SMC_FINISHED)) {

		/* Queued -> Executing -> Finished */
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_FINISHED);

	} else if (STATE_CHANGE(SMC_QUEUED, SMC_STAGE_OUT)) {

		/* Queued -> Executing -> Stage-out */
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_STAGE_OUT);

	} else if (STATE_CHANGE(SMC_QUEUED, SMC_SUSPENDED)) {

		/* Queued -> Executing -> Suspended */
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_SUSPENDED);

	} else if (STATE_CHANGE(SMC_HELD, SMC_FAILED)) {

		/* Held -> Queued -> Executing -> Failed */
		ENTER_STATE(SMC_QUEUED);
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_FAILED);

	} else if (STATE_CHANGE(SMC_STAGE_IN, SMC_STAGE_OUT)) {

		/* Stage-in -> Queued -> Executing -> Stage-out */
		ENTER_STATE(SMC_QUEUED);
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_STAGE_OUT);

	} else if (STATE_CHANGE(SMC_UNSPECIFIED_STATE, SMC_STAGE_OUT)) {

		/* NEW -> Queued -> Executing -> Stage-out */
		ENTER_STATE(SMC_QUEUED);
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_STAGE_OUT);

	} else if (STATE_CHANGE(SMC_UNSPECIFIED_STATE, SMC_FINISHED)) {

		/* NEW -> Queued -> Executing -> Finished */
		ENTER_STATE(SMC_QUEUED);
		ENTER_STATE(SMC_EXECUTING);
		ENTER_STATE(SMC_FINISHED);

	} else {
		smc_job_enter_state(this);
	}
}

sm_ret_code_t
smc_job_save(smc_concrete_job_t this)
{
	sm_status_ret_code_t (*save_func)(smc_concrete_job_t) = NULL;
	sm_ret_code_t ret;

	CHECK_THIS;

	SM_DEBUG_ENTER_STR(this->id);

	if (this->readonly)
		SM_ASSERT(false, "You cannot save an object that is read only");

	if (this->error.code)
		return this->error;

	if (!this->modified) {
		if (this->in_transaction) {
			odbc->commit(odbc);
			this->in_transaction = false;
		}
		return sm_ret_code_create(SMC_JOB_OK, NULL);
	}

	smc_job_add_estimated_time(this);

	ret = smc_job_process_jsdl(this);
	if (ret.code)
		return ret;

	if (this->new_entry) {
		if (SMC_JOB_STATE_TERMINAL(this->state))
			save_func = smc_job_insert_acc;
		else
			save_func = smc_job_insert;
	} else {
		if (SMC_JOB_STATE_TERMINAL(this->state)) {
			/* This may only happen once since after moving to
			 * jobs_acc one can only get a read only copy of the
			 * job object. */
			save_func = smc_job_move_to_acc;
		} else
			save_func = smc_job_update;
	}

	if (save_func(this) != SM_OK)
		return sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, "Failed to save job object to database");

	if (this->previous_state != this->state)
		smc_job_process_state_change(this);

	this->modified = false;
	this->error = sm_ret_code_create(SMC_JOB_INTERNAL_FAULT, "Cannot save the same job object twice. You need to fetch a new one.");

	return sm_ret_code_create(SMC_JOB_OK, NULL);
}

int
smc_job_count(void)
{
	double result = 0;

	SM_DEBUG_ENTER();

	odbc->begin_transaction(odbc);

	SM_TRY_NOT_OK(count_stmt->execute(count_stmt, NULL));
	SM_TRY_NOT_OK(count_stmt->fetch_row(count_stmt));
	SM_TRY_NOT_OK(count_stmt->get_double(count_stmt, 1, &result));

	count_stmt->close_cursor(count_stmt);
	odbc->commit(odbc);

	SM_ASSERT(result <= INT_MAX, "smc_job_count needs refactoring ;-)");

	return (int) result;

fault:
	count_stmt->close_cursor(count_stmt);
	odbc->rollback(odbc);
	return -1;
}

#ifdef TEST
static void
test_prepare(void)
{
	sm_odbc_stmt_t stmt;

	SM_DEBUG_ENTER();

	assert(odbc->begin_transaction(odbc) == SM_OK);

	stmt = sm_odbc_stmt_create(odbc, "DELETE FROM jobs");
	assert(stmt);
	assert(stmt->execute(stmt, NULL) == SM_OK);
	sm_odbc_stmt_free(&stmt);

	stmt = sm_odbc_stmt_create(odbc, "DELETE FROM jobs_acc");
	assert(stmt);
	assert(stmt->execute(stmt, NULL) == SM_OK);
	sm_odbc_stmt_free(&stmt);

	assert(odbc->commit(odbc) == SM_OK);
}

static const char *jsdl = "<jsdl:JobDefinition xmlns:jsdl=\"http://schemas.ggf.org/jsdl/2005/11/jsdl\" \
xmlns:jsdl-qcg-comp-ares=\"http://schemas.qoscosgrid.org/comp/2011/04/jsdl/ares\"><jsdl:JobDescription>\
<jsdl:Application><jsdl-hpcpa:HPCProfileApplication \
xmlns:jsdl-hpcpa=\"http://schemas.ggf.org/jsdl/2006/07/jsdl-hpcpa\"><jsdl-hpcpa:Executable>/bin/date</jsdl-hpcpa:Executable>\
<jsdl-hpcpa:UserName>root</jsdl-hpcpa:UserName>\
</jsdl-hpcpa:HPCProfileApplication></jsdl:Application>\
<jsdl:Resources><jsdl-qcg-comp-ares:Reservation><jsdl-qcg-comp-ares:ReservationIdentifier xmlns:wsa5=\"http://www.w3.org/2005/08/addressing\">\
<wsa5:Address>http://localhost:19000/</wsa5:Address><wsa5:ReferenceParameters>\
<qcg-comp-ares:ReservationId xmlns:qcg-comp-ares=\"http://schemas.qoscosgrid.org/comp/2011/04/ares\">\
urn:uuid:9b7183a3-420e-45b7-863e-915dc380594d</qcg-comp-ares:ReservationId></wsa5:ReferenceParameters>\
</jsdl-qcg-comp-ares:ReservationIdentifier></jsdl-qcg-comp-ares:Reservation></jsdl:Resources></jsdl:JobDescription></jsdl:JobDefinition>";

static const char *jsdl_wo_username = "<jsdl:JobDefinition xmlns:jsdl=\"http://schemas.ggf.org/jsdl/2005/11/jsdl\" \
xmlns:jsdl-qcg-comp-ares=\"http://schemas.qoscosgrid.org/comp/2011/04/jsdl/ares\"><jsdl:JobDescription>\
<jsdl:Application><jsdl-hpcpa:HPCProfileApplication \
xmlns:jsdl-hpcpa=\"http://schemas.ggf.org/jsdl/2006/07/jsdl-hpcpa\"><jsdl-hpcpa:Executable>/bin/date</jsdl-hpcpa:Executable>\
</jsdl-hpcpa:HPCProfileApplication></jsdl:Application>\
<jsdl:Resources><jsdl-qcg-comp-ares:Reservation><jsdl-qcg-comp-ares:ReservationIdentifier xmlns:wsa5=\"http://www.w3.org/2005/08/addressing\">\
<wsa5:Address>http://localhost:19000/</wsa5:Address><wsa5:ReferenceParameters>\
<qcg-comp-ares:ReservationId xmlns:qcg-comp-ares=\"http://schemas.qoscosgrid.org/comp/2011/04/ares\">\
urn:uuid:9b7183a3-420e-45b7-863e-915dc380594d</qcg-comp-ares:ReservationId></wsa5:ReferenceParameters>\
</jsdl-qcg-comp-ares:ReservationIdentifier></jsdl-qcg-comp-ares:Reservation></jsdl:Resources></jsdl:JobDescription></jsdl:JobDefinition>";

static void
test_minimal_save_and_find(void)
{
	smc_job_t job = NULL;
	char *tmp = NULL;
	sm_ret_code_t ret;
	sm_list_t result = NULL;
	time_t now;
	struct tm now_tm;
	struct tm db_tm;

	SM_DEBUG_ENTER();

	now = time(NULL);
	now_tm = *localtime(&now);

	assert(smc_job_count() == 0);

	ret = smc_job_get_active_readonly(&result);
	assert(!ret.code);
	assert(sm_list_count(result) == 0);

	sm_debug("Creating");
	job = smc_job_create();

	job->set_id(job, (tmp = smc_generate_id(NULL, NULL)));
	job->set_drms_id(job, "100");
	job->set_submitted_jsdl(job, jsdl);
	job->set_jsdl(job, jsdl);
	job->set_subject(job, "subject");
	job->set_mapped_user(job, "root");
	job->set_state(job, SMC_QUEUED);
	job->set_start_time(job, now_tm);
	job->set_exit_status(job, 100);
	job->set_core_dumped(job, true);
	job->set_posix_signal_name(job, "SIGKILL");
	job->set_cpu_usage(job, 5);
	job->set_mem_usage(job, 10);
	job->set_vmem_usage(job, 15);
	job->set_queue(job, "default");
	job->set_project(job, "Manhattan");
	job->set_walltime(job, 120);
	job->set_nodes(job, "hosta:1|hostb:2");

	assert(!strcmp("9b7183a3-420e-45b7-863e-915dc380594d", job->get_reservation_id(job)));

	sm_debug("Saving");
	job->save(job);
	assert(smc_job_count() == 1);
	smc_job_free(&job);

	/* Get for potential modifications. */
	sm_debug("Getting");
	ret = smc_job_get(tmp, &result);
	assert(!ret.code);
	assert(sm_list_count(result) == 1);

	job = (smc_job_t) result->data;
	assert(!strcmp(tmp, job->get_id(job)));
	assert(!strcmp("100", job->get_drms_id(job)));
	assert(!strcmp(jsdl, job->get_submitted_jsdl(job)));
	/* JSDL is modified so we cannot do a strcmp() here. */
	assert(!strcmp("subject", job->get_subject(job)));
	assert(!strcmp("root", job->get_mapped_user(job)));
	assert(job->get_state(job) == SMC_QUEUED);
	assert(job->get_exit_status(job) == 100);
	assert(job->get_core_dumped(job) == true);
	assert(!strcmp("SIGKILL", job->get_posix_signal_name(job)));
	assert(job->get_cpu_usage(job) == 5);
	assert(job->get_mem_usage(job) == 10);
	assert(job->get_vmem_usage(job) == 15);
	assert(!strcmp(job->get_queue(job), "default"));
	assert(!strcmp(job->get_project(job), "Manhattan"));
	assert(job->get_walltime(job) == 120);
	assert(!strcmp(job->get_nodes(job), "hosta:1|hostb:2"));
	assert(!strcmp("9b7183a3-420e-45b7-863e-915dc380594d", job->get_reservation_id(job)));

	/* This should do nothing. */
	sm_debug("Saving after no modifications");
	job->save(job);
	assert(smc_job_count() == 1);

	sm_list_free_user(result, (sm_list_free_user_function) smc_job_free);

	/* Get as readonly. */
	sm_debug("Getting read only");
	ret = smc_job_get_by_drms_id_readonly("100", &job);
	assert(!ret.code);

	assert(!strcmp(tmp, job->get_id(job)));
	assert(!strcmp("100", job->get_drms_id(job)));
	assert(!strcmp(jsdl, job->get_submitted_jsdl(job)));
	/* JSDL is modified so we cannot do a strcmp() here. */
	assert(!strcmp("subject", job->get_subject(job)));
	assert(!strcmp("root", job->get_mapped_user(job)));
	assert(job->get_state(job) == SMC_QUEUED);

	db_tm = job->get_start_time(job);
	assert(mktime(&db_tm) == now);

	assert(job->has_exit_status(job));
	assert(job->get_exit_status(job) == 100);
	assert(job->has_core_dumped(job));
	assert(job->get_core_dumped(job) == true);
	assert(!strcmp("SIGKILL", job->get_posix_signal_name(job)));
	assert(job->get_cpu_usage(job) == 5);
	assert(job->get_mem_usage(job) == 10);
	assert(job->get_vmem_usage(job) == 15);
	assert(!strcmp(job->get_queue(job), "default"));
	assert(!strcmp(job->get_project(job), "Manhattan"));
	assert(job->get_walltime(job) == 120);
	assert(!strcmp(job->get_nodes(job), "hosta:1|hostb:2"));
	assert(!strcmp("9b7183a3-420e-45b7-863e-915dc380594d", job->get_reservation_id(job)));

	smc_job_free(&job);
	assert(smc_job_count() == 1);

	/* Get active readonly. */
	sm_debug("Getting active");
	ret = smc_job_get_active_readonly(&result);
	assert(!ret.code);
	assert(sm_list_count(result) == 1);

	job = (smc_job_t) result->data;

	assert(!strcmp(tmp, job->get_id(job)));
	assert(!strcmp("100", job->get_drms_id(job)));
	assert(!strcmp(jsdl, job->get_submitted_jsdl(job)));
	/* JSDL is modified so we cannot do a strcmp() here. */
	assert(!strcmp("subject", job->get_subject(job)));
	assert(!strcmp("root", job->get_mapped_user(job)));
	assert(job->get_state(job) == SMC_QUEUED);

	db_tm = job->get_start_time(job);
	assert(mktime(&db_tm) == now);

	assert(job->get_exit_status(job) == 100);
	assert(job->get_core_dumped(job) == true);
	assert(!strcmp("SIGKILL", job->get_posix_signal_name(job)));
	assert(job->get_cpu_usage(job) == 5);
	assert(job->get_mem_usage(job) == 10);
	assert(job->get_vmem_usage(job) == 15);
	assert(!strcmp(job->get_queue(job), "default"));
	assert(!strcmp(job->get_project(job), "Manhattan"));
	assert(job->get_walltime(job) == 120);
	assert(!strcmp(job->get_nodes(job), "hosta:1|hostb:2"));
	assert(!strcmp("9b7183a3-420e-45b7-863e-915dc380594d", job->get_reservation_id(job)));

	sm_list_free_user(result, (sm_list_free_user_function) smc_job_free);
	assert(smc_job_count() == 1);

	SM_FREE(tmp);
}

static void
test_add_username_to_jsdl(void)
{
	smc_job_t job = NULL;
	char *tmp = NULL;
	sm_ret_code_t ret;
	sm_list_t result = NULL;

	SM_DEBUG_ENTER();

	job = smc_job_create();

	job->set_id(job, (tmp = smc_generate_id(NULL, NULL)));
	job->set_drms_id(job, "100");
	job->set_submitted_jsdl(job, jsdl_wo_username);
	job->set_jsdl(job, jsdl_wo_username);
	job->set_subject(job, "subject");
	job->set_mapped_user(job, "root");
	job->set_state(job, SMC_QUEUED);

	assert(strstr(job->get_jsdl(job), "UserName"));
	assert(strstr(job->get_jsdl(job), "root"));

	job->save(job);
	smc_job_free(&job);

	ret = smc_job_get_readonly(tmp, &result);
	assert(!ret.code);
	assert(sm_list_count(result) == 1);

	job = (smc_job_t) result->data;
	assert(!strcmp(tmp, job->get_id(job)));

	assert(strstr(job->get_jsdl(job), "UserName"));
	assert(strstr(job->get_jsdl(job), "root"));
	assert(!job->has_exit_status(job));
	assert(!job->has_core_dumped(job));

	sm_list_free_user(result, (sm_list_free_user_function) smc_job_free);

	SM_FREE(tmp);
}

static void
test_accounting(void)
{
	smc_job_state_t states[] = { SMC_FINISHED, SMC_FAILED, SMC_CANCELLED };
	size_t i;

	SM_DEBUG_ENTER();

	for (i = 0; i < sizeof(states) / sizeof(states[0]); i++) {
		smc_job_t job;
		char *tmp;
		char drms_id[128];
		sm_ret_code_t ret;

		job = smc_job_create();

		job->set_id(job, (tmp = smc_generate_id(NULL, NULL)));
		snprintf(drms_id, sizeof(drms_id), "%d", i);
		job->set_drms_id(job, drms_id);
		job->set_submitted_jsdl(job, jsdl);
		job->set_jsdl(job, jsdl);
		job->set_subject(job, "subject");
		job->set_mapped_user(job, "root");
		job->set_state(job, SMC_QUEUED);
		job->set_nodes(job, "hosta:1|hostb:2");

		job->save(job);
		smc_job_free(&job);

		ret = smc_job_get_by_drms_id(drms_id, &job);
		assert(!ret.code);

		assert(!strcmp(tmp, job->get_id(job)));

		job->set_state(job, states[i]);

		job->save(job);
		smc_job_free(&job);

		/* Cannot get the job for modification once it has been put into accounting table. */
		ret = smc_job_get_by_drms_id(drms_id, &job);
		assert(ret.code == SMC_JOB_NOT_FOUND);

		/* Getting the job object for reading should work on the other hand. */
		ret = smc_job_get_by_drms_id_readonly(drms_id, &job);
		assert(!ret.code);

		assert(!strcmp(tmp, job->get_id(job)));
		assert(job->get_state(job) == states[i]);
		assert(!strcmp(job->get_nodes(job), "hosta:1|hostb:2"));

		smc_job_free(&job);

		SM_FREE(tmp);
	}
}

int
main()
{
	sm_odbc_handle_t db;

	assert(sm_library_setup() == SM_OK);
	assert(sm_log_init_str(NULL, "DEBUG") == SM_OK);

	db = sm_odbc_connect("qcg_comp_test", "qcg_comp", "qcg_comp");
	if (!db) {
		sm_notice("Could not connect to database. Prepare the test database first.");
		sleep(1);
		return 0;
	}

	smc_job_setup(db, false);

	test_prepare();
	test_minimal_save_and_find();

	test_prepare();
	test_add_username_to_jsdl();

	test_prepare();
	test_accounting();

	smc_job_cleanup();
	sm_odbc_disconnect(&db);
	sm_library_cleanup();

	return 0;
}
#endif
