/* $Id: qcg_comp_ares.c 145 2010-07-15 13:37:19Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <limits.h>

#include <qcg-core/sm.h>
#include <qcg-core/access.h>
#include <qcg-core/logger.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>

#include <ardl.h>
#include <common.h>
#include <faults.h>
#include <icomm.h>
#include <identifiers.h>
#include <rsvn.h>
#include <user.h>
#include <include/ardl_filter.h>
#include <stub/qcgcompStub.h>
#include <stub/qcgcompH.h>

__SM_ID("$Id: qcg_comp_ares.c 145 2010-07-15 13:37:19Z mamonski $");

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "qcg_comp_ares"
	}
};

extern struct Namespace qcgcomp_namespaces[];
extern const char *smc_reservations_administrator;
extern uid_t smc_reservations_administrator_uid;

static sm_status_ret_code_t
get_rsvn_from_epr(sm_transport_t transport, sm_transport_fault_t fault, struct wsa5__EndpointReferenceType *ReservationIdentifier, bool readonly, smc_rsvn_t *result)
{
	const char *id = NULL;
	sm_ret_code_t ret;
	bool return_neg;

	SM_MANDATORY_P(transport);
	SM_MANDATORY_P(ReservationIdentifier);
	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER();

	return_neg = false;
	*result = NULL;

	if (smc_parse_reservation_epr(ReservationIdentifier, &id) != SM_OK)
		THROW_INTERNAL_FAULT("Cannot parse reservation EPR");

	if (readonly)
		ret = smc_rsvn_get_readonly(id, result);
	else
		ret = smc_rsvn_get(id, result);

	if (ret.code == SMC_RSVN_NOT_FOUND) {
		return_neg = true;
		THROW_APP_FAULT(qcg_comp_ares_UnknownReservationIdentifierFault, id);
	} else if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	return SM_OK;

fault:
	if (return_neg)
		return SM_NEG;
	else
		return SM_FAULT;
}

static void
kill_rsvn(uid_t uid, const char *drms_id)
{
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	if (!drms_id)
		return;

	SM_DEBUG_ENTER_STR(drms_id);

	sm_notice("Killing reservation %s due to a severe processing error", drms_id);

	params = sm_value_create(1, SM_VALUE_STRING, drms_id);
	if (!params)
		goto fault;

	if (smc_icomm_request_TerminateReservation(uid, params, &result) != SM_OK)
		goto fault;

	if (!SM_VALUE_RET_CODE_OK(result[0]))
		goto fault;

	sm_value_free(&params);
	sm_value_free(&result);
	return;

fault:
	sm_value_free(&params);
	sm_value_free(&result);
	sm_crit("Failed to kill reservation %s. Reservation orphaned.", drms_id);
	return;
}

int
__qcg_comp_ares__CreateReservation(
	struct soap *soap,
	struct qcg_comp_ares__CreateReservationType *qcg_comp_ares__CreateReservation,
	struct qcg_comp_ares__CreateReservationResponseType *qcg_comp_ares__CreateReservationResponse
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	sm_ret_code_t ret;
	_ardl__ReservationDefinition *ardl = NULL;
	const char *ardl_str = NULL;
	const char *ardl_str_after_filters = NULL;
	sm_alloc_tracked_t tracked = NULL;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	const char *subject = NULL;
	char *username = NULL;
	size_t id_idx = 0;
	uid_t uid;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "CreateReservation", fault) != SM_OK)
		goto fault;

	if (transport->access_check(transport, SM_TRUE, "CreateReservation", fault) != SM_ACCESS_GRANTED)
		goto fault;

	ardl = qcg_comp_ares__CreateReservation->qcg_comp_ares__ReservationDocument->ardl__ReservationDefinition;

	ret = smc_user_handle_submission(smc_ardl_get_username(ardl), transport->atc_token, &username, &uid);
	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	SM_TRY_NOT(ardl_str = smc_ardl_str(ardl, (smc_strdup_function) soap_strdup, soap));

	if (strlen(ardl_str) > SMC_MAX_ARDL_LENGTH)
		THROW_INTERNAL_FAULT("ARDL size exceeds predefined limit");

	ret = smc_ardl_handle_filters(ardl_str, &ardl_str_after_filters, (smc_strdup_function) soap_strdup, soap);

	switch (ret.code) {
		case ARDL_FILTER_OK:
			SM_TRY_NOT(tracked = smc_ardl_from_str(ardl_str_after_filters, ardl));
			break;
		default:
			THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));
			break;
	}

	if (smc_reservations_administrator) {
		if (smc_ardl_get_username(ardl) == NULL) {
			SM_TRY_FAULT(smc_ardl_set_username(ardl, username, (smc_malloc_function)soap_malloc, (smc_strdup_function)soap_strdup, soap));
			SM_TRY_NOT(ardl_str_after_filters = smc_ardl_str(ardl, (smc_strdup_function) soap_strdup, soap));
		}
		uid = smc_reservations_administrator_uid;
	}

	subject = transport->atc_token->get_last_oneline(transport->atc_token);

	SM_TRY_NOT(params = sm_value_create(4,
		SM_VALUE_STRING, subject,
		SM_VALUE_STRING, username,
		SM_VALUE_STRING, ardl_str,
		SM_VALUE_STRING, ardl_str_after_filters));

	if (smc_icomm_request_CreateReservation(uid, params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	id_idx = sm_value_count(result) - 1;

	if (!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT_FROM_VAL(result);

	assert(SM_VALUE_IS_STRING(result[sm_value_count(result) - 1]));
	if (smc_create_reservation_epr(transport, &qcg_comp_ares__CreateReservationResponse->ReservationIdentifier, result[id_idx]->vstring) != SM_OK) {
		kill_rsvn(uid, result[1]->vstring);
		THROW_INTERNAL_FAULT("Reservation created but failed to create EPR");
	}

	sm_info("Created reservation: id = %s (drms_id = %s)", result[id_idx]->vstring, result[1]->vstring);

	sm_alloc_tracked_free(&tracked);
	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->exit(transport, NULL);

fault:
	sm_alloc_tracked_free(&tracked);
	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->fault(transport, fault);
}

int
__qcg_comp_ares__TerminateReservation(
	struct soap *soap,
	struct qcg_comp_ares__TerminateReservationType *qcg_comp_ares__TerminateReservation,
	struct qcg_comp_ares__TerminateReservationResponseType *qcg_comp_ares__TerminateReservationResponse sm_unused
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	smc_rsvn_t rsvn = NULL;
	sm_status_ret_code_t status;
	sm_ret_code_t ret;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	uid_t uid;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "TerminateReservation", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_FALSE, "TerminateReservation", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	SM_TRY_NOT_OK(get_rsvn_from_epr(transport, fault, &qcg_comp_ares__TerminateReservation->ReservationIdentifier, false, &rsvn));

	SM_TRY_FAULT(status = smc_user_authorize(SMC_USER_AUTH_TERMINATE, rsvn->get_subject(rsvn), transport->atc_token->get_last_oneline(transport->atc_token)));
	if (status == SM_NEG)
		THROW_APP_FAULT(qcg_comp_ares_NotAuthorizedFault, "You are not authorized to terminate this reservation");

	if (!smc_reservations_administrator) {
		ret = smc_user_get_uid(rsvn->get_mapped_user(rsvn), &uid);
		if (ret.code)
			THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));
	} else {
		uid = smc_reservations_administrator_uid;
	}

	rsvn->terminate(rsvn);

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, rsvn->get_drms_id(rsvn)));

	if (smc_icomm_request_TerminateReservation(uid, params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if (SM_VALUE_RET_CODE_C(result[0]) == qcg_comp_ares_UnknownReservationIdentifierFault)
		THROW_APP_FAULT(qcg_comp_ares_UnknownReservationIdentifierFault, rsvn->get_id(rsvn));

	if (!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT_FROM_VAL(result);

	ret = rsvn->save(rsvn);
	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	sm_info("Terminated reservation: id = %s (drms_id = %s)", rsvn->get_id(rsvn), rsvn->get_drms_id(rsvn));

	sm_value_free(&params);
	sm_value_free(&result);
	smc_rsvn_free(&rsvn);
	return transport->exit(transport, NULL);

fault:
	sm_value_free(&params);
	sm_value_free(&result);
	smc_rsvn_free(&rsvn);
	return transport->fault(transport, fault);
}

int
__qcg_comp_ares__GetReservationDocument(
	struct soap *soap,
	struct qcg_comp_ares__GetReservationDocumentType *qcg_comp_ares__GetReservationDocument,
	struct qcg_comp_ares__GetReservationDocumentResponseType *response
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	smc_rsvn_t rsvn = NULL;
	sm_status_ret_code_t status;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetReservationDocument", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_FALSE, "GetReservationDocument", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	SM_TRY_NOT_OK(get_rsvn_from_epr(transport, fault, &qcg_comp_ares__GetReservationDocument->ReservationIdentifier, true, &rsvn));

	SM_TRY_FAULT(status = smc_user_authorize(SMC_USER_AUTH_GET_DOCUMENT, rsvn->get_subject(rsvn), transport->atc_token->get_last_oneline(transport->atc_token)));
	if (status == SM_NEG)
		THROW_APP_FAULT(qcg_comp_ares_NotAuthorizedFault, "You are not authorized to get document of this reservation");

	SM_TRY_NOT(response->qcg_comp_ares__ReservationDocument = transport->transport_malloc(transport, sizeof(*response->qcg_comp_ares__ReservationDocument)));
	SM_TRY_NOT(response->qcg_comp_ares__ReservationDocument->ardl__ReservationDefinition = transport->transport_malloc(transport, sizeof(*response->qcg_comp_ares__ReservationDocument->ardl__ReservationDefinition)));

	if (sm_gsoap_buffer_get_alloc_soap(rsvn->get_ardl(rsvn), response->qcg_comp_ares__ReservationDocument->ardl__ReservationDefinition,
		ardl__ARDLReservationDefinition_USCOREType, "ardl:ReservationDefinition", qcgcomp_namespaces, soap) != SM_OK) {

		THROW_INTERNAL_FAULT("Unable to parse ARDL document from database");
	}

	SM_TRY_NOT(response->qcg_comp_ares__LocalReservationId = transport->transport_strdup(transport, rsvn->get_drms_id(rsvn)));

	smc_rsvn_free(&rsvn);
	return transport->exit(transport, NULL);

fault:
	smc_rsvn_free(&rsvn);
	return transport->fault(transport, fault);
}

int
__qcg_comp_ares__GetActiveReservations(
	struct soap *soap,
	struct qcg_comp_ares__GetActiveReservationsType *qcg_comp_ares__GetActiveReservations sm_unused,
	struct qcg_comp_ares__GetActiveReservationsResponseType *qcg_comp_ares__GetActiveReservationsResponse
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	sm_value_t *ids = NULL;
	size_t i, count;
	smc_rsvn_t rsvn = NULL;
	sm_ret_code_t ret;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetActiveReservations", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_FALSE, "GetActiveReservations", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	params = SM_VALUE_NONE_SINGLETON;

	if (smc_icomm_request_GetActiveReservations( (smc_reservations_administrator ? smc_reservations_administrator_uid : getuid()), params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if (!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT_FROM_VAL(result);

	count = sm_value_count(result);

	for (i = 1; i < count; i++) {
		ret = smc_rsvn_get_by_drms_id_readonly(result[i]->vstring, &rsvn);

		if (ret.code) {
			if (ret.code == SMC_RSVN_NOT_FOUND)
				continue;

			THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));
		}

		SM_TRY_NOT(sm_value_append(&ids, 1, SM_VALUE_STRING, rsvn->get_id(rsvn)));
		smc_rsvn_free(&rsvn);
	}

	count = sm_value_count(ids);

	/* Just for fun ;-) */
	if (((size_t) -1) / sizeof(struct wsa5__EndpointReferenceType) < count)
		THROW_INTERNAL_FAULT("Out of memory");

	SM_TRY_NOT(qcg_comp_ares__GetActiveReservationsResponse->ReservationIdentifier = transport->transport_malloc(transport, count * sizeof(struct wsa5__EndpointReferenceType)));
	qcg_comp_ares__GetActiveReservationsResponse->__sizeReservationIdentifier = count;

	for (i = 0; i < count; i++)
		SM_TRY_NOT_OK(smc_create_reservation_epr(transport, &qcg_comp_ares__GetActiveReservationsResponse->ReservationIdentifier[i], ids[i]->vstring));

	sm_value_free(&params);
	sm_value_free(&result);
	sm_value_free(&ids);
	return transport->exit(transport, NULL);

fault:
	smc_rsvn_free(&rsvn);
	sm_value_free(&params);
	sm_value_free(&result);
	sm_value_free(&ids);
	return transport->fault(transport, fault);
}

int
__qcg_comp_ares__GetReservationStatus(
		struct soap* soap,
		struct qcg_comp_ares__GetReservationStatusType *qcg_comp_ares__GetReservationStatus,
		struct qcg_comp_ares__GetReservationStatusResponseType *qcg_comp_ares__GetReservationStatusResponse)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	sm_value_t *activities = NULL;
	sm_list_t jobs = NULL;
	smc_rsvn_t rsvn = NULL;
	sm_list_t l;
	size_t i;
	size_t activities_count, reserved_hosts_count;
	sm_value_t *reserved_hosts = NULL;
	uid_t uid;
	sm_status_ret_code_t status;
	sm_ret_code_t ret;
	struct qcg_comp_ares__ARESReservationStatus_USCOREType *res;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetReservationStatus", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_FALSE, "GetReservationStatus", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	SM_TRY_NOT_OK(get_rsvn_from_epr(transport, fault, &qcg_comp_ares__GetReservationStatus->ReservationIdentifier, true, &rsvn));

	SM_TRY_FAULT(status = smc_user_authorize(SMC_USER_AUTH_GET_STATUS, rsvn->get_subject(rsvn), transport->atc_token->get_last_oneline(transport->atc_token)));
	if (status == SM_NEG)
		THROW_APP_FAULT(qcg_comp_ares_NotAuthorizedFault, "You are not authorized to get status of this reservation");

	ret = smc_user_get_uid(rsvn->get_mapped_user(rsvn), &uid);
	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, rsvn->get_drms_id(rsvn)));

	if (smc_icomm_request_GetReservationStatus((smc_reservations_administrator ? smc_reservations_administrator_uid : uid) , params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if (SM_VALUE_RET_CODE_C(result[0]) == qcg_comp_ares_UnknownReservationIdentifierFault)
		THROW_APP_FAULT(qcg_comp_ares_UnknownReservationIdentifierFault, rsvn->get_id(rsvn));

	if (!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT_FROM_VAL(result);

	SM_TRY_NOT(qcg_comp_ares__GetReservationStatusResponse->qcg_comp_ares__ReservationStatus = transport->transport_malloc(transport, sizeof(*res)));
	res = qcg_comp_ares__GetReservationStatusResponse->qcg_comp_ares__ReservationStatus;

	ret = smc_job_get_active_readonly(&jobs);
	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	for (l = jobs; l; l = l->next) {
		smc_job_t j = l->data;

		if (!sm_strcmp(j->get_reservation_id(j), rsvn->get_id(rsvn)))
			SM_TRY_NOT(sm_value_append(&activities, 1, SM_VALUE_STRING, j->get_id(j)));
	}

	activities_count = sm_value_count(activities);

	if (activities_count > 0) {

		SM_TRY_NOT(res->AssociatedActivities = transport->transport_malloc(transport, sizeof(struct qcg_comp_ares__ARESAssociatedActivities_USCOREType)));
		SM_TRY_NOT(res->AssociatedActivities->ActivityIdentifier = transport->transport_malloc(transport, activities_count * sizeof(struct wsa5__EndpointReferenceType)));
		res->AssociatedActivities->__sizeActivityIdentifier = activities_count;
		for (i = 0; i < activities_count; i++)
			SM_TRY_NOT_OK(smc_create_activity_epr(transport, &res->AssociatedActivities->ActivityIdentifier[i], activities[i]->vstring));
	} else {
		res->AssociatedActivities = NULL;
	}

	reserved_hosts = result[1]->varray;
	reserved_hosts_count = sm_value_count(reserved_hosts);

	if (!reserved_hosts_count)
		THROW_INTERNAL_FAULT("Reservation module didn't return a list of reserved hosts");

	SM_TRY_NOT(res->ReservedHost = transport->transport_malloc(transport, reserved_hosts_count * sizeof(struct qcg_comp_ares__ARESReservedHost_USCOREType)));
	res->__sizeReservedHost = reserved_hosts_count;

	for (i = 0; i < reserved_hosts_count; i++) {
		char *hostname = NULL;
		uint64_t *reserved_slots = NULL;
		uint64_t *total_slots = NULL;
		uint64_t *used_slots = NULL;
		char *ctxt = NULL;

		SM_TRY_NOT(reserved_slots = soap_malloc(soap, sizeof(*reserved_slots)));
		SM_TRY_NOT(total_slots = soap_malloc(soap, sizeof(*total_slots)));
		SM_TRY_NOT(used_slots = soap_malloc(soap, sizeof(*used_slots)));

		hostname = strtok_r(reserved_hosts[i]->vstring, ":", &ctxt);
		*total_slots = atoi(strtok_r(NULL, ":", &ctxt));
		*reserved_slots = atoi(strtok_r(NULL, ":", &ctxt));
		*used_slots = atoi(strtok_r(NULL, ":", &ctxt));

		SM_TRY_NOT(hostname = soap_strdup(soap, hostname));

		sm_debug("reserved_host[%zu] = %s:%d:%d:%d", i, hostname, (int)*total_slots, (int)*reserved_slots, (int)*used_slots);

		res->ReservedHost[i].HostName = hostname;
		res->ReservedHost[i].ReservedSlotsCount = reserved_slots;
		res->ReservedHost[i].TotalSlotsCount = total_slots;
		res->ReservedHost[i].UsedSlotsCount = used_slots;
	}

	SM_LIST_FREE_USER(jobs, smc_job_free);
	smc_rsvn_free(&rsvn);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->exit(transport, NULL);

fault:
	SM_LIST_FREE_USER(jobs, smc_job_free);
	smc_rsvn_free(&rsvn);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->fault(transport, fault);
}
