#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include <qcg-core/logger.h>
#include <qcg-core/sm.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>
#include <faults.h>
#include <user.h>
#include <icomm.h>

#ifdef WITH_LIBRSYNC
#  define SERVER
#  include <rsync/types.h>
#  include <rsync/interface.h>
#  include <rsync/attachment.h>
#  include <qcg-core/privsep.h>
#  define min(a,b) ((a)<(b)?(a):(b))
#endif

#include <stub/qcgcompStub.h>


extern struct Namespace qcgcomp_namespaces[];

SM_INTERNAL_GENERAL_MODULE = {
        .header = {
                .name = "qcg-comp-rsync"
        }
};

int
__qcg_comp_rsync__GetFiles(
		struct soap* soap,
		struct qcg_comp_rsync__GetFilesType *request,
		struct qcg_comp_rsync__GetFilesResponseType *response)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_ret_code_t ret;
	char *username = NULL;
	uid_t uid;
	const char *basedir = request->Common->BaseDir;
	sm_transport_t transport = soap->user;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetFiles", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "GetFiles", fault)
			!= SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

#ifndef WITH_LIBRSYNC
	sm_error("RSYNC port not supported. Recompile QCG Comp with librsync.");
	RSYNC_CHECK(RSyncInternalError);
#else
	ret = smc_user_handle_submission(request->Common->UserName, transport->atc_token,
			&username, &uid);

	sm_info("GetFiles: basedir='%s'", basedir);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, basedir));
	if (smc_icomm_request_GetFiles(uid, params, &result) != SM_OK)
		goto fault;

	if(!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT(result[0]->vretcode.code, transport->transport_strdup(transport, result[0]->vretcode.string));

	SM_TRY_NOT(response->Files = transport->transport_malloc(transport,sizeof(struct qcg_comp_rsync__FilesType)));
	response->Files->__sizeFile = result[1]->vuint32;
	SM_TRY_NOT(response->Files->File = transport->transport_malloc(transport, sizeof(struct qcg_comp_rsync__FileType)*response->Files->__sizeFile));

	for(int i=0;i<response->Files->__sizeFile;++i) {
		response->Files->File[i].MTime=result[4*i+2]->vuint32;
		response->Files->File[i].isDirectory = result[4*i+1+2]->vbool;
		response->Files->File[i].isExecutable = result[4*i+2+2]->vbool;
		SM_TRY_NOT(response->Files->File[i].Path = transport->transport_strdup(transport, result[4*i+3+2]->vstring));
	}

	SM_FREE(username);
	SM_DEBUG_LEAVE_VOID();
	return transport->exit(transport, NULL);
#endif
fault:
	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	SM_DEBUG_LEAVE_VOID();
	return transport->fault(transport, fault);
}

int
__qcg_comp_rsync__GetSignature(struct soap* soap,
		struct qcg_comp_rsync__GetSignatureType *request,
		struct qcg_comp_rsync__GetSignatureResponseType *response)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_ret_code_t ret;
	char *username = NULL;
	uid_t uid;
	const char *basedir = request->Common->BaseDir;
	sm_transport_t transport = soap->user;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetSignature", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "GetSignature", fault)
			!= SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);
#ifndef WITH_LIBRSYNC
	sm_error("RSYNC port not supported. Recompile QCG Comp with librsync.");
	RSYNC_CHECK(RSyncInternalError);
#else
	ret = smc_user_handle_submission(request->Common->UserName, transport->atc_token,
			&username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	sm_info("GetSignature: basedir=%s, path=%s", basedir, request->Path);

	SM_TRY_NOT(params = sm_value_create(2,
			SM_VALUE_STRING, basedir,
			SM_VALUE_STRING, request->Path
			));

	if (smc_icomm_request_GetSignature(uid, params, &result) != SM_OK) {
		sm_debug("GetSignature request failed");
		goto fault;
	}

	if(!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, result[0]->vretcode.string));

	response->Signature=transport->transport_malloc(transport, sizeof(struct qcg_comp_rsync__SignatureType));

	if(!icomm_to_signature(transport, result+1, response->Signature))
		THROW_INTERNAL_FAULT("Insufficient memory in GetSignature in Rsync module");

	sm_value_free(&params);
	SM_FREE(username);
	SM_DEBUG_LEAVE_VOID();
	return transport->exit(transport, NULL);
#endif
fault:
	sm_value_free(&params);
	SM_FREE(username);
	SM_DEBUG_LEAVE_VOID();
	return transport->fault(transport, fault);
}

#ifdef WITH_LIBRSYNC
struct rsync_stream_data_s
{
	int fd;
	void *handle;

	char *buf;
	size_t buf_len;
	size_t buf_size;
};

static void* rsync_stream_data_in_open(struct sm_transport_s *transport sm_unused, const char *id sm_unused, void *handle)
{
	SM_DEBUG_ENTER();
	return SM_DEBUG_LEAVE_PTR(handle);
}

static int rsync_stream_data_in_write(struct sm_transport_s *transport sm_unused, void *handle, const char *buf, size_t len)
{
	struct rsync_stream_data_s *context=(struct rsync_stream_data_s*)handle;
	sm_value_t *params;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER();

	if(context->buf_size < len) {
		size_t extend = 1024*1024; //1MB

		if(extend < len)
			extend = len;

		context->buf_size += extend;
		context->buf = sm_realloc(context->buf, context->buf_size);

		sm_debug("Buffer size %lu", (unsigned long)context->buf_size);
	}

	if(context->buf_len + len < context->buf_size && len != 0) {
		memmove(context->buf + context->buf_len, buf, len);
		context->buf_len += len;
	}
	else {
		do {
			int error, status;

			sm_debug("Writing %lu bytes of delta", (unsigned long)context->buf_len);

			if(context->buf_len > 0) {
				SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_BYTES, context->buf_len, context->buf));
				context->buf_len = 0;
			}
			else
				SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_INT32, 0));

			error = smc_icomm_request_PutDeltaProcess(context->fd, params, &result);
			sm_value_free(&params);

			if(error != SM_OK)
				goto fault;

			sm_debug("PutDeltaProcess = %i", result[0]->vint32);

			error = result[0]->vint32;
			status = result[1]->vint32;

			sm_value_free(&result);

			if(error != RSyncNoError)
				goto fault;

			if(status == RS_DONE)
				break;
		}
		while(len == 0);

		if(len > 0) {
			memmove(context->buf, buf, len);
			context->buf_len = len;
		}
	}

	sm_debug("Buffer %lu/%lu", (unsigned long)context->buf_len, (unsigned long)context->buf_size);

quit:
	return SM_DEBUG_LEAVE_INT(SM_OK); //see stream_in_write in qcg_comp_staging.c for explanation
fault:
	goto quit;
}

static void rsync_stream_data_in_close(struct sm_transport_s *transport, void *handle)
{
	struct rsync_stream_data_s *context=(struct rsync_stream_data_s*)handle;

	SM_DEBUG_ENTER();

	rsync_stream_data_in_write(transport, handle, NULL, 0);	//flush buffered data

	sm_free(context->buf);

	smc_icomm_PutDeltaRelease(&context->fd,&context->handle);

	SM_DEBUG_LEAVE_VOID();
}
#endif

int
__qcg_comp_rsync__PutDelta(
		struct soap* soap,
		struct qcg_comp_rsync__PutDeltaType *request,
		struct qcg_comp_rsync__PutDeltaResponseType *response sm_unused)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_ret_code_t ret;
	char *username =  NULL;
	uid_t uid;
	const char *basedir = request->Common->BaseDir;
	sm_transport_t transport = soap->user;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	struct rsync_stream_data_s *stream_context;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "PutDelta", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "PutDelta", fault)
			!= SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);
#ifndef WITH_LIBRSYNC
	sm_error("RSYNC port not supported. Recompile QCG Comp with librsync.");
	RSYNC_CHECK(RSyncInternalError);
#else
	ret = smc_user_handle_submission(request->Common->UserName, transport->atc_token,
			&username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	sm_info("PutDelta: basedir='%s' path='%s'", basedir, request->Path);

	SM_TRY_NOT(params = sm_value_create(2,
			SM_VALUE_STRING, basedir,
			SM_VALUE_STRING, request->Path
	));

	stream_context = transport->transport_malloc(transport, sizeof(struct rsync_stream_data_s));
	stream_context->buf = NULL;
	stream_context->buf_len = 0;
	stream_context->buf_size = 0;

	if (smc_icomm_request_PutDeltaBegin(uid, params, &result, &stream_context->fd, &stream_context->handle) != SM_OK)
		goto fault;

	if(!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT(result[0]->vretcode.code, transport->transport_strdup(transport, result[0]->vretcode.string));

	transport->stream_data_in_open = rsync_stream_data_in_open;
	transport->stream_data_in_write = rsync_stream_data_in_write;
	transport->stream_data_in_close = rsync_stream_data_in_close;

	transport->stream_data_in(transport, stream_context);

	sm_value_free(&result);
	sm_value_free(&params);
	SM_FREE(username);
	SM_DEBUG_LEAVE_VOID();
	return transport->exit(transport, NULL);
#endif
fault:
	sm_value_free(&result);
	sm_value_free(&params);
	SM_FREE(username);
	SM_DEBUG_LEAVE_VOID();
	return transport->fault(transport, fault);
}

#ifdef WITH_LIBRSYNC
static int rsync_stream_data_out_read(struct sm_transport_s *transport sm_unused, void *handle, char *buf, size_t len)
{
	struct rsync_stream_data_s *stream_params=(struct rsync_stream_data_s*)handle;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER();

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_UINT32, len));

	sm_debug("Requesting GetDeltaProcess");
	smc_icomm_request_GetDeltaProcess(stream_params->fd, params, &result);
	sm_debug("GetDeltaProcess: %lu", (unsigned long)sm_value_count(result));

	if(!SM_VALUE_RET_CODE_OK(result[0]))
		goto fault;

	if(sm_value_count(result)==2 && SM_VALUE_IS_BYTES(result[1])) {
		len = min (len, result[1]->size);
		memcpy(buf, result[1]->vbytes, len);
	}
	else
		len = 0;

	sm_value_free(&params);
	sm_value_free(&result);

	return SM_DEBUG_LEAVE_UINT((unsigned int)len);

fault:
	sm_value_free(&params);
	sm_value_free(&result);

	return SM_DEBUG_LEAVE_INT(-1);
}
static void* rsync_stream_data_out_open(struct sm_transport_s *transport sm_unused, const char *id sm_unused, void *handle)
{
	return handle;
}

static void rsync_stream_data_out_close(struct sm_transport_s *transport sm_unused, void *handle)
{
	struct rsync_stream_data_s *stream_params=(struct rsync_stream_data_s*)handle;

	SM_DEBUG_ENTER();
	smc_icomm_GetDeltaRelease(&stream_params->fd, &stream_params->handle);
	SM_DEBUG_LEAVE_VOID();
}
#endif

int
__qcg_comp_rsync__GetDelta(
		struct soap* soap,
		struct qcg_comp_rsync__GetDeltaType *request,
		struct qcg_comp_rsync__GetDeltaResponseType *response sm_unused)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_ret_code_t ret;
	char *username = NULL;
	uid_t uid;
	const char *basedir = request->Common->BaseDir;
	sm_transport_t transport = soap->user;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	struct rsync_stream_data_s *stream_context = NULL;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "GetDelta", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "GetDelta", fault)
			!= SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);
#ifndef WITH_LIBRSYNC
	sm_error("RSYNC port not supported. Recompile QCG Comp with librsync.");
	RSYNC_CHECK(RSyncInternalError);
#else
	ret = smc_user_handle_submission(request->Common->UserName, transport->atc_token,
			&username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	sm_info("GetDelta: basedir=%s, path=%s", basedir, request->Path);

	SM_TRY_NOT(params = sm_value_create(2, SM_VALUE_STRING, basedir, SM_VALUE_STRING, request->Path));

	SM_TRY_NOT(signature_to_icomm(transport,request->Signature,&params));

	{
		char *__ptr = sm_value_debug(params); \
		sm_debug("%s", __ptr); \
		SM_FREE(__ptr); \
	}

	SM_TRY_NOT(stream_context = transport->transport_malloc(transport, sizeof(struct rsync_stream_data_s)));

	if (smc_icomm_request_GetDeltaBegin(uid, params, &result, &stream_context->fd, &stream_context->handle) != SM_OK)
		goto fault;

	sm_info("GetDeltaBegin requested");
	if(!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, result[0]->vretcode.string));

	transport->stream_data_out_open = rsync_stream_data_out_open;
	transport->stream_data_out_read = rsync_stream_data_out_read;
	transport->stream_data_out_close = rsync_stream_data_out_close;

	transport->stream_data_out(transport, stream_context, 0, "delta");

	sm_value_free(&params);
	SM_FREE(username);
	SM_DEBUG_LEAVE_VOID();
	return transport->exit(transport, NULL);
#endif
fault:
	sm_value_free(&params);
	SM_FREE(username);
	SM_DEBUG_LEAVE_VOID();
	return transport->fault(transport, fault);
}
