/* $Id: qcg_comp_service.c 271 2010-08-16 13:52:15Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <pwd.h>

#include <qcg-core/sm.h>
#include <qcg-core/event_timer.h>
#include <qcg-core/logger.h>
#include <qcg-core/privsep.h>
#include <qcg-core/signal.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/atz_module.h>
#include <qcg-core/modules/ecm_service_module.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>

#include <db.h>
#include <faults.h>
#include <icomm.h>
#include <job.h>
#include <jsdl.h>
#include <ardl.h>
#include <rsvn.h>
#include <session_process.h>
#include <include/namespaces.h>
#include <include/extensions.h>
#include <stub/qcgcomp.nsmap>
#include <stub/qcgcompStub.h>
#include <include/submission_module.h>
#include <qcg_comp_socks.h>

#define DEFAULT_UNPRIVILEGED_USER	"nobody"

__SM_ID("$Id: qcg_comp_service.c 271 2010-08-16 13:52:15Z mamonski $");

static int qcg_comp_setup(sm_module_t this);
static void qcg_comp_cleanup(sm_module_t this);
static sm_value_t *reservation_vacuum(void *object sm_unused, sm_value_t *params sm_unused);

static const void *
qcg_comp_service_params(void)
{
	static struct ecm_gsoap_transport_params_s params = {
		.serve = qcgcomp_serve,
		.namespaces = qcgcomp_namespaces,
		.get_fault_class = smc_get_fault_class,
		.get_fault_msg = smc_get_fault_msg,
		.create_fault_detail = smc_create_fault_detail,
		.free_fault_data = smc_free_fault_data,
		.fault_code2s = smc_fault_code2s
	};

	return &params;
}

SM_ECM_SERVICE_MODULE = {

	.header = {
		.entry_size = SM_ECM_SERVICE_MODULE_ENTRY_SIZE,
		.api_type = SM_API_ECM_SERVICE,

		.name = "qcg-comp-service",
		.descr = "QCG Computing gSOAP bindings",

		.version_major = 0,
		.version_minor = 1,

		.setup = qcg_comp_setup,
		.cleanup = qcg_comp_cleanup
	},

	.params = qcg_comp_service_params,

	.have_wsa_header = SM_TRUE,
	.have_wsse_header = SM_TRUE,
};

/* WS-Addressing plugin */
#include <qcg-core/modules/ecm_gsoap/gsoap/wsaapi.c>

/* WS-Security plugin */
#include <qcg-core/modules/ecm_gsoap/gsoap/wsseapi.c>
#include <qcg-core/modules/ecm_gsoap/gsoap/smdevp.c>
#include <qcg-core/modules/ecm_gsoap/gsoap/threads.c>

static bool smc_setuid_enabled = true;
static unsigned int smc_reservation_vacuum_period = 300;

const char *smc_attributes_common_name = NULL;
const char *smc_attributes_long_description = NULL;
const char *smc_reservations_administrator = NULL;
uid_t smc_reservations_administrator_uid;
char *smc_attributes_local_resource_manager_type = NULL;
char *smc_factory_attributes = NULL;
char *smc_attributes_contained_resources = NULL;
SM_BOOL smc_return_list_of_activities = SM_TRUE;
SM_BOOL smc_return_list_of_contained_resources = SM_TRUE;
SM_BOOL smc_only_owner_can_terminate = SM_TRUE;
SM_BOOL smc_only_owner_can_get_status = SM_TRUE;
SM_BOOL smc_only_owner_can_get_document = SM_TRUE;
submission_module_t smc_sumbission_module = NULL;
sm_event_timer_subscription_t reservation_vacuum_subscription;
sm_atz_module_t smc_mapfile_module = NULL;

static int
qcg_comp_setup(sm_module_t this)
{
	const char *user = NULL;
	struct passwd *pwd = NULL;

	SM_DEBUG_ENTER();

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:UnprivilegedUser", &user) == SM_FAULT)
		goto fault;

	if (sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:SetuidEnabled", (SM_BOOL*) &smc_setuid_enabled) == SM_FAULT)
		goto fault;

	if (sm_conf_node_uint(this->conf_ctxt, SM_FALSE, "/conf:ReservationVacuumPeriod", &smc_reservation_vacuum_period) == SM_FAULT)
		goto fault;

	if (sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:ReturnListOfActivities", &smc_return_list_of_activities) == SM_FAULT)
		goto fault;

	if (sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:ReturnListOfContainedResources", &smc_return_list_of_contained_resources) == SM_FAULT)
		goto fault;

	if (sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:OnlyOwnerCanTerminate", &smc_only_owner_can_terminate) == SM_FAULT)
		goto fault;

	if (sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:OnlyOwnerCanGetStatus", &smc_only_owner_can_get_status) == SM_FAULT)
		goto fault;

	if (sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:OnlyOwnerCanGetDocument", &smc_only_owner_can_get_document) == SM_FAULT)
		goto fault;

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:ReservationsAdministrator", &smc_reservations_administrator) == SM_FAULT)
		goto fault;

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:FactoryAttributes/conf:CommonName", &smc_attributes_common_name) == SM_FAULT)
		goto fault;

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:FactoryAttributes/conf:LongDescription", &smc_attributes_long_description) == SM_FAULT)
		goto fault;

	if (sm_conf_node_strdup(this->conf_ctxt, SM_FALSE, "/conf:FactoryAttributes/conf:LocalResourceManagerType", &smc_attributes_local_resource_manager_type) == SM_FAULT)
		goto fault;

	if (sm_conf_node_raw(this->conf_ctxt, SM_FALSE, "/conf:FactoryAttributes/conf:ContainedResources", &smc_attributes_contained_resources) == SM_FAULT)
		goto fault;

	if (sm_conf_node_raw(this->conf_ctxt, SM_FALSE, "/conf:FactoryAttributes", &smc_factory_attributes) == SM_FAULT)
		goto fault;

	if (!user)
		user = DEFAULT_UNPRIVILEGED_USER;

	if (!(pwd = getpwnam(user))) {
		sm_perror("getpwnam");
		sm_error("Couldn't get uid/gid for user %s", user);
		goto fault;
	}

	if (sm_privsep_setup(pwd->pw_uid, smc_session_process, !smc_setuid_enabled) != SM_OK) {
		sm_error("Failed to set up privilege separation");
		goto fault;
	}

	if (smc_reservations_administrator) {
		if (!(pwd = getpwnam(smc_reservations_administrator))) {
			sm_perror("getpwnam");
			sm_error("Couldn't get uid/gid for user %s", smc_reservations_administrator);
			goto fault;
		}

		smc_reservations_administrator_uid = pwd->pw_uid;
	}

	/* Running as unprivileged user from now on. */

	if (smc_db_setup(this->conf_ctxt) != SM_OK) {
		sm_error("Failed to set up database module");
		goto fault;
	}

	if (smc_job_setup(smc_db_get_handle(), false) != SM_OK) {
		sm_error("Failed to set up job module");
		goto fault;
	}

	if (smc_rsvn_setup(smc_db_get_handle()) != SM_OK) {
		sm_error("Failed to set up rsvn module");
		goto fault;
	}

	if (smc_icomm_setup(this) != SM_OK) {
		sm_error("Failed to set up internal communication module");
		goto fault;
	}

	if (smc_jsdl_setup() != SM_OK) {
		sm_error("Failed to set up jsdl module");
		goto fault;
	}

	if (smc_ardl_setup() != SM_OK) {
		sm_error("Failed to set up ardl module");
		goto fault;
	}

	if (!(reservation_vacuum_subscription = sm_event_timer_subscribe(sm_closure_create(reservation_vacuum, NULL, NULL), smc_reservation_vacuum_period))) {
		sm_error("Failed to start reservation vacuum");
		goto fault;
	}

	/* HPCP BP Basic Filter extension is always supported. */
	smc_extensions_add(SMC_NAMESPACE_HPCP_BP_BASIC_FILTER);

	/* Advertise all of our extensions compared to plain BES service. */
	smc_extensions_add(SMC_NAMESPACE_SERVICE);

	/* Setup BES FactoryAttributes xsd:any type handler */
	if (sm_signal_register("smc_factory_attributes", 0, (sm_value_type_t []) { SM_VALUE_STRING }, 1, (sm_value_type_t []) { SM_VALUE_STRING }, 1) !=  SM_OK)
		goto fault;

	/* Setup BES FactoryAttributes ContainedResources type handler */
	if (sm_signal_register("smc_contained_resources", 0, (sm_value_type_t []) { SM_VALUE_STRING }, 1, (sm_value_type_t []) { SM_VALUE_STRING }, 1) !=  SM_OK)
		goto fault;

	if ((smc_sumbission_module = (submission_module_t)this->module_manager->find_by_api_first(this->module_manager, SUBMISSION_MODULE_API)) == NULL) {
		sm_error("Submission module not found");
		goto fault;
	}

	if (smc_socks_setup(this) == SM_FAULT) {
		sm_error("SOCKS server failed to start");
		goto fault;
	}

	smc_mapfile_module = (sm_atz_module_t)this->module_manager->find_by_name_first(this->module_manager, "atz_mapfile");

	return SM_OK;

fault:
	smc_icomm_cleanup(this);
	smc_job_cleanup();
	smc_db_cleanup();
	sm_privsep_cleanup();
	return SM_FAULT;
}

static void
qcg_comp_cleanup(sm_module_t this)
{
	SM_DEBUG_ENTER();

	sm_event_timer_unsubscribe(&reservation_vacuum_subscription);

	smc_jsdl_cleanup();
	smc_ardl_cleanup();
	smc_icomm_cleanup(this);
	smc_job_cleanup();
	smc_db_cleanup();
	smc_socks_cleanup();

	sm_signal_unregister("smc_factory_attributes");
	sm_signal_unregister("smc_contained_resources");

	SM_FREE(smc_attributes_local_resource_manager_type);
	SM_FREE(smc_attributes_contained_resources);
	SM_FREE(smc_factory_attributes);
	sm_privsep_cleanup();
}

static sm_value_t *
reservation_vacuum(void *object sm_unused, sm_value_t *p sm_unused)
{
	sm_list_t rsvns = NULL;
	sm_list_t l;
	sm_ret_code_t ret;

	SM_DEBUG_ENTER();

	ret = smc_rsvn_get_active_readonly(&rsvns);
	if (ret.code)
		return NULL;

	for (l = rsvns; l; l = l->next) {
		smc_rsvn_t rsvn = l->data;
		struct tm t;
		const char *id;
		const char *drms_id;

		id = rsvn->get_id(rsvn);
		drms_id = rsvn->get_drms_id(rsvn);
		t = rsvn->get_end_time(rsvn);

		if (time(NULL) > mktime(&t)) {
			smc_rsvn_t rw;

			ret = smc_rsvn_get(id, &rw);
			if (!ret.code) {
				sm_debug("Moving reservation %s (drms_id = %s) to accounting table", id, drms_id);

				rw->archive(rw);
				rw->save(rw);
				smc_rsvn_free(&rw);
			}
		}
	}

	SM_LIST_FREE_USER(rsvns, smc_rsvn_free);
	return NULL;
}
