/* $Id: qcg_comp_staging.c 294 2010-08-30 08:51:28Z smaug $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <qcg-core/logger.h>
#include <qcg-core/sm.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/modules/ecm_gsoap/gsoap/gsoap_buffer.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>

#include <faults.h>
#include <common.h>
#include <icomm.h>
#include <jsdl.h>
#include <user.h>

#include <stub/qcgcompH.h>
#include <stub/qcgcompStub.h>

__SM_ID("$Id: qcg_comp_staging.c 294 2010-08-30 08:51:28Z smaug $");

SM_INTERNAL_GENERAL_MODULE = {
	.header = {
		.name = "qcg_comp_staging"
	}
};

extern struct Namespace qcg_comp_namespaces[];

#define BUFFER_SIZE 1024*1024

typedef struct smc_staging_context_s {
	const char *attachemnt_id;
	int session_context;
	sm_value_t *error_detail;
	SM_BOOL error_state;
	/* only stage out fields */
	size_t files_count;
	void *session_handle;

	char *buffer;
	size_t buffer_len;
	size_t buffer_size;
} *smc_staging_context_t;

static void *
stream_in_open(struct sm_transport_s *transport, const char *id, void *handle)
{
	smc_staging_context_t ctxt = (smc_staging_context_t)handle;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER_STR(id);

	if (ctxt->error_state)
		goto fault;

	if (!(ctxt->attachemnt_id = transport->transport_strdup(transport, id)))
		goto fault;

	if (!(params = sm_value_create(1, SM_VALUE_STRING, id)))
		goto fault;

	if (smc_icomm_request_StageInOpen(ctxt->session_context, params, &result) != SM_OK)
		goto fault;

	if (!SM_VALUE_RET_CODE_OK(result[0])) {
		ctxt->error_detail = result;
		goto fault;
	}

	ctxt->buffer_size = BUFFER_SIZE;
	ctxt->buffer = sm_malloc(ctxt->buffer_size); /*JP where is coresponding sm_free */

	sm_value_free(&params);
	sm_value_free(&result);
	return handle;

fault:
	ctxt->error_state = SM_TRUE;
	sm_value_free(&params);
	return handle;
}

static int
stream_in_write(struct sm_transport_s *transport sm_unused, void *handle, const char *buf, size_t len)
{
	smc_staging_context_t ctxt = (smc_staging_context_t)handle;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER_STR(ctxt->attachemnt_id);

	if (ctxt->error_state)
		goto fault;

	if(ctxt->buffer_size < len) {
		ctxt->buffer_size = len;
		ctxt->buffer = sm_realloc(ctxt->buffer, ctxt->buffer_size);
	}

	if((ctxt->buffer_len + len >= ctxt->buffer_size) || //there is no free room for next chunk in buffer
			(ctxt->buffer_len > 0 && len == 0)) {	//or we wants to flush buffer

		sm_debug("Sending %lu bytes", (unsigned long)ctxt->buffer_len);

		if (!(params = sm_value_create(2,
				SM_VALUE_STRING, ctxt->attachemnt_id,
				SM_VALUE_BYTES, ctxt->buffer_len, ctxt->buffer
				)))
			goto fault;

		if (smc_icomm_request_StageInWrite(ctxt->session_context, params, &result)!= SM_OK)
			goto fault;

		if (!SM_VALUE_RET_CODE_OK(result[0])) {
			ctxt->error_detail = result;
			goto fault;
		}

		ctxt->buffer_len = 0;

		sm_value_free(&params);
		sm_value_free(&result);
	}

	if(len > 0) {	//there is enough place because we either resized buffer or sent out data
		memcpy(ctxt->buffer + ctxt->buffer_len, buf, len);
		ctxt->buffer_len += len;
		sm_debug("Buffered %lu bytes, buffer: %lu/%lu", (unsigned long)len, (unsigned long)ctxt->buffer_len, (unsigned long)ctxt->buffer_size);
	}

	return SM_OK;

fault:
	ctxt->error_state = SM_TRUE;
	sm_value_free(&params);
	return SM_OK; /* yes, we must return SM_OK otherwise user will not get meaningful error message */
}

static void
stream_in_close(struct sm_transport_s *transport sm_unused, void *handle)
{
	smc_staging_context_t ctxt = (smc_staging_context_t)handle;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER_STR(ctxt->attachemnt_id);

	if (ctxt->error_state)
		goto fault;

	stream_in_write(transport, handle, NULL, 0);	//flush buffer

	if (!(params = sm_value_create(1, SM_VALUE_STRING, ctxt->attachemnt_id)))
		goto fault;

	if (smc_icomm_request_StageInClose(ctxt->session_context, params, &result) != SM_OK)
		goto fault;

	if (!SM_VALUE_RET_CODE_OK(result[0])) {
		ctxt->error_detail = result;
		goto fault;
	}

	sm_value_free(&params);
	sm_value_free(&result);
	return;

fault:
	ctxt->error_state = SM_TRUE;
	sm_value_free(&params);
}

static void *
stream_out_open(struct sm_transport_s *transport, const char *id, void *handle)
{
	smc_staging_context_t ctxt = (smc_staging_context_t)handle;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER_STR(id);

	if (ctxt->error_state)
		goto fault;

	if (!(ctxt->attachemnt_id = transport->transport_strdup(transport, id)))
		goto fault;

	if (!(params = sm_value_create(1, SM_VALUE_STRING, id)))
		goto fault;

	if (smc_icomm_request_StageOutOpen(ctxt->session_context, params, &result) != SM_OK)
		goto fault;

	if (!SM_VALUE_RET_CODE_OK(result[0])) {
		ctxt->error_detail = result;
		goto fault;
	}

	sm_value_free(&params);
	sm_value_free(&result);
	return handle;

fault:
	ctxt->error_state = SM_TRUE;

	if (!ctxt->error_detail) {
		sm_value_free(&result);
		smc_icomm_request_StageOutDestroyContext(&ctxt->session_context, &ctxt->session_handle, SM_VALUE_NONE_SINGLETON, &result);
	} else {
		smc_icomm_request_StageOutRelease(&ctxt->session_context, &ctxt->session_handle);
	}

	sm_value_free(&params);
	sm_value_free(&result);
	return NULL;
}

static int
stream_out_read(struct sm_transport_s *transport sm_unused, void *handle, char *buf, size_t len)
{
	smc_staging_context_t ctxt = (smc_staging_context_t)handle;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER_STR(ctxt->attachemnt_id);

	if (ctxt->error_state)
		goto fault;

	if (!(params = sm_value_create(2, SM_VALUE_STRING, ctxt->attachemnt_id, SM_VALUE_UINT32, len)))
		goto fault;

	if (smc_icomm_request_StageOutRead(ctxt->session_context, params, &result)!= SM_OK)
		goto fault;

	if (!SM_VALUE_RET_CODE_OK(result[0])) {
		ctxt->error_detail = result;
		goto fault;
	}

	assert(SM_VALUE_IS_BYTES(result[1]));
	assert(result[1]->size == len);

	memcpy(buf, result[1]->vbytes, result[1]->size);

	sm_value_free(&params);
	sm_value_free(&result);
	return len;

fault:
	ctxt->error_state = SM_TRUE;

	if (!ctxt->error_detail) {
		sm_value_free(&result);
		smc_icomm_request_StageOutDestroyContext(&ctxt->session_context, &ctxt->session_handle, SM_VALUE_NONE_SINGLETON, &result);
	} else {
		smc_icomm_request_StageOutRelease(&ctxt->session_context, &ctxt->session_handle);
	}

	sm_value_free(&params);
	sm_value_free(&result);
	return -1;
}

static void
stream_out_close(struct sm_transport_s *transport sm_unused, void *handle)
{
	smc_staging_context_t ctxt = (smc_staging_context_t)handle;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER_STR(ctxt->attachemnt_id);

	ctxt->files_count--;

	if (ctxt->error_state)
		goto fault;

	if (!(params = sm_value_create(1, SM_VALUE_STRING, ctxt->attachemnt_id)))
		goto fault;

	if (smc_icomm_request_StageOutClose(ctxt->session_context, params, &result) != SM_OK)
		goto fault;

	sm_value_free(&params);
	sm_value_free(&result);

	if (ctxt->files_count == 0) {
		/* don't forget to say "bye bye" to session process */
		if (smc_icomm_request_StageOutDestroyContext(&ctxt->session_context, &ctxt->session_handle, SM_VALUE_NONE_SINGLETON, &result) != SM_OK) {
			sm_error("Failed to send smc_icomm_request_StageOutDestroyContext");
			goto fault;
		}

		if (!SM_VALUE_RET_CODE_OK(result[0])) {
			ctxt->error_detail = result;
			goto fault;
		}

		sm_value_free(&params);
		sm_value_free(&result);
	}

	return;

fault:
	ctxt->error_state = SM_TRUE;

	if (!ctxt->error_detail) {
		sm_value_free(&result);
		smc_icomm_request_StageOutDestroyContext(&ctxt->session_context, &ctxt->session_handle, SM_VALUE_NONE_SINGLETON, &result);
	} else {
		smc_icomm_request_StageOutRelease(&ctxt->session_context, &ctxt->session_handle);
	}

	sm_value_free(&params);
	sm_value_free(&result);
}

int
__qcg_comp_staging__StageFilesIn(
	struct soap *soap,
	struct qcg_comp_staging__StageFilesInType *qcg_comp_staging__StageFilesIn,
	struct qcg_comp_staging__StageFilesInResponseType *qcg_comp_staging__StageFilesInResponse sm_unused
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	char *username = NULL;
	uid_t uid = 0;
	sm_ret_code_t ret;
	_jsdl__JobDefinition *jsdl = NULL;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	const char *jsdl_str = NULL;
	int session_context = -1;
	void *handle = NULL;
	struct smc_staging_context_s staging_context;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "StageFilesIn", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "StageFilesIn", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	jsdl = qcg_comp_staging__StageFilesIn->jsdl__JobDefinition;

	if (!jsdl->jsdl__JobDescription->__sizeDataStaging)
		sm_info("No DataStaging elements found");

	SM_TRY_NOT(jsdl_str = smc_jsdl_str(jsdl, (smc_strdup_function) soap_strdup, soap));

	/* get user mapping */
	ret = smc_user_handle_submission(smc_jsdl_get_username(qcg_comp_staging__StageFilesIn->jsdl__JobDefinition), transport->atc_token, &username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	transport->stream_data_in_open = stream_in_open;
	transport->stream_data_in_write = stream_in_write;
	transport->stream_data_in_close = stream_in_close;

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, jsdl_str));

	/* start session process */
	if (smc_icomm_request_StageInCreateContext(uid, params, &result, &session_context, &handle) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if (!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT_FROM_VAL(result);

	sm_value_free(&params);
	sm_value_free(&result);

	sm_debug("StageIn context created");

	staging_context.attachemnt_id = NULL;
	staging_context.session_context = session_context;
	staging_context.error_detail = NULL;
	staging_context.error_state = SM_FALSE;
	staging_context.buffer_size = 0;
	staging_context.buffer = NULL;
	staging_context.buffer_len = 0;

	if (transport->stream_data_in(transport, &staging_context) != SM_OK || staging_context.error_state) {
		if (staging_context.error_detail) {
			sm_debug("Error detail: %s", SM_VALUE_RET_CODE_STRING(staging_context.error_detail[0]));
			/* error detail available => session process decided to terminate */
			THROW_APP_FAULT_FROM_VAL(staging_context.error_detail);
		} else {
			/* not sure whether session process terminated */
			smc_icomm_request_StageInDestroyContext(&session_context, &handle, SM_VALUE_NONE_SINGLETON, &result);
			THROW_INTERNAL_FAULT("Streaming failed");
		}
	}

	/* don't forget to say "bye bye" to session process */
	if (smc_icomm_request_StageInDestroyContext(&session_context, &handle, SM_VALUE_NONE_SINGLETON, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->exit(transport, NULL);

fault:
	smc_icomm_request_StageInRelease(&session_context, &handle);
	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->fault(transport, fault);
}

int
__qcg_comp_staging__StageFilesOut(
        struct soap *soap,
        struct qcg_comp_staging__StageFilesOutType *qcg_comp_staging__StageFilesOut,
        struct qcg_comp_staging__StageFilesOutResponseType *qcg_comp_staging__StageFilesOutResponse sm_unused
)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s faultstruct, *fault;
	char *username = NULL;
	uid_t uid = 0;
	sm_ret_code_t ret;
	_jsdl__JobDefinition *jsdl = NULL;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	const char *jsdl_str = NULL;
	int session_context = -1;
	void *handle = NULL;
	smc_staging_context_t staging_context;
	size_t i;

	SMC_FAULT_INIT;

	if (transport->enter(transport, "StageFilesOut", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "StageFilesOut", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	SM_TRY_NOT(staging_context = transport->transport_malloc(transport, sizeof(*staging_context)));

	jsdl = qcg_comp_staging__StageFilesOut->jsdl__JobDefinition;

	if (!jsdl->jsdl__JobDescription->__sizeDataStaging) {
		sm_info("No DataStaging elements found");
		return transport->exit(transport, NULL);
	}

	SM_TRY_NOT(jsdl_str = smc_jsdl_str(jsdl, (smc_strdup_function) soap_strdup, soap));

	/* get user mapping */
	ret = smc_user_handle_submission(smc_jsdl_get_username(qcg_comp_staging__StageFilesOut->jsdl__JobDefinition), transport->atc_token, &username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	transport->stream_data_out_open = stream_out_open;
	transport->stream_data_out_read = stream_out_read;
	transport->stream_data_out_close = stream_out_close;

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, jsdl_str));

	/* start session process */
	if (smc_icomm_request_StageOutCreateContext(uid, params, &result, &session_context, &handle) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if (!SM_VALUE_RET_CODE_OK(result[0]))
		THROW_APP_FAULT_FROM_VAL(result);

	sm_debug("StageOut context created");

	staging_context->attachemnt_id = NULL;
	staging_context->session_context = session_context;
	staging_context->error_detail = NULL;
	staging_context->error_state = SM_FALSE;
	staging_context->files_count = sm_value_count(result) - 1;

	if (staging_context->files_count == 0) {
		sm_value_free(&result);

		if (smc_icomm_request_StageOutDestroyContext(&staging_context->session_context, &staging_context->session_handle, SM_VALUE_NONE_SINGLETON, &result) != SM_OK) {
			sm_error("Failed to send smc_icomm_request_StageOutDestroyContext");
			goto fault;
		}
	} else {
		for (i = 1; i <= staging_context->files_count; i++) {
			char *file_uuid = NULL;
			size_t file_size = 0;
			sm_value_t *file_info = NULL;

			assert(SM_VALUE_IS_ARRAY(result[i]));

			file_info = result[i]->varray;
			assert(SM_VALUE_IS_STRING(file_info[0]));
			assert(SM_VALUE_IS_UINT64(file_info[1]));

			file_uuid = file_info[0]->vstring;
			file_size = (size_t) file_info[1]->vuint64;

			sm_debug("Staging out file: %s (size = %zu)", file_uuid, file_size);

			SM_TRY_NOT(file_uuid = soap_strdup(soap, file_uuid));

			if (transport->stream_data_out(transport, staging_context, file_size, file_uuid) != SM_OK)
				THROW_INTERNAL_FAULT("stream_data_out failed"); /* actual staging is done after leaving this function */
		}
	}

	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->exit(transport, NULL);

fault:
	smc_icomm_request_StageOutRelease(&session_context, &handle);
	SM_FREE(username);
	sm_value_free(&params);
	sm_value_free(&result);
	return transport->fault(transport, fault);
}

int
__qcg_comp_staging__ListDirectory(struct soap* soap, struct qcg_comp_staging__ListDirectoryType *request, struct qcg_comp_staging__ListDirectoryResponseType *response)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_transport_t transport = soap->user;
	sm_ret_code_t ret;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	int count, i;
	char *username = request->UserName;
	uid_t uid = 0;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "ListDirectory", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "ListDirectory", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	ret = smc_user_handle_submission(NULL, transport->atc_token, &username, &uid);

	if (ret.code)
			THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, request->Directory));

	if(smc_icomm_request_ListDirectory(uid, params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if(result[0]->vretcode.code != 0) {
		THROW_APP_FAULT_FROM_VAL(result);
	}

	count = result[1]->size;

	SM_TRY_NOT(response->jsdl__JobDefinition = transport->transport_malloc(transport, sizeof(*response->jsdl__JobDefinition)));
	sm_debug("JobDefinition");
	SM_TRY_NOT(response->jsdl__JobDefinition->jsdl__JobDescription = transport->transport_malloc(transport, sizeof(*response->jsdl__JobDefinition->jsdl__JobDescription)));
	sm_debug("JobDescription");
	SM_TRY_NOT(response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__Application = transport->transport_malloc(transport, sizeof(*response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__Application)));
	sm_debug("JobApplication");
	SM_TRY_NOT(response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication = transport->transport_malloc(transport, sizeof(*response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication)));
	sm_debug("HPCProfileApplication");
	SM_TRY_NOT(response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__WorkingDirectory = transport->transport_malloc(transport, sizeof(*response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__WorkingDirectory)));
	sm_debug("WorkingDirectory");

	response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__WorkingDirectory->__item = transport->transport_strdup(transport, result[2]->vstring);
	sm_debug("Full working directory: %s", response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__WorkingDirectory->__item);

	response->jsdl__JobDefinition->jsdl__JobDescription->__sizeDataStaging = count;
	SM_TRY_NOT(response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__DataStaging = transport->transport_malloc(transport, count*sizeof(*response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__DataStaging)));

	for(i=0 ; i<count ; ++i) {
		sm_debug("%i-th file: %s", i, result[1]->vstrings[i]);

		SM_TRY_NOT(response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__DataStaging[i].jsdl__FileName = transport->transport_strdup(transport, result[1]->vstrings[i]));
		response->jsdl__JobDefinition->jsdl__JobDescription->jsdl__DataStaging[i].jsdl__CreationFlag = jsdl__CreationFlagEnumeration__overwrite;
	}

	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->exit(transport, NULL));

fault:
	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->fault(transport, fault));
}


int
__qcg_comp_staging__StatFile(struct soap* soap, struct qcg_comp_staging__StatFileType *request, struct qcg_comp_staging__StatFileResponseType *response)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_transport_t transport = soap->user;
	sm_ret_code_t ret;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	char *username = request->UserName;
	uid_t uid;
	int i = 0;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "StatFile", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "StatFile", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	ret = smc_user_handle_submission(NULL, transport->atc_token, &username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	response->__sizeFiles = request->__sizePath;
	SM_TRY_NOT(response->Files = transport->transport_malloc(transport, sizeof(*response->Files) * response->__sizeFiles));

	for(i=0 ; i<request->__sizePath ; ++i) {

		SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, request->Path[i]));

		if(smc_icomm_request_StatFile(uid, params, &result) != SM_OK)
			THROW_INTERNAL_FAULT("Failed to send request to session process");

		sm_value_free(&params);

		if(result[0]->vretcode.code != 0) {
			THROW_APP_FAULT_FROM_VAL(result);
		}

		response->Files[i].ATime = result[1]->vuint32;
		response->Files[i].CTime = result[2]->vuint32;
		response->Files[i].GID = transport->transport_strdup(transport, result[3]->vstring);
		response->Files[i].MTime = result[4]->vuint32;
		response->Files[i].Mode = result[5]->vuint32;
		response->Files[i].NLink = result[6]->vuint32;
		response->Files[i].Size = result[7]->vuint32;
		response->Files[i].UID = transport->transport_strdup(transport, result[8]->vstring);
	}

	sm_value_free(&result);
	return SM_DEBUG_LEAVE_INT(transport->exit(transport, NULL));

fault:
	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->fault(transport, fault));
}

int
__qcg_comp_staging__Delete(struct soap* soap, struct qcg_comp_staging__DeleteType *request, struct qcg_comp_staging__DeleteResponseType *response sm_unused)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_transport_t transport = soap->user;
	sm_ret_code_t ret;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	char *username = request->UserName;
	uid_t uid;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "Delete", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "Delete", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	ret = smc_user_handle_submission(NULL, transport->atc_token, &username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	sm_debug("Deleting %s", request->Path);

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, request->Path));

	if(smc_icomm_request_Delete(uid, params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if(!SM_VALUE_RET_CODE_OK(result[0])) {
		THROW_APP_FAULT_FROM_VAL(result);
	}

	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->exit(transport, NULL));

fault:
	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->fault(transport, fault));
}

int
__qcg_comp_staging__ChMod(struct soap* soap, struct qcg_comp_staging__ChModType *request, struct qcg_comp_staging__ChModResponseType *response sm_unused)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_transport_t transport = soap->user;
	sm_ret_code_t ret;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	char *username = request->UserName;
	uid_t uid;
	mode_t mode = 0644;
	char *endptr = NULL;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "ChMod", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "ChMod", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	ret = smc_user_handle_submission(NULL, transport->atc_token, &username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	sm_debug("path=%s mode=%s", request->Path, request->Mode);

	mode = strtol(request->Mode, &endptr, 8);

	if(endptr[0]!='\0' || request->Mode[0]=='\0')
		THROW_INTERNAL_FAULT("Invalid mode specified. Mode must be specified as a string with octal representation of mode, with or without leading 0.");

	sm_info("Changing mode of %s to %o", request->Path, mode);

	SM_TRY_NOT(params = sm_value_create(2, SM_VALUE_STRING, request->Path, SM_VALUE_INT32, mode));

	if(smc_icomm_request_ChMod(uid, params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if(!SM_VALUE_RET_CODE_OK(result[0])) {
		THROW_APP_FAULT_FROM_VAL(result);
	}

	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->exit(transport, NULL));

fault:
	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->fault(transport, fault));
}

int
__qcg_comp_staging__MkDir(struct soap* soap, struct qcg_comp_staging__MkDirType *request, struct qcg_comp_staging__MkDirResponseType *response sm_unused)
{
	struct sm_transport_fault_s faultstruct, *fault;
	sm_transport_t transport = soap->user;
	sm_ret_code_t ret;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	char *username = request->UserName;
	uid_t uid;
	mode_t mode = 0755;
	char *endptr = NULL;

	SM_DEBUG_ENTER();
	SMC_FAULT_INIT;

	if (transport->enter(transport, "MkDir", fault) != SM_OK)
		return transport->fault(transport, fault);

	if (transport->access_check(transport, SM_TRUE, "MkDir", fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, fault);

	ret = smc_user_handle_submission(NULL, transport->atc_token, &username, &uid);

	if (ret.code)
		THROW_INTERNAL_FAULT(transport->transport_strdup(transport, ret.string));

	sm_debug("path=%s mode=%s", request->Path, request->Mode);

	if(request->Mode != NULL) {
		mode = strtol(request->Mode, &endptr, 8);

		if(endptr[0]!='\0' || request->Mode[0]=='\0')
			THROW_INTERNAL_FAULT("Invalid mode specified. Mode, if specified, must be specified as a string with octal representation of mode, with or without leading 0. If mode is unspecified the default mode 0755 is used.");
	}
	else
		mode = 0755;

	sm_info("Creating directory %s with mode %o", request->Path, mode);

	SM_TRY_NOT(params = sm_value_create(2, SM_VALUE_STRING, request->Path, SM_VALUE_INT32, mode));

	if(smc_icomm_request_MkDir(uid, params, &result) != SM_OK)
		THROW_INTERNAL_FAULT("Failed to send request to session process");

	if(!SM_VALUE_RET_CODE_OK(result[0])) {
		THROW_APP_FAULT_FROM_VAL(result);
	}

	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->exit(transport, NULL));

fault:
	sm_value_free(&result);
	sm_value_free(&params);
	return SM_DEBUG_LEAVE_INT(transport->fault(transport, fault));
}
