/* $Id: session_process.c 310 2010-10-13 14:04:19Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdbool.h>
#include <signal.h>
#include <pwd.h>

#include <qcg-core/sm.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/privsep.h>
#include <qcg-core/signal.h>
#include <qcg-core/thread.h>
#include <qcg-core/utils.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <libxml/hash.h>

#include <db.h>
#include <faults.h>
#include <icomm.h>
#include <identifiers.h>
#include <job.h>
#include <jsdl.h>
#include <rsvn.h>
#include <session_process.h>
#include <staging.h>
#include <include/submission_module.h>
#include <include/reservation_module.h>
#include <include/staging_module.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>

#include <stub/qcgcompStub.h>

#include <rsync/interface.h>

#define min(a,b) ((a)<(b)?(a):(b))

__SM_ID("$Id: session_process.c 310 2010-10-13 14:04:19Z mamonski $");
SM_INTERNAL_GENERAL_MODULE =
{ .header =
{
	.name = "session_process"
} };

static unsigned int session_process_timeout = 10;
static unsigned int submitted_jobs_hash_table = 4096;
static unsigned int job_status_polling_period = 10;
static SM_BOOL job_status_full_polling = SM_TRUE;

static sm_conf_ctxt_t conf;
static sm_module_manager_t module_manager;

static submission_module_t submission_module;
static reservation_module_t reservation_module;

bool have_waiting_thread;
bool have_polling_thread;
static sm_thread_t submission_waiting_thread;
static sm_thread_t submission_polling_thread;

static xmlHashTablePtr submitted_jobs;
static sm_mutex_t submitted_jobs_lock;
static sm_cond_t submitted_jobs_cond;

static sm_mutex_t staging_in_progress_lock;
static int staging_in_progress;
static sm_value_t *local_resource_manager_type;

static void
kill_rsvn(const char *drms_id)
{
	sm_value_t *params = NULL;
	sm_value_t *subresult = NULL;

	if (!drms_id)
		return;

	SM_DEBUG_ENTER_STR(drms_id);

	sm_notice("Killing reservation %s due to a severe processing error", drms_id);

	params = sm_value_create(1, SM_VALUE_STRING, drms_id);
	if (!params)
		goto fault;

	subresult = reservation_module->terminate((sm_module_t) reservation_module, params);
	if (!subresult || !SM_VALUE_RET_CODE_OK(subresult[0]))
		goto fault;

	sm_value_free(&subresult);
	sm_value_free(&params);
	sm_debug("Killed reservation %s", drms_id);
	return;

fault:
	sm_value_free(&subresult);
	sm_value_free(&params);
	sm_crit("Failed to kill reservation %s. Reservation orphaned.", drms_id);
	return;
}

static void
handle_CreateReservation(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *params = NULL;
	sm_value_t *subresult = NULL;
	smc_rsvn_t rsvn = NULL;
	const char *subject = NULL;
	const char *username = NULL;
	const char *submitted_ardl = NULL;
	const char *ardl = NULL;
	const char *drms_id = NULL;
	sm_ret_code_t ret;
	sm_value_t *response = NULL;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_CreateReservation(fd, &req) != SM_OK) {
		sm_error("Failed to receive CreateReservation request");
		goto fault;
	}

	subject = req[0]->vstring;
	username = req[1]->vstring;
	submitted_ardl = req[2]->vstring;
	ardl = req[3]->vstring;

	if (!reservation_module) {
		response = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "No reservation module loaded");
		goto fault;
	}

	SM_TRY_NOT(rsvn = smc_rsvn_create());
	rsvn->set_subject(rsvn, subject);
	rsvn->set_mapped_user(rsvn, username);
	rsvn->set_submitted_ardl(rsvn, submitted_ardl);
	rsvn->set_ardl(rsvn, ardl);

	/* We use rsvn->get_ardl to get more "canonical" form of ARDL. */
	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, rsvn->get_ardl(rsvn)));
	subresult = reservation_module->create((sm_module_t) reservation_module, params);
	if (!subresult) {
		sm_error("Reservation module didn't return any result");
		goto fault;
	}

	if (SM_VALUE_RET_CODE_OK(subresult[0])) {
		drms_id = subresult[1]->vstring;

		if (sm_value_count(subresult) == 4) {
			time_t start_time = 0;
			time_t end_time = 0;
			struct tm start_time_tm;
			struct tm end_time_tm;

			assert(SM_VALUE_IS_UINT32(subresult[2]));
			assert(SM_VALUE_IS_UINT32(subresult[3]));

			start_time = (time_t)subresult[2]->vuint32;
			end_time = (time_t)subresult[3]->vuint32;

			rsvn->set_start_time(rsvn, *localtime_r(&start_time, &start_time_tm));
			rsvn->set_end_time(rsvn, *localtime_r(&end_time, &end_time_tm));
		}

		rsvn->set_drms_id(rsvn, drms_id);

		ret = rsvn->save(rsvn);
		if (ret.code) {
			sm_error("Failed to save reservation in database: %d:%s", ret.code, ret.string);
			goto fault;
		}

		SM_TRY_NOT(sm_value_append(&subresult, 1, SM_VALUE_STRING, rsvn->get_id(rsvn)));
	}

	if (smc_icomm_respond_CreateReservation(fd, subresult) != SM_OK) {
		sm_error("Failed to send CreateReservation response");
		goto fault;
	}

	sm_value_free(&req);
	sm_value_free(&params);
	smc_rsvn_free(&rsvn);
	return;

fault:
	kill_rsvn(drms_id);
	sm_value_free(&req);
	sm_value_free(&params);
	smc_rsvn_free(&rsvn);
	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Internal fault in session process");
	if (smc_icomm_respond_CreateReservation(fd, response) != SM_OK)
		sm_error("Failed to send CreateReservation response");
}

static void
handle_TerminateReservation(int fd)
{
	sm_value_t *params = NULL;
	sm_value_t *subresult = NULL;
	sm_value_t *response = NULL;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_TerminateReservation(fd, &params) != SM_OK) {
		sm_error("Failed to receive TerminateReservation request");
		goto fault;
	}

	if (!reservation_module) {
		response = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "No reservation module loaded");
		goto fault;
	}

	subresult = reservation_module->terminate((sm_module_t) reservation_module, params);
	if (!subresult) {
		sm_error("Reservation module didn't return any result");
		goto fault;
	}

	if (smc_icomm_respond_TerminateReservation(fd, subresult) != SM_OK) {
		sm_error("Failed to send TerminateReservation response");
		goto fault;
	}

	sm_value_free(&params);
	return;

fault:
	sm_value_free(&params);
	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Internal fault in session process");
	if (smc_icomm_respond_TerminateReservation(fd, response) != SM_OK)
		sm_error("Failed to send TerminateReservation response");
}

static void
handle_GetReservationStatus(int fd)
{
	sm_value_t *params = NULL;
	sm_value_t *subresult = NULL;
	sm_value_t *response = NULL;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_GetReservationStatus(fd, &params) != SM_OK) {
		sm_error("Failed to receive GetReservationStatus request");
		goto fault;
	}

	if (!reservation_module) {
		response = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "No reservation module loaded");
		goto fault;
	}

	subresult = reservation_module->get_status((sm_module_t) reservation_module, params);
	if (!subresult) {
		sm_error("Reservation module didn't return any result");
		goto fault;
	}

	if (smc_icomm_respond_GetReservationStatus(fd, subresult) != SM_OK) {
		sm_error("Failed to send GetReservationStatus response");
		goto fault;
	}

	sm_value_free(&params);
	return;

fault:
	sm_value_free(&params);
	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Internal fault in session process");
	if (smc_icomm_respond_GetReservationStatus(fd, response) != SM_OK)
		sm_error("Failed to send GetReservationStatus response");
}

static void
handle_GetActiveReservations(int fd)
{
	sm_value_t *empty = NULL;
	sm_value_t *subresult = NULL;
	sm_value_t *response = NULL;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_GetActiveReservations(fd, &empty) != SM_OK) {
		sm_error("Failed to receive GetActiveReservations request");
		goto fault;
	}

	if (!reservation_module) {
		response = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "No reservation module loaded");
		goto fault;
	}

	subresult = reservation_module->get_active((sm_module_t) reservation_module);
	if (!subresult) {
		sm_error("Reservation module didn't return any result");
		goto fault;
	}

	if (smc_icomm_respond_GetActiveReservations(fd, subresult) != SM_OK) {
		sm_error("Failed to send GetActiveReservations response");
		goto fault;
	}

	sm_value_free(&empty);
	return;

fault:
	sm_value_free(&empty);
	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, RESERVATION_INTERNAL_FAULT, "Internal fault in session process");
	if (smc_icomm_respond_GetActiveReservations(fd, response) != SM_OK)
		sm_error("Failed to send GetActiveReservations response");
}

static void
handle_StageIn(int fd)
{
	sm_value_t *parameters = NULL;
	sm_value_t *response = NULL;
	smc_icomm_request_t request = -1;
	smc_icomm_request_t last_request = smc_icomm_StageInCreateContext;
	sm_alloc_tracked_t tracked = NULL;
	_jsdl__JobDefinition jsdl;
	const char *working_dir = NULL;
	const char *local_file_name = NULL;
	const char *file_uuid = NULL;
	struct jsdl__DataStaging_USCOREType *ds = NULL;
	FILE *fhandle = NULL;
	SM_BOOL skip_file = SM_FALSE;
	char errmsg[512] = "";

	SM_DEBUG_ENTER();

	/* create context **/
	if (smc_icomm_receive_StageInCreateContext(fd, &parameters) != SM_OK) {
		sm_error("Failed to receive StageInCreateContext request");
		goto fault;
	}

	assert(SM_VALUE_IS_STRING(parameters[0]));

	SM_TRY_NOT(tracked = smc_jsdl_from_str(parameters[0]->vstring, &jsdl));

	working_dir = smc_jsdl_get_working_dir(&jsdl, SM_TRUE);

	if (smc_icomm_respond_StageIn(fd, sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK)) != SM_OK) {
		sm_error("Failed to send StageInCreateContext response");
		goto fault;
	}

	/* open/write/close loop */
	do {
		if (smc_icomm_begin_receive(fd, &request) != SM_OK) {
			sm_error("smc_icomm_begin_receive failed.");
			goto fault;
		}

		sm_value_free(&parameters);
		response = NULL;

		switch (request) {

		case smc_icomm_StageInOpen:
		{
			const char *fopen_mode = NULL;
			int i = 0;

			ds = NULL;
			skip_file = SM_FALSE;

			assert(!fhandle);
			assert(last_request == smc_icomm_StageInCreateContext || last_request == smc_icomm_StageInClose);

			last_request = smc_icomm_StageInOpen;

			if (smc_icomm_receive_StageInOpen(fd, &parameters) != SM_OK) {
				sm_error("Failed to receive StageInOpen request");
				goto fault;
			}

			assert(SM_VALUE_IS_STRING(parameters[0]));

			SM_TRY_NOT(file_uuid = sm_alloc_tracked_strdup(tracked, parameters[0]->vstring));

			sm_debug("smc_icomm_StageInOpen (file_uuid = %s)", file_uuid);

			for (i = 0; i < jsdl.jsdl__JobDescription->__sizeDataStaging; i++) {
				if (jsdl.jsdl__JobDescription->jsdl__DataStaging[i].jsdl__Source &&
						sm_strcmp(jsdl.jsdl__JobDescription->jsdl__DataStaging[i].jsdl__Source->jsdl__URI, file_uuid) == 0)
					ds = &(jsdl.jsdl__JobDescription->jsdl__DataStaging[i]);
			}

			if (!ds) {
				snprintf(errmsg, sizeof(errmsg), "File with UUID not found: %s", file_uuid);
				sm_debug("%s", errmsg);
				response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
				goto fault;
			}

			if (!ds->jsdl__FileName) {
				snprintf(errmsg, sizeof(errmsg), "Duplicated attachment id: %s", file_uuid);
				sm_debug("%s", errmsg);
				response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
				goto fault;
			}

			if (working_dir)
				local_file_name = sm_alloc_tracked_asprintf(tracked, "%s/%s", working_dir, ds->jsdl__FileName);
			else
				local_file_name = ds->jsdl__FileName;

			/* mark file as already handled */
			ds->jsdl__FileName = NULL;

			sm_debug("Staging in: %s -> %s", file_uuid, local_file_name);

			if (ds->jsdl__CreationFlag == jsdl__CreationFlagEnumeration__dontOverwrite) {
				struct stat st;

				if (!stat(local_file_name, &st)) {
					sm_debug("File %s already exists, not staging in", local_file_name);
					skip_file = SM_TRUE;

					if (smc_icomm_respond_StageIn(fd, sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_OK)) != SM_OK) {
						sm_error("Failed to send StageInOpen response");
						goto fault;
					}

					continue;
				}
			}

			switch (ds->jsdl__CreationFlag) {
				case jsdl__CreationFlagEnumeration__dontOverwrite:
				case jsdl__CreationFlagEnumeration__overwrite:
					fopen_mode = "wb";
					break;

				case jsdl__CreationFlagEnumeration__append:
					fopen_mode = "ab";
					break;
			}

			if (!(fhandle = fopen(local_file_name, fopen_mode))) {
				snprintf(errmsg, sizeof(errmsg), "Unable to open file %s for writing: %s", local_file_name, strerror(errno));
				sm_debug("%s", errmsg);
				response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
				goto fault;
			}

			if(ds->qcgcomp1__FileMode) {
				mode_t mode;
				char *endptr;

				mode = strtol(ds->qcgcomp1__FileMode, &endptr, 8);

				if(!(ds->qcgcomp1__FileMode[0]!='\0' && endptr[0]=='\0') ||
						ds->qcgcomp1__FileMode[0]=='\0')
					mode = 0644;

				sm_debug("Setting mode %o (was %s) for file %s", mode, ds->qcgcomp1__FileMode, local_file_name);

				if(chmod(local_file_name, mode) != 0) {
					snprintf(errmsg, sizeof(errmsg), "Unable to set mode %o for file %s: %s", mode, local_file_name, strerror(errno));
					sm_debug("%s", errmsg);
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
					goto fault;
				}

			}

			if (smc_icomm_respond_StageIn(fd, sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK)) != SM_OK) {
				sm_error("Failed to send StageInOpen response");
				goto fault;
			}

			break;
		} /* smc_icomm_StageInOpen */

		case smc_icomm_StageInWrite:
		{
			char *buf = NULL;
			int buf_len;

			if (smc_icomm_receive_StageInWrite(fd, &parameters) != SM_OK) {
				sm_error("Failed to receive StageInWrite request");
				goto fault;
			}

			assert(last_request == smc_icomm_StageInWrite || last_request == smc_icomm_StageInOpen);
			assert(fhandle || skip_file);
			assert(SM_VALUE_IS_STRING(parameters[0]));
			assert(SM_VALUE_IS_BYTES(parameters[1]) || SM_VALUE_IS_NONE(parameters[1]));
			assert(sm_strcmp(parameters[0]->vstring, file_uuid) == 0);

			last_request = smc_icomm_StageInWrite;

			if (!skip_file && SM_VALUE_IS_BYTES(parameters[1])) {
				buf_len = parameters[1]->size;
				buf = parameters[1]->vbytes;

				sm_debug("Write %d bytes (file_uuid = %s)", buf_len, file_uuid);

				if (fwrite(buf, buf_len, 1, fhandle) != 1) {
					snprintf(errmsg, sizeof(errmsg), "Write to file %s failed: %s", local_file_name, strerror(errno));
					sm_debug("%s", errmsg);
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
					goto fault;
				}
			}

			if (smc_icomm_respond_StageIn(fd, sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_OK)) != SM_OK) {
				sm_error("Failed to send StageInWrite response");
				goto fault;
			}

			break;
		} /* smc_icomm_StageInWrite */

		case smc_icomm_StageInClose:
		{
			if (smc_icomm_receive_StageInClose(fd, &parameters) != SM_OK) {
				sm_error("Failed to receive StageInClose request");
				goto fault;
			}

			assert(last_request == smc_icomm_StageInWrite || last_request == smc_icomm_StageInOpen);
			assert(fhandle || skip_file);
			assert(SM_VALUE_IS_STRING(parameters[0]));
			SM_ASSERT(sm_strcmp(parameters[0]->vstring, file_uuid) == 0, "%s != %s", parameters[0]->vstring, file_uuid);

			last_request = smc_icomm_StageInClose;

			if (!skip_file && (fclose(fhandle) != 0)) {
				snprintf(errmsg, sizeof(errmsg), "Writing to file %s failed: %s", local_file_name, strerror(errno));
				sm_debug("%s", errmsg);
				response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
				goto fault;
			}

			fhandle = NULL;

			if (smc_icomm_respond_StageIn(fd, sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_OK)) != SM_OK) {
				sm_error("Failed to send StageInClose response");
				goto fault;
			}

			break;
		} /* smc_icomm_StageInClose */

		case smc_icomm_StageInDestroyContext:
		{
			if (smc_icomm_receive_StageInDestroyContext(fd, &parameters) != SM_OK) {
				sm_error("Failed to receive StageInOpen request");
				goto fault;
			}

			assert(SM_VALUE_IS_NONE(parameters[0]));

			if (smc_icomm_respond_StageIn(fd, sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK)) != SM_OK) {
				sm_error("Failed to send StageInDestroy response");
				goto fault;
			}

			break;
		} /* smc_icomm_StageInDestroyContext*/

		default:
			SM_ASSERT(0, "Unknown request (%d) in %s", request, __FUNCTION__);
		}

	} while (request != smc_icomm_StageInDestroyContext);

	sm_debug("Terminating StageIn process");

	sm_value_free(&parameters);
	sm_alloc_tracked_free(&tracked);
	return;

fault:
	sm_value_free(&parameters);
	sm_alloc_tracked_free(&tracked);

	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_InternalFault, "Internal fault in session process");

	if (smc_icomm_respond_StageIn(fd, response) != SM_OK)
		sm_error("Failed to send StageIn response");
}

static void
handle_StageOut(int fd)
{
	sm_value_t *parameters = NULL;
	sm_value_t *response = NULL;
	smc_icomm_request_t request = -1;
	smc_icomm_request_t last_request = smc_icomm_StageOutCreateContext;
	sm_alloc_tracked_t tracked = NULL;
	_jsdl__JobDefinition jsdl;
	const char *working_dir = NULL;
	const char *local_file_name = NULL;
	const char *file_uuid = NULL;
	char errmsg[512] = "";
	struct jsdl__DataStaging_USCOREType *ds = NULL;
	FILE *fhandle = NULL;
	int i;
	off_t off;

	SM_DEBUG_ENTER();

	/* create context */
	if (smc_icomm_receive_StageOutCreateContext(fd, &parameters) != SM_OK) {
		sm_error("Failed to receive StageOutCreateContext request");
		goto fault;
	}

	assert(SM_VALUE_IS_STRING(parameters[0]));

	SM_TRY_NOT(tracked = smc_jsdl_from_str(parameters[0]->vstring, &jsdl));

	working_dir = smc_jsdl_get_working_dir(&jsdl, SM_FALSE);

	SM_TRY_NOT(response = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK));

	for (i = 0; i < jsdl.jsdl__JobDescription->__sizeDataStaging; i ++) {
		ds = &(jsdl.jsdl__JobDescription->jsdl__DataStaging[i]);

		if (ds->jsdl__Target && ds->jsdl__Target->jsdl__URI && strstr(ds->jsdl__Target->jsdl__URI, "cid:")) {
			sm_value_t *file_info = NULL;
			struct stat st;

			if (working_dir)
				local_file_name = sm_alloc_tracked_asprintf(tracked, "%s/%s", working_dir, ds->jsdl__FileName);
			else
				local_file_name = ds->jsdl__FileName;

			if(ds->qcgcomp1__Offset)
				off = *(ds->qcgcomp1__Offset);
			else
				off = 0;

			if (!local_file_name) {
				sm_value_free(&response);
				goto fault;
			}

			if (stat(local_file_name, &st) == -1) {
				int saved_errno = errno;

				sm_debug("stat on file %s failed: %d", local_file_name, saved_errno);
				sm_value_free(&response);

				if (saved_errno == ENOENT || saved_errno == ENOTDIR) {
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_FileNotFoundFault, local_file_name);
				} else if (saved_errno == EACCES) {
					snprintf(errmsg, sizeof(errmsg), "User with uid %u is not allowed to access file %s", getuid(), local_file_name);
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_NotAuthorizedFault, errmsg);
				} else {
					snprintf(errmsg, sizeof(errmsg), "Stat on file %s failed: %s", local_file_name, strerror(saved_errno));
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
				}

				goto fault;
			}

			if (access(local_file_name, R_OK) == -1) {
				int saved_errno = errno;

				sm_debug("access(R_OK) on file %s failed: %s", local_file_name, strerror(saved_errno));
				sm_value_free(&response);

				if (saved_errno == EACCES) {
					snprintf(errmsg, sizeof(errmsg), "User with uid %u is not allowed to read file %s", getuid(), local_file_name);
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_NotAuthorizedFault, errmsg);
				} else {
					snprintf(errmsg, sizeof(errmsg), "Failed to verify permission to read file %s: %s", local_file_name, strerror(saved_errno));
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
				}

				goto fault;
			}

			if (ds->jsdl__DeleteOnTermination && (*ds->jsdl__DeleteOnTermination) == xsd__boolean__true_) {
				char *dir_name;

				SM_TRY_NOT(dir_name = sm_dirname(local_file_name));

				sm_debug("Checking whether user has write access to directory (DeleteOnTermination): %s", dir_name);

				if (access(dir_name, W_OK) == -1) {
					int saved_errno = errno;

					snprintf(errmsg, sizeof(errmsg), "Getting access rights for directory %s failed: %s", dir_name, strerror(saved_errno));
					sm_debug("%s", errmsg);
					sm_value_free(&response);

					if (saved_errno == EACCES)
						response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_DeleteOnTerminationFault, local_file_name);
					else
						response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);

					SM_FREE(dir_name);
					goto fault;
				}

				SM_FREE(dir_name);
			}

			file_info = sm_value_create(2, SM_VALUE_STRING, ds->jsdl__Target->jsdl__URI, SM_VALUE_UINT64, (unsigned long long) st.st_size-off);
			sm_debug("File %s has size %llu", ds->jsdl__Target->jsdl__URI, (unsigned long long) st.st_size-off);

			if (!file_info) {
				sm_value_free(&response);
				goto fault;
			}

			SM_TRY_NOT(response = sm_value_append(&response, 1, SM_VALUE_ARRAY, file_info));
		} else if (ds->jsdl__DeleteOnTermination && (*ds->jsdl__DeleteOnTermination) == xsd__boolean__true_) {
			if (working_dir)
				local_file_name = sm_alloc_tracked_asprintf(tracked, "%s/%s",working_dir, ds->jsdl__FileName);
			else
				local_file_name = ds->jsdl__FileName;

			if (!local_file_name) {
				sm_value_free(&response);

				goto fault;
			}

			sm_debug("Deleting file: %s", local_file_name);

			if (unlink(local_file_name) == -1 && errno != ENOENT) {
				int saved_errno = errno;

				snprintf(errmsg, sizeof(errmsg), "Deleting file %s failed: %s", local_file_name, strerror(saved_errno));
				sm_debug("%s", errmsg);
				sm_value_free(&response);

				if (saved_errno == EACCES)
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_DeleteOnTerminationFault, local_file_name);
				else
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);

				goto fault;
			}

		}
	}

	if (smc_icomm_respond_StageOut(fd, response) != SM_OK) {
		sm_error("Failed to send StageOutCreateContext response");
		goto fault;
	}

	response = NULL;

	/* open/write/close loop */
	do {
		if (smc_icomm_begin_receive(fd, &request) != SM_OK) {
			sm_error("smc_icomm_begin_receive failed");
			goto fault;
		}

		sm_value_free(&parameters);
		response = NULL;

		switch (request) {

		case smc_icomm_StageOutOpen:
		{
			ds = NULL;

			assert(!fhandle);
			assert(last_request == smc_icomm_StageOutCreateContext || last_request == smc_icomm_StageOutClose);

			last_request = smc_icomm_StageOutOpen;

			if (smc_icomm_receive_StageOutOpen(fd, &parameters) != SM_OK) {
				sm_error("Failed to receive StageOutOpen request");
				goto fault;
			}

			assert(SM_VALUE_IS_STRING(parameters[0]));

			SM_TRY_NOT(file_uuid = sm_alloc_tracked_strdup(tracked, parameters[0]->vstring));

			sm_debug("smc_icomm_StageOutOpen (file_uuid = %s)", file_uuid);

			for (i = 0; i < jsdl.jsdl__JobDescription->__sizeDataStaging; i++) {
				if (jsdl.jsdl__JobDescription->jsdl__DataStaging[i].jsdl__Target &&
						sm_strcmp(jsdl.jsdl__JobDescription->jsdl__DataStaging[i].jsdl__Target->jsdl__URI, file_uuid) == 0)
					ds = &(jsdl.jsdl__JobDescription->jsdl__DataStaging[i]);
			}

			SM_ASSERT(ds, "File identified by UUID not found: %s", file_uuid);

			if (working_dir)
				local_file_name = sm_alloc_tracked_asprintf(tracked, "%s/%s",working_dir, ds->jsdl__FileName);
			else
				local_file_name = ds->jsdl__FileName;

			sm_debug("Staging out: %s -> %s", local_file_name, file_uuid);

			if (!(fhandle = fopen(local_file_name, "rb"))) {
				snprintf(errmsg, sizeof(errmsg), "Failed to open file %s: %s", local_file_name, strerror(errno));
				sm_notice("%s", errmsg);
				response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
				goto fault;
			}

			if(ds->qcgcomp1__Offset) {
				sm_debug("Skipping %u bytes in file %s", *(ds->qcgcomp1__Offset), local_file_name);

				if(fseek(fhandle, *(ds->qcgcomp1__Offset), SEEK_SET) != 0) {
					snprintf(errmsg, sizeof(errmsg), "Unable to seek in file stream %s: %s", local_file_name, strerror(errno));
					sm_debug("%s", errmsg);
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
					goto fault;
				}
			}

			if (smc_icomm_respond_StageOut(fd, sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK)) != SM_OK) {
				sm_error("Failed to send StageOutOpen response");
				goto fault;
			}

			break;
		} /* smc_icomm_StageOutOpen */

		case smc_icomm_StageOutRead:
		{
			char *buf = NULL;
			size_t buf_len;

			if (smc_icomm_receive_StageOutRead(fd, &parameters) != SM_OK) {
				sm_error("Failed to receive StageOutRead request");
				goto fault;
			}

			assert(last_request == smc_icomm_StageOutOpen || last_request == smc_icomm_StageOutRead);
			assert(fhandle);
			assert(SM_VALUE_IS_STRING(parameters[0]));
			assert(SM_VALUE_IS_UINT32(parameters[1]));
			assert(strcmp(parameters[0]->vstring, file_uuid) == 0);

			last_request = smc_icomm_StageOutRead;

			buf_len = (size_t) parameters[1]->vuint32;
			SM_TRY_NOT(buf = sm_malloc(buf_len));

			sm_debug("read %zu bytes (file_uuid = %s)", buf_len, file_uuid);

			if (fread(buf, buf_len, 1, fhandle) != 1) {
				snprintf(errmsg, sizeof(errmsg), "Reading from file %s failed: %s", local_file_name, strerror(errno));
				sm_notice("%s", errmsg);
				response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
				goto fault;
			}

			if (smc_icomm_respond_StageOut(fd, sm_value_create(2, SM_VALUE_RET_CODE, qcg_comp_staging_OK, SM_VALUE_BYTES | SM_VALUE_DONT_COPY_FLAG, buf_len, buf)) != SM_OK) {
				sm_error("Failed to send StageOutRead response");
				goto fault;
			}

			break;
		} /* smc_icomm_StageOutRead */

		case smc_icomm_StageOutClose:
		{
			if (smc_icomm_receive_StageOutClose(fd, &parameters) != SM_OK) {
				sm_error("Failed to receive StageOutClose request");
				goto fault;
			}

			assert(last_request == smc_icomm_StageOutOpen || last_request == smc_icomm_StageOutRead);
			assert(fhandle);
			assert(SM_VALUE_IS_STRING(parameters[0]));
			assert(strcmp(parameters[0]->vstring, file_uuid) == 0);

			last_request = smc_icomm_StageOutClose;

			fclose(fhandle);
			fhandle = NULL;

			if (ds->jsdl__DeleteOnTermination && *ds->jsdl__DeleteOnTermination == xsd__boolean__true_) {
				sm_debug("Deleting file: %s", local_file_name);

				if (unlink(local_file_name) != 0) {
					sm_notice("Deleting file %s failed: %s", local_file_name, strerror(errno));
					response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_DeleteOnTerminationFault, local_file_name);
					goto fault;
				}
			}

			if (smc_icomm_respond_StageOut(fd, sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_OK)) != SM_OK) {
				sm_error("Failed to send StageOutClose response");
				goto fault;
			}

			break;
		} /* smc_icomm_StageOutClose */

		case smc_icomm_StageOutDestroyContext:
		{
			if (smc_icomm_receive_StageOutDestroyContext(fd, &parameters) != SM_OK) {
				sm_error("Failed to receive StageOutDestroyContext request");
				goto fault;
			}

			assert(SM_VALUE_IS_NONE(parameters[0]));

			if (smc_icomm_respond_StageOut(fd, sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK)) != SM_OK) {
				sm_error("Failed to send StageOutDestroyContext response");
				goto fault;
			}

			break;
		} /* smc_icomm_StageOutDestroyContext */

		default:
			SM_ASSERT(0, "Unknown request (%d) in %s", request, __FUNCTION__);
		}

	} while (request != smc_icomm_StageOutDestroyContext);

	sm_debug("Terminating StageOut process");

	sm_value_free(&parameters);
	sm_alloc_tracked_free(&tracked);
	return;

fault:
	sm_value_free(&parameters);
	sm_alloc_tracked_free(&tracked);

	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_InternalFault, "Internal fault in session process");

	if (smc_icomm_respond_StageOut(fd, response) != SM_OK)
		sm_error("Failed to send StageOut response");
}

static void
kill_job(const char *drms_id)
{
	sm_value_t *params = NULL;
	sm_value_t *subresult = NULL;

	if (!drms_id)
		return;

	SM_DEBUG_ENTER_STR(drms_id);

	sm_notice("Killing job %s due to a severe processing error", drms_id);

	params = sm_value_create(1, SM_VALUE_STRING, drms_id);
	if (!params)
		goto fault;

	subresult = submission_module->terminate((sm_module_t) submission_module, params);
	if (!subresult || !SM_VALUE_RET_CODE_OK(subresult[0]))
		goto fault;

	sm_value_free(&subresult);
	sm_value_free(&params);
	sm_debug("Killed job %s", drms_id);
	return;

fault:
	sm_value_free(&subresult);
	sm_value_free(&params);
	sm_crit("Failed to kill job %s. Job orphaned.", drms_id);
	return;
}

void
smc_session_process_stager_new(void)
{
	sm_mutex_lock(&staging_in_progress_lock);
	staging_in_progress++;
	sm_mutex_unlock(&staging_in_progress_lock);
}

void
smc_session_process_stager_done(void)
{
	sm_mutex_lock(&staging_in_progress_lock);
	staging_in_progress--;
	sm_mutex_unlock(&staging_in_progress_lock);
}

#define SAVE_JOB() \
	do { \
		ret = job->save(job); \
		if (ret.code) { \
			sm_error("Failed to save job in database: %d:%s", ret.code, ret.string); \
			goto fault; \
		} \
	} while (0)

sm_value_t *
smc_session_process_submit_job(smc_job_t job, sm_value_t *params)
{
	const char *drms_id = NULL;
	sm_value_t *result = NULL;
	smc_job_state_t *state = NULL;

	SM_MANDATORY_P(params);
	SM_MANDATORY_P(job);

	SM_ASSERT(job->get_id(job), "Job must have an id if we're here");

	SM_DEBUG_ENTER();

	if (!submission_module)
		return sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INTERNAL_FAULT, "No submission module loaded");

	/* jsdl, arid */
	assert(SM_VALUE_IS_STRING(params[0]));
	assert(sm_value_count(params) == 1 || (sm_value_count(params) == 3 && SM_VALUE_IS_STRING(params[1]) && SM_VALUE_IS_UINT32(params[2])));

	result = submission_module->submit((sm_module_t) submission_module, params);
	if (!result) {
		sm_error("Submission module didn't return any result");
		goto fault;
	}

	if (SM_VALUE_RET_CODE_OK(result[0])) {
		sm_ret_code_t ret;

		drms_id = result[1]->vstring;
		SM_TRY_NOT(state = sm_malloc(sizeof(*state)));

		/* Let the polling scanner retrieve the initial state. */
		*state = SMC_UNSPECIFIED_STATE;

		job->set_drms_id(job, drms_id);
		job->set_state(job, SMC_QUEUED);
		SAVE_JOB();

		SM_TRY_NOT(sm_value_append(&result, 1, SM_VALUE_STRING, job->get_id(job)));

		sm_mutex_lock(&submitted_jobs_lock);
		if (xmlHashAddEntry(submitted_jobs, (xmlChar *) drms_id, state) != 0) {
			sm_mutex_unlock(&submitted_jobs_lock);
			goto fault;
		}
		sm_cond_broadcast(&submitted_jobs_cond);
		sm_mutex_unlock(&submitted_jobs_lock);

		/* This should not be freed from now on. */
		state = NULL;
	}

	return result;

fault:
	kill_job(drms_id);
	sm_value_free(&result);
	SM_FREE(state);
	return NULL;
}

static void*
transport_malloc(rsync_transport_t this, size_t size)
{
	if(this->data==NULL)
		this->data=sm_alloc_tracked_create();

	return sm_alloc_tracked_malloc(this->data,size);
}
static char*
transport_strdup(rsync_transport_t this, const char *string)
{
	if(this->data==NULL)
			this->data=sm_alloc_tracked_create();

	return sm_alloc_tracked_strdup(this->data,string);
}

static void
free_transport(rsync_transport_t this)
{
	sm_alloc_tracked_free(&this->data);
}

static struct sm_transport_s init_transport = {
		.transport_malloc = transport_malloc,
		.transport_strdup = transport_strdup,
		.data = NULL
};

static sm_status_ret_code_t
handle_GetFiles(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	Files files;
	sm_status_ret_code_t result = SM_FAULT;
	int fun_result;
	struct sm_transport_s transport = init_transport;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_GetFiles(fd, &req) != SM_OK) {
		sm_error("Failed to receive GetFiles request");
		goto quit;
	}

	fun_result = GetFiles(&transport, req[0]->vstring, &files);
	SM_TRY_NOT(response = sm_value_create(1, SM_VALUE_RET_CODE, fun_result, "GetFiles failed"));	//beware, the last argument may not be used in sm_value_create, because

	if (fun_result == RSyncNoError) {
		SM_TRY_NOT(sm_value_append(&response, 1, SM_VALUE_UINT32, files.__sizeFile));
		for (int i = 0; i < files.__sizeFile; ++i)
			SM_TRY_NOT(sm_value_append(&response, 4, SM_VALUE_UINT32, files.File[i].MTime, SM_VALUE_BOOL, files.File[i].isDirectory ,SM_VALUE_BOOL, files.File[i].isExecutable, SM_VALUE_STRING, files.File[i].Path));
	}

	if (smc_icomm_respond_GetFiles(fd, response) != SM_OK) {
		sm_error("Failed to send GetFiles response");
		goto quit;
	}

	result = SM_OK;

quit:
	free_transport(&transport);
	sm_value_free(&req);
	return SM_DEBUG_LEAVE_INT(result);
fault:
	goto quit;
}

static sm_status_ret_code_t
handle_GetSignature(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	sm_status_ret_code_t result = SM_FAULT;
	Signature sig;
	int fun_result;
	struct sm_transport_s transport = init_transport;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_GetSignature(fd, &req) != SM_OK) {
		sm_error("Failed to receive GetSignature request");
		goto fault;
	}

	fun_result = GetSignature(&transport, req[0]->vstring, req[1]->vstring, &sig);
	SM_TRY_NOT(response = sm_value_create(1, SM_VALUE_RET_CODE, fun_result, "Error in GetSignature"));

	if (fun_result == RSyncNoError)
		SM_TRY_NOT(signature_to_icomm(&transport, &sig, &response));

quit:

	if(!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, RSyncInternalError, "Internal error in GetSignature in rsync module");

	if (smc_icomm_respond_GetSignature(fd, response) != SM_OK)
		sm_error("Failed to send GetSignature response");
	else
		result = SM_OK;

	free_transport(&transport);
	sm_value_free(&req);

	return SM_DEBUG_LEAVE_INT(result);

fault:
	goto quit;
}

#ifdef WITH_LIBRSYNC
static sm_status_ret_code_t
handle_GetDelta(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	smc_icomm_request_t request;
	sm_status_ret_code_t result = SM_FAULT;
	int fun_result;
	Signature sig;
	struct sm_transport_s transport = init_transport;
	struct rsync_delta_context *context;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_GetDeltaBegin(fd, &req) != SM_OK) {
		sm_error("Failed to receive GetDelta request");
		goto fault;
	}

	if(!icomm_to_signature(&transport, req + 2, &sig)) {
		sm_error("Invalid request, can't create signature");
		goto fault;
	}

	fun_result = GetDeltaBegin(&transport, req[0]->vstring, req[1]->vstring, &sig, &context);
	sm_value_free(&req);

	SM_TRY_NOT(response = sm_value_create(1, SM_VALUE_RET_CODE, fun_result, "GetDeltaBegin failed"));

	if (smc_icomm_respond_GetDeltaBegin(fd, response) != SM_OK) {
		sm_error("Failed to send GetDeltaBegin response");
		goto fault;
	}

	do {
		char *ptr;
		size_t len;

		fun_result = GetDeltaProcess(context);
		sm_debug("GetDeltaProcess=%i, delta=%p, length=%lu", fun_result, context->delta, (unsigned long)context->delta_length);

		ptr=context->delta;
		len=context->delta_length;

		while(len>0) {
			size_t send_len;

			if (smc_icomm_begin_receive(fd, &request) != SM_OK) {
				sm_error("smc_icomm_begin_receive failed.");
				goto fault;
			}

			sm_debug("Request %i",request);

			if (smc_icomm_receive_GetDeltaProcess(fd, &req) != SM_OK) {
				sm_error("Failed to receive GetDeltaProcess request");
				goto fault;
			}

			send_len=min(req[0]->vuint32, len);
			sm_value_free(&req);

			if(fun_result != 0)
				SM_TRY_NOT(response = sm_value_create(2, SM_VALUE_RET_CODE, fun_result, "GetDeltaProcess failed", SM_VALUE_BYTES, send_len, ptr));
			else
				SM_TRY_NOT(response = sm_value_create(2, SM_VALUE_RET_CODE, 0, SM_VALUE_BYTES, send_len, ptr));

			if (smc_icomm_respond_GetDeltaProcess(fd, response) != SM_OK) {
				sm_error("Failed to send GetDeltaProcess response");
				goto fault;
			}
			ptr+=send_len;
			len-=send_len;
		}
		if(fun_result != RSyncNoError)
			break;
	}
	while(context->status==RS_BLOCKED);
	GetDeltaFinish(context);

	sm_debug("Contex free, waiting for last request");
	if (smc_icomm_begin_receive(fd, &request) != SM_OK) {
		sm_error("smc_icomm_begin_receive failed.");
		goto fault;
	}
	sm_debug("Request %i",request);
	if (smc_icomm_receive_GetDeltaProcess(fd, &req) != SM_OK) {
		sm_error("Failed to receive GetDeltaProcess request");
		goto fault;
	}
	sm_value_free(&req);
	if (smc_icomm_respond_GetDeltaProcess(fd, sm_value_create(1, SM_VALUE_RET_CODE, 0)) != SM_OK) {
		sm_error("Failed to send GetDeltaProcess response");
		goto fault;
	}
	sm_debug("Last request responded, dying.");

	result = SM_OK;
quit:
	free_transport(&transport);
	return SM_DEBUG_LEAVE_INT(result);
fault:
	GetDeltaFinish(context);
	goto quit;
}

static sm_status_ret_code_t
handle_PutDelta(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	sm_status_ret_code_t result = SM_FAULT;
	int fun_result = -1;
	struct sm_transport_s transport = init_transport;
	struct rsync_delta_context *context;
	smc_icomm_request_t request;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_PutDeltaBegin(fd, &req) != SM_OK) {
		sm_error("Failed to receive PutDeltaBegin request");
		goto fault;
	}

	fun_result = PutDeltaBegin(&transport, req[0]->vstring, req[1]->vstring, &context);

	SM_TRY_NOT(response = sm_value_create(1, SM_VALUE_RET_CODE, fun_result, "PutDeltaBegin failed"));

	if (smc_icomm_respond_PutDeltaBegin(fd, response) != SM_OK) {
		sm_error("Failed to send PutDeltaBegin response");
		goto fault;
	}

	if(fun_result != RSyncNoError)
		goto fault;

	sm_value_free(&req);

	do {
		if (smc_icomm_begin_receive(fd, &request) != SM_OK) {
			sm_error("smc_icomm_begin_receive failed.");
			goto fault;
		}
		sm_debug("Request %i",request);

		if (smc_icomm_receive_PutDeltaProcess(fd, &req) != SM_OK) {
			sm_error("Failed to receive PutDeltaProcess request");
			goto fault;
		}

		if(SM_VALUE_IS_BYTES(req[0])) {

			sm_debug("Part of delta size = %lu", (unsigned long)req[0]->size);

			if(context->delta_max_length < context->delta_length+req[0]->size) {
				context->delta_max_length += req[0]->size;
				context->delta = sm_realloc(context->delta, context->delta_max_length);
			}

			sm_debug("New delta = %p, delta length = %lu, max delta length = %lu", context->delta, (unsigned long)context->delta_length, (unsigned long)context->delta_max_length);

			memmove(context->delta + context->delta_length, req[0]->vbytes, req[0]->size);
			context->delta_length += req[0]->size;
		}

		sm_value_free(&req);

		fun_result = PutDeltaProcess(context);

		sm_debug("PutDeltaProcess = %i, left = %lu", fun_result, (unsigned long)context->delta_length);

		if(fun_result != 0)
			SM_TRY_NOT(response = sm_value_create(2, SM_VALUE_RET_CODE, fun_result, "PutDeltaProcess failed", SM_VALUE_INT32, context->status));
		else
			SM_TRY_NOT(response = sm_value_create(2, SM_VALUE_RET_CODE, 0, SM_VALUE_INT32, context->status));

		sm_debug("response = %p", response);

		if (smc_icomm_respond_PutDeltaProcess(fd, response) != SM_OK) {
			sm_error("Failed to send PutDelta response");
			goto quit;
		}
	}
	while(context->status == RS_BLOCKED);

	sm_free(context->delta);
	PutDeltaFinish(context);

	result = SM_OK;

quit:
	free_transport(&transport);
	return SM_DEBUG_LEAVE_INT(result);

fault:

	if(context)
		sm_free(context->delta);

	PutDeltaFree(context);
	goto quit;
}
#endif

static sm_status_ret_code_t
handle_ListDirectory(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	const char *directory = NULL;
	DIR *dir = NULL;
	struct dirent *ent = NULL;
	int count = 0, i = 0;
	char **files = NULL;
	sm_status_ret_code_t result = SM_FAULT;
	char *wd = NULL;
	size_t wd_len = 0;
	char *jsdl_wd = NULL;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_ListDirectory(fd, &req) != SM_OK) {
		sm_error("Failed to receive ListDirectory request");
		goto fault;
	}

	directory = req[0]->vstring;

	sm_info("Listing directory %s", directory);

	dir = opendir(directory);

	if(!dir) {
		char errmsg[256] = "";
		int saved_errno = errno;
		sm_perror("opendir");

		if (saved_errno == ENOENT || saved_errno == ENOTDIR) {
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_FileNotFoundFault, directory);
		} else if (saved_errno == EACCES) {
			snprintf(errmsg, sizeof(errmsg), "User with uid %u is not allowed to access dir %s", getuid(), directory);
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_NotAuthorizedFault, errmsg);
		} else {
			snprintf(errmsg, sizeof(errmsg), "opendir on %s failed: %s", directory, strerror(saved_errno));
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
		}

		goto fault;
	}

	while((ent = readdir(dir)) != NULL) {
		++count;
	}

	rewinddir(dir);

	SM_TRY_NOT(files = sm_calloc(count, sizeof(char*)));

	sm_debug("Found %i files", count);

	while((ent = readdir(dir)) != NULL) {
		SM_TRY_NOT(files[i] = sm_strdup(ent->d_name));
		sm_debug("Found file %s", files[i]);
		i++;
	}

	wd_len = 100;
	SM_TRY_NOT(wd = sm_malloc(wd_len));

	while(getcwd(wd, wd_len) == NULL) {
		int stored_errno = errno;

		sm_perror("getcwd");

		if(stored_errno != ERANGE)
			goto fault;

		wd_len += 100;
		wd = sm_realloc(wd, wd_len);
	}

	sm_debug("Working directory: %s", wd);

	if(strcmp(directory, ".") == 0)
		SM_TRY_NOT(jsdl_wd = sm_strdup(wd));
	else
		SM_TRY_NOT(jsdl_wd = sm_asprintf("%s/%s", wd, directory));


	response = sm_value_create(3, SM_VALUE_RET_CODE, 0, SM_VALUE_STRINGS| SM_VALUE_DONT_COPY_FLAG, count, files, SM_VALUE_STRING|SM_VALUE_DONT_COPY_FLAG, jsdl_wd);
	files = NULL;	//if sm_value_create fails files has been released

	if(!response)
		goto fault;

quit:

	sm_free(wd);

	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in ListDirectory in session process");

	if (smc_icomm_respond_ListDirectory(fd, response) != SM_OK)
		sm_error("Failed to send ListDirectory response");
	else
		result = SM_OK;

	sm_value_free(&req);

	if(dir)
		closedir(dir);

	return SM_DEBUG_LEAVE_INT(result);

fault:

	if(files) {

		for(i=0 ; i<count ; ++i)
			sm_free(files[i]);	//not checking if files[i] is ok, because it's done in sm_free, and not used areas are NULL because sm_calloc memsets memory

		sm_free(files);
	}

	sm_free(jsdl_wd);

	goto quit;
}

static sm_status_ret_code_t
handle_StatFile(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	const char *path = NULL;
	char *user = NULL, *group = NULL;
	struct stat buf;
	sm_status_ret_code_t result = SM_FAULT;
	struct passwd *user_st = NULL;
	struct group *group_st = NULL;
	

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_StatFile(fd, &req) != SM_OK) {
		sm_error("Failed to receive StatFile request");
		goto fault;
	}

	path = req[0]->vstring;

	sm_info("Stat on file: %s", path);

	if (stat(path, &buf) != 0) {
		char errmsg[256] = "";
		int saved_errno = errno;

		sm_perror("stat");

		if (saved_errno == ENOENT || saved_errno == ENOTDIR) {
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_FileNotFoundFault, path);
		} else if (saved_errno == EACCES) {
			snprintf(errmsg, sizeof(errmsg), "User with uid %u is not allowed to access file %s", getuid(), path);
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_NotAuthorizedFault, errmsg);
		} else {
			snprintf(errmsg, sizeof(errmsg), "Stat on file %s failed: %s", path, strerror(saved_errno));
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
		}

		goto fault;
	}

	if ((group_st = getgrgid(buf.st_gid)))
		SM_TRY_NOT(group = sm_strdup(group_st->gr_name));
	else
		SM_TRY_NOT(group = sm_asprintf("%d",buf.st_gid));

	if ((user_st = getpwuid(buf.st_uid)))
		SM_TRY_NOT(user = sm_strdup(user_st->pw_name));
	else
		SM_TRY_NOT(user = sm_asprintf("%d", buf.st_uid));

	sm_debug("File owner %s:%s", user, group);

	SM_TRY_NOT(response = sm_value_create(9,
			SM_VALUE_RET_CODE, 0,
			SM_VALUE_UINT32, (unsigned int)buf.st_atime,
			SM_VALUE_UINT32, (unsigned int)buf.st_ctime,
			SM_VALUE_STRING|SM_VALUE_DONT_COPY_FLAG, group,
			SM_VALUE_UINT32, (unsigned int)buf.st_mtime,
			SM_VALUE_UINT32, (unsigned int)buf.st_mode,
			SM_VALUE_UINT32, (unsigned int)buf.st_nlink,
			SM_VALUE_UINT32, (unsigned int)buf.st_size,
			SM_VALUE_STRING|SM_VALUE_DONT_COPY_FLAG, user
			));

quit:

	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in StatFile in session process");

	if (smc_icomm_respond_StatFile(fd, response) != SM_OK)
		sm_error("Failed to send StatFile response");
	else
		result = SM_OK;

	sm_value_free(&req);

	return SM_DEBUG_LEAVE_INT(result);

fault:
	goto quit;
}

static sm_status_ret_code_t
handle_Delete(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	const char *path = NULL;
	sm_status_ret_code_t result = SM_FAULT;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_Delete(fd, &req) != SM_OK) {
		sm_error("Failed to receive Delete request");
		goto fault;
	}

	path = req[0]->vstring;

	sm_info("Delete: %s", path);

	if(remove(path) != 0) {
		char errmsg[256] = "";
		int saved_errno = errno;

		sm_perror("remove");

		if (saved_errno == ENOENT || saved_errno == ENOTDIR) {
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_FileNotFoundFault, path);
		} else if (saved_errno == EACCES) {
			snprintf(errmsg, sizeof(errmsg), "User with uid %u is not allowed to delete file %s", getuid(), path);
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_NotAuthorizedFault, errmsg);
		} else {
			snprintf(errmsg, sizeof(errmsg), "Deleting of file %s failed: %s", path, strerror(saved_errno));
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
		}

		goto fault;
	}

	SM_TRY_NOT(response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_OK));

quit:

	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in StatFile in session process");

	if (smc_icomm_respond_Delete(fd, response) != SM_OK)
		sm_error("Failed to send Delete response");
	else
		result = SM_OK;

	sm_value_free(&req);

	return SM_DEBUG_LEAVE_INT(result);

fault:
	goto quit;
}

static sm_status_ret_code_t
handle_ChMod(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	const char *path = NULL;
	mode_t mode = 0644;
	sm_status_ret_code_t result = SM_FAULT;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_ChMod(fd, &req) != SM_OK) {
		sm_error("Failed to receive ChMod request");
		goto fault;
	}

	path = req[0]->vstring;
	mode = req[1]->vint32;

	sm_info("Changing mode %s to %o", path, mode);

	if(chmod(path, mode) != 0) {
		char errmsg[256] = "";
		int saved_errno = errno;

		sm_perror("chmod");

		if (saved_errno == ENOENT || saved_errno == ENOTDIR) {
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_FileNotFoundFault, path);
		} else if (saved_errno == EACCES) {
			snprintf(errmsg, sizeof(errmsg), "User with uid %u is not allowed to change mode of file %s", getuid(), path);
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_NotAuthorizedFault, errmsg);
		} else {
			snprintf(errmsg, sizeof(errmsg), "Changing mode of file %s failed: %s", path, strerror(saved_errno));
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
		}

		goto fault;
	}

	SM_TRY_NOT(response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_OK));

quit:

	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in ChMod in session process");

	if (smc_icomm_respond_ChMod(fd, response) != SM_OK)
		sm_error("Failed to send ChMod response");
	else
		result = SM_OK;

	sm_value_free(&req);

	return SM_DEBUG_LEAVE_INT(result);

fault:
	goto quit;
}

static int 
recursive_mkdir(const char *pathname, mode_t mode)
{
	int result = -1;
	SM_BOOL first_try = SM_TRUE;
	char *p = NULL;

	sm_info("recursive_mkdir:%s", pathname);
retry:
	result = mkdir(pathname, mode);

	if (result == -1 && errno == ENOENT && first_try && ( (p = strrchr(pathname, '/')) != NULL)) {
		first_try = SM_FALSE;
		*p = '\0';
		recursive_mkdir(pathname, mode);
		*p = '/';
		goto retry;
	} 

	return result;
}

static sm_status_ret_code_t
handle_MkDir(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *response = NULL;
	const char *path = NULL;
	mode_t mode = 0644;
	sm_status_ret_code_t result = SM_FAULT;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_MkDir(fd, &req) != SM_OK) {
		sm_error("Failed to receive MkDir request");
		goto fault;
	}

	path = req[0]->vstring;
	mode = req[1]->vint32;

	sm_info("Creating directory %s with mode %o", path, mode);

	if(recursive_mkdir(path, mode) != 0) {
		char errmsg[256] = "";
		int saved_errno = errno;

		sm_perror("mkdir");

		if (saved_errno == ENOENT || saved_errno == ENOTDIR) {
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_FileNotFoundFault, path);
		} else if (saved_errno == EACCES) {
			snprintf(errmsg, sizeof(errmsg), "User with uid %u is not allowed to create directory %s", getuid(), path);
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_NotAuthorizedFault, errmsg);
		} else {
			snprintf(errmsg, sizeof(errmsg), "Creating directory of file %s failed: %s", path, strerror(saved_errno));
			response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_StagingFault, errmsg);
		}

		goto fault;
	}

	SM_TRY_NOT(response = sm_value_create(1, SM_VALUE_RET_CODE, qcg_comp_staging_OK));

quit:

	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in MkDir in session process");

	if (smc_icomm_respond_MkDir(fd, response) != SM_OK)
		sm_error("Failed to send MkDir response");
	else
		result = SM_OK;

	sm_value_free(&req);

	return SM_DEBUG_LEAVE_INT(result);

fault:
	goto quit;
}

static sm_status_ret_code_t
handle_CreateActivity(int fd)
{
	sm_value_t *req = NULL;
	sm_value_t *params = NULL;
	sm_value_t *subresult = NULL;
	smc_job_t job = NULL;
	smc_job_res_info_t res_info = NULL;
	char *id = NULL;
	char *drms_id = NULL;
	const char *subject = NULL;
	const char *username = NULL;
	const char *submitted_jsdl = NULL;
	const char *jsdl = NULL;
	sm_ret_code_t ret;
	sm_value_t *response = NULL;
	sm_value_t *atc_token = NULL;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_CreateActivity(fd, &req) != SM_OK) {
		sm_error("Failed to receive CreateActivity request");
		goto fault;
	}

	subject = req[0]->vstring;
	username = req[1]->vstring;
	submitted_jsdl = req[2]->vstring;
	jsdl = req[3]->vstring;

	if (sm_value_count(req) > 4)
		atc_token = req[4]->varray;

	SM_TRY_NOT(id = smc_generate_id((smc_strdup_function) NULL, NULL));
	SM_TRY_NOT(job = smc_job_create());
	job->set_id(job, id);
	job->set_subject(job, subject);
	job->set_mapped_user(job, username);
	job->set_submitted_jsdl(job, submitted_jsdl);
	job->set_jsdl(job, jsdl);

	if (atc_token)
		job->set_atc_token(job, atc_token);

	/*create working directory */
	smc_jsdl_get_working_dir(job->get_jsdl_parsed(job), SM_TRUE);

	if (job->get_reservation_id(job)) {
		res_info = job->get_reservation_info(job);

		if (!res_info) {
			SM_TRY_NOT(subresult = sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INVALID_REQUEST_MESSAGE_FAULT, "Unknown reservation identifier"));
			goto reply;
		}

		if (res_info->cancelled) {
			SM_TRY_NOT(subresult = sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INVALID_REQUEST_MESSAGE_FAULT, "Reservation identifier refers to cancelled reservation"));
			goto reply;
		}

		if (res_info->expired) {
			SM_TRY_NOT(subresult = sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INVALID_REQUEST_MESSAGE_FAULT, "Reservation identifier refers to expired reservation"));
			goto reply;
		}
	}

	if (smc_jsdl_stage_in_requested(job->get_jsdl_parsed(job))) {

		SM_TRY_NOT_OK(smc_staging_start_stage_in(job, res_info, module_manager));
		SM_TRY_NOT(subresult = sm_value_create(3, SM_VALUE_RET_CODE, SUBMISSION_OK, SM_VALUE_NONE, SM_VALUE_STRING, id));

	} else if (smc_jsdl_has_application(job->get_jsdl_parsed(job))) {
		/* We use job->get_jsdl(job) and not jsdl to have WorkingDirectory
		 * element properly set by job module. */
		if (res_info) {
			SM_TRY_NOT(params = sm_value_create(3, SM_VALUE_STRING, job->get_jsdl(job), SM_VALUE_STRING, res_info->drms_id, SM_VALUE_UINT32, res_info->duration));
		} else {
			SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, job->get_jsdl(job)));
		}

		SM_TRY_NOT(subresult = smc_session_process_submit_job(job, params));
		if (SM_VALUE_RET_CODE_OK(subresult[0]))
			drms_id = sm_strdup(subresult[1]->vstring);

	} else if (smc_jsdl_stage_out_requested(job->get_jsdl_parsed(job))) {
		sm_debug("Activity has no application defined. No need to stage in, proceeding directly to stage out.");
		SM_TRY_NOT_OK(smc_staging_start_stage_out(job, module_manager));
		SM_TRY_NOT(subresult = sm_value_create(3, SM_VALUE_RET_CODE, SUBMISSION_OK, SM_VALUE_NONE, SM_VALUE_STRING, id));
	} else {
		sm_debug("Empty Activity, finishing.");
		job->set_state(job, SMC_FINISHED);
		SAVE_JOB();
		SM_TRY_NOT(subresult = sm_value_create(3, SM_VALUE_RET_CODE, SUBMISSION_OK, SM_VALUE_NONE, SM_VALUE_STRING, id));
	}

reply:
	if (smc_icomm_respond_CreateActivity(fd, subresult) != SM_OK) {
		kill_job(drms_id);
		sm_error("Failed to send CreateActivity response");
		goto fault;
	}

	sm_value_free(&req);
	sm_value_free(&params);
	smc_job_free(&job);

	if (res_info) {
		SM_FREE(res_info->drms_id);
		SM_FREE(res_info);
	}

	SM_FREE(id);
	SM_FREE(drms_id);
	return SM_OK;

fault:
	sm_value_free(&req);
	sm_value_free(&params);
	smc_job_free(&job);

	if (res_info) {
		SM_FREE(res_info->drms_id);
		SM_FREE(res_info);
	}

	SM_FREE(id);
	if (!response)
		response = sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INTERNAL_FAULT, "Internal fault in session process");
	if (smc_icomm_respond_CreateActivity(fd, response) != SM_OK)
		sm_error("Failed to send CreateActivity response");
	return SM_FAULT;
}

static sm_status_ret_code_t
handle_GetActivityStatuses(int fd)
{
	sm_value_t *params = NULL;
	sm_value_t *subresult = NULL;
	sm_list_t extra_info_list = NULL;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_GetActivityStatuses(fd, &params) != SM_OK) {
		sm_error("Failed to receive GetActivityStatuses request");
		goto fault;
	}

	if (!submission_module) {
		subresult = sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INTERNAL_FAULT, "No submission module loaded");
	} else {
		subresult = submission_module->get_statuses((sm_module_t) submission_module, params);
		if (!subresult) {
			sm_error("Submission module didn't return any result");
			goto fault;
		}
	}

	/* XXX: multiple identifiers */
	sm_signal_emit("smc_job_info", local_resource_manager_type[2]->vstring, params[0]->vstring, &extra_info_list);
	SM_ASSERT((!extra_info_list || sm_list_count(extra_info_list) == 1), "at most one module can be connected to the smc_job_info signal");
	if (extra_info_list) {
		sm_value_append(&subresult, 1, SM_VALUE_ARRAY, (sm_value_t *)extra_info_list->data);
	}

	if (smc_icomm_respond_GetActivityStatuses(fd, subresult) != SM_OK) {
		sm_error("Failed to send GetActivityStatuses response");
		goto fault;
	}

	/* sm_signal_emit already freed params: sm_value_free(&params); */
	return SM_OK;

fault:
	sm_value_free(&params);
	if (smc_icomm_respond_GetActivityStatuses(fd, sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INTERNAL_FAULT, "Internal fault in session process")) != SM_OK)
		sm_error("Failed to send GetActivityStatuses response");
	return SM_FAULT;
}

static sm_status_ret_code_t
handle_TerminateActivities(int fd)
{
	sm_value_t *params = NULL;
	sm_value_t *subresult = NULL;
	smc_job_t *jobs = NULL;
	size_t i, jobs_count = 0;
	sm_ret_code_t ret;

	SM_DEBUG_ENTER();

	if (smc_icomm_receive_TerminateActivities(fd, &params) != SM_OK) {
		sm_error("Failed to receive TerminateActivities request");
		goto fault;
	}

	if (!submission_module) {
		subresult = sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INTERNAL_FAULT, "No submission module loaded");
	} else {
		jobs_count = sm_value_count(params);

		SM_TRY_NOT(jobs = sm_calloc(jobs_count, sizeof(smc_job_t)));

		for (i =0; i < jobs_count; i++) {
			SM_TRY_FAULT(smc_job_get_by_drms_id(params[i]->vstring, &jobs[i]).code);
		}

		subresult = submission_module->terminate((sm_module_t) submission_module, params);

		if (!subresult) {
			sm_error("Submission module didn't return any result");
			goto fault;
		}

		SM_ASSERT(sm_value_count(subresult) == jobs_count, "sm_value_count(subresult) == jobs_count");

		/* Clean up the files here because later (when submission_drmaa returns
		 * in the waiting thread) we're in a situation that the state is already
		 * set and saved in database so there's a possibility of a race
		 * condition if someone decides to submit the same job immediately. */
		for (i = 0; i < jobs_count; i++) {
			smc_job_t job = jobs[i];
			if (job && SM_RET_CODE_OK(subresult[i]->vretcode)) {
				smc_staging_delete_on_termination(job, true);
				job->set_state(job, SMC_CANCELLED);
				SAVE_JOB();
			}
		}

	}

	if (smc_icomm_respond_TerminateActivities(fd, subresult) != SM_OK) {
		sm_error("Failed to send TerminateActivities response");
		goto fault;
	}

	if (jobs) {
		for (i = 0; i < jobs_count; i++)
			smc_job_free(&jobs[i]);

		sm_free(jobs);
	}
	sm_value_free(&params);
	return SM_OK;

fault:
	if (jobs) {
		for (i = 0; i < jobs_count; i++)
			smc_job_free(&jobs[i]);

		sm_free(jobs);
	}
	sm_value_free(&params);
	if (smc_icomm_respond_TerminateActivities(fd, sm_value_create(1, SM_VALUE_RET_CODE, SUBMISSION_INTERNAL_FAULT, "Internal fault in session process")) != SM_OK)
		sm_error("Failed to send TerminateActivities response");
	return SM_FAULT;
}

static void
submitted_jobs_deallocator(void *payload, xmlChar *name sm_unused)
{
	SM_FREE(payload);
}

static void *
submitted_jobs_copier(void *payload, xmlChar *name sm_unused)
{
	smc_job_state_t *state;

	state = sm_malloc(sizeof(*state));
	if (!state)
		return NULL;

	*state = *((smc_job_state_t *) payload);
	return state;
}

static void
parse_rusage(sm_value_t *array, smc_job_t job)
{
	sm_value_t *result = NULL;
	size_t i;

	SM_DEBUG_ENTER();

	/* Sun Grid Engine */
	if (local_resource_manager_type[1]->vint32 == SUBMISSION_DRMS_GE) {

		for (i = 0; i < sm_value_count(array); i++) {
			char name[32] = "";
			char value[32] = "";

			SM_ASSERT(SM_VALUE_IS_STRING(array[i]), "rusage array contains non-string element");

			if (sscanf(array[i]->vstring, "%31[^=]=%31s", name, value) == 2) {

				if (!strcmp(name, "cpu"))
					job->set_cpu_usage(job, atof(value));
				else if (!strcmp(name, "mem"))
					job->set_mem_usage(job, atof(value));
				else if (!strcmp(name, "start_time")) {
					time_t timet = atof(value);
					struct tm timetm;

					localtime_r(&timet, &timetm);
					job->set_start_time(job, timetm);
				}

				/* We don't use submission_time and end_time as
				 * these values should include staging and thus
				 * are set by job module. */
			}
		}
	} else if (local_resource_manager_type[1]->vint32 == SUBMISSION_DRMS_TORQUE) {
			sm_info("parsing torque rusage");

			for (i = 0; i < sm_value_count(array); i++) {
				int slen = strlen(array[i]->vstring) + 1;
				char name[32] = "";
				char *value = sm_malloc(slen * sizeof(char));

				SM_ASSERT(SM_VALUE_IS_STRING(array[i]), "rusage array contains non-string element");

				if (value && sscanf(array[i]->vstring, "%31[^=]=%s", name, value) == 2) {
					sm_info("rusage: %s=%s", name, value);

					if (!strcmp(name, "cpu"))
						job->set_cpu_usage(job, atof(value));
					else if (!strcmp(name, "mem"))
						job->set_mem_usage(job, atof(value));
					else if (!strcmp(name, "vmem"))
						job->set_vmem_usage(job, atof(value));
					else if (!strcmp(name, "walltime"))
						job->set_walltime(job, atoi(value));
					else if (!strcmp(name, "hosts"))
						job->set_nodes(job, value);
					else if (!strcmp(name, "queue"))
						job->set_queue(job, value);
					else if (!strcmp(name, "project"))
						job->set_project(job, value);
					else if (!strcmp(name, "start_time")) {
						time_t timet = atof(value);
						struct tm timetm;

						localtime_r(&timet, &timetm);
						job->set_start_time(job, timetm);
					}


					/* We don't use submission_time and end_time as
					 * these values should include staging and thus
					 * are set by job module. */
				} else {
					sm_error("failed to parse rusage: %s", array[i]->vstring);
				}

				sm_free(value);
			}
	}

	sm_value_free(&result);
}

static void *
submission_waiting_thread_loop(void *arg sm_unused)
{
	sm_value_t *result = NULL;

	assert(submission_module);

	SM_DEBUG_ENTER();

	for (;;) {
		const char *jobid = NULL;
		smc_job_t job = NULL;
		sm_ret_code_t ret;

		sm_mutex_lock(&submitted_jobs_lock);

		while (submitted_jobs && xmlHashSize(submitted_jobs) == 0) {
			sm_debug("No submitted jobs, submission waiting thread is idle");
			sm_cond_wait(&submitted_jobs_cond, &submitted_jobs_lock);
		}

		sm_mutex_unlock(&submitted_jobs_lock);

		if (!submitted_jobs) {
			sm_value_free(&result);
			break;
		}

		result = submission_module->wait((sm_module_t) submission_module);

		if (!result || SM_VALUE_RET_CODE_C(result[0])) {
			sm_error("Submission module failed to wait (%s). Retrying in 5 seconds...", SM_VALUE_RET_CODE_STRING(result[0]));
			sm_value_free(&result);
			sleep(5);
			continue;
		}

		assert(SM_VALUE_IS_STRING(result[2]));
		jobid = result[2]->vstring;

		ret = smc_job_get_by_drms_id(jobid, &job);
		if (ret.code) {
			if (ret.code != SMC_JOB_NOT_FOUND) {
				sm_error("Job %s done but couldn't retrieve object from database (%d:%s)", jobid, ret.code, ret.string);
			} else {
				ret = smc_job_get_by_drms_id_readonly(jobid, &job);

				if (ret.code) {
					sm_error("Job %s done but couldn't retrieve object from database, accounting table (%d:%s)", jobid, ret.code, ret.string);
				} else {
					/* This job was cancelled because we found it in the
					 * accounting table. There's nothing we can do now. */
				}
			}

			goto dispose;
		}

		if (result[1]->vint32 == SUBMISSION_FAILED) {
			job->set_core_dumped(job, result[3]->vbool);
			if (!SM_VALUE_IS_NONE(result[4]))
				job->set_posix_signal_name(job, result[4]->vstring);

			job->set_state(job, SMC_FAILED);
			smc_staging_delete_on_termination(job, true);
			SAVE_JOB();
		} else {

			job->set_exit_status(job, result[3]->vint32);
			if (!SM_VALUE_IS_NONE(result[4]))
				parse_rusage(result[4]->varray, job);

			if (smc_jsdl_stage_out_requested(job->get_jsdl_parsed(job))) {

				if (smc_staging_start_stage_out(job, module_manager) == SM_FAULT)
					sm_error("Failed to initiate stage out process for job %s", job->get_id(job));

			} else {
				job->set_state(job, SMC_FINISHED);
				smc_staging_delete_on_termination(job, false);
				SAVE_JOB();
			}
		}

fault:
dispose:
		sm_mutex_lock(&submitted_jobs_lock);
		xmlHashRemoveEntry(submitted_jobs, (xmlChar *) jobid, submitted_jobs_deallocator);
		sm_cond_broadcast(&submitted_jobs_cond);
		sm_mutex_unlock(&submitted_jobs_lock);

		smc_job_free(&job);
		sm_value_free(&result);
	}

	sm_thread_exit(NULL);
	return NULL;
}

static void
polling_scanner(void *payload, void *data sm_unused, xmlChar *name)
{
	smc_job_state_t *state = payload;
	smc_job_state_t *original_state;
	sm_value_t *params = NULL;
	sm_value_t *result = NULL;
	smc_job_t job = NULL;
	sm_ret_code_t ret;

	/*
	 * If full polling is turned off, we're only monitoring jobs that
	 * haven't entered the running state yet.
	 *
	 * In case of full polling we're also monitoring the job if it is
	 * already running -- in case it is suspended externally (local user or
	 * admin).
	 */
	if (!job_status_full_polling) {
		if (*state != SMC_UNSPECIFIED_STATE && *state != SMC_QUEUED && *state != SMC_HELD)
			return;
	}

	SM_DEBUG_ENTER_STR(name);

	ret = smc_job_get_by_drms_id((const char *) name, &job);
	if (ret.code)
		goto fault;

	if (job->get_state(job) == SMC_CANCELLED)
		goto out;

	SM_TRY_NOT(params = sm_value_create(1, SM_VALUE_STRING, name));
	SM_TRY_NOT(result = submission_module->get_statuses((sm_module_t) submission_module, params));

	if (SM_VALUE_RET_CODE_OK(result[0])) {

		job->set_state(job, result[1]->vint32);

		/* This shouldn't hurt if there's actually no state change. */
		ret = job->save(job);
		if (ret.code)
			sm_error("Failed to update job %s in database (%d:%s)", job->get_drms_id(job), ret.code, ret.string);

		/* We are working on a copy of submitted_jobs, so we need to
		 * update state in the original hash table. */
		sm_mutex_lock(&submitted_jobs_lock);
		if ((original_state = xmlHashLookup(submitted_jobs, name)))
			*original_state = result[1]->vint32;
		sm_mutex_unlock(&submitted_jobs_lock);
	}

out:
fault:
	sm_value_free(&params);
	sm_value_free(&result);
	smc_job_free(&job);
}

static void *
submission_polling_thread_loop(void *arg sm_unused)
{
	unsigned int i;
	xmlHashTablePtr submitted_jobs_polling_copy = NULL;

	assert(submission_module);

	SM_DEBUG_ENTER();

	for (;;) {
		sm_mutex_lock(&submitted_jobs_lock);

		while (submitted_jobs && xmlHashSize(submitted_jobs) == 0) {
			sm_debug("No submitted jobs, submission polling thread is idle");
			sm_cond_wait(&submitted_jobs_cond, &submitted_jobs_lock);
		}

		if (!submitted_jobs) {
			sm_mutex_unlock(&submitted_jobs_lock);
			break;
		}

		/* We want to work on a copy of the table so not to block job
		 * submission during a polling phase. */
		submitted_jobs_polling_copy = xmlHashCopy(submitted_jobs, submitted_jobs_copier);
		sm_mutex_unlock(&submitted_jobs_lock);

		if (submitted_jobs_polling_copy) {
			xmlHashScan(submitted_jobs_polling_copy, polling_scanner, NULL);
			xmlHashFree(submitted_jobs_polling_copy, submitted_jobs_deallocator);
		}

		for (i = 0; i < job_status_polling_period; i++) {
			bool done;

			sleep(1);

			sm_mutex_lock(&submitted_jobs_lock);
			done = !submitted_jobs;
			sm_mutex_unlock(&submitted_jobs_lock);

			if (done)
				break;
		}
	}

	sm_thread_exit(NULL);
	return NULL;
}

static sm_status_ret_code_t
session_process_setup(sm_value_t *data, SM_BOOL one_shot)
{
	sm_module_api_t api;
	sm_module_mode_t mode;
	sm_status_ret_code_t status;
	struct passwd *pwd;
	int ret;

	SM_MANDATORY_P(data);

	SM_DEBUG_ENTER_INT(one_shot);

	errno = 0;
	if (!(pwd = getpwuid(getuid()))) {
		sm_perror("getpwuid");
		goto fault;
	}

	/* Fail-safe */
	ret = chdir("/tmp");

	if (!pwd->pw_dir) {
		char buf[256];
		snprintf(buf, sizeof(buf), "User (%s) doesn't have a home directory?", pwd->pw_name);
		sm_error("%s", buf);
		goto fault;
	}

	if (chdir(pwd->pw_dir)) {
		char buf[256];
		sm_perror("chdir");
		snprintf(buf, sizeof(buf), "Failed to change directory to user's (%s) home directory: %s", pwd->pw_name, pwd->pw_dir);
		sm_notice("%s", buf);
	}

	if (sm_setenv("HOME", pwd->pw_dir) || sm_setenv("PWD", pwd->pw_dir) || sm_setenv("USER", pwd->pw_name)) {
		sm_error("sm_setenv failed");
		goto fault;
	}

	sm_mutex_init(&staging_in_progress_lock);

	sm_mutex_init(&submitted_jobs_lock);
	sm_cond_init(&submitted_jobs_cond);

	conf = sm_conf_setup_from_memory(data[0]->vstring, data[1]->vstring);
	if (!conf) {
		sm_error("Failed to read configuration");
		goto fault;
	}

	sm_conf_ctxt_set_default_prefix(conf, data[2]->vstring);
	sm_conf_ctxt_set_service_prefix(conf, data[2]->vstring);

	if (smc_db_setup(conf) != SM_OK) {
		sm_error("Failed to set up database module");
		goto fault;
	}

	if (smc_job_setup(smc_db_get_handle(), true) != SM_OK) {
		sm_error("Failed to set up job module");
		goto fault;
	}

	if (smc_rsvn_setup(smc_db_get_handle()) != SM_OK) {
		sm_error("Failed to set up rsvn module");
		goto fault;
	}

	if (!(module_manager = sm_module_manager_create(conf))) {
		sm_error("Failed to create module manager");
		goto fault;
	}

	status = sm_conf_node_uint(conf, SM_FALSE, "/conf:SessionProcessTimeout", &session_process_timeout);
	status |= sm_conf_node_uint(conf, SM_FALSE, "/conf:JobStatusPollingPeriod", &job_status_polling_period);
	status |= sm_conf_node_bool(conf, SM_FALSE, "/conf:JobStatusFullPolling", &job_status_full_polling);
	status |= sm_conf_node_uint(conf, SM_FALSE, "/conf:SubmittedJobsHashTable", &submitted_jobs_hash_table);

	if (status & SM_FAULT) {
		sm_error("Failed to read configuration");
		goto fault;
	}

	if (submitted_jobs_hash_table < 1024)
		submitted_jobs_hash_table = 1024;

	sm_debug("session_process_timeout = %u", session_process_timeout);
	sm_debug("job_status_polling_period = %u", job_status_polling_period);
	sm_debug("job_status_full_polling = %u", job_status_full_polling);
	sm_debug("submitted_jobs_hash_table = %u", submitted_jobs_hash_table);

	if (!(submitted_jobs = xmlHashCreate(submitted_jobs_hash_table))) {
		sm_error("Failed to create submitted jobs hash table. Too big?");
		goto fault;
	}

	/* all modules */
	api = SM_API_ALL;

	/* excluding those defined in qcg-core */
	api &= ~SM_API_ALL_QCG_CORE;

	/* but we do want general modules*/
	api |= SM_API_GENERAL;
	/*and client modules (e.g. for staging) */
	api |= (SM_API_ECM_CLIENT | SM_API_ECM | SM_API_ATC_TRANSPORT | SM_API_ATC_MSG);
	/* XXX: consider enhancing module manager api with "module mode awareness"*/

	mode = SM_MODULE_CLIENT_MODE | SM_MODULE_UNKNOWN_MODE;

	if (module_manager->load_all(module_manager, &api, &mode) != SM_OK) {
		sm_error("Failed to load modules");
		goto fault;
	}

	submission_module = (submission_module_t) module_manager->find_by_api_first(module_manager, SUBMISSION_MODULE_API);
	reservation_module = (reservation_module_t) module_manager->find_by_api_first(module_manager, RESERVATION_MODULE_API);

	if (submission_module && !one_shot) {
		if (sm_application_thread_create(&submission_waiting_thread, NULL, submission_waiting_thread_loop, NULL) != SM_OK) {
			sm_error("Failed to create submission waiting thread");
			goto fault;
		}

		have_waiting_thread = true;

		if (job_status_polling_period > 0) {
			if (sm_application_thread_create(&submission_polling_thread, NULL, submission_polling_thread_loop, NULL) != SM_OK) {
				sm_error("Failed to create submission polling thread");
				goto fault;
			}

			have_polling_thread = true;
		}
	}

	local_resource_manager_type = submission_module->get_drms((sm_module_t)submission_module);

	if (! SM_VALUE_RET_CODE_OK( local_resource_manager_type[0]) ) {
		sm_value_free(&local_resource_manager_type);
		sm_error("get_drms() failed");
		goto fault;
	} else {
		int drms_type = local_resource_manager_type[1]->vint32;
		char *drms_namespace = local_resource_manager_type[2]->vstring;
		sm_info("drms type = %d (ns=%s)" , drms_type, drms_namespace);
	}

	/* Setup job info handler (monitoring DRMS) */
	if (sm_signal_register("smc_job_info", 0, (sm_value_type_t []) { SM_VALUE_STRING, SM_VALUE_STRING}, 2, (sm_value_type_t []) { SM_VALUE_STRING }, 1) !=  SM_OK)
		goto fault;

	return SM_OK;

fault:
	sm_module_manager_free(&module_manager);
	sm_conf_ctxt_free(&conf);
	return SM_FAULT;
}

static void
session_process_cleanup(void)
{
	int count;

	SM_DEBUG_ENTER();

	sm_mutex_lock(&submitted_jobs_lock);

	while ((count = xmlHashSize(submitted_jobs)) > 0) {
		sm_debug("Submission waiting thread still has %d jobs to wait for, waiting...", count);
		sm_cond_wait(&submitted_jobs_cond, &submitted_jobs_lock);
	}

	sm_debug("No jobs, cleaning up");

	xmlHashFree(submitted_jobs, NULL);
	submitted_jobs = NULL;

	sm_debug("Waking up the threads");
	sm_cond_broadcast(&submitted_jobs_cond);

	sm_mutex_unlock(&submitted_jobs_lock);

	/* We need to wait for all the threads using our locks. */
	if (have_waiting_thread) {
		sm_debug("Waiting for waiting thread");
		sm_thread_join(submission_waiting_thread, NULL);
	}

	if (have_polling_thread && job_status_polling_period) {
		sm_debug("Waiting for polling thread");
		sm_thread_join(submission_polling_thread, NULL);
	}

	sm_cond_destroy(&submitted_jobs_cond);
	sm_mutex_destroy(&submitted_jobs_lock);
	sm_mutex_destroy(&staging_in_progress_lock);

	smc_job_cleanup();
	smc_db_cleanup();

	sm_module_manager_free(&module_manager);
	sm_conf_ctxt_free(&conf);

	sm_signal_unregister("smc_job_info");

	sm_value_free(&local_resource_manager_type);

	SM_DEBUG_LEAVE_VOID();
}

static void
handle_one_shot(int fd)
{
	SM_DEBUG_ENTER();

	smc_icomm_request_t request;

	if (smc_icomm_begin_receive(fd, &request) != SM_OK) {
		sm_error("Failed to receive one shot request");
		return;
	}

	switch (request) {
		case smc_icomm_CreateReservation:
			handle_CreateReservation(fd);
			break;
		case smc_icomm_TerminateReservation:
			handle_TerminateReservation(fd);
			break;
		case smc_icomm_GetReservationStatus:
			handle_GetReservationStatus(fd);
			break;
		case smc_icomm_GetActiveReservations:
			handle_GetActiveReservations(fd);
			break;
		case smc_icomm_StageInCreateContext:
			handle_StageIn(fd);
			break;
		case smc_icomm_StageOutCreateContext:
			handle_StageOut(fd);
			break;
		case smc_icomm_GetDeltaBegin:
			handle_GetDelta(fd);
			break;
		case smc_icomm_PutDeltaBegin:
			handle_PutDelta(fd);
			break;
		default:
			SM_ASSERT(0, "Unknown request (%d) in %s", request, __FUNCTION__);
	}
}

void
smc_session_process(int fd, int controlling_fd, SM_BOOL one_shot, sm_value_t *data)
{
	sm_status_ret_code_t ret;
	SM_BOOL pending_request;
	struct timeval tv;
	time_t last_activity;

	sm_debug("Session process for uid = %u starting (one_shot = %d)", getuid(), one_shot);

	if (session_process_setup(data, one_shot) != SM_OK) {
		sm_error("Failed to set up session process for uid = %u", getuid());
		return;
	}

	if (one_shot) {
		handle_one_shot(fd);
		goto out;
	}

	pending_request = SM_FALSE;
	last_activity = 0;

	for (;;) {
		fd_set readfds;
		int status;

		FD_ZERO(&readfds);

		if (fd != -1 && controlling_fd != -1) {
			FD_SET(fd, &readfds);
			FD_SET(controlling_fd, &readfds);
		}

		tv.tv_sec = 1;
		tv.tv_usec = 0;

		status = select(MAX(fd, controlling_fd) + 1, &readfds, NULL, NULL, &tv);

		if (status == -1) {
			if (errno == EINTR)
				continue;
			sm_perror("select");
			break;
		}

		if (status == 0) {
			SM_BOOL terminating;

			/* We've agreed to accept a request but haven't served it yet. */
			if (pending_request)
				continue;

			/* We haven't served any request yet. */
			if (!last_activity)
				continue;

			sm_mutex_lock(&submitted_jobs_lock);
			terminating = (xmlHashSize(submitted_jobs) == 0);
			sm_mutex_unlock(&submitted_jobs_lock);

			if (terminating) {
				sm_mutex_lock(&staging_in_progress_lock);
				terminating = (staging_in_progress == 0);
				if (!terminating)
					sm_debug("No jobs to wait for but staging_in_progress = %d", staging_in_progress);
				sm_mutex_unlock(&staging_in_progress_lock);
			}

			if (!terminating) {
				last_activity = time(NULL);
			} else {
				time_t passed = time(NULL) - last_activity;

				if (passed < (time_t) session_process_timeout && fd != -1 && controlling_fd != -1)
					continue;

				sm_debug("Terminating session process after %u seconds of inactivity", (unsigned int) passed);
				break;
			}

			continue;
		}

		if (fd != -1 && FD_ISSET(fd, &readfds)) {
			smc_icomm_request_t request;

			last_activity = time(NULL);
			pending_request = SM_FALSE;

			if (smc_icomm_begin_receive(fd, &request) != SM_OK) {
				sm_debug("Lost connection to the application");

				/* This is not a critical error. Session process
				 * might have been orphaned. We continue until
				 * we wait for the last job and all stagers
				 * finish. */
				SM_CLOSE(fd);
				SM_CLOSE(controlling_fd);
				continue;
			}

			switch (request)
			{
			case smc_icomm_CreateActivity:
				ret = handle_CreateActivity(fd);
				break;
			case smc_icomm_GetActivityStatuses:
				ret = handle_GetActivityStatuses(fd);
				break;
			case smc_icomm_TerminateActivities:
				ret = handle_TerminateActivities(fd);
				break;
			case smc_icomm_GetFiles:
				ret = handle_GetFiles(fd);
				break;
			case smc_icomm_GetSignature:
				ret = handle_GetSignature(fd);
				break;
			case smc_icomm_ListDirectory:
				ret = handle_ListDirectory(fd);
				break;
			case smc_icomm_StatFile:
				ret = handle_StatFile(fd);
				break;
			case smc_icomm_Delete:
				ret = handle_Delete(fd);
				break;
			case smc_icomm_ChMod:
				ret = handle_ChMod(fd);
				break;
			case smc_icomm_MkDir:
				ret = handle_MkDir(fd);
				break;
			default:
				SM_ASSERT(0, "Unknown request (%d) in %s", request, __FUNCTION__);
			}

			if (ret != SM_OK) {
				/* We are detaching and will be exiting as soon
				 * as possible -- when all jobs and stagers
				 * finish. */
				SM_CLOSE(fd);
				SM_CLOSE(controlling_fd);
				continue;
			}
		}

		if (controlling_fd != -1 && FD_ISSET(controlling_fd, &readfds)) {
			ret = sm_privsep_handle_controlling_fd(controlling_fd);

			if (ret != SM_OK) {
				sm_debug("Lost connection to the spawner");

				/* This is not a critical error. Session process
				 * might have been orphaned. We continue until
				 * we wait for the last job and all stagers
				 * finish. */
				SM_CLOSE(fd);
				SM_CLOSE(controlling_fd);
				continue;
			}

			assert(!pending_request);
			pending_request = SM_TRUE;
		}
	}

out:
	/* Close all the descriptors sooner so that the spawner knows we're not available. */
	SM_CLOSE(fd);
	SM_CLOSE(controlling_fd);
	session_process_cleanup();
	sm_debug("Session process for uid = %u shutting down", getuid());
}
