/* $Id: qcg_coordinator_client.c 2 2009-10-05 10:09:52Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <assert.h>
#include <stdlib.h>

#include <qcg-core/sm.h>
#include <qcg-core/closure.h>
#include <qcg-core/client.h>
#include <qcg-core/event_timer.h>
#include <qcg-core/logger.h>
#include <qcg-core/metric.h>
#include <qcg-core/python.h>
#include <qcg-core/signal.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>

#include <qcg-core/modules/atc_transport_module.h>
#include <qcg-core/modules/ecm_module.h>
#include <qcg-core/modules/ecm_client_module.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>

#include "stub/qcgcoordinator.nsmap"
#include "stub/qcgcoordinatorStub.h"
#include "qcg_coordinator_client.h"

#define OP_UNDEFINED	0
#define OP_GET		1
#define OP_PUT		2
#define OP_ELECT	3

#define OP_GET_TOKEN	"get"
#define OP_PUT_TOKEN	"put"
#define OP_ELECT_TOKEN	"elect"

#define PROCESS_ENTRY_KEY(entry)	((entry)->qcgcoordinator__ProcessEntryHeader->qcgcoordinator__Key)
#define PROCESS_ENTRY_URL(entry)	((entry)->qcgcoordinator__ProcessEntryHeader->qcgcoordinator__GridCoordinatorUrl)
#define PROCESS_ENTRY_DATA(entry)	((entry)->qcgcoordinator__ProcessData)

__SM_ID("$Id: qcg_coordinator_client.c 2 2009-10-05 10:09:52Z mamonski $");

static void print_usage(FILE *f);
static sm_status_ret_code_t args_handler(int argc, char **args);
static sm_status_ret_code_t opt_handler(char opt, char *optvalue);

static int start_client(sm_client_transport_t transport);
static void print_fault(sm_client_transport_t transport);
static const void *client_params(void);
static int client_setup(sm_module_t this);
static void client_cleanup(sm_module_t this);
static int smc_coordinator_GetProcessEntry(sm_client_transport_t, int*, smc_coordinator_process_entry_t*, int, smc_coordinator_process_entry_t**,int*);
static int smc_coordinator_PutProcessEntry(sm_client_transport_t, smc_coordinator_process_entry_t*, int);

static int op = OP_UNDEFINED;
static char** op_params = NULL;
static int op_params_length = 0;
static int op_params_dimensions = 0;
static char* grid_coordinator_url;
static int timeout_global = -1;

static struct smc_coordinator_client_api_s client_api;

SM_ECM_CLIENT_MODULE = {

	.header = {
		.entry_size = SM_ECM_CLIENT_MODULE_ENTRY_SIZE,
		.api_type = SM_API_ECM_CLIENT,

		.name = "qcg_coordinator_client",
		.descr = "QCG Coordinator gSOAP client module",

		.version_major = 0,
		.version_minor = 1,

		.setup = client_setup,
		.cleanup = client_cleanup
	},

	.print_usage = print_usage,
	.opt_handler = opt_handler,
	.args_handler = args_handler,
	.start_client = start_client,
	.params = client_params,
	.opt_string = "t:",

	.client_api = &client_api,

	.have_wsa_header = SM_TRUE,
	.have_wsse_header = SM_TRUE
};

/* WS-Addressing plugin */
#include <qcg-core/modules/ecm_gsoap/gsoap/wsaapi.c>

/* WS-Security plugin */
#include <qcg-core/modules/ecm_gsoap/gsoap/wsseapi.c>
#include <qcg-core/modules/ecm_gsoap/gsoap/smdevp.c>
#include <qcg-core/modules/ecm_gsoap/gsoap/threads.c>

void print_usage(FILE *f)
{
	fprintf(f,"qcg-coordinator [-t timeout] <operation> <options> \n");
	fprintf(f,"where possible operations and corresponding options are:\n");
	fprintf(f,"  get <key>\n");
	fprintf(f,"  put <key> [N1 value1 value2 ... valueN1 N2 value1 value2 ... valueN2 ...]\n");
	fprintf(f,"  elect <key>\n");
}

sm_status_ret_code_t
opt_handler(char opt, char *optvalue)
{
	switch (opt) {
		case 't':
		{
			if (sscanf(optvalue, "%d",&timeout_global) != 1) {
				sm_error("Illegal -t option value: %s", optvalue);
				return SM_NEG;
			}

			if (timeout_global < 0) {
				sm_error("Illegal -t option value: %s", optvalue);
				return SM_NEG;
			}

			break;
		}
		default:
		{
			sm_error("Illegal option: -%c", opt);
			return SM_NEG;
		}
	}

	return SM_OK;
}

sm_status_ret_code_t
args_handler(int argc, char **args)
{
	int n;
	int tmp;

	SM_DEBUG_ENTER_INT(argc);

	if (argc < 2) {
		sm_error("Illegal number of arguments: %d", argc);
		return SM_NEG;
	}

	if (!strcmp(args[0],OP_GET_TOKEN)) {
		op = OP_GET;

		if (argc != 2) {
			sm_error("Illegal number of arguments: %d",argc);
			return SM_NEG;
		}
	} else if (!strcmp(args[0],OP_PUT_TOKEN)) {
		op = OP_PUT;

		if (argc > 2) {
			n = 2;
			for (;;) {
				if (sscanf(args[n],"%d",&tmp) != 1) {
					sm_error("Argument %s is not a number",args[n]);
					return SM_NEG;
				}

				n += (tmp + 1);
				if (n > argc) {
					sm_error("Not enough arguments given for last dimension");
					return SM_NEG;
				}

				op_params_dimensions++;

				if (n == argc) {
					break;
				}
			}
		}
	} else if (!strcmp(args[0],OP_ELECT_TOKEN)) {
		op = OP_ELECT;

		if (argc != 2) {
			sm_error("Illegal number of arguments: %d",argc);
			return SM_NEG;
		}
	} else {
		sm_error("Unknown operation: %s",args[0]);
		return SM_NEG;
	}

	op_params = args + 1;
	op_params_length = argc - 1;

	grid_coordinator_url = getenv("GRID_COORDINATOR_URL");

	return SM_OK;
}

static int
smc_coordinator_GetProcessEntry(sm_client_transport_t transport,
		int* timeout,
		smc_coordinator_process_entry_t* in,
		int size_in,
		smc_coordinator_process_entry_t** out,
		int* size_out)
{
	struct _qcgcoordinator__GetProcessEntry request;
	struct qcgcoordinator__GetProcessEntryResponse response;
	struct qcgcoordinator__ProcessEntry* entry;
	sm_value_t* value;
	int ret;
	int i;
	int j;
	ULONG64 t;

	SM_MANDATORY_P(in);
	SM_MANDATORY_P(out);
	SM_MANDATORY_P(size_out);
	SM_ASSERT(size_in > 0,"Input array may not be empty");

	if (timeout != NULL) {
		t = *timeout;
		request.qcgcoordinator__Timeout = &t;
	} else {
		request.qcgcoordinator__Timeout = NULL;
	}

	request.__sizeProcessEntryHeader = size_in;
	request.qcgcoordinator__ProcessEntryHeader = transport->transport_malloc(transport,sizeof(struct qcgcoordinator__ProcessEntryHeader) * size_in);

	for (i = 0; i < size_in; i++) {
		request.qcgcoordinator__ProcessEntryHeader[i].qcgcoordinator__Key = in[i]->key_name;
		request.qcgcoordinator__ProcessEntryHeader[i].qcgcoordinator__GridCoordinatorUrl = in[i]->master_url;
	}

	if (transport->pre_call_setup(transport, "http://schemas.qoscosgrid.org/coordinator/2009/04/service", "GetProcessEntry", NULL, NULL, NULL) != SM_OK) {
		return 1;
	}

	ret = soap_call___qcgcoordinator__GetProcessEntry(((ecm_gsoap_client_transport_t)transport)->soap, transport->service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		if (response.__sizeProcessEntry > 0) {
			*size_out = response.__sizeProcessEntry;
			*out = sm_malloc(sizeof(smc_coordinator_process_entry_t) * (*size_out));

			for (i = 0; i < (*size_out); i++) {
				entry = &response.qcgcoordinator__ProcessEntry[i];
				value = NULL;

				if (PROCESS_ENTRY_DATA(entry) != NULL) {
					value = sm_value_create(0);
					for (j = 0; j < PROCESS_ENTRY_DATA(entry)->__sizeitems; j++) {
						sm_value_append(&value,1,SM_VALUE_STRINGS|SM_VALUE_COPY_FLAG,
								PROCESS_ENTRY_DATA(entry)->items[j].__sizeitems,
								PROCESS_ENTRY_DATA(entry)->items[j].items);
					}
				}

				(*out)[i] = smc_coordinator_process_entry_create(PROCESS_ENTRY_KEY(entry),PROCESS_ENTRY_URL(entry),value);
			}
		} else {
			*out = NULL;
			*size_out = 0;
		}

		return SM_OK;
	} else {
		return SM_FAULT;
	}
}

static int
smc_coordinator_PutProcessEntry(sm_client_transport_t transport,
		smc_coordinator_process_entry_t* in,
		int size_in)
{
	struct _qcgcoordinator__PutProcessEntry request;
	struct qcgcoordinator__PutProcessEntryResponse response;
	struct qcgcoordinator__ProcessEntry* entry;
	int ret;
	int i;
	int j;

	request.__sizeProcessEntry = size_in;
	request.qcgcoordinator__ProcessEntry = transport->transport_malloc(transport,sizeof(struct qcgcoordinator__ProcessEntry) * size_in);

	for (i = 0; i < size_in; i++) {
		entry = &request.qcgcoordinator__ProcessEntry[i];
		entry->qcgcoordinator__ProcessEntryHeader = transport->transport_malloc(transport,sizeof(struct qcgcoordinator__ProcessEntryHeader));
		PROCESS_ENTRY_KEY(entry) = in[i]->key_name;
		PROCESS_ENTRY_URL(entry) = in[i]->master_url;
		if (in[i]->value != NULL) {
			PROCESS_ENTRY_DATA(entry) = transport->transport_malloc(transport,sizeof(*PROCESS_ENTRY_DATA(entry)));
			PROCESS_ENTRY_DATA(entry)->__sizeitems = sm_value_count(in[i]->value);
			PROCESS_ENTRY_DATA(entry)->items = transport->transport_malloc(transport,sizeof(struct qcgcoordinator__StringVector) * PROCESS_ENTRY_DATA(entry)->__sizeitems);
			for (j = 0; j < PROCESS_ENTRY_DATA(entry)->__sizeitems; j++) {
				PROCESS_ENTRY_DATA(entry)->items[j].__sizeitems = in[i]->value[j]->size;
				PROCESS_ENTRY_DATA(entry)->items[j].items = in[i]->value[j]->vstrings;
			}
		} else {
			PROCESS_ENTRY_DATA(entry) = NULL;
		}
	}

	if (transport->pre_call_setup(transport, "http://schemas.qoscosgrid.org/coordinator/2009/04/service", "PutProcessEntry", NULL, NULL, NULL) != SM_OK) {
		return 1;
	}

	ret = soap_call___qcgcoordinator__PutProcessEntry(((ecm_gsoap_client_transport_t)transport)->soap, transport->service_url, NULL, &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK) {
		return SM_OK;
	} else {
		return SM_FAULT;
	}
}

static int
smc_coordinator_Elect(sm_client_transport_t transport,
		char* key,
		SM_BOOL* result)
{
	struct _qcgcoordinator__Elect request;
	struct qcgcoordinator__ElectResponse response;
	int ret;

	request.Key = key;

	if (transport->pre_call_setup(transport, "http://schemas.qoscosgrid.org/coordinator/2009/04/service", "Elect", NULL, NULL, NULL) != SM_OK) {
		return 1;
	}

	ret = soap_call___qcgcoordinator__Elect(((ecm_gsoap_client_transport_t)transport)->soap, transport->service_url, NULL, &request, &response);


	if (transport->post_call_check(transport, ret) == SM_OK) {
		*result = response.Result;
		return SM_OK;
	} else {
		return SM_FAULT;
	}
}

int
start_client(sm_client_transport_t transport)
{
	int ret;
	smc_coordinator_process_entry_t in;

	SM_DEBUG_ENTER();

	if (op == OP_GET) {
		int* t = timeout_global < 0 ? NULL : &timeout_global;
		smc_coordinator_process_entry_t* out;
		int size;

		in = smc_coordinator_process_entry_create(op_params[0],grid_coordinator_url,NULL);
		ret = smc_coordinator_GetProcessEntry(transport,t,&in,1,&out,&size);

		if (ret == SM_OK) {
			if (size == 1) {
				ret = 0;

				if ((*out)->value != NULL) {
					int count = sm_value_count((*out)->value);
					int count2;
					int i;
					int j;

					for (i = 0; i < count; i++) {
						count2 = (*out)->value[i]->size;

						for (j = 0; j < count2; j++) {
							printf("\"%s\" ",(*out)->value[i]->vstrings[j]);
						}

						printf("\n");
					}

				}

				smc_coordinator_process_entry_destroy(out,SM_TRUE);
				sm_free(out);
			} else {
				ret = 2;
			}
		} else {
			ret = 1;
		}

		smc_coordinator_process_entry_destroy(&in,SM_FALSE);
	} else if (op == OP_PUT) {
		sm_value_t* value = sm_value_create(0);
		int i;
		int n;
		int tmp;

		for (i = 0, n = 1; i < op_params_dimensions; i++, n += tmp + 1) {
			sscanf(op_params[n],"%d",&tmp);
			sm_value_append(&value,1,SM_VALUE_STRINGS,tmp,op_params + n + 1);
		}

		in = smc_coordinator_process_entry_create(op_params[0],grid_coordinator_url,value);
		ret = smc_coordinator_PutProcessEntry(transport,&in,1);
		smc_coordinator_process_entry_destroy(&in,SM_TRUE);
	} else if (op == OP_ELECT) {
		SM_BOOL result;

		ret = smc_coordinator_Elect(transport,op_params[0],&result);

		if (ret == SM_OK) {
			printf("%s\n",result == SM_TRUE ? "true" : "false");
		}
	} else {
		sm_error("Unknown operation mode. This shouldn't have happened!");
		ret = 1;
	}

	return ret;
}

void
print_fault(sm_client_transport_t transport)
{
	struct soap *soap = ((ecm_gsoap_client_transport_t)transport)->soap;

	soap_print_fault(soap, stderr);
}

const void *
client_params(void)
{
	static struct ecm_gsoap_client_transport_params_s params = {
		.namespaces = qcgcoordinator_namespaces,
		.print_fault = print_fault
	};

	return &params;
}

int
client_setup(sm_module_t this sm_unused)
{
	client_api.get = &smc_coordinator_GetProcessEntry;
	client_api.put = &smc_coordinator_PutProcessEntry;
	client_api.elect = &smc_coordinator_Elect;

	return SM_OK;
}

void
client_cleanup(sm_module_t this sm_unused)
{
	/*nothing*/
}
