/* $Id: qcg_coordinator_service.c 2 2009-10-05 10:09:52Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * $Id: qcg_coordinator_service.c 2 2009-10-05 10:09:52Z mamonski $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <assert.h>
#include <stdlib.h>
#include <time.h>

#include <qcg-core/sm.h>
#include <qcg-core/client.h>
#include <qcg-core/closure.h>
#include <qcg-core/event_timer.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/metric.h>
#include <qcg-core/python.h>
#include <qcg-core/signal.h>
#include <qcg-core/thread.h>
#include <qcg-core/value.h>
#include <qcg-core/xmalloc.h>

#include <qcg-core/modules/atc_transport_module.h>
#include <qcg-core/modules/ecm_module.h>
#include <qcg-core/modules/ecm_service_module.h>
#include <qcg-core/modules/ecm_client_module.h>
#include <qcg-core/modules/ecm_gsoap/ecm_gsoap.h>

#include "stub/qcgcoordinator.nsmap"
#include "stub/qcgcoordinatorStub.h"

#include "../qcg_coordinator_client/qcg_coordinator_client.h"

__SM_ID("$Id: qcg_coordinator_service.c 2 2009-10-05 10:09:52Z mamonski $");

static int qcg_coordinator_setup(sm_module_t this);
static void qcg_coordinator_cleanup(sm_module_t this);
static int post_transport_setup_callback(sm_transport_t transport, void *data sm_unused);

static const void *
qcg_coordinator_service_params(void)
{
	static struct ecm_gsoap_transport_params_s params = {
		.serve = qcgcoordinator_serve,
		.namespaces = qcgcoordinator_namespaces,
		.get_fault_class = NULL,
		.get_fault_msg = NULL,
		.create_fault_detail = NULL,
		.fault_code2s = NULL
	};

	return &params;
}

SM_ECM_SERVICE_MODULE = {

	.header = {
		.entry_size = SM_ECM_SERVICE_MODULE_ENTRY_SIZE,
		.api_type = SM_API_ECM_SERVICE,

		.name = "qcg_coordinator_service",
		.descr = "QCG Coordinator Service",

		.version_major = 0,
		.version_minor = 1,

		.setup = qcg_coordinator_setup,
		.cleanup = qcg_coordinator_cleanup
	},

	.params = qcg_coordinator_service_params,

	.pre_request_callback = NULL,
	.post_request_callback = NULL,

	.have_wsa_header = SM_TRUE,
	.have_wsse_header = SM_TRUE
};

/* WS-Addressing plugin */
#include <qcg-core/modules/ecm_gsoap/gsoap/wsaapi.c>

/* WS-Security plugin */
#include <qcg-core/modules/ecm_gsoap/gsoap/wsseapi.c>
#include <qcg-core/modules/ecm_gsoap/gsoap/smdevp.c>
#include <qcg-core/modules/ecm_gsoap/gsoap/threads.c>

#define PROCESS_ENTRY_KEY(entry)	((entry)->qcgcoordinator__ProcessEntryHeader.qcgcoordinator__Key)
#define PROCESS_ENTRY_URL(entry)	((entry)->qcgcoordinator__ProcessEntryHeader.qcgcoordinator__GridCoordinatorUrl)
#define PROCESS_ENTRY_DATA(entry)	((entry)->qcgcoordinator__ProcessData)

#define GET_OP_TOKEN	"GetProcessEntry"
#define PUT_OP_TOKEN	"PutProcessEntry"
#define ELECT_OP_TOKEN	"Elect"

struct smc_coordinator_manager_s;

/*
 * Used for notifying GET request of available data.
 */
typedef struct smc_coordinator_wait_s {
	sm_cond_t cond;
	SM_BOOL active;
} *smc_coordinator_wait_t;

/*
 * Data key with wait objects list to notify
 * when data for a key is available. Creation
 * time used by GC to remove old data.
 */
typedef struct smc_coordinator_key_s {
	char* name;
	sm_list_t wait_objs;
	time_t created;
	SM_BOOL elected;
} *smc_coordinator_key_t;

/*
 * Key data.
 */
typedef struct smc_coordinator_data_s {
	smc_coordinator_key_t key;
	sm_value_t* value;
} *smc_coordinator_data_t;

/*
 * GET request for single key.
 */
typedef struct smc_coordinator_op_get_s {
	char* key_name;
	char* master_url;
	time_t expires_after;
} *smc_coordinator_op_get_t;

/*
 * PUT request for single key.
 */
typedef struct smc_coordinator_op_put_s {
	char* key_name;
	sm_value_t* value;
	char* master_url;
} *smc_coordinator_op_put_t;

/*
 * Used in URL threads to accumulate operations
 * for single endpoint.
 */
typedef struct smc_coordinator_queue_s {
	sm_list_t op_get_list;
	sm_list_t op_put_list;
} *smc_coordinator_queue_t;

/*
 * Thread making calls to single endpoint.
 */
typedef struct smc_coordinator_url_thread_s {
	struct smc_coordinator_manager_s* manager;
	char* master_url;
	smc_coordinator_queue_t queue;
	sm_thread_t thread;
	sm_mutex_t lock;
	SM_BOOL active;
	sm_client_transport_t client_transport;
	const char* _backup_service_url;
} *smc_coordinator_url_thread_t;

/*
 * Room for data stored locally.
 */
typedef struct smc_coordinator_store_s {
	sm_list_t keys;
	sm_list_t data;
} *smc_coordinator_store_t;

/*
 * Centre control point.
 */
typedef struct smc_coordinator_manager_s {
	char* default_master;
	smc_coordinator_store_t local_store;
	sm_list_t url_threads;
	sm_mutex_t lock;
} *smc_coordinator_manager_t;

static smc_coordinator_manager_t smc_coordinator_manager_create(char* default_master);
static void smc_coordinator_manager_destroy(smc_coordinator_manager_t* manager);
static void* smc_coordinator_url_thread_process(void* data);
static sm_value_t* smc_coordinator_gc(void* data, sm_value_t* params);

static sm_value_t* service_urls;
static int polling_period = 5; /* 5 s */
static int get_request_timeout = 120; /* 2 min */
static int expires_period = 3600; /* 1h */
static int polling_timeout = 1;

static sm_event_timer_subscription_t gc_subscription;
static smc_coordinator_manager_t coordinator_manager;
static sm_list_t inactive_wait_objs;
static sm_ecm_client_module_t remote_coordinator_module;
static const char* client_protocol;

static int
qcg_coordinator_setup(sm_module_t this)
{
	int gc_period = 600; /* 10 min */
	const char* master_coordinator_url = NULL;
	sm_list_t l, modules;

	SM_DEBUG_ENTER();

	if (sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:MasterCoordinatorURL", &master_coordinator_url) == SM_FAULT)
		return SM_FAULT;

	sm_info("Default master coordinator URL: %s", master_coordinator_url);

	if (sm_conf_node_int(this->conf_ctxt, SM_FALSE, "/conf:PollingPeriod", &polling_period) == SM_FAULT)
		return SM_FAULT;

	sm_info("Polling period: %d", polling_period);

	if (sm_conf_node_int(this->conf_ctxt, SM_FALSE, "/conf:PollingTimeout", &polling_timeout) == SM_FAULT)
		return SM_FAULT;

	sm_info("Polling timeout: %d", polling_timeout);

	if (sm_conf_node_int(this->conf_ctxt, SM_FALSE, "/conf:GetRequestTimeout", &get_request_timeout) == SM_FAULT)
			return SM_FAULT;

	sm_info("GET request timeout: %d", get_request_timeout);

	if (sm_conf_node_int(this->conf_ctxt, SM_FALSE, "/conf:GCPeriod", &gc_period) == SM_FAULT)
			return SM_FAULT;

	sm_info("Garbage collector period: %d", gc_period);

	if (sm_conf_node_int(this->conf_ctxt, SM_FALSE, "/conf:ExpiresPeriod", &expires_period) == SM_FAULT)
			return SM_FAULT;

	sm_info("Expires period: %d", expires_period);

	coordinator_manager = smc_coordinator_manager_create((char*) master_coordinator_url);

	gc_subscription = sm_event_timer_subscribe(sm_closure_create(smc_coordinator_gc, coordinator_manager, NULL), gc_period);

	modules = this->module_manager->find_by_api2(this->module_manager, SM_API_ECM, SM_MODULE_SERVICE_MODE);

        for (l = modules; l; l = l->next) {
            sm_ecm_module_t ecm_module = (sm_ecm_module_t) l->data;
			ecm_module->add_post_setup_callback((sm_module_t) ecm_module, post_transport_setup_callback, NULL);
        }

	sm_list_free(modules,SM_FALSE);

	remote_coordinator_module = (sm_ecm_client_module_t) this->module_manager->find_by_name_first(this->module_manager, "qcg_coordinator_client");
	if (!remote_coordinator_module) {
                sm_error("Failed to find qcg_coordinator_client module");
                return SM_FAULT;
        }

	return SM_OK;
}

static void
qcg_coordinator_cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();

	sm_event_timer_unsubscribe(&gc_subscription);
	smc_coordinator_manager_destroy(&coordinator_manager);
	sm_value_free(&service_urls);
}

static int
post_transport_setup_callback(sm_transport_t transport,
		void *data sm_unused)
{
	sm_atc_transport_module_t transport_module;
	sm_client_transport_t client_transport;

        transport_module = (sm_atc_transport_module_t) transport->atc_transport_modules->data;
        sm_value_append(&service_urls,1,SM_VALUE_STRING|SM_VALUE_DONT_COPY_FLAG,transport_module->get_service_url((sm_module_t) transport_module, transport));

	client_transport = sm_client_create_transport(remote_coordinator_module);

	if (!strncmp("http://",client_transport->service_url,7)) {
		client_protocol = "http://";
	} else if (!strncmp("https://",client_transport->service_url,8)) {
		client_protocol = "https://";
	} else {
		sm_error("Unknown transport protocol");
		return SM_FAULT;
	}

	sm_client_free_transport(&client_transport);

        return SM_OK;
}

static smc_coordinator_wait_t
smc_coordinator_wait_create(void)
{
	smc_coordinator_wait_t result;

	result = sm_malloc(sizeof(*result));

	if (result == NULL) {
		return NULL;
	}

	sm_cond_init(&result->cond);
	result->active = SM_TRUE;

	return result;
}

static void
smc_coordinator_wait_signal(smc_coordinator_wait_t wait)
{
	SM_MANDATORY_P(wait);

	SM_DEBUG_ENTER();

	if (wait->active) {
		sm_cond_signal(&wait->cond);
	}

	SM_DEBUG_LEAVE_VOID();
}

static void
smc_coordinator_wait_dispose(smc_coordinator_wait_t wait)
{
	SM_MANDATORY_P(wait);

	wait->active = SM_FALSE;
	sm_list_add(&inactive_wait_objs,wait,0);
}

static void
smc_coordinator_wait_destroy(smc_coordinator_wait_t* wait)
{
	SM_MANDATORY_P(wait);

	sm_cond_destroy(&(*wait)->cond);
	sm_free(*wait);
	*wait = NULL;
}

static smc_coordinator_key_t
smc_coordinator_key_create(char* key_name)
{
	smc_coordinator_key_t key;

	SM_ASSERT(!SM_IS_EMPTY_STR(key_name),"Empty mandatory key name");

	SM_DEBUG_ENTER_STR(key_name);

	key = sm_malloc(sizeof(*key));
	key->name = sm_strdup(key_name);
	key->wait_objs = NULL;
	key->created = time(NULL);
	key->elected = SM_FALSE;

	SM_DEBUG_LEAVE_PTR(key);

	return key;
}

static SM_BOOL
smc_coordinator_key_elect(smc_coordinator_key_t key)
{
	SM_BOOL result;

	SM_MANDATORY_P(key);

	result = !key->elected;
	if (result) key->elected = SM_TRUE;

	return result;
}

static void
smc_coordinator_key_subscribe(smc_coordinator_key_t key,
		smc_coordinator_wait_t wait)
{
	SM_MANDATORY_P(key);
	SM_MANDATORY_P(wait);

	SM_DEBUG_ENTER_PTR(key);

	sm_list_add(&key->wait_objs,wait,0);
}

static void
smc_coordinator_key_signal(smc_coordinator_key_t key)
{
	sm_list_t cur;
	smc_coordinator_wait_t wait_obj;

	SM_MANDATORY_P(key);

	SM_DEBUG_ENTER_PTR(key);

	cur = key->wait_objs;
	while (cur != NULL) {
		wait_obj = (smc_coordinator_wait_t) cur->data;
		smc_coordinator_wait_signal(wait_obj);
		cur = cur->next;
	}

	sm_list_free(key->wait_objs,SM_FALSE);
	key->wait_objs = NULL;

	SM_DEBUG_LEAVE_VOID();
}

static void
smc_coordinator_key_clean(smc_coordinator_key_t key)
{
	sm_list_t wait_to_remove = NULL;
	sm_list_t cur;
	smc_coordinator_wait_t wait;

	SM_MANDATORY_P(key);

	for (cur = key->wait_objs; cur; cur = cur->next) {
		wait = (smc_coordinator_wait_t) cur->data;
		if (!wait->active) {
			sm_list_add(&wait_to_remove,wait,0);
		}
	}

	for (cur = wait_to_remove; cur; cur = cur->next) {
		sm_list_remove(&key->wait_objs,cur->data,SM_FALSE);
	}

	sm_list_free(wait_to_remove,SM_FALSE);
}

static void
smc_coordinator_key_destroy(smc_coordinator_key_t* key)
{
	SM_MANDATORY_P(key);

	if (*key == NULL) {
		return;
	}

	sm_free((*key)->name);
	sm_free(*key);
	*key = NULL;
}

static smc_coordinator_data_t
smc_coordinator_data_create(smc_coordinator_key_t key,
		sm_value_t* value)
{
	smc_coordinator_data_t data;

	SM_MANDATORY_P(key);

	data = sm_malloc(sizeof(*data));

	data->key = key;
	data->value = value;

	return data;
}

static void
smc_coordinator_data_destroy(smc_coordinator_data_t* data)
{
	SM_MANDATORY_P(data);

	if (*data == NULL) {
		return;
	}

	sm_value_free(&(*data)->value);

	sm_free(*data);
	*data = NULL;
}

static smc_coordinator_op_get_t
smc_coordinator_op_get_create(char* key_name,
		char* master_url)
{
	smc_coordinator_op_get_t op_get;

	SM_MANDATORY_P(key_name);

	op_get = sm_malloc(sizeof(*op_get));
	op_get->key_name = strdup(key_name);
	op_get->master_url = (master_url == NULL) ? NULL : strdup(master_url);

	return op_get;
}

static smc_coordinator_op_get_t
smc_coordinator_op_get_clone(smc_coordinator_op_get_t op_get)
{
	smc_coordinator_op_get_t result;

	SM_MANDATORY_P(op_get);

	result = sm_malloc(sizeof(*result));
	result->key_name = strdup(op_get->key_name);
	result->master_url = (op_get->master_url == NULL) ? NULL : strdup(op_get->master_url);
	result->expires_after = op_get->expires_after;

	return result;
}

static void
smc_coordinator_op_get_destroy(smc_coordinator_op_get_t* op_get)
{
	SM_MANDATORY_P(op_get);

	if (*op_get == NULL) {
		return;
	}

	sm_free((*op_get)->key_name);
	sm_free((*op_get)->master_url);
	sm_free(*op_get);
	*op_get = NULL;
}

static smc_coordinator_op_put_t
smc_coordinator_op_put_create(char* key_name,
		sm_value_t* value,
		char* master_url)
{
	smc_coordinator_op_put_t op_put;

	SM_MANDATORY_P(key_name);

	op_put = sm_malloc(sizeof(*op_put));
	op_put->key_name = strdup(key_name);
	op_put->master_url = (master_url == NULL) ? NULL : sm_strdup(master_url);
	op_put->value = value;

	return op_put;
}

static smc_coordinator_op_put_t
smc_coordinator_op_put_clone(smc_coordinator_op_put_t op_put)
{
	smc_coordinator_op_put_t result;

	SM_MANDATORY_P(op_put);

	result = sm_malloc(sizeof(*result));
	result->key_name = strdup(op_put->key_name);
	result->master_url = (op_put->master_url == NULL) ? NULL : strdup(op_put->master_url);
	result->value = op_put->value;

	return result;
}

static void
smc_coordinator_op_put_destroy(smc_coordinator_op_put_t* op_put)
{
	SM_MANDATORY_P(op_put);

	if (*op_put == NULL) {
		return;
	}

	sm_free((*op_put)->master_url);
	sm_free(*op_put);
	*op_put = NULL;
}

static smc_coordinator_queue_t
smc_coordinator_queue_create(void)
{
	smc_coordinator_queue_t queue;

	queue = sm_malloc(sizeof(*queue));
	queue->op_get_list = NULL;
	queue->op_put_list = NULL;

	return queue;
}

static void
smc_coordinator_queue_add_get(smc_coordinator_queue_t queue,
		smc_coordinator_op_get_t op_get)
{
	SM_MANDATORY_P(queue);
	SM_MANDATORY_P(op_get);

	sm_list_add(&queue->op_get_list,op_get,0);
}

static void
smc_coordinator_queue_add_put(smc_coordinator_queue_t queue,
		smc_coordinator_op_put_t op_put)
{
	SM_MANDATORY_P(queue);
	SM_MANDATORY_P(op_put);

	sm_list_add(&queue->op_put_list,op_put,0);
}

static void
smc_coordinator_queue_clear_get(smc_coordinator_queue_t queue)
{
	sm_list_t cur;
	smc_coordinator_op_get_t op_get;

	SM_MANDATORY_P(queue);

	cur = queue->op_get_list;
	while (cur != NULL) {
		op_get = (smc_coordinator_op_get_t) cur->data;
		smc_coordinator_op_get_destroy(&op_get);
		cur = cur->next;
	}
	sm_list_free(queue->op_get_list,SM_FALSE);
	queue->op_get_list = NULL;
}

static void
smc_coordinator_queue_clear_put(smc_coordinator_queue_t queue)
{
	sm_list_t cur;
	smc_coordinator_op_put_t op_put;

	SM_MANDATORY_P(queue);

	cur = queue->op_put_list;
	while (cur != NULL) {
		op_put = (smc_coordinator_op_put_t) cur->data;
		smc_coordinator_op_put_destroy(&op_put);
		cur = cur->next;
	}
	sm_list_free(queue->op_put_list,SM_FALSE);
	queue->op_put_list = NULL;
}

static void
smc_coordinator_queue_clear(smc_coordinator_queue_t queue)
{
	SM_MANDATORY_P(queue);

	smc_coordinator_queue_clear_get(queue);
	smc_coordinator_queue_clear_put(queue);
}

static smc_coordinator_op_get_t
smc_coordinator_queue_remove_get(smc_coordinator_queue_t queue, smc_coordinator_key_t key)
{
	sm_list_t cur;
	smc_coordinator_op_get_t op_get;

	SM_MANDATORY_P(queue);
	SM_MANDATORY_P(key);

	cur = queue->op_get_list;
	while (cur != NULL) {
		op_get = (smc_coordinator_op_get_t) cur->data;
		if (!strcmp(op_get->key_name,key->name)) {
			sm_list_remove(&queue->op_get_list,op_get,SM_FALSE);
			return op_get;
		}
		cur = cur->next;
	}

	return NULL;
}

static void
smc_coordinator_queue_destroy(smc_coordinator_queue_t* queue)
{
	SM_MANDATORY_P(queue);

	if (*queue == NULL) {
		return;
	}

	smc_coordinator_queue_clear(*queue);

	sm_free(*queue);
	*queue = NULL;
}

static smc_coordinator_url_thread_t
smc_coordinator_url_thread_create(smc_coordinator_manager_t manager,
		char* master_url)
{
	smc_coordinator_url_thread_t url_thread;

	SM_MANDATORY_P(manager);
	SM_ASSERT(!SM_IS_EMPTY_STR(master_url),"Master URL cannot be empty");

	SM_DEBUG_ENTER_STR(master_url);

	url_thread = sm_malloc(sizeof(*url_thread));
	url_thread->manager = manager;
	url_thread->master_url = sm_strdup(master_url);
	url_thread->queue = smc_coordinator_queue_create();
	sm_mutex_init(&url_thread->lock);
	url_thread->active = SM_TRUE;
	sm_thread_create(&url_thread->thread,NULL,smc_coordinator_url_thread_process,url_thread);
	url_thread->client_transport = sm_client_create_transport(remote_coordinator_module);
	url_thread->_backup_service_url = url_thread->client_transport->service_url;
	url_thread->client_transport->service_url = url_thread->master_url;

	SM_DEBUG_LEAVE_PTR(url_thread);

	return url_thread;
}

static void
smc_coordinator_url_thread_lock(smc_coordinator_url_thread_t url_thread)
{
	SM_MANDATORY_P(url_thread);

	sm_mutex_lock(&url_thread->lock);
}

static void
smc_coordinator_url_thread_unlock(smc_coordinator_url_thread_t url_thread)
{
	SM_MANDATORY_P(url_thread);

	sm_mutex_unlock(&url_thread->lock);
}

static void
smc_coordinator_url_thread_add_get(smc_coordinator_url_thread_t url_thread,
		smc_coordinator_op_get_t op_get)
{
	smc_coordinator_op_get_t new_op_get;

	SM_MANDATORY_P(url_thread);
	SM_MANDATORY_P(op_get);

	SM_DEBUG_ENTER();

	new_op_get = smc_coordinator_op_get_clone(op_get);

	smc_coordinator_url_thread_lock(url_thread);
	sm_debug("add_get Locked thread");
	smc_coordinator_queue_add_get(url_thread->queue,new_op_get);
	smc_coordinator_url_thread_unlock(url_thread);
	sm_debug("add_get Unlocked thread");

	SM_DEBUG_LEAVE_VOID();
}

static void
smc_coordinator_url_thread_add_put(smc_coordinator_url_thread_t url_thread,
		smc_coordinator_op_put_t op_put)
{
	smc_coordinator_op_put_t new_op_put;

	SM_MANDATORY_P(url_thread);
	SM_MANDATORY_P(op_put);

	SM_DEBUG_ENTER();

	new_op_put = smc_coordinator_op_put_clone(op_put);

	smc_coordinator_url_thread_lock(url_thread);
	sm_debug("add_put Locked thread");
	smc_coordinator_queue_add_put(url_thread->queue,new_op_put);
	smc_coordinator_url_thread_unlock(url_thread);
	sm_debug("add_put Unlocked thread");

	SM_DEBUG_LEAVE_VOID();
}

static void
smc_coordinator_url_thread_end(smc_coordinator_url_thread_t url_thread)
{
	void* ptr;

	smc_coordinator_url_thread_lock(url_thread);
	url_thread->active = SM_FALSE;
	smc_coordinator_url_thread_unlock(url_thread);
	sm_thread_join(url_thread->thread,&ptr);
}

static void
smc_coordinator_url_thread_destroy(smc_coordinator_url_thread_t* url_thread)
{
	SM_MANDATORY_P(url_thread);

	if (*url_thread == NULL) {
		return;
	}

	sm_free((*url_thread)->master_url);
	smc_coordinator_queue_destroy(&(*url_thread)->queue);
	sm_mutex_destroy(&(*url_thread)->lock);
	(*url_thread)->client_transport->service_url = (*url_thread)->_backup_service_url;
	sm_client_free_transport(&(*url_thread)->client_transport);
	sm_free(*url_thread);
	*url_thread = NULL;
}


static smc_coordinator_store_t
smc_coordinator_store_create(void)
{
	smc_coordinator_store_t store;

	store = sm_malloc(sizeof(*store));
	store->keys = NULL;
	store->data = NULL;

	return store;
}

static smc_coordinator_key_t
smc_coordinator_store_get_key(smc_coordinator_store_t store,
		char* key_name)
{
	sm_list_t cur;
	smc_coordinator_key_t key;

	SM_MANDATORY_P(store);
	SM_ASSERT(!SM_IS_EMPTY_STR(key_name),"Key name cannot be empty");

	cur = store->keys;
	while (cur != NULL) {
		key = (smc_coordinator_key_t) cur->data;
		if (!strcmp(key_name,key->name)) {
			goto end;
		}
		cur = cur->next;
	}

	key = smc_coordinator_key_create(key_name);
	sm_list_add(&store->keys,key,0);

end:
	return key;
}

static smc_coordinator_data_t
smc_coordinator_store_get_data(smc_coordinator_store_t store,
		smc_coordinator_key_t key)
{
	sm_list_t cur;
	smc_coordinator_data_t data;

	SM_MANDATORY_P(store);
	SM_MANDATORY_P(key);

	data = NULL;
	cur = store->data;
	while (cur != NULL) {
		data = (smc_coordinator_data_t) cur->data;

		if (data->key == key) {
			return data;
		}

		cur = cur->next;
	}

	return NULL;
}

static void
smc_coordinator_store_remove_data(smc_coordinator_store_t store,
		smc_coordinator_data_t data)
{
	SM_MANDATORY_P(store);
	SM_MANDATORY_P(data);

	sm_list_remove(&store->data,data,SM_FALSE);
	smc_coordinator_data_destroy(&data);
}

static void
smc_coordinator_store_put_data(smc_coordinator_store_t store,
		smc_coordinator_data_t data)
{
	smc_coordinator_data_t old_data;
	SM_MANDATORY_P(store);
	SM_MANDATORY_P(data);

	if ((old_data = smc_coordinator_store_get_data(store, data->key)) != NULL) {
		sm_notice("Duplicated entry: %s. Removing old one.", data->key->name);
		smc_coordinator_store_remove_data(store, old_data);
	}

	sm_list_add(&store->data,data,0);
	smc_coordinator_key_signal(data->key);
}

static void
smc_coordinator_store_remove_key(smc_coordinator_store_t store,
		smc_coordinator_key_t key)
{
	smc_coordinator_data_t data;

	SM_MANDATORY_P(store);
	SM_MANDATORY_P(key);

	SM_DEBUG_ENTER_STR(key->name);

	data = smc_coordinator_store_get_data(store,key);
	if (data != NULL) {
		smc_coordinator_store_remove_data(store,data);
	}
	sm_list_remove(&store->keys,key,SM_FALSE);
	smc_coordinator_key_destroy(&key);
}

static void
smc_coordinator_store_destroy(smc_coordinator_store_t* store)
{
	smc_coordinator_key_t key;
	smc_coordinator_data_t data;
	sm_list_t cur;

	SM_MANDATORY_P(store);

	if (*store == NULL) {
		return;
	}

	cur = (*store)->keys;
	while (cur != NULL) {
		key = (smc_coordinator_key_t) cur->data;
		smc_coordinator_key_destroy(&key);
		cur = cur->next;
	}

	sm_list_free((*store)->keys,SM_FALSE);

	cur = (*store)->data;
	while (cur != NULL) {
		data = (smc_coordinator_data_t) cur->data;
		smc_coordinator_data_destroy(&data);
		cur = cur->next;
	}

	sm_list_free((*store)->data,SM_FALSE);
	sm_free(*store);
	*store = NULL;
}

static smc_coordinator_manager_t
smc_coordinator_manager_create(char* default_master)
{
	smc_coordinator_manager_t manager;

	SM_DEBUG_ENTER_STR(default_master);

	manager = sm_malloc(sizeof(*manager));

	if (default_master) {
		manager->default_master = sm_strdup(default_master);
	} else {
		manager->default_master = NULL;
	}

	manager->local_store = smc_coordinator_store_create();
	manager->url_threads = NULL;
	sm_mutex_init(&manager->lock);

	SM_DEBUG_LEAVE_PTR(manager);

	return manager;
}

static smc_coordinator_key_t
smc_coordinator_manager_get_key(smc_coordinator_manager_t manager,
		char* key_name)
{
	smc_coordinator_key_t key;

	SM_MANDATORY_P(manager);
	SM_MANDATORY_P(key_name);

	SM_DEBUG_ENTER_STR(key_name);

	key = smc_coordinator_store_get_key(manager->local_store,key_name);

	SM_DEBUG_LEAVE_PTR(key);

	return key;
}

static void
smc_coordinator_manager_lock(smc_coordinator_manager_t manager)
{
	SM_MANDATORY_P(manager);

	sm_mutex_lock(&manager->lock);
}

static void
smc_coordinator_manager_unlock(smc_coordinator_manager_t manager)
{
	SM_MANDATORY_P(manager);

	sm_mutex_unlock(&manager->lock);
}

static smc_coordinator_url_thread_t
smc_coordinator_manager_get_url_thread(smc_coordinator_manager_t manager,
		char* master_url)
{
	smc_coordinator_url_thread_t url_thread;
	sm_list_t cur;

	SM_MANDATORY_P(manager);
	SM_ASSERT(!SM_IS_EMPTY_STR(master_url),"Master URL cannot be empty");

	cur = manager->url_threads;
	while (cur != NULL) {
		url_thread = (smc_coordinator_url_thread_t) cur->data;
		if (!strcmp(url_thread->master_url,master_url)) {
			goto end;
		}

		cur = cur->next;
	}

	url_thread = smc_coordinator_url_thread_create(manager,master_url);
	sm_list_add(&manager->url_threads,url_thread,0);

end:
	return url_thread;
}

static char*
smc_coordinator_manager_master_url(smc_coordinator_manager_t manager,
		char* master_url)
{
	char* url;
	int size;
	int i;

	SM_MANDATORY_P(manager);

	if (master_url && strncmp(master_url,client_protocol,strlen(client_protocol))) {
		sm_error("Coordinator URL passed by client has bad transport protocol");
		return NULL;
	}

	url = SM_IS_EMPTY_STR(master_url) ? manager->default_master : master_url;

	size = sm_value_count(service_urls);
	for (i = 0; i < size; i++) {
		if (url != NULL && !strcmp(url,service_urls[i]->vstring)) {
			return NULL;
		}
	}

	return url;
}

static void
smc_coordinator_manager_insert(smc_coordinator_manager_t manager,
		smc_coordinator_data_t data)
{
	SM_MANDATORY_P(manager);
	SM_MANDATORY_P(data);

	SM_DEBUG_ENTER();

	smc_coordinator_store_put_data(manager->local_store,data);

	SM_DEBUG_LEAVE_VOID();
}

static SM_BOOL
smc_coordinator_manager_elect(smc_coordinator_manager_t manager,
		char* key_name)
{
	smc_coordinator_key_t key;
	SM_BOOL result;

	SM_MANDATORY_P(manager);
	SM_MANDATORY_P(key_name);

	SM_DEBUG_ENTER();

	key = smc_coordinator_store_get_key(manager->local_store,key_name);
	result = smc_coordinator_key_elect(key);

	SM_DEBUG_LEAVE_VOID();

	return result;
}

static void
smc_coordinator_manager_destroy(smc_coordinator_manager_t* manager)
{
	sm_list_t cur;
	smc_coordinator_url_thread_t thread;

	SM_MANDATORY_P(manager);

	if (*manager == NULL) {
		return;
	}

	for (cur = (*manager)->url_threads; cur; cur = cur->next) {
		thread = (smc_coordinator_url_thread_t) cur->data;
		smc_coordinator_url_thread_end(thread);
		smc_coordinator_url_thread_destroy(&thread);
	}

	sm_list_free((*manager)->url_threads,SM_FALSE);
	sm_free((*manager)->default_master);
	smc_coordinator_store_destroy(&(*manager)->local_store);
	sm_mutex_destroy(&(*manager)->lock);
	sm_free(*manager);
	*manager = NULL;
}

static void
smc_coordinator_manager_put_data(smc_coordinator_manager_t manager,
		smc_coordinator_op_put_t* op_put,
		int size)
{
	smc_coordinator_key_t key;
	smc_coordinator_data_t data;
	char* url;
	int i;

	SM_MANDATORY_P(manager);
	SM_MANDATORY_P(op_put);
	SM_ASSERT(size > 0, "op_put set cannot be empty");

	SM_DEBUG_ENTER();

	for (i = 0; i < size; i++) {
		key = smc_coordinator_manager_get_key(manager,op_put[i]->key_name);
		data = smc_coordinator_data_create(key,op_put[i]->value);
		smc_coordinator_store_put_data(manager->local_store,data);

		url = smc_coordinator_manager_master_url(manager,op_put[i]->master_url);

		if (url != NULL) {
			smc_coordinator_url_thread_t url_thread;

			url_thread = smc_coordinator_manager_get_url_thread(manager,url);
			smc_coordinator_url_thread_add_put(url_thread,op_put[i]);
		}
	}

	SM_DEBUG_LEAVE_VOID();
}

int
qcgcoordinator__PutProcessEntry(struct soap* soap,
		int size,
		struct qcgcoordinator__ProcessEntry *entries,
		struct qcgcoordinator__PutProcessEntryResponse *response sm_unused)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s fault;
	int i;
	int j;
	sm_value_t* value;
	smc_coordinator_op_put_t* op_put;

	if (transport->enter(transport, PUT_OP_TOKEN, &fault) != SM_OK)
		return transport->fault(transport, &fault);

	if (transport->access_check(transport, SM_FALSE, PUT_OP_TOKEN, &fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, &fault);

	op_put = sm_malloc(sizeof(smc_coordinator_op_put_t) * size);

	for (i = 0; i < size; i++) {
		struct qcgcoordinator__ProcessEntry* entry = &entries[i];

		if (PROCESS_ENTRY_DATA(entry) != NULL) {
			value = sm_value_create(0);
			for (j = 0; j < PROCESS_ENTRY_DATA(entry)->__size; j++) {
				sm_value_append(&value,1,SM_VALUE_STRINGS|SM_VALUE_COPY_FLAG,PROCESS_ENTRY_DATA(entry)->items[j].__size,PROCESS_ENTRY_DATA(entry)->items[j].items);
			}
		} else {
			value = NULL;
		}

		op_put[i] = smc_coordinator_op_put_create(PROCESS_ENTRY_KEY(entry),value,PROCESS_ENTRY_URL(entry));
	}

	smc_coordinator_manager_lock(coordinator_manager);
	sm_debug("PutProcessEntry Locked manager");
	smc_coordinator_manager_put_data(coordinator_manager,op_put,size);
	smc_coordinator_manager_unlock(coordinator_manager);
	sm_debug("PutProcessEntry Unlocked manager");

	for (i = 0; i < size; i++) {
		smc_coordinator_op_put_destroy(&op_put[i]);
	}

	sm_free(op_put);

	return transport->exit(transport, NULL);
}

static smc_coordinator_wait_t
smc_coordinator_manager_queue(smc_coordinator_manager_t manager,
		smc_coordinator_op_get_t* op_get,
		int size,
		time_t expires_after)
{
	smc_coordinator_wait_t wait_obj;
	smc_coordinator_key_t key;
	smc_coordinator_url_thread_t url_thread;
	char* url;
	int i;

	SM_DEBUG_ENTER();

	wait_obj = smc_coordinator_wait_create();

	for (i = 0; i < size; i++) {
		url = smc_coordinator_manager_master_url(manager,op_get[i]->master_url);
		key = smc_coordinator_manager_get_key(manager,op_get[i]->key_name);
		smc_coordinator_key_subscribe(key,wait_obj);

		if (url != NULL) {
			op_get[i]->expires_after = expires_after;
			url_thread = smc_coordinator_manager_get_url_thread(manager,url);
			smc_coordinator_url_thread_add_get(url_thread,op_get[i]);
		}
	}

	SM_DEBUG_LEAVE_PTR(wait_obj);

	return wait_obj;
}

static void
smc_coordinator_manager_local_get_data(smc_coordinator_manager_t manager,
		smc_coordinator_op_get_t* op_get,
		int size,
		sm_list_t* result)
{
	smc_coordinator_data_t data;
	smc_coordinator_key_t key;
	int i;

	SM_MANDATORY_P(manager);
	SM_MANDATORY_P(op_get);
	SM_MANDATORY_P(result);

	SM_DEBUG_ENTER();

	for (i = 0; i < size; i++) {
		key = smc_coordinator_manager_get_key(manager,op_get[i]->key_name);
		data = smc_coordinator_store_get_data(manager->local_store,key);

		if (data != NULL) {
			sm_list_add(result,data,0);
		}
	}

	SM_DEBUG_LEAVE_VOID();
}

static int
smc_coordinator_manager_get_data(smc_coordinator_manager_t manager,
		smc_coordinator_op_get_t* op_get,
		int size,
		smc_coordinator_data_t** result,
		int* result_size,
		time_t timeout)
{
	smc_coordinator_wait_t wait_obj;
	sm_list_t result_list = NULL;
	int result_list_size;
	int i;
	time_t expiration;
	int ret;
	struct timespec ts;

	SM_MANDATORY_P(manager);
	SM_MANDATORY_P(op_get);
	SM_MANDATORY_P(result);
	SM_MANDATORY_P(result_size);
	SM_ASSERT(size > 0, "op_get set cannot be empty");

	SM_DEBUG_ENTER();

	expiration = time(NULL) + timeout;
	ts.tv_sec = expiration;
	ts.tv_nsec = 0;

	smc_coordinator_manager_local_get_data(manager,op_get,size,&result_list);

	if (result_list == NULL && timeout > 0) {
		wait_obj = smc_coordinator_manager_queue(manager,op_get,size,expiration);

		do {
			sm_debug("get_data Unlocked manager due to wait");
			ret = sm_cond_timedwait(&wait_obj->cond,&manager->lock,&ts);
			sm_debug("get_data Locked manager due to get_data wait signal");
			smc_coordinator_manager_local_get_data(manager,op_get,size,&result_list);
		} while (result_list == NULL && ret == SM_TRUE);

		smc_coordinator_wait_dispose(wait_obj);
	}

	if (result_list == NULL) {
		*result = NULL;
		*result_size = 0;
	} else {
		sm_list_t cur = result_list;

		result_list_size = sm_list_count(result_list);
		*result = sm_malloc(sizeof(smc_coordinator_data_t) * result_list_size);
		*result_size = result_list_size;

		for (i = 0; i < result_list_size; i++) {
			(*result)[i] = (smc_coordinator_data_t) cur->data;
			cur = cur->next;
		}
	}

	SM_DEBUG_LEAVE_VOID();

	return SM_OK;
}

static struct qcgcoordinator__ProcessEntry*
create_process_entry_array(sm_transport_t transport,
		int size)
{
	struct qcgcoordinator__ProcessEntry* res;
	struct qcgcoordinator__ProcessEntry* cur;
	int i;

	res = transport->transport_malloc(transport, sizeof(struct qcgcoordinator__ProcessEntry) * size);

	if (res == NULL) {
		return NULL;
	}

	for (i = 0, cur = res; i < size; i++, cur++) {
		PROCESS_ENTRY_KEY(cur) = NULL;
		PROCESS_ENTRY_URL(cur) = NULL;
		PROCESS_ENTRY_DATA(cur) = NULL;
	}

	return res;
}

static struct qcgcoordinator__ProcessData*
create_process_data(sm_transport_t transport,
		int size)
{
	struct qcgcoordinator__ProcessData* res;
	struct qcgcoordinator__StringVector* cur;
	int i;

	if (transport != NULL) {
		res = transport->transport_malloc(transport, sizeof(struct qcgcoordinator__ProcessData));
	} else {
		res = sm_malloc(sizeof(struct qcgcoordinator__ProcessData));
	}

	if (res == NULL) {
		return NULL;
	}

	res->__size = size;

	if (transport != NULL) {
		res->items = transport->transport_malloc(transport, sizeof(struct qcgcoordinator__StringVector) * size);
	} else {
		res->items = sm_malloc(sizeof(struct qcgcoordinator__StringVector) * size);
	}

	for (i = 0, cur = res->items; i < size; i++, cur++) {
		cur->__size = 0;
		cur->items = NULL;
	}

	return res;
}

static void
smc_coordinator_get_data_response(sm_transport_t transport,
		struct qcgcoordinator__GetProcessEntryResponse* response,
		smc_coordinator_data_t* data,
		int res_size)
{
	int i;
	int j;
	int k;

	response->__size = res_size;
	response->qcgcoordinator__ProcessEntry = create_process_entry_array(transport, res_size);

	for (i = 0; i < res_size; i++) {
		struct qcgcoordinator__ProcessEntry* entry = &response->qcgcoordinator__ProcessEntry[i];
		int len;

		PROCESS_ENTRY_KEY(entry) = transport->transport_strdup(transport,data[i]->key->name);
		len = sm_value_count(data[i]->value);
		if (len == 0) continue;
		PROCESS_ENTRY_DATA(entry) = create_process_data(transport,len);

		for (j = 0; j < len; j++) {
			int len2;
			char** val = data[i]->value[j]->vstrings;

			len2 = data[i]->value[j]->size;
			PROCESS_ENTRY_DATA(entry)->items[j].__size = len2;
			PROCESS_ENTRY_DATA(entry)->items[j].items = transport->transport_malloc(transport,sizeof(char*) * len2);

			for (k = 0; k < len2; k++) {
				PROCESS_ENTRY_DATA(entry)->items[j].items[k] = transport->transport_strdup(transport,val[k]);
			}
		}
	}
}

int
qcgcoordinator__GetProcessEntry(struct soap* soap,
		ULONG64* timeout,
		int size,
		struct qcgcoordinator__ProcessEntryHeader *headers,
		struct qcgcoordinator__GetProcessEntryResponse *response)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s fault;
	int i;
	time_t t;
	smc_coordinator_op_get_t* op_get;
	int res_size;
	smc_coordinator_data_t* data;

	if (transport->enter(transport, GET_OP_TOKEN, &fault) != SM_OK)
		return transport->fault(transport, &fault);

	if (transport->access_check(transport, SM_FALSE, GET_OP_TOKEN, &fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, &fault);


	t = (timeout == NULL ? (unsigned) get_request_timeout : *timeout);
	op_get = sm_malloc(sizeof(smc_coordinator_op_get_t) * size);

	for (i = 0; i < size; i++) {
		op_get[i] = smc_coordinator_op_get_create(headers[i].qcgcoordinator__Key,headers[i].qcgcoordinator__GridCoordinatorUrl);
	}

	smc_coordinator_manager_lock(coordinator_manager);
	sm_debug("GetProcessEntry Locked manager");
	smc_coordinator_manager_get_data(coordinator_manager,op_get,size,&data,&res_size,t);
	smc_coordinator_get_data_response(transport,response,data,res_size);
	smc_coordinator_manager_unlock(coordinator_manager);
	sm_debug("GetProcessEntry Unlocked manager");

	for (i = 0; i < size; i++) {
		smc_coordinator_op_get_destroy(&op_get[i]);
	}


	sm_free(op_get);
	sm_free(data);

	return transport->exit(transport, NULL);
}

int
qcgcoordinator__Elect(struct soap* soap,
		char* Key,
		struct qcgcoordinator__ElectResponse *response)
{
	sm_transport_t transport = soap->user;
	struct sm_transport_fault_s fault;

	if (transport->enter(transport, ELECT_OP_TOKEN, &fault) != SM_OK)
		return transport->fault(transport, &fault);

	if (transport->access_check(transport, SM_FALSE, ELECT_OP_TOKEN, &fault) != SM_ACCESS_GRANTED)
		return transport->fault(transport, &fault);


	smc_coordinator_manager_lock(coordinator_manager);
	response->Result = smc_coordinator_manager_elect(coordinator_manager,Key);
	smc_coordinator_manager_unlock(coordinator_manager);

	return transport->exit(transport, NULL);
}

static void
smc_coordinator_url_thread_make_get(smc_coordinator_url_thread_t url_thread)
{
	sm_list_t cur;
	smc_coordinator_op_get_t op_get;
	int size;
	int i;
	smc_coordinator_key_t key;
	smc_coordinator_data_t data;
	time_t now;
	sm_list_t keys_to_remove = NULL;
	smc_coordinator_process_entry_t* entries;
	smc_coordinator_process_entry_t* result;
	int result_size;
	int ret;

	SM_MANDATORY_P(url_thread);

	SM_DEBUG_ENTER();

	smc_coordinator_url_thread_lock(url_thread);
	sm_debug("make_get Locked thread");
	size = sm_list_count(url_thread->queue->op_get_list);

	if (size == 0) {
		smc_coordinator_url_thread_unlock(url_thread);
		sm_debug("make_get Unlocked thread");
		return;
	}

	entries = sm_malloc(sizeof(smc_coordinator_process_entry_t) * size);
	for (i = 0, cur = url_thread->queue->op_get_list; i < size; i++, cur = cur->next) {
		op_get = (smc_coordinator_op_get_t) cur->data;
		entries[i] = smc_coordinator_process_entry_create(op_get->key_name,op_get->master_url,NULL);
	}
	smc_coordinator_url_thread_unlock(url_thread);
	sm_debug("make_get Unlocked thread");

	ret = ((smc_coordinator_client_api_t)remote_coordinator_module->client_api)->get(url_thread->client_transport, &polling_timeout, entries, size, &result, &result_size);

	if (ret != SM_OK) {
		return;
	}

	smc_coordinator_manager_lock(url_thread->manager);
	sm_debug("make_get Locked manager");
	smc_coordinator_url_thread_lock(url_thread);
	sm_debug("make_get Locked thread");
	/*
	 * Remove expired GET requests from queue.
	 */
	now = time(NULL);
	for (cur = url_thread->queue->op_get_list; cur; cur = cur->next) {
		op_get = (smc_coordinator_op_get_t) cur->data;
		if (op_get->expires_after <= now) {
			key = smc_coordinator_manager_get_key(url_thread->manager,op_get->key_name);
			sm_list_add(&keys_to_remove,key,0);
		}
	}

	for (cur = keys_to_remove; cur; cur = cur->next) {
		key = (smc_coordinator_key_t) cur->data;
		smc_coordinator_queue_remove_get(url_thread->queue,key);
	}

	sm_list_free(keys_to_remove,SM_FALSE);

	for (i = 0; i < result_size; i++) {
		key = smc_coordinator_manager_get_key(url_thread->manager,result[i]->key_name);
		data = smc_coordinator_data_create(key,result[i]->value);
		smc_coordinator_manager_insert(url_thread->manager,data);
		op_get = smc_coordinator_queue_remove_get(url_thread->queue,key);
		smc_coordinator_op_get_destroy(&op_get);
		smc_coordinator_process_entry_destroy(&result[i],SM_FALSE);
	}
	smc_coordinator_manager_unlock(url_thread->manager);
	sm_debug("make_get Unlocked manager");
	smc_coordinator_url_thread_unlock(url_thread);
	sm_debug("make_get Unlocked thread");

	for (i = 0; i < size; i++) {
		smc_coordinator_process_entry_destroy(&entries[i],SM_FALSE);
	}

	sm_free(entries);
	sm_free(result);
}

static void
smc_coordinator_url_thread_make_put(smc_coordinator_url_thread_t url_thread)
{
	sm_list_t cur;
	smc_coordinator_op_put_t op_put;
	int size;
	int i;
	smc_coordinator_process_entry_t* entries;
	int ret;

	SM_MANDATORY_P(url_thread);

	SM_DEBUG_ENTER();

	smc_coordinator_url_thread_lock(url_thread);
	sm_debug("make_put Locked thread");
	size = sm_list_count(url_thread->queue->op_put_list);

	if (size == 0) {
		smc_coordinator_url_thread_unlock(url_thread);
		sm_debug("make_put Unlocked thread");
		return;
	}

	entries = sm_malloc(sizeof(smc_coordinator_process_entry_t) * size);
	for (i = 0, cur = url_thread->queue->op_put_list; i < size; i++, cur = cur->next) {
		op_put = (smc_coordinator_op_put_t) cur->data;
		entries[i] = smc_coordinator_process_entry_create(op_put->key_name,op_put->master_url,op_put->value);
	}
	smc_coordinator_queue_clear_put(url_thread->queue);
	smc_coordinator_url_thread_unlock(url_thread);
	sm_debug("make_put Unlocked thread");

	ret = ((smc_coordinator_client_api_t)remote_coordinator_module->client_api)->put(url_thread->client_transport, entries, size);
	for (i = 0; i < size; i++) {
		smc_coordinator_process_entry_destroy(&entries[i],SM_FALSE);
	}

	sm_free(entries);
}

static void*
smc_coordinator_url_thread_process(void* data)
{
	smc_coordinator_url_thread_t url_thread;

	SM_MANDATORY_P(data);

	url_thread = (smc_coordinator_url_thread_t) data;

	sm_info("Starting URL thread (%s)",url_thread->master_url);

	for (;;) {
		if (url_thread->active) {
			smc_coordinator_url_thread_make_put(url_thread);
			smc_coordinator_url_thread_make_get(url_thread);
		} else {
			sm_info("Finishing URL thread (%s)",url_thread->master_url);
			sm_thread_exit(NULL);
		}

		sleep(polling_period);
	}

}

static void
smc_coordinator_store_gc(smc_coordinator_store_t store)
{
	sm_list_t cur;
	time_t expires_before;
	smc_coordinator_key_t key;
	smc_coordinator_data_t data;
	sm_list_t key_to_remove = NULL;

	SM_MANDATORY_P(store);

	expires_before = time(NULL) - expires_period;

	for (cur = store->keys; cur; cur = cur->next) {
		key = (smc_coordinator_key_t) cur->data;
		if (key->created <= expires_before) {
			data = smc_coordinator_store_get_data(store,key);
			if (data == NULL) {
				sm_list_add(&key_to_remove,key,0);
			}
		} else {
			smc_coordinator_key_clean(key);
		}
	}

	for (cur = key_to_remove; cur; cur = cur->next) {
		key = (smc_coordinator_key_t) cur->data;
		smc_coordinator_store_remove_key(store,key);
	}
	sm_list_free(key_to_remove,SM_FALSE);
}

static void
smc_coordinator_wait_objs_gc(void)
{
	sm_list_t cur;
	smc_coordinator_wait_t wait;

	cur = inactive_wait_objs;
	while (cur != NULL) {
		wait = (smc_coordinator_wait_t) cur->data;
		smc_coordinator_wait_destroy(&wait);
		cur = cur->next;
	}

	sm_list_free(inactive_wait_objs,SM_FALSE);
	inactive_wait_objs = 0;
}

static sm_value_t*
smc_coordinator_gc(void* data,
		sm_value_t* params sm_unused)
{
	smc_coordinator_manager_t manager;

	SM_MANDATORY_P(data);

	SM_DEBUG_ENTER();

	manager = (smc_coordinator_manager_t) data;
	smc_coordinator_manager_lock(manager);
	sm_debug("gc Locked manager");
	smc_coordinator_store_gc(manager->local_store);
	smc_coordinator_wait_objs_gc();
	smc_coordinator_manager_unlock(manager);
	sm_debug("gc Unlocked manager");

	SM_DEBUG_LEAVE_VOID();

	return NULL;
}
