/* $Id: staging_curl.c 314 2010-10-14 12:54:06Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _FILE_OFFSET_BITS	64

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdbool.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>

#include <curl/curl.h>

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/value.h>
#include <qcg-core/utils.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <xmlstructs/xmlstructs.h>
#include <include/namespaces.h>
#include <include/staging_module.h>
#include <include/extensions.h>

__SM_ID("$Id: staging_curl.c 314 2010-10-14 12:54:06Z mamonski $");

static int setup(sm_module_t);
static void cleanup(sm_module_t);
static sm_value_t *stage_in(sm_module_t, _jsdl__DataStaging *, size_t, const char *);
static sm_value_t *stage_out(sm_module_t, _jsdl__DataStaging *, size_t, const char *);

STAGING_MODULE = {

	.header = {
		.entry_size = STAGING_MODULE_ENTRY_SIZE,
		.api_type = STAGING_MODULE_API,

		.name = "staging_curl",
		.descr = "libcurl data staging",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	},

	.stage_in = stage_in,
	.stage_out = stage_out,
};

#define PROGRESS_PERIOD	10

static sm_value_t *supported_protocols = NULL;
static bool verbose = false;
static bool largefile = false;
static bool show_progress = true;
static bool show_summary = true;
static bool ssl_verify_host = true;
static const char *x509_cert_file = NULL;
static const char *x509_key_file = NULL;
static const char *x509_key_pass = NULL;
static const char *trusted_ca_file = NULL;
static const char *trusted_ca_path = NULL;

static int
setup(sm_module_t this)
{
	int i;
	curl_version_info_data *data;

	SM_DEBUG_ENTER();

	if (this && sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:ShowProgress", (SM_BOOL *) &show_progress) == SM_FAULT)
		return SM_FAULT;

	if (this && sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:ShowSummary", (SM_BOOL *) &show_summary) == SM_FAULT)
		return SM_FAULT;

	if (this && sm_conf_node_bool(this->conf_ctxt, SM_FALSE, "/conf:SSLVerifyHost", (SM_BOOL *) &ssl_verify_host) == SM_FAULT)
		return SM_FAULT;

	if (this && sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:X509CertFile", &x509_cert_file) == SM_FAULT)
		return SM_FAULT;

	if (this && sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:X509KeyFile", &x509_key_file) == SM_FAULT)
		return SM_FAULT;

	if (this && sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:X509KeyPass", &x509_key_pass) == SM_FAULT)
		return SM_FAULT;

	if (this && sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:TrustedCAFile", &trusted_ca_file) == SM_FAULT)
		return SM_FAULT;

	if (this && sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:TrustedCAPath", &trusted_ca_path) == SM_FAULT)
		return SM_FAULT;

	if (curl_global_init(CURL_GLOBAL_ALL)) {
		sm_error("Failed to initialize libcurl");
		return SM_FAULT;
	}

	data = curl_version_info(CURLVERSION_NOW);

	if (data->features & CURL_VERSION_LARGEFILE) {
		sm_debug("CURL_VERSION_LARGEFILE");
		largefile = true;
	}

	for (i = 0; data->protocols[i]; i++) {
		if (!strcmp(data->protocols[i], "ftp"))
			smc_extensions_add(SMC_NAMESPACE_HPCP_FS_PROTO "/ftp");
		else if (!strcmp(data->protocols[i], "http"))
			smc_extensions_add(SMC_NAMESPACE_HPCP_FS_PROTO "/http/v11");
		else if (!strcmp(data->protocols[i], "scp"))
			smc_extensions_add(SMC_NAMESPACE_HPCP_FS_PROTO "/scp");
		else {
			char buf[1024];

			/* Protocols not defined in GFD.135 */
			snprintf(buf, sizeof(buf), SMC_NAMESPACE_HPCP_FS_PROTO "/%s", data->protocols[i]);
			smc_extensions_add(buf);
		}

		sm_value_append(&supported_protocols, 1, SM_VALUE_STRING, data->protocols[i]);
	}

	smc_extensions_add(SMC_NAMESPACE_HPCP_FS_TOKEN "/username");
	return SM_OK;
}

static void
cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();

	curl_global_cleanup();
	sm_value_free(&supported_protocols);
}

typedef struct {
	time_t last;
	const char *url;
} progress_data_t;

static int
progress_callback(void *clientp, double dltotal, double dlnow, double ultotal, double ulnow)
{
	double now = 0;
	double total = 0;
	progress_data_t *p = (progress_data_t *) clientp;

	if (dltotal || dlnow) {
		now = dlnow;
		total = dltotal;
	} else if (ultotal || ulnow) {
		now = ulnow;
		total = ultotal;
	}

	if (time(NULL) - p->last < PROGRESS_PERIOD)
		return 0;

	p->last = time(NULL);

	if (show_progress) {
		if (total == 0)
			sm_info("Transferring %s", p->url);
		else
			sm_info("Transferring %s: %d%% (%.2lf/%.2lf KB)", p->url, (int) (now / total * 100), now / 1024, total / 1024);
	} else {
		if (total == 0)
			sm_debug("Transferring %s", p->url);
		else
			sm_debug("Transferring %s: %d%% (%.2lf/%.2lf KB)", p->url, (int) (now / total * 100), now / 1024, total / 1024);
	}

	return 0;
}

#define GET_HPCP_FS_CREDENTIAL() \
	do { \
		if (data_staging->hpcp_fs__Credential) { \
			_hpcp_fs__Credential *cred = data_staging->hpcp_fs__Credential; \
			char __buf[4096] = ""; \
			int __i; \
			\
			for (__i = 0; __i < cred->__size; __i++) { \
				struct soap_dom_element *dom = cred->__any + __i; \
				\
				if (dom->type == SOAP_TYPE__wsse__UsernameToken) { \
					_wsse__UsernameToken *t = (_wsse__UsernameToken *) dom->node; \
					char *__tmp; \
					\
					if (t->Username) { \
						SM_TRY_NOT(__tmp = curl_easy_escape(handles[i], t->Username, 0)); \
						strlcat(__buf, t->Username, sizeof(__buf)); \
					} \
					\
					if (t->Password) { \
						strlcat(__buf, ":", sizeof(__buf)); \
						strlcat(__buf, t->Password->__item, sizeof(__buf)); \
					} \
				} else { \
					result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Unsupported element in hpcp_fs__Credential");\
					goto fault; \
				} \
			} \
			\
			if (strlen(__buf)) \
				cret |= curl_easy_setopt(handles[i], CURLOPT_USERPWD, __buf); \
		} \
	} while(0)

#define CHECK_ELEMENT(element, ignored) \
	do { \
		struct soap_dom_attribute *__attr; \
		\
		if (!(element)) \
			break; \
		\
		if ((element)->__size != 0) { \
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Unsupported " SM_STR(element) " element extension in staging_curl"); \
			goto fault; \
		} \
		\
		for (__attr = &(element)->__anyAttribute; __attr; __attr = __attr->next) { \
			if (!__attr->name) continue; \
			if (!strncmp(__attr->name, "xmlns", strlen("xmlns"))) continue; \
			if (*ignored && !strcmp(__attr->name, ignored)) continue; \
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Unsupported " SM_STR(element) " attribute extension in staging_curl"); \
			goto fault; \
		} \
	} while (0)

#define CHECK_EXTENSIONS()   \
	do { \
		_jsdl__DataStaging *jsdl__DataStaging = data_staging; \
		\
		if (data_staging->jsdl__FilesystemName) { \
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "jsdl__DataStaging->jsdl__FilesystemName element not supported in staging_curl"); \
			goto fault; \
		} \
		\
		CHECK_ELEMENT(jsdl__DataStaging, "name"); \
		CHECK_ELEMENT(jsdl__DataStaging->jsdl__Source, ""); \
		CHECK_ELEMENT(jsdl__DataStaging->jsdl__Target, ""); \
	} while(0)

#define CLEANUP_HANDLES() \
	do { \
		for (i = 0; i < size; i++) { \
			if (!handles[i]) \
				continue; \
			curl_easy_cleanup(handles[i]); \
			handles[i] = NULL; \
		} \
		handles = NULL; \
		if (multi) \
			curl_multi_cleanup(multi); \
		multi = NULL; \
	} while(0)

static bool
is_supported(const char *uri)
{
	char *buf = NULL;
	char *p;
	size_t i, count;

	SM_MANDATORY_PARAM(uri, NULL);

	SM_DEBUG_ENTER();

	SM_TRY_NOT(buf = sm_strdup(uri));

	if (!(p = strstr(buf, "://"))) {
		sm_notice("Malformed URI in DataStaging element: %s", uri);
		goto fault;
	}

	*p = 0;

	count = sm_value_count(supported_protocols);

	for (i = 0; i < count; i++) {
		if (!strcmp(supported_protocols[i]->vstring, buf)) {
			SM_FREE(buf);
			return true;
		}
	}

fault:
	SM_FREE(buf);
	return false;
}

static size_t
get_handle_idx(CURL **handles, size_t size, CURL *handle)
{
	size_t i;

	for (i = 0; i < size; i++) {
		if (handles[i] == handle)
			return i;
	}

	return 0;
}

static RETSIGTYPE
sig_handler(int sig sm_unused)
{
	return;
}

static sm_value_t *
stage_in(sm_module_t this sm_unused, _jsdl__DataStaging *ds, size_t size, const char *working_dir)
{
	CURLcode cret = 0;
	sm_value_t *result = NULL;
	size_t i;
	sm_alloc_tracked_t tracked = NULL;
	CURLM *multi = NULL;
	CURL **handles = NULL;
	FILE **files = NULL;
	const char **file_names = NULL;
	const char **urls = NULL;
	char **errbufs = NULL;
	progress_data_t *progress_data = NULL;
	int prev_running = 0;
	int running = 0;
	struct sigaction sa;

	SM_MANDATORY_P(ds);

	SM_DEBUG_ENTER();

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = sig_handler;
	sigaction(SIGUSR1, &sa, NULL);

	SM_TRY_NOT(tracked = sm_alloc_tracked_create());
	SM_TRY_NOT(handles = sm_alloc_tracked_calloc(tracked, size, sizeof(CURL *)));
	SM_TRY_NOT(files = sm_alloc_tracked_calloc(tracked, size, sizeof(FILE *)));
	SM_TRY_NOT(file_names = sm_alloc_tracked_calloc(tracked, size, sizeof(const char *)));
	SM_TRY_NOT(urls = sm_alloc_tracked_calloc(tracked, size, sizeof(const char *)));
	SM_TRY_NOT(errbufs = sm_alloc_tracked_calloc(tracked, size, sizeof(char *)));
	for (i = 0; i < size; i++)
		SM_TRY_NOT(errbufs[i] = sm_alloc_tracked_malloc(tracked, CURL_ERROR_SIZE));
	SM_TRY_NOT(progress_data = sm_alloc_tracked_calloc(tracked, size, sizeof(progress_data_t)));

	for (i = 0; i < size; i++) {
		const char *url = NULL;
		const char *fopen_mode = NULL;
		_jsdl__DataStaging *data_staging;
		char *file_name = NULL;

		data_staging = ds + i;

		if (!data_staging->jsdl__FileName)
			continue;
		if (!data_staging->jsdl__Source || !data_staging->jsdl__Source->jsdl__URI)
			continue;

		file_name = data_staging->jsdl__FileName;
		url = data_staging->jsdl__Source->jsdl__URI;

		if (!is_supported(url))
			continue;

		/* Mark it as being handled by this module. */
		data_staging->jsdl__FileName = NULL;

		CHECK_EXTENSIONS();

		if (working_dir)
			SM_TRY_NOT(file_name = sm_alloc_tracked_asprintf(tracked, "%s/%s", working_dir, file_name));
		else
			SM_TRY_NOT(file_name = sm_alloc_tracked_strdup(tracked, file_name));

		urls[i] = url;
		file_names[i] = file_name;

		if (data_staging->jsdl__CreationFlag == jsdl__CreationFlagEnumeration__dontOverwrite) {
			struct stat st;

			if (!stat(file_name, &st)) {
				sm_debug("File %s already exists, not staging in", file_name);
				continue;
			}

			/* On error condition, e.g. access denied, we try to write to
			 * the file and we'll eventually throw a fault later anyway. */
		}

		switch (data_staging->jsdl__CreationFlag) {
			case jsdl__CreationFlagEnumeration__dontOverwrite:
			case jsdl__CreationFlagEnumeration__overwrite:
				fopen_mode = "wb";
				break;
			case jsdl__CreationFlagEnumeration__append:
				fopen_mode = "ab";
				break;
		}

		if (!(files[i] = fopen(file_name, fopen_mode))) {
			char buf[256] = "";
			char err[256] = "";
			int saved_errno;

			saved_errno = errno;
			sm_strerror_r(errno, err, sizeof(err));

			snprintf(buf, sizeof(buf), "Unable to open file %s for writing: %s", file_name, err);

			if (saved_errno == EACCES)
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_NOT_AUTHORIZED_FAULT, buf);
			else
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, buf);

			goto fault;
		}

		sm_debug("Staging in: %s -> %s", url, file_name);

		if ((handles[i] = curl_easy_init()) == NULL) {
			sm_error("curl_easy_init failed");
			goto fault;
		}

		cret |= curl_easy_setopt(handles[i], CURLOPT_URL, url);
		cret |= curl_easy_setopt(handles[i], CURLOPT_ERRORBUFFER, errbufs[i]);
		cret |= curl_easy_setopt(handles[i], CURLOPT_FAILONERROR, 1);

		cret |= curl_easy_setopt(handles[i], CURLOPT_WRITEDATA, files[i]);

		if (verbose)
			cret |= curl_easy_setopt(handles[i], CURLOPT_VERBOSE, 1);

		cret |= curl_easy_setopt(handles[i], CURLOPT_AUTOREFERER, 1);
		cret |= curl_easy_setopt(handles[i], CURLOPT_FOLLOWLOCATION, 1);
		cret |= curl_easy_setopt(handles[i], CURLOPT_NOSIGNAL, 1);
		cret |= curl_easy_setopt(handles[i], CURLOPT_NOPROGRESS, 0);

		progress_data[i].last = time(NULL);
		progress_data[i].url = url;
		cret |= curl_easy_setopt(handles[i], CURLOPT_PROGRESSDATA, &progress_data[i]);
		cret |= curl_easy_setopt(handles[i], CURLOPT_PROGRESSFUNCTION, progress_callback);
		cret |= curl_easy_setopt(handles[i], CURLOPT_USERAGENT, "QCG Computing using curl");
		cret |= curl_easy_setopt(handles[i], CURLOPT_HTTPAUTH, CURLAUTH_ANY);
		cret |= curl_easy_setopt(handles[i], CURLOPT_MAXREDIRS, 5);

		cret |= curl_easy_setopt(handles[i], CURLOPT_SSL_VERIFYHOST, ssl_verify_host ? (long) 2 : (long) 0);

		if (x509_cert_file)
			cret |= curl_easy_setopt(handles[i], CURLOPT_SSLCERT, x509_cert_file);
		if (x509_key_file)
			cret |= curl_easy_setopt(handles[i], CURLOPT_SSLKEY, x509_key_file);
		if (x509_key_pass)
			cret |= curl_easy_setopt(handles[i], CURLOPT_KEYPASSWD, x509_key_pass);
		if (trusted_ca_file)
			cret |= curl_easy_setopt(handles[i], CURLOPT_CAINFO, trusted_ca_file);
		if (trusted_ca_path)
			cret |= curl_easy_setopt(handles[i], CURLOPT_CAPATH, trusted_ca_path);

		/* XXX: To be strict with GFD.135 we should ignore any user/password
		 * that is specified in URL. */
		GET_HPCP_FS_CREDENTIAL();

		if (cret) {
			sm_error("curl_easy_setopt failed");
			goto fault;
		}
	}

	if ((multi = curl_multi_init()) == NULL) {
		sm_error("curl_multi_init failed");
		goto fault;
	}

	(void) curl_multi_setopt(multi, CURLMOPT_PIPELINING, (long) 1);

	for (i = 0; i < size; i++) {
		if (!handles[i])
			continue;

		if (curl_multi_add_handle(multi, handles[i]) != CURLM_OK) {
			sm_error("curl_multi_add_handle failed");
			goto fault;
		}
	}

	while (1) {
		CURLMcode mret;
		fd_set rfds, wfds, xfds;
		int max_fds;
		struct timespec tv;
		long timeout;
		int ret;
		sigset_t sig_set;

		prev_running = running;

		do {
			mret = curl_multi_perform(multi, &running);
		} while (mret == CURLM_CALL_MULTI_PERFORM);

		if (running == 0 || running < prev_running) {
			CURLMsg *msg;

			if (prev_running > running)
				sm_debug("%d transfers finished", prev_running - running);

			while (1) {
				size_t handle_idx;
				int msgs;

				msg = curl_multi_info_read(multi, &msgs);
				if (!msg)
					break;

				handle_idx = get_handle_idx(handles, size, msg->easy_handle);

				if (msg->data.result != CURLE_OK) {
					switch (msg->data.result) {
						case CURLE_REMOTE_FILE_NOT_FOUND:
						case CURLE_COULDNT_RESOLVE_HOST:
							result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_FILE_NOT_FOUND_FAULT, errbufs[handle_idx]);
							break;
						case CURLE_LOGIN_DENIED:
						case CURLE_REMOTE_ACCESS_DENIED:
							result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_NOT_AUTHORIZED_FAULT, errbufs[handle_idx]);
							break;
						case CURLE_FAILED_INIT:
						case CURLE_OUT_OF_MEMORY:
						case CURLE_WRITE_ERROR:
						case CURLE_FILE_COULDNT_READ_FILE:
							result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, errbufs[handle_idx]);
							break;
						default:
							result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_REMOTE_STAGING_FAULT, errbufs[handle_idx]);
							break;
					}

					goto fault;

				} else if (show_summary) {
					double value;

					if (curl_easy_getinfo(msg->easy_handle, CURLINFO_SPEED_DOWNLOAD, &value) == CURLE_OK)
						sm_info("Downloaded %s -> %s (%.2lf KB/s)", urls[handle_idx], file_names[handle_idx], value / 1024);
				}
			}
		}

		if (running == 0)
			break;

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_ZERO(&xfds);

		if (curl_multi_fdset(multi, &rfds, &wfds, &xfds, &max_fds) != CURLM_OK) {
			sm_error("curl_multi_fdset failed");
			goto fault;
		}

		if (max_fds == -1)
			continue;

		if (curl_multi_timeout(multi, (long *) &timeout) != CURLM_OK) {
			sm_error("curl_multi_timeout failed");
			goto fault;
		}

		if (timeout == 0)
			continue;

		if (timeout == -1)
			timeout = 500;

		tv.tv_sec = timeout / 1000;
		tv.tv_nsec = (timeout % 1000) * 1000;

		assert(tv.tv_sec >= 0 && tv.tv_nsec >= 0);

		/* This thread blocks all the signals, because it was
		 * created with sm_application_thread_create(). We want to know
		 * when the master thread in session process tells us to
		 * terminate. */
		sigfillset(&sig_set);
		sigdelset(&sig_set, SIGUSR1);

		ret = pselect(max_fds + 1, &rfds, &wfds, &xfds, &tv, &sig_set);

		if (ret == -1) {
			if (errno == EINTR) {
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Staging cancelled");
				goto fault;
			}

			sm_perror("pselect()");
			goto fault;
		}
	}

	CLEANUP_HANDLES();

	for (i = 0; i < size; i++) {
		if (files[i] && fclose(files[i])) {
			char buf[256] = "";
			char err[256] = "";

			sm_strerror_r(errno, err, sizeof(err));
			snprintf(buf, sizeof(buf), "Failed to close file %s: %s", file_names[i], err);
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, buf);
			goto fault;
		}

		files[i] = NULL;
	}

	sm_alloc_tracked_free(&tracked);
	return sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK);

fault:
	CLEANUP_HANDLES();

	for (i = 0; i < size; i++) {
		if (files[i]) {
			fclose(files[i]);
			files[i] = NULL;
		}
	}

	sm_alloc_tracked_free(&tracked);

	if (result) {
		sm_notice("staging (IN) fault: %s",  SM_VALUE_RET_CODE_STRING(result[0]));
		return result;
	} else
		return sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in staging_curl module");
}

static sm_value_t *
stage_out(sm_module_t this sm_unused, _jsdl__DataStaging *ds, size_t size, const char *working_dir)
{
	CURLcode cret = 0;
	sm_value_t *result = NULL;
	size_t i;
	sm_alloc_tracked_t tracked = NULL;
	CURLM *multi = NULL;
	CURL **handles = NULL;
	FILE **files = NULL;
	const char **file_names = NULL;
	const char **urls = NULL;
	char **errbufs = NULL;
	progress_data_t *progress_data = NULL;
	int prev_running = 0;
	int running = 0;
	struct sigaction sa;

	SM_MANDATORY_P(ds);

	SM_DEBUG_ENTER();

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = sig_handler;
	sigaction(SIGUSR1, &sa, NULL);

	SM_TRY_NOT(tracked = sm_alloc_tracked_create());
	SM_TRY_NOT(handles = sm_alloc_tracked_calloc(tracked, size, sizeof(CURL *)));
	SM_TRY_NOT(files = sm_alloc_tracked_calloc(tracked, size, sizeof(FILE *)));
	SM_TRY_NOT(file_names = sm_alloc_tracked_calloc(tracked, size, sizeof(const char *)));
	SM_TRY_NOT(urls = sm_alloc_tracked_calloc(tracked, size, sizeof(const char *)));
	SM_TRY_NOT(errbufs = sm_alloc_tracked_calloc(tracked, size, sizeof(char *)));
	for (i = 0; i < size; i++)
		SM_TRY_NOT(errbufs[i] = sm_alloc_tracked_malloc(tracked, CURL_ERROR_SIZE));
	SM_TRY_NOT(progress_data = sm_alloc_tracked_calloc(tracked, size, sizeof(progress_data_t)));

	for (i = 0; i < size; i++) {
		const char *url = NULL;
		_jsdl__DataStaging *data_staging;
		char *file_name = NULL;
		struct stat st;

		data_staging = ds + i;

		if (!data_staging->jsdl__FileName)
			continue;
		if (!data_staging->jsdl__Target || !data_staging->jsdl__Target->jsdl__URI)
			continue;

		file_name = data_staging->jsdl__FileName;
		url = data_staging->jsdl__Target->jsdl__URI;

		if (!is_supported(url))
			continue;

		/* Mark it as being handled by this module. */
		data_staging->jsdl__FileName = NULL;

		CHECK_EXTENSIONS();

		if (data_staging->jsdl__CreationFlag == jsdl__CreationFlagEnumeration__dontOverwrite) {
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "\"dontOverwrite\" flag not supported when staging out");
			goto fault;
		}

		if (data_staging->jsdl__CreationFlag == jsdl__CreationFlagEnumeration__append && strncmp(url, "ftp://", 6)) {
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "\"append\" flag only supported for ftp protocol when staging out");
			goto fault;
		}

		if (working_dir)
			SM_TRY_NOT(file_name = sm_alloc_tracked_asprintf(tracked, "%s/%s", working_dir, file_name));
		else
			SM_TRY_NOT(file_name = sm_alloc_tracked_strdup(tracked, file_name));

		urls[i] = url;
		file_names[i] = file_name;

		if (!(files[i] = fopen(file_name, "rb")) || stat(file_name, &st)) {
			char buf[256] = "";
			char err[256] = "";
			int saved_errno;

			saved_errno = errno;
			sm_strerror_r(errno, err, sizeof(err));

			snprintf(buf, sizeof(buf), "Unable to open file %s for reading: %s", file_name, err);

			if (saved_errno == EACCES)
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_NOT_AUTHORIZED_FAULT, buf);
			else
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, buf);
			/*TODO implement fault tolerant semantics */
			goto fault;
		}

		if (!(S_ISREG(st.st_mode) || S_ISLNK(st.st_mode))) {
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, "Only regular files and symbolic links can be staged out");
			goto fault;
		}

		sm_debug("Staging out: %s -> %s", file_name, url);

		if ((handles[i] = curl_easy_init()) == NULL) {
			sm_error("curl_easy_init failed");
			goto fault;
		}

		cret |= curl_easy_setopt(handles[i], CURLOPT_UPLOAD, 1);
		cret |= curl_easy_setopt(handles[i], CURLOPT_FTP_CREATE_MISSING_DIRS, 1);

		if (data_staging->jsdl__CreationFlag == jsdl__CreationFlagEnumeration__append) {
			/* This only makes sense to ftp transfers. */
			cret |= curl_easy_setopt(handles[i], CURLOPT_APPEND, 1);
		}

		if (largefile) {
			SM_ASSERT(sizeof(curl_off_t) == 8, "curl_off_t should be 64 bits, check for needed definitions for sys/types.h");
			cret |= curl_easy_setopt(handles[i], CURLOPT_INFILESIZE_LARGE, (curl_off_t) st.st_size);
		} else
			cret |= curl_easy_setopt(handles[i], CURLOPT_INFILESIZE, (long) st.st_size);

		cret |= curl_easy_setopt(handles[i], CURLOPT_URL, url);
		cret |= curl_easy_setopt(handles[i], CURLOPT_ERRORBUFFER, errbufs[i]);
		cret |= curl_easy_setopt(handles[i], CURLOPT_FAILONERROR, 1);

		cret |= curl_easy_setopt(handles[i], CURLOPT_READDATA, files[i]);

		if (verbose)
			cret |= curl_easy_setopt(handles[i], CURLOPT_VERBOSE, 1);

		cret |= curl_easy_setopt(handles[i], CURLOPT_AUTOREFERER, 1);
		cret |= curl_easy_setopt(handles[i], CURLOPT_FOLLOWLOCATION, 1);
		cret |= curl_easy_setopt(handles[i], CURLOPT_NOSIGNAL, 1);
		cret |= curl_easy_setopt(handles[i], CURLOPT_NOPROGRESS, 0);

		progress_data[i].last = time(NULL);
		progress_data[i].url = url;
		cret |= curl_easy_setopt(handles[i], CURLOPT_PROGRESSDATA, &progress_data[i]);
		cret |= curl_easy_setopt(handles[i], CURLOPT_PROGRESSFUNCTION, progress_callback);
		cret |= curl_easy_setopt(handles[i], CURLOPT_USERAGENT, "QCG Computing using curl");
		cret |= curl_easy_setopt(handles[i], CURLOPT_HTTPAUTH, CURLAUTH_ANY);
		cret |= curl_easy_setopt(handles[i], CURLOPT_MAXREDIRS, 5);

		cret |= curl_easy_setopt(handles[i], CURLOPT_SSL_VERIFYHOST, ssl_verify_host ? (long) 2 : (long) 0);

		if (x509_cert_file)
			cret |= curl_easy_setopt(handles[i], CURLOPT_SSLCERT, x509_cert_file);
		if (x509_key_file)
			cret |= curl_easy_setopt(handles[i], CURLOPT_SSLKEY, x509_key_file);
		if (x509_key_pass)
			cret |= curl_easy_setopt(handles[i], CURLOPT_KEYPASSWD, x509_key_pass);
		if (trusted_ca_file)
			cret |= curl_easy_setopt(handles[i], CURLOPT_CAINFO, trusted_ca_file);
		if (trusted_ca_path)
			cret |= curl_easy_setopt(handles[i], CURLOPT_CAPATH, trusted_ca_path);

		/* XXX: To be strict with GFD.135 we should ignore any user/password
		 * that is specified in URL. */
		GET_HPCP_FS_CREDENTIAL();

		if (cret) {
			sm_error("curl_easy_setopt failed");
			goto fault;
		}
	}

	if ((multi = curl_multi_init()) == NULL) {
		sm_error("curl_multi_init failed");
		goto fault;
	}

	(void) curl_multi_setopt(multi, CURLMOPT_PIPELINING, (long) 1);

	for (i = 0; i < size; i++) {
		if (!handles[i])
			continue;

		if (curl_multi_add_handle(multi, handles[i]) != CURLM_OK) {
			sm_error("curl_multi_add_handle failed");
			goto fault;
		}
	}

	while (1) {
		CURLMcode mret;
		fd_set rfds, wfds, xfds;
		int max_fds;
		struct timespec tv;
		long timeout;
		int ret;
		sigset_t sig_set;

		prev_running = running;

		do {
			mret = curl_multi_perform(multi, &running);
		} while (mret == CURLM_CALL_MULTI_PERFORM);

		if (running == 0 || running < prev_running) {
			CURLMsg *msg;

			if (prev_running > running)
				sm_debug("%d transfers finished", prev_running - running);

			while (1) {
				size_t handle_idx;
				int msgs;

				msg = curl_multi_info_read(multi, &msgs);
				if (!msg)
					break;

				handle_idx = get_handle_idx(handles, size, msg->easy_handle);

				if (msg->data.result != CURLE_OK) {
					switch (msg->data.result) {
						case CURLE_REMOTE_FILE_NOT_FOUND:
						case CURLE_COULDNT_RESOLVE_HOST:
							result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_FILE_NOT_FOUND_FAULT, errbufs[handle_idx]);
							break;
						case CURLE_LOGIN_DENIED:
						case CURLE_REMOTE_ACCESS_DENIED:
							result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_NOT_AUTHORIZED_FAULT, errbufs[handle_idx]);
							break;
						case CURLE_FAILED_INIT:
						case CURLE_OUT_OF_MEMORY:
						case CURLE_READ_ERROR:
						case CURLE_FILE_COULDNT_READ_FILE:
							result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, errbufs[handle_idx]);
							break;
						default:
							result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_REMOTE_STAGING_FAULT, errbufs[handle_idx]);
							break;
					}

					goto fault;

				} else if (show_summary) {
					double value;

					if (curl_easy_getinfo(msg->easy_handle, CURLINFO_SPEED_UPLOAD, &value) == CURLE_OK)
						sm_info("Uploaded %s -> %s (%.2lf KB/s)", file_names[handle_idx], urls[handle_idx], value / 1024);
				}
			}
		}

		if (running == 0)
			break;

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_ZERO(&xfds);

		if (curl_multi_fdset(multi, &rfds, &wfds, &xfds, &max_fds) != CURLM_OK) {
			sm_error("curl_multi_fdset failed");
			goto fault;
		}

		if (max_fds == -1)
			continue;

		if (curl_multi_timeout(multi, (long *) &timeout) != CURLM_OK) {
			sm_error("curl_multi_timeout failed");
			goto fault;
		}

		if (timeout == 0)
			continue;

		if (timeout == -1)
			timeout = 500;

		tv.tv_sec = timeout / 1000;
		tv.tv_nsec = (timeout % 1000) * 1000;

		assert(tv.tv_sec >= 0 && tv.tv_nsec >= 0);

		/* This thread blocks almost all the signals, because it was
		 * created with sm_application_thread_create(). We want to know
		 * when the master thread in session process tells us to
		 * terminate. */
		sigfillset(&sig_set);
		sigdelset(&sig_set, SIGUSR1);

		ret = pselect(max_fds + 1, &rfds, &wfds, &xfds, &tv, &sig_set);

		if (ret == -1) {
			if (errno == EINTR) {
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Staging cancelled");
				goto fault;
			}

			sm_perror("pselect()");
			goto fault;
		}
	}

	CLEANUP_HANDLES();

	for (i = 0; i < size; i++) {
		if (files[i]) {
			fclose(files[i]);
			files[i] = NULL;
		}
	}

	sm_alloc_tracked_free(&tracked);
	return sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK);

fault:
	CLEANUP_HANDLES();

	for (i = 0; i < size; i++) {
		if (files[i]) {
			fclose(files[i]);
			files[i] = NULL;
		}
	}

	sm_alloc_tracked_free(&tracked);

	if (result) {
		sm_notice("staging (OUT) fault: %s",  SM_VALUE_RET_CODE_STRING(result[0]));
		return result;
	} else
		return sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in staging_curl module");
}

#ifdef TEST
static void
test_download(char *argv[])
{
	sm_value_t *result;
	int i;
	size_t j;
	_jsdl__DataStaging ds[128];
	_jsdl__Source source[128];

	memset(&ds, 0, sizeof(ds));
	memset(&source, 0, sizeof(source));

	for (i = 0, j = 0; argv[i] && j < 128; i += 2, j++) {
		const char *file;
		const char *url;

		file = argv[i];
		url = argv[i + 1];

		if (!file || !url) {
			sm_error("Please specify file and URL");
			return;
		}

		ds[j].jsdl__FileName = (char *) file;
		ds[j].jsdl__CreationFlag = jsdl__CreationFlagEnumeration__overwrite;
		ds[j].jsdl__Source = &source[j];
		ds[j].jsdl__Source->jsdl__URI = (char *) url;
	}

	result = stage_in(NULL, ds, j, NULL);

	SM_VALUE_DEBUG(result);
	sm_value_free(&result);
}

static void
test_upload(char *argv[])
{
	sm_value_t *result;
	int i;
	size_t j;
	_jsdl__DataStaging ds[128];
	_jsdl__Target source[128];

	memset(&ds, 0, sizeof(ds));
	memset(&source, 0, sizeof(source));

	for (i = 0, j = 0; argv[i] && j < 128; i += 2, j++) {
		const char *file;
		const char *url;

		file = argv[i];
		url = argv[i + 1];

		if (!file || !url) {
			sm_error("Please specify file and URL");
			return;
		}

		ds[j].jsdl__FileName = (char *) file;
		ds[j].jsdl__CreationFlag = jsdl__CreationFlagEnumeration__overwrite;
		ds[j].jsdl__Target = &source[j];
		ds[j].jsdl__Target->jsdl__URI = (char *) url;
	}

	result = stage_out(NULL, ds, j, NULL);

	SM_VALUE_DEBUG(result);
	sm_value_free(&result);
}

int
main(int argc, char *argv[])
{
	argc--;

	if (argc < 3) {
		fprintf(stderr, "%s: <d|u> (<file> <URL>) ...\n", argv[0]);
		return -1;
	}

	assert(sm_library_setup() == SM_OK);
	assert(sm_log_init_str(NULL, "DEBUG") == SM_OK);

	setup(NULL);

	if (getenv("VERBOSE"))
		verbose = true;
	else
		verbose = false;

	ssl_verify_host = false;

	if (!strcmp(argv[1], "d"))
		test_download(argv + 2);
	else if (!strcmp(argv[1], "u"))
		test_upload(argv + 2);

	return 0;
}
#endif
