/* $Id: staging_s3.c 54 2010-01-21 18:36:32Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdbool.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>

#include <libs3.h>

#include <qcg-core/sm.h>
#include <qcg-core/logger.h>
#include <qcg-core/value.h>
#include <qcg-core/utils.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>

#include <xmlstructs/xmlstructs.h>
#include <include/namespaces.h>
#include <include/staging_module.h>
#include <include/extensions.h>

__SM_ID("$Id: staging_s3.c 54 2010-01-21 18:36:32Z mamonski $");

static int setup(sm_module_t);
static void cleanup(sm_module_t);
static sm_value_t *stage_in(sm_module_t, _jsdl__DataStaging *, size_t, const char *);
static sm_value_t *stage_out(sm_module_t, _jsdl__DataStaging *, size_t, const char *);

STAGING_MODULE = {

	.header = {
		.entry_size = STAGING_MODULE_ENTRY_SIZE,
		.api_type = STAGING_MODULE_API,

		.name = "staging_s3",
		.descr = "Amazon S3 staging module",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	},

	.stage_in = stage_in,
	.stage_out = stage_out,
};


static const char *aws_key_id = NULL;
static const char *aws_access_key = NULL;

static S3Status write_callback(int bufferSize, const char *buffer, void *callbackData);
static int read_callback(int bufferSize, char *buffer, void *callbackData);

static void complete_callback(S3Status status, const S3ErrorDetails *errorDetails, void *callbackData);
static S3Status properties_callback(const S3ResponseProperties *properties sm_unused, void *callbackData sm_unused);

static int parse_url(sm_alloc_tracked_t handle, const char *object_url, S3BucketContext *bucket, char **key);

typedef struct smc_staging_context_s {
	const char *file_name;
	FILE *fhandle;
	sm_value_t *fault;
	size_t file_size;
} *smc_staging_context_t;

static int download_file(sm_alloc_tracked_t handle, const char *object_url, smc_staging_context_t ctxt);
static int upload_file(sm_alloc_tracked_t handle, const char *object_url, smc_staging_context_t ctxt);

static int
setup(sm_module_t this)
{
	S3Status status;

	SM_DEBUG_ENTER();

	if (this && sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:AWSKeyId", &aws_key_id) != SM_OK)
		return SM_FAULT;

	if (this && sm_conf_node_str(this->conf_ctxt, SM_FALSE, "/conf:AWSAccessKey", &aws_access_key) != SM_OK)
		return SM_FAULT;

	if ((status  = S3_initialize("QCG Computing", S3_INIT_ALL)) != S3StatusOK) {
		sm_error("S3_initialize failed: %s", S3_get_status_name(status));
		return SM_FAULT;
	}

	return SM_OK;
}

static void
cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();

	S3_deinitialize();
}

#define CHECK_ELEMENT(element, ignored) \
	do { \
		struct soap_dom_attribute *__attr; \
		\
		if (!(element)) \
			break; \
		\
		if ((element)->__size != 0) { \
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Unsupported " SM_STR(element) " element extension in staging_s3"); \
			goto fault; \
		} \
		\
		for (__attr = &(element)->__anyAttribute; __attr; __attr = __attr->next) { \
			if (!__attr->name) continue; \
			if (!strncmp(__attr->name, "xmlns", strlen("xmlns"))) continue; \
			if (*ignored && !strcmp(__attr->name, ignored)) continue; \
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Unsupported " SM_STR(element) " attribute extension in staging_s3"); \
			goto fault; \
		} \
	} while (0)

#define CHECK_EXTENSIONS()   \
	do { \
		_jsdl__DataStaging *jsdl__DataStaging = data_staging; \
		\
		if (data_staging->jsdl__FilesystemName) { \
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "jsdl__DataStaging->jsdl__FilesystemName element not supported in staging_s3"); \
			goto fault; \
		} \
		\
		CHECK_ELEMENT(jsdl__DataStaging, "name"); \
		CHECK_ELEMENT(jsdl__DataStaging->jsdl__Source, ""); \
		CHECK_ELEMENT(jsdl__DataStaging->jsdl__Target, ""); \
	} while(0)


static sm_value_t *
stage_in(sm_module_t this sm_unused, _jsdl__DataStaging *ds, size_t size, const char *working_dir)
{
	sm_value_t *result = NULL;
	size_t i;
	sm_alloc_tracked_t tracked = NULL;
	struct smc_staging_context_s ctxt;

	SM_MANDATORY_P(ds);

	SM_DEBUG_ENTER();

	SM_TRY_NOT(tracked = sm_alloc_tracked_create());

	for (i = 0; i < size; i++) {
		const char *url = NULL;
		const char *fopen_mode = NULL;
		_jsdl__DataStaging *data_staging;
		char *file_name = NULL;

		data_staging = ds + i;

		if (!data_staging->jsdl__FileName)
			continue;
		if (!data_staging->jsdl__Source || !data_staging->jsdl__Source->jsdl__URI)
			continue;

		file_name = data_staging->jsdl__FileName;
		url = data_staging->jsdl__Source->jsdl__URI;

		if ( !url || (strncmp(url, "s3://", 5))) {
			continue;
		}

		/* Mark it as being handled by this module. */
		data_staging->jsdl__FileName = NULL;

		CHECK_EXTENSIONS();

		if (working_dir)
			SM_TRY_NOT(file_name = sm_alloc_tracked_asprintf(tracked, "%s/%s", working_dir, file_name));
		else
			SM_TRY_NOT(file_name = sm_alloc_tracked_strdup(tracked, file_name));

		if (data_staging->jsdl__CreationFlag == jsdl__CreationFlagEnumeration__dontOverwrite) {
			struct stat st;

			if (!stat(file_name, &st)) {
				sm_debug("File %s already exists, not staging in", file_name);
				continue;
			}

			/* On error condition, e.g. access denied, we try to write to
			 * the file and we'll eventually throw a fault later anyway. */
		}

		switch (data_staging->jsdl__CreationFlag) {
			case jsdl__CreationFlagEnumeration__dontOverwrite:
			case jsdl__CreationFlagEnumeration__overwrite:
				fopen_mode = "wb";
				break;
			case jsdl__CreationFlagEnumeration__append:
				fopen_mode = "ab";
				break;
		}

		ctxt.file_name = file_name;
		ctxt.file_size = 0;
		ctxt.fault = NULL;

		if (!(ctxt.fhandle = fopen(file_name, fopen_mode))) {
			char buf[256] = "";
			char err[256] = "";
			int saved_errno;

			saved_errno = errno;
			sm_strerror_r(errno, err, sizeof(err));

			snprintf(buf, sizeof(buf), "Unable to open file %s for writing: %s", file_name, err);
			sm_notice("%s", buf);

			if (saved_errno == EACCES)
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_NOT_AUTHORIZED_FAULT, buf);
			else
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, buf);

			goto fault;
		}

		sm_info("Staging in: %s -> %s", url, file_name);

		/* TODO: use asyn api */
		download_file(tracked, url, &ctxt);

		if (ctxt.fault) {
			fclose(ctxt.fhandle);
			result = ctxt.fault;
			goto fault;
		}

		fclose(ctxt.fhandle);
	}

	sm_alloc_tracked_free(&tracked);
	return sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK);

fault:


	sm_alloc_tracked_free(&tracked);

	if (result)
		return result;
	else
		return sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in staging_curl module");
}

static sm_value_t *
stage_out(sm_module_t this sm_unused, _jsdl__DataStaging *ds, size_t size, const char *working_dir)
{
	sm_value_t *result = NULL;
	size_t i;
	sm_alloc_tracked_t tracked = NULL;
	struct smc_staging_context_s ctxt;

	SM_MANDATORY_P(ds);

	SM_DEBUG_ENTER();

	SM_TRY_NOT(tracked = sm_alloc_tracked_create());

	for (i = 0; i < size; i++) {
		const char *url = NULL;
		_jsdl__DataStaging *data_staging;
		char *file_name = NULL;
		struct stat st;

		data_staging = ds + i;

		if (!data_staging->jsdl__FileName)
			continue;
		if (!data_staging->jsdl__Target || !data_staging->jsdl__Target->jsdl__URI)
			continue;

		file_name = data_staging->jsdl__FileName;
		url = data_staging->jsdl__Target->jsdl__URI;

		if ( !url || (strncmp(url, "s3://", 5))) {
			continue;
		}

		/* Mark it as being handled by this module. */
		data_staging->jsdl__FileName = NULL;

		CHECK_EXTENSIONS();

		if (data_staging->jsdl__CreationFlag == jsdl__CreationFlagEnumeration__dontOverwrite) {
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "\"dontOverwrite\" flag not supported when staging out");
			goto fault;
		}

		if (data_staging->jsdl__CreationFlag == jsdl__CreationFlagEnumeration__append) {
			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "\"append\" flag not supported when staging out");
			goto fault;
		}

		if (working_dir)
			SM_TRY_NOT(file_name = sm_alloc_tracked_asprintf(tracked, "%s/%s", working_dir, file_name));
		else
			SM_TRY_NOT(file_name = sm_alloc_tracked_strdup(tracked, file_name));

		if (((ctxt.fhandle = fopen(file_name, "rb")) == NULL) || (stat(file_name, &st) == -1)) {
			char buf[256] = "";
			char err[256] = "";
			int saved_errno;

			saved_errno = errno;
			sm_strerror_r(errno, err, sizeof(err));

			snprintf(buf, sizeof(buf), "Unable to open file %s for reading: %s", file_name, err);
			sm_notice("%s", buf);

			if (result)
				sm_value_free(&result);

			if (saved_errno == EACCES)
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_NOT_AUTHORIZED_FAULT, buf);
			else
				result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, buf);

			continue;
		}

		ctxt.file_name = file_name;
		ctxt.fault = NULL;
		ctxt.file_size = st.st_size;

		if (!(S_ISREG(st.st_mode) || S_ISLNK(st.st_mode))) {
			if (result)
				sm_value_free(&result);

			result = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, "Only regular files and symbolic links can be staged out");
			fclose(ctxt.fhandle);
			continue;
		}

		sm_info("Staging out: %s -> %s", file_name, url);

		/* TODO: use asyn api */
		upload_file(tracked, url, &ctxt);

		if (ctxt.fault) {
			if (result)
				sm_value_free(&result);
			result = ctxt.fault;
			fclose(ctxt.fhandle);
			continue;
		}

		fclose(ctxt.fhandle);
	}

	if (result)
		goto fault;

	sm_alloc_tracked_free(&tracked);
	return sm_value_create(1, SM_VALUE_RET_CODE, STAGING_OK);

fault:

	sm_alloc_tracked_free(&tracked);

	if (result)
		return result;
	else
		return sm_value_create(1, SM_VALUE_RET_CODE, STAGING_INTERNAL_FAULT, "Internal fault in staging_curl module");
}

static S3Status
write_callback(int bufferSize, const char *buffer, void *callbackData) {
	smc_staging_context_t ctxt = callbackData;

	SM_DEBUG_ENTER_INT(bufferSize);

	if (fwrite(buffer, bufferSize, 1, ctxt->fhandle) != 1) {
		char buf[256] = "";
		char err[256] = "";

		sm_strerror_r(errno, err, sizeof(err));
		sm_perror("fwrite failed");

		snprintf(buf, sizeof(buf), "Unable to open file %s for reading: %s", ctxt->file_name, err);

		ctxt->fault = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, buf);
		return S3StatusAbortedByCallback;
	}

	return S3StatusOK;
}

static int
read_callback(int bufferSize, char *buffer, void *callbackData) {
	smc_staging_context_t ctxt = callbackData;
	int ret;

	SM_DEBUG_ENTER_INT(bufferSize);

	if ((ret = fread(buffer, bufferSize, 1, ctxt->fhandle)) == 1) {
		return bufferSize;
	} else if (ret == 0) {
		sm_debug("EOF");
		return 0;
	} else {
		char buf[256] = "";
		char err[256] = "";

		sm_strerror_r(errno, err, sizeof(err));
		sm_perror("fread failed");

		snprintf(buf, sizeof(buf), "Unable to open file %s for reading: %s", ctxt->file_name, err);
		ctxt->fault = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_LOCAL_STAGING_FAULT, buf);

		return -1;
	}

}

static void
complete_callback(S3Status status, const S3ErrorDetails *errorDetails, void *callbackData) {
	smc_staging_context_t ctxt = callbackData;

	SM_DEBUG_ENTER();

	if (status == S3StatusOK) {
		sm_debug("File transfer completed");
	} else {
		sm_notice("File transfer failed: %s ", errorDetails->message);

		if (!ctxt->fault)
			ctxt->fault = sm_value_create(1, SM_VALUE_RET_CODE, STAGING_REMOTE_STAGING_FAULT, (errorDetails->message ? errorDetails->message : (S3_get_status_name(status))) );
	}

}

static S3Status
properties_callback(const S3ResponseProperties *properties sm_unused, void *callbackData sm_unused) {

	SM_DEBUG_ENTER();

	//ignore
	return S3StatusOK;
}

static int
parse_url(sm_alloc_tracked_t handle, const char *object_url, S3BucketContext *bucket, char **key) {
	char *url = NULL;
	char *p = NULL;
	char *kp = NULL;

	SM_DEBUG_ENTER_STR(object_url);

	bucket->accessKeyId = aws_key_id;
	bucket->secretAccessKey = aws_access_key;
	bucket->protocol = S3ProtocolHTTPS;
	bucket->uriStyle = S3UriStyleVirtualHost;

	p = url = sm_strdup(object_url);

	if (strncmp("s3://", url, 5) == 0) {
		p+=5;
	} else {
		return SM_FAULT;
	}

	if ((kp = strstr(p,"/")) == NULL) {
		return SM_FAULT;
	}

	*kp = '\0';
	kp++;

	bucket->bucketName = sm_alloc_tracked_strdup(handle, p);
	*key = sm_alloc_tracked_strdup(handle, kp);

	if (!bucket->bucketName || !*key)
		return SM_FAULT;

	sm_debug("%s resolved to: bucket =  %s, key = %s", object_url, bucket->bucketName, *key);

	sm_free(url);

	return 0;
}

static int
download_file(sm_alloc_tracked_t handle, const char *object_url, smc_staging_context_t ctxt) {
	S3GetObjectHandler handler;
	char *key = NULL;
	S3BucketContext bucketCxt;

	SM_DEBUG_ENTER_STR(object_url);

	handler.responseHandler.propertiesCallback = properties_callback;
	handler.responseHandler.completeCallback = complete_callback;
	handler.getObjectDataCallback = write_callback;

	if (parse_url(handle, object_url, &bucketCxt, &key) != 0) {
		return 1;
	}

	S3_get_object(&bucketCxt, key, NULL, 0, 0, NULL, &handler, ctxt);

	return SM_OK;
}

static int
upload_file(sm_alloc_tracked_t handle, const char *object_url, smc_staging_context_t ctxt) {
	S3PutObjectHandler handler;
	char *key = NULL;
	S3BucketContext bucketCxt;
	S3PutProperties putProperties;

	SM_DEBUG_ENTER_STR(object_url);

	handler.responseHandler.propertiesCallback = properties_callback;
	handler.responseHandler.completeCallback = complete_callback;
	handler.putObjectDataCallback = read_callback;

	if (parse_url(handle, object_url, &bucketCxt, &key) != 0) {
		return 1;
	}

	memset(&putProperties, 0, sizeof(putProperties));

	putProperties.cannedAcl = S3CannedAclAuthenticatedRead;

	S3_put_object(&bucketCxt, key, ctxt->file_size, &putProperties, NULL, &handler, ctxt);

	return SM_OK;
}


#ifdef TEST
static void
test_download(char *argv[])
{
	sm_value_t *result;
	int i;
	size_t j;
	_jsdl__DataStaging ds[128];
	_jsdl__Source source[128];

	memset(&ds, 0, sizeof(ds));
	memset(&source, 0, sizeof(source));

	for (i = 0, j = 0; argv[i] && j < 128; i += 2, j++) {
		const char *file;
		const char *url;

		file = argv[i];
		url = argv[i + 1];

		if (!file || !url) {
			sm_error("Please specify file and URL");
			return;
		}

		ds[j].jsdl__FileName = (char *) file;
		ds[j].jsdl__CreationFlag = jsdl__CreationFlagEnumeration__overwrite;
		ds[j].jsdl__Source = &source[j];
		ds[j].jsdl__Source->jsdl__URI = (char *) url;
	}

	result = stage_in(NULL, ds, j, NULL);

	SM_VALUE_DEBUG(result);
	sm_value_free(&result);
}

static void
test_upload(char *argv[])
{
	sm_value_t *result;
	int i;
	size_t j;
	_jsdl__DataStaging ds[128];
	_jsdl__Target source[128];

	memset(&ds, 0, sizeof(ds));
	memset(&source, 0, sizeof(source));

	for (i = 0, j = 0; argv[i] && j < 128; i += 2, j++) {
		const char *file;
		const char *url;

		file = argv[i];
		url = argv[i + 1];

		if (!file || !url) {
			sm_error("Please specify file and URL");
			return;
		}

		ds[j].jsdl__FileName = (char *) file;
		ds[j].jsdl__CreationFlag = jsdl__CreationFlagEnumeration__overwrite;
		ds[j].jsdl__Target = &source[j];
		ds[j].jsdl__Target->jsdl__URI = (char *) url;
	}

	result = stage_out(NULL, ds, j, NULL);

	SM_VALUE_DEBUG(result);
	sm_value_free(&result);
}

int
main(int argc, char *argv[])
{
	argc--;

	if (argc < 3) {
		fprintf(stderr, "%s: <d|u> (<file> <URL>) ...\n", argv[0]);
		return -1;
	}

	assert(sm_library_setup() == SM_OK);
	assert(sm_log_init_str(NULL, "DEBUG") == SM_OK);

	aws_key_id = getenv("S3_ACCESS_KEY_ID");
	aws_access_key = getenv("S3_SECRET_ACCESS_KEY");

	sm_debug("key_id=%s", aws_key_id);
	sm_debug("access_key=%s", aws_access_key);

	setup(NULL);

	if (!strcmp(argv[1], "d"))
		test_download(argv + 2);
	else if (!strcmp(argv[1], "u"))
		test_upload(argv + 2);

	return 0;
}
#endif
