/* $Id: wd_sandboxing.c 49 2010-01-05 11:10:31Z mamonski $ */

/*
 * Copyright 2006-2009 QCG Project Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>

#include <qcg-core/sm.h>
#include <qcg-core/compat.h>
#include <qcg-core/list.h>
#include <qcg-core/logger.h>
#include <qcg-core/signal.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/modules.h>
#include <qcg-core/uuid.h>

#include <include/jsdl_filter.h>
#include <xmlstructs/xmlstructs.h>

__SM_ID("$Id: wd_sandboxing.c 49 2010-01-05 11:10:31Z mamonski $");

static int setup(sm_module_t);
static void cleanup(sm_module_t);
static sm_value_t *handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params);

SM_GENERAL_MODULE = {

	.header = {
		.entry_size = SM_GENERAL_MODULE_ENTRY_SIZE,
		.api_type = SM_API_GENERAL,

		.name = "wd_sandboxing",
		.descr = "working directory sandboxing",

		.version_major = 0,
		.version_minor = 1,

		.setup = setup,
		.cleanup = cleanup,

		.multiple_instances_supported = SM_FALSE
	},
};

static int
setup(sm_module_t this sm_unused)
{

	SM_DEBUG_ENTER();

	SM_TRY_FAULT(sm_signal_connect("smc_jsdl_filter", handler));

	return SM_OK;

fault:

	return SM_FAULT;
}

static void
cleanup(sm_module_t this sm_unused)
{
	SM_DEBUG_ENTER();
}


static sm_value_t *
handler(sm_signal_t signal, sm_value_t *params, const sm_value_t *chained_params)
{
	sm_alloc_tracked_t mh;
	_jsdl__JobDefinition jsdl;
	const char *input_jsdl;
	char *output_jsdl;
	char directory_name[128] = "";
	sm_value_t *result = NULL;

	SM_DEBUG_ENTER();

	if (sm_previous_signal_failed(chained_params))
		return sm_value_clone(chained_params);
	else if (sm_signal_first(chained_params))
		input_jsdl = params[0]->vstring;
	else
		input_jsdl = chained_params[1]->vstring;

	SM_TRY_NOT(mh = sm_gsoap_buffer_get(input_jsdl, &jsdl, jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", xmlstructs_namespaces));

	if (!(jsdl.jsdl__JobDescription->jsdl__Application &&
			jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication &&
			jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__WorkingDirectory &&
			jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__WorkingDirectory->__item) /* 1. don't have working directory set */
			&&
				((jsdl.jsdl__JobDescription->jsdl__Application && /*2.a) have executable...*/
				jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication &&
				jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__Executable &&
				jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__Executable->__item)
				||
				(jsdl.jsdl__JobDescription->jsdl__Application && /*2.b) or application name defined */
				jsdl.jsdl__JobDescription->jsdl__Application->jsdl__ApplicationName &&
				jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication /*we deal only with HPCProfileApplication */))
			)
		{
		uuid_t wd_uuid;
		char wd_uuid_str[64];
		struct jsdl_hpcpa__DirectoryName_USCOREType hpcpa_wd;

		uuid_generate(wd_uuid);
		uuid_unparse(wd_uuid, wd_uuid_str);

		if (jsdl.jsdl__JobDescription->jsdl__JobIdentification &&
				jsdl.jsdl__JobDescription->jsdl__JobIdentification->jsdl__JobName) {
			snprintf(directory_name, sizeof(directory_name), "%s-%s", wd_uuid_str, jsdl.jsdl__JobDescription->jsdl__JobIdentification->jsdl__JobName);
		} else {
			snprintf(directory_name, sizeof(directory_name), "%s", wd_uuid_str);
		}

		sm_debug("sandbox working directory: %s", directory_name);

		memset(&hpcpa_wd, 0, sizeof(hpcpa_wd));
		jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__WorkingDirectory = &hpcpa_wd;
		jsdl.jsdl__JobDescription->jsdl__Application->jsdl_hpcpa__HPCProfileApplication->jsdl_hpcpa__WorkingDirectory->__item = directory_name;

		SM_TRY_NOT(output_jsdl = sm_gsoap_buffer_put(&jsdl, jsdl__JobDefinition_USCOREType, "jsdl:JobDefinition", xmlstructs_namespaces));

		result = sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_OK, SM_VALUE_STRING, output_jsdl);
		SM_FREE(output_jsdl);
		SM_TRY_NOT(result);

	} else {
		SM_TRY_NOT(result = sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_OK, SM_VALUE_STRING, input_jsdl));
	}

	sm_alloc_tracked_free(&mh);
	return result;

fault:
	sm_alloc_tracked_free(&mh);

	signal->stop_emission(signal);

	if (result)
		return result;
	else
		return sm_value_create(2, SM_VALUE_RET_CODE, JSDL_FILTER_INTERNAL_FAULT, "Internal fault in working directory sandboxing JSDL module", SM_VALUE_NONE);
}
