#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stddef.h>
#include <sys/stat.h>
#include <librsync-config.h>
#include <librsync.h>
#include <assert.h>

#include <qcg-core/sm.h>
#include <qcg-core/utils.h>
#include <qcg-core/logger.h>
#include <qcg-core/modules/general_module.h>
#include <qcg-core/xmalloc.h>

#include <rsync/types.h>
#include <rsync/sumset.h>
#include <rsync/interface.h>

#include <modules/qcg_comp_service/faults.h>

#define BUFFER_SIZE 10*1024*1024

SM_INTERNAL_GENERAL_MODULE = {
        .header = {
                .name = "rsync-base"
        }
};

static char*
join_path(rsync_transport_t transport,const char *dir,const char *fn)
{
	const char sep[]="/";
	char *result = NULL;
	size_t len = 0;
	size_t result_len = 0;

	SM_DEBUG_ENTER_STR2(dir,fn);

	result_len = strlen(dir) + strlen(fn) + 2;
	SM_TRY_NOT(result = transport->transport_malloc(transport, result_len));
	strncpy(result, dir, strlen(dir));
	len = strlen(result);

	if((len>0 && result[len-1]!=sep[0]) || len==0)
		strlcat(result, sep, result_len);

	while(fn[0]==sep[0])
		++fn;

	strlcat(result, fn, result_len);

	return SM_DEBUG_LEAVE_STR(result);

fault:
	return SM_DEBUG_LEAVE_PTR(NULL);
}


#define JOB_OK(a) ((a==RS_BLOCKED) || (a==RS_DONE))

static int
copy_rsig_to_sig(rsync_transport_t transport,const rs_signature_t *sigs,Signature *s)
{
	Length i;

	SM_DEBUG_ENTER();

	SM_MANDATORY_P(s);

	s->Length=sigs->flength;
	s->__sizeSignatures=sigs->count;
	s->Remainder=sigs->remainder;
	s->BlockLength=sigs->block_len;
	s->StrongSumLength=sigs->strong_sum_len;
	s->Signatures=transport->transport_malloc(transport,sigs->count*sizeof(BlockSignature));

	if(sigs->count>0 && s->Signatures==NULL)
		return SM_DEBUG_LEAVE_INT(RSyncNoMemoryError);

	for(i=0; i<sigs->count; ++i) {
		char buf[sigs->strong_sum_len*4/3+1];

		s->Signatures[i].Weak = sigs->block_sigs[i].weak_sum;
		rs_base64(sigs->block_sigs[i].strong_sum, sigs->strong_sum_len, buf);
		s->Signatures[i].Strong = transport->transport_strdup(transport, buf);

		if(!s->Signatures[i].Strong)
			return SM_DEBUG_LEAVE_INT(RSyncNoMemoryError);
	}

	assert(sigs->targets == NULL);
	assert(sigs->tag_table == NULL);

	return SM_DEBUG_LEAVE_INT(RSyncNoError);
}

static int
copy_sig_to_rsig(const Signature* s, rs_signature_t** _sig)
{
	rs_signature_t *sig;
	size_t buf_len = 0;
	char *buf = NULL;
	size_t n;

	SM_DEBUG_ENTER();

	SM_MANDATORY_P(s);

	// intensional use of malloc, because we need to invoke rs_free_sumset later to release memory allocated by rs_build_hash_table
	sig = sm_malloc(sizeof(rs_signature_t));
	if(sig==NULL)
		return SM_DEBUG_LEAVE_INT(RSyncNoMemoryError);
	memset(sig,0,sizeof(rs_signature_t));

	sig->flength = s->Length;
	sig->count = s->__sizeSignatures;
	sig->remainder = s->Remainder;
	sig->block_len = s->BlockLength;
	sig->strong_sum_len = s->StrongSumLength;

	if(sig->count>0) {
		Length i;

		sig->block_sigs = sm_calloc(sig->count,sizeof(rs_block_sig_t));

		if(!sig->block_sigs) {
			sm_free(sig);
			return SM_DEBUG_LEAVE_INT(RSyncNoMemoryError);
		}

		buf_len = sig->strong_sum_len*4/3+2;
		buf = sm_malloc(buf_len);

		if(!buf) {
			sm_free(sig);
			sm_free(sig->block_sigs);
			return SM_DEBUG_LEAVE_INT(RSyncNoMemoryError);
		}

		for(i=0;i<sig->count;++i) {
			sig->block_sigs[i].i = i+1;
			sig->block_sigs[i].weak_sum = s->Signatures[i].Weak;
			strlcpy(buf, s->Signatures[i].Strong, buf_len);
			n = rs_unbase64(buf);
			memcpy(sig->block_sigs[i].strong_sum, buf, n);
		}

		sm_free(buf);
	}

	*_sig=sig;

	return SM_DEBUG_LEAVE_INT(RSyncNoError);
}


static int
get_files(rsync_transport_t transport, const char *basedir, const char *path,Files *result, size_t *max)
{
	int error = RSyncInternalError;
	struct dirent **de = NULL;
	int i = 0;
	int n = 0;
	char *fullpath = NULL;

	SM_DEBUG_ENTER_STR2(basedir, path);

	SM_MANDATORY_P(transport);
	SM_MANDATORY_P(result);
	SM_MANDATORY_P(max);
	SM_MANDATORY_P(path);

	fullpath = join_path(transport, basedir, path);

	if(!fullpath) {
		error=RSyncNoMemoryError;
		goto error;
	}

	mkdir(fullpath,0777);	//it's ok if it fails, because scandir handles it
	n = scandir(fullpath, &de, NULL, NULL);

	if (n == -1) {
		sm_perror("scandir");
		return RSyncInvalidPathError;
	}

	for(i=0; i<n; ++i) {
		File *f = NULL;
		const char *fn = NULL;
		struct stat buf;

		if(*max <= (size_t)result->__sizeFile) {
			int newMax = (*max+1)*2;
			File *new = NULL;

			new = transport->transport_malloc(transport, sizeof(File)*newMax);
			if(new==NULL) {
				error=RSyncNoMemoryError;
				goto error;
			}
			memcpy(new, result->File, sizeof(File)* (*max));
			*max = newMax;
			result->File = new;
		}

		f=&(result->File[result->__sizeFile]);
		fn=de[i]->d_name;

		if(sm_strcmp(fn,".")==0 || sm_strcmp(fn,"..")==0)
			goto clear;

		f->Path=join_path(transport,path,fn);
		fn=join_path(transport,basedir,f->Path);

		if(fn==NULL) {
			error=RSyncNoMemoryError;
			goto error;
		}

		if (stat(fn, &buf) != 0) {
			sm_perror("stat");
			error=RSyncIOError;
			goto error;
		}

		if(S_ISREG(buf.st_mode)) {
			f->MTime=buf.st_mtime;
			f->isDirectory = 0;
			++result->__sizeFile;

			if((buf.st_mode && S_IXUSR)) {
				f->isExecutable = 1;
			} else {
				f->isExecutable = 0;
			}
		} else if(S_ISDIR(buf.st_mode)) {
			int size_before = result->__sizeFile;
			error=get_files(transport, basedir, f->Path, result, max);
			if(error!=RSyncNoError)
				goto error;

			if (!size_before - result->__sizeFile) { /*empty directory */
				f->MTime=buf.st_mtime;
				f->isDirectory = 1;
				f->isExecutable = 0;
				++result->__sizeFile;
			}
		}
clear:
		sm_free(de[i]);
		de[i]=NULL;
	}
	sm_free(de);
	return SM_DEBUG_LEAVE_INT(RSyncNoError);
error:
	for(i=0;i<n;++i)
		sm_free(de[i]);
	sm_free(de);
	return SM_DEBUG_LEAVE_INT(error);
}

static int
files_cmp(const void *a,const void *b)
{
	return sm_strcmp(((const File*)a)->Path,((const File*)b)->Path);
}

int
GetFiles(rsync_transport_t transport, const char *path, Files *result)
{
	size_t max = 0;
	int error = RSyncNoError;

	SM_DEBUG_ENTER_STR(path);

	SM_MANDATORY_P(result);

	result->__sizeFile = 0;
	result->File = NULL;
	error=get_files(transport, path, "", result, &max);

	if(error != RSyncNoError)
		return SM_DEBUG_LEAVE_INT(error);

	qsort(result->File, result->__sizeFile, sizeof(File), files_cmp);

	return SM_DEBUG_LEAVE_INT(RSyncNoError);
}

int
GetSignature(rsync_transport_t transport, const char *basedir, const char *path, Signature *result)
{
	int error=RSyncNoError;
	rs_job_t *sig_job = NULL;
	rs_job_t *load_job = NULL;
	rs_signature_t *sigs = NULL;
	int fd = -1;
	const char *fn = NULL;
	rs_result sig_result = RS_BLOCKED;
	rs_result load_result = RS_BLOCKED;
	char *input = NULL;
	char *blob = NULL;

	SM_DEBUG_ENTER_STR2(basedir, path);

	fn = join_path(transport, basedir, path);

	if(!fn) {
		error = RSyncNoMemoryError;
		goto quit;
	}

	fd = open(fn, O_RDONLY);	//file may not exists and it's handled down there
	sig_job = rs_sig_begin(RS_DEFAULT_BLOCK_LEN, RS_DEFAULT_STRONG_LEN);

	if (!sig_job) {
		sm_debug("rs_sig_begin failed");
		error = RSyncInternalError;
		goto quit;
	}

	load_job=rs_loadsig_begin(&sigs);

	if(!load_job) {
		sm_debug("rs_loadsig_begin failed");
		error = RSyncInternalError;
		goto quit;
	}

	input=sm_malloc(BUFFER_SIZE);
	blob=sm_malloc(BUFFER_SIZE);

	if(!input || !blob) {
		error = RSyncNoMemoryError;
		goto quit;
	}

	do {
		rs_buffers_t data={.next_in=NULL,.avail_in=0,.eof_in=1};
		ssize_t s;

		if(sig_result!=RS_DONE) {

			if(fd>=0) {
				s = read(fd, input, BUFFER_SIZE);

				if (s < 0) {
					sm_perror("wrapper_read");
					error=RSyncIOError;
					goto quit;
				}

				if(s!=0) {
					data.next_in=input;
					data.avail_in=s;
					data.eof_in=0;
				}

			}
			data.next_out=blob;
			data.avail_out=BUFFER_SIZE;
			sig_result=rs_job_iter(sig_job,&data);

			if(!JOB_OK(sig_result)) {
				sm_error("rs_job_iter failed");
				error=RSyncInternalError;
				goto quit;
			}

			data.next_in=blob;
			data.avail_in=BUFFER_SIZE-data.avail_out;
			data.eof_in=0;
		}
		data.next_out=NULL;
		data.avail_out=0;
		load_result=rs_job_iter(load_job,&data);

		if(!JOB_OK(load_result)) {
			sm_error("Job load_result is NOT ok");
			error=RSyncInternalError;
			goto quit;
		}

	}
	while(sig_result!=RS_DONE || load_result!=RS_DONE);
	error=copy_rsig_to_sig(transport,sigs,result);
quit:
	sm_free(blob);
	sm_free(input);

	if(sigs)
		rs_free_sumset(sigs);

	if(sig_job)
		rs_job_free(sig_job);

	if(load_job)
		rs_job_free(load_job);

	if(fd>=0)
		close(fd);

	return SM_DEBUG_LEAVE_INT(error);
}

int
GetDeltaBegin(rsync_transport_t transport, const char *basedir, const char *path,
		const Signature *signature,struct rsync_delta_context **pcontext)
{
	int error = RSyncNoMemoryError;
	int result = RSyncNoError;
	rs_signature_t *sig=NULL;
	const char *fn;
	struct rsync_delta_context *context;

	SM_DEBUG_ENTER_STR2(basedir, path);
	SM_MANDATORY_P(transport);
	SM_MANDATORY_P(basedir);
	SM_MANDATORY_P(path);
	SM_MANDATORY_P(signature);
	SM_MANDATORY_P(pcontext);

	SM_TRY_NOT(context = sm_malloc(sizeof(struct rsync_delta_context)));

	result = copy_sig_to_rsig(signature,&sig);

	if(result != RSyncNoError) {
		sm_debug("copy_sig_to_rsig failed");
		error = result;
		goto fault;
	}

	rs_build_hash_table(sig);
	context->job=rs_delta_begin(sig);

	if(!context->job) {
		sm_debug("rs_delta_begin failed");
		error=RSyncInternalError;
		goto fault;
	}

	SM_TRY_NOT(fn = join_path(transport,basedir,path));

	context->delta_max_length=BUFFER_SIZE;
	context->delta_length=0;
	SM_TRY_NOT(context->delta = sm_malloc(context->delta_max_length));

	context->fd = open(fn,O_RDONLY);	//result value is handled in GetDeltaProcess

	context->input_max_length=BUFFER_SIZE;
	context->input_length=0;
	SM_TRY_NOT(context->input = sm_malloc(context->input_max_length));

	*pcontext = context;

	return SM_DEBUG_LEAVE_INT(RSyncNoError);

fault:

	if(context) {
		if(context->input)
			sm_free(context->input);

		if(context->delta)
			sm_free(context->delta);

		sm_free(context);
	}

	*pcontext=NULL;
	return SM_DEBUG_LEAVE_INT(error);
}

int
GetDeltaProcess(struct rsync_delta_context* context)
{
	int error = RSyncNoError;
	ssize_t s = 0;
	rs_buffers_t data = {
		.next_in=NULL,
		.avail_in=0,
		.eof_in=1,
		.next_out=context->delta,
		.avail_out=context->delta_max_length
	};

	SM_DEBUG_ENTER_PTR(context);

	if(context->fd>=0) {
		s = read(context->fd, context->input+context->input_length,context->input_max_length-context->input_length);
		if (s < 0) {
			sm_perror("wrapper_read");
			error=RSyncIOError;
			goto quit;
		}

		if(s>0) {
			data.next_in = context->input;
			data.avail_in = s;
			data.eof_in = 0;
		}
		//s = 0 is ignored because data is set to eof state by default
	}

	context->status = rs_job_iter(context->job,&data);

	if(!JOB_OK(context->status)) {
		sm_debug("GetDelta job is NOT ok");
		error=RSyncInternalError;
		goto quit;
	}

	context->delta_length = data.next_out-context->delta;

quit:
	return SM_DEBUG_LEAVE_INT(error);
}

void
GetDeltaFinish(struct rsync_delta_context* context)
{
	SM_DEBUG_ENTER_PTR(context);

	sm_free(context->input);
	sm_free(context->delta);

	if(context->sig)
		rs_free_sumset(context->sig);

	if(context->job)
		rs_job_free(context->job);

	if(context->fd>=0)
		close(context->fd);

	sm_free(context);

	return SM_DEBUG_LEAVE_VOID();
}

struct callback_params
{
	int fd;
	uid_t uid;
};

static rs_result
put_delta_callback(void *opaque, rs_long_t pos, size_t *len,
		void **buf)
{
	struct callback_params *params = (struct callback_params*) opaque;
	int fd = params->fd;
	ssize_t r;

	SM_DEBUG_ENTER_INT(pos);

	sm_info("lseek(%d, %d)", fd, pos);
	if(lseek(fd, pos, SEEK_SET) != pos) {
		sm_perror("wrapper_lseek");
		return SM_DEBUG_LEAVE_INT(RS_IO_ERROR);
	}

	r = read(fd, *buf, *len);

	if (r == 0)
		return SM_DEBUG_LEAVE_INT(RS_INPUT_ENDED);

	if(r<0) {
		sm_perror("wrapper_read");
		return SM_DEBUG_LEAVE_INT(RS_IO_ERROR);
	}

	*len = r;
	return SM_DEBUG_LEAVE_INT(RS_DONE);
}

static char*
get_dir_part(rsync_transport_t transport,const char *path)
{
	char *result;
	char *ptr;

	SM_DEBUG_ENTER_STR(path);
	result = transport->transport_strdup(transport,path);

	if(result == NULL)
		return SM_DEBUG_LEAVE_STR(NULL);

	ptr=strrchr(result,'/');

	if(ptr)
		*(ptr+1)='\0';

	return SM_DEBUG_LEAVE_STR(result);
}


static int
mkdir_p(const char *path) {
	char *p = NULL;
	char *x = NULL;
	int result = -1;

	SM_DEBUG_ENTER_STR(path);
	p = sm_strdup(path);

	if (!p) {
		errno = ENOMEM;
		return SM_DEBUG_LEAVE_INT(-1);
	}

	x = p + 1;

	while ((x = strchr(x, '/')) != NULL) {
		*x = '\0';

		result = mkdir(p, 0777);
		sm_debug("mkdir(%s) = %i", p, result);

		if (result != 0 && errno != EEXIST) {
			sm_perror("mkdir");
			goto quit;
		}

		*x = '/';
		x++;
	}

	result = 0;
quit:
	sm_free(p);
	return SM_DEBUG_LEAVE_INT(result);
}

int
PutDeltaBegin(rsync_transport_t transport, const char *basedir, const char *path,
		struct rsync_delta_context **context)
{
	SM_DEBUG_ENTER_STR2(basedir, path);

	SM_MANDATORY_P(transport);
	SM_MANDATORY_P(basedir);
	SM_MANDATORY_P(path);
	SM_MANDATORY_P(context);

	int error = RSyncNoMemoryError;
	struct callback_params *params = NULL;

	SM_TRY_NOT(params = transport->transport_malloc(transport, sizeof(struct callback_params)));

	SM_TRY_NOT(*context=sm_malloc(sizeof(struct rsync_delta_context)));
	(*context)->rfd = -1;
	(*context)->wfd = -1;

	SM_TRY_NOT(path = join_path(transport, basedir, path));

	SM_TRY_NOT((*context)->path = sm_strdup(path));

	SM_TRY_NOT((*context)->output_fn = sm_asprintf("%s.XXXXXX", path));

	sm_debug("Temp path template: %s", (*context)->output_fn);

	if(mkdir_p(get_dir_part(transport, path)) != 0) {
		error = RSyncInvalidPathError;
		goto fault;
	}

	(*context)->rfd = open(path, O_RDONLY); // it's ok if rfd is invalid - base file may simply does not exists

	(*context)->wfd = mkstemp((*context)->output_fn);

	if ((*context)->wfd < 0) {
		sm_perror("wrapper_mkstemp");
		error=RSyncIOError;
		goto fault;
	}

	params->fd = (*context)->rfd;
	(*context)->job = rs_patch_begin(put_delta_callback, params);

	if(!(*context)->job) {
		sm_debug("rs_patch_begin failed");
		error=RSyncInternalError;
		goto quit;
	}

	error = RSyncNoError;

quit:
	return SM_DEBUG_LEAVE_INT(error);

fault:

	PutDeltaFree(*context);
	*context = NULL;

	goto quit;
}

int
PutDeltaProcess(struct rsync_delta_context *context)
{
	int error = RSyncNoMemoryError;
	char *buf = NULL;
	size_t amount;

	SM_DEBUG_ENTER();

	SM_TRY_NOT(buf = sm_malloc(BUFFER_SIZE));

	rs_buffers_t data= {
		.next_in = context->delta,
	    .avail_in = context->delta_length,
	    .eof_in = (context->delta_length == 0),
	    .next_out = buf,
	    .avail_out = BUFFER_SIZE
	};

	context->status = rs_job_iter(context->job,&data);

	if(!JOB_OK(context->status)) {
		sm_debug("Job in PutDelta is NOT ok");
		error=RSyncInternalError;
		goto fault;
	}

	amount=BUFFER_SIZE - data.avail_out;

	if(write(context->wfd, buf, amount) != (ssize_t)amount) {
		sm_perror("wrapper_write");
		error=RSyncIOError;
		goto fault;
	}

	memmove(context->delta, data.next_in, data.avail_in);
	context->delta_length = data.avail_in;

	error = RSyncNoError;

quit:
	sm_free(buf);
	return SM_DEBUG_LEAVE_INT(error);

fault:
	goto quit;
}

void
PutDeltaFree(struct rsync_delta_context *context)
{
	SM_DEBUG_ENTER();

	if(context) {

		if (context->wfd >= 0)
			close(context->wfd);

		if (context->rfd >= 0)
			close(context->rfd);

		if (context->job)
			rs_job_free(context->job);

		sm_free(context->output_fn);
		sm_free(context->path);
		sm_free(context);
	}

	SM_DEBUG_LEAVE_VOID();
}

int
PutDeltaFinish(struct rsync_delta_context *context)
{
	int error = RSyncNoError;

	SM_DEBUG_ENTER_STR(context->path);

	unlink(context->path);	//it's ok if it fails, because path may be invalid

	sm_debug("rename(%s,%s)", context->output_fn, context->path);
	if (rename(context->output_fn, context->path) != 0) {
		sm_perror("rename");
		error=RSyncIOError;
		unlink(context->output_fn);	//it doesn't matter if it fails
	}

	PutDeltaFree(context);

	return SM_DEBUG_LEAVE_INT(error);
}
