#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include "stub/rsyncstructsStub.h"
#include "types.h"
#include "interface.h"
#include "attachment.h"
#include "client.h"
#include <fnmatch.h>
#include <include/namespaces.h>
#include <qcg-core/xmalloc.h>
#include <qcg-core/modules/general_module.h>
#include <qcg-core/logger.h>

#include <modules/qcg_comp_service/faults.h>

#define TRANSPORT_SETUP(action) \
	do \
	{ \
		if(transport->pre_call_setup(transport, SMC_NAMESPACE_RSYNC, action, NULL, NULL, NULL)!=SM_OK) \
		{ \
			sm_error("pre_call_setup failed"); \
			goto fault; \
		} \
	} \
	while(0)

extern struct Namespace qcgcomp_namespaces[];

SM_INTERNAL_GENERAL_MODULE = { .header = {
	.name = "rsync-client"
} };

static struct qcg_comp_rsync__CommonType*
make_params(rsync_transport_t transport, const char *basedir, const char *username)
{
	struct qcg_comp_rsync__CommonType *params = NULL;

	SM_TRY_NOT(params = transport->transport_malloc(transport, sizeof(struct qcg_comp_rsync__CommonType)));
	SM_TRY_NOT(params->BaseDir = transport->transport_strdup(transport, basedir));

	if(username)
		SM_TRY_NOT(params->UserName = transport->transport_strdup(transport, username));
	else
		params->UserName = NULL;

	return params;

fault:
	return NULL;
}

static int
PutDelta(rsync_transport_t transport, const char *basedir, const char *path, const Delta delta)
{
	int result, tmp;
	struct rsync_delta_context *context = NULL;

	result = PutDeltaBegin(transport, basedir, path, &context);

	if (result != RSyncNoError)
		goto fault;

	context->delta = delta.data;
	context->delta_length = delta.length;

	result = PutDeltaProcess(context);

	if (result != RSyncNoError)
		goto fault;

quit:
	return PutDeltaFinish(context);

fault:
	PutDeltaFree(context);
	return result;
}

/*@client*/
static int
get_delta(rsync_transport_t transport, const char *service, const char *username, const char *local_basedir, const char *remote_basedir, const char *path)
{
	Signature local_sig;
	struct qcg_comp_rsync__GetDeltaType request;
	struct qcg_comp_rsync__GetDeltaResponseType result;
	Delta remote_delta;
	int ret;

	SM_DEBUG_ENTER_STR(service);

	sm_debug("Getting delta for `%s'\n", path);

	if(GetSignature(transport, local_basedir, path, &local_sig) != RSyncNoError) {
		sm_error("GetSignature");
		return SM_DEBUG_LEAVE_INT(0);
	}

	TRANSPORT_SETUP("GetDelta");

	SM_TRY_NOT(request.Common = make_params(transport, remote_basedir, username));
	request.Path = (char*) path;
	request.Signature = &local_sig;

	ret = soap_call___qcg_comp_rsync__GetDelta(RSYNC_GET_SOAP(transport), service, NULL, &request, &result);

	if (transport->post_call_check(transport, ret) != SM_OK) {
		sm_error("GetDelta: %s\n", RSYNC_GET_SOAP(transport)->fault->faultstring);
		return SM_DEBUG_LEAVE_INT(0);
	}

	if (unpack_delta(transport, &remote_delta) != RSyncNoError) {
		sm_error("Unpacking delta failed\n");
		return SM_DEBUG_LEAVE_INT(0);
	}

	sm_debug("Recived delta of size %lu\n", (unsigned long)remote_delta.length);
	if (PutDelta(transport, local_basedir, path, remote_delta) != RSyncNoError)
		sm_error("PutDelta failed\n");

	return SM_DEBUG_LEAVE_INT(1);

fault:
	sm_error("Internal error in get_delta");

	return SM_DEBUG_LEAVE_INT(0);
}

#define BUFFER_SIZE 10*1024*1024

static int
GetDelta(rsync_transport_t transport, const char *basedir, const char *path, const Signature *signature, Delta *result)
{
	struct rsync_delta_context *context = NULL;
	int fun_result = RSyncNoError;

	char *base = NULL;
	char *output = NULL;
	size_t avail = 0;
	size_t size = 0;

	fun_result = GetDeltaBegin(transport, basedir, path, signature, &context);
	if (fun_result != RSyncNoError)
		goto quit;

	do {
		fun_result = GetDeltaProcess(context);

		if (fun_result != RSyncNoError)
			goto quit;

		while (avail < context->delta_length) {
			ptrdiff_t len = output - base;
			size_t extend = BUFFER_SIZE;

			if (extend < context->delta_length)
				extend += context->delta_length;

			size += extend;
			avail += extend;
			base = sm_realloc(base, size);
			output = base + len;
		}

		memmove(output, context->delta, context->delta_length);
		output += context->delta_length;
		avail -= context->delta_length;
	} while (context->status == RS_BLOCKED);

	result->length = output - base;
	result->data = transport->transport_malloc(transport, result->length);
	memmove(result->data, base, result->length);

	sm_debug("Created delta length %lu", (unsigned long)result->length);

quit:
	GetDeltaFinish(context);
	sm_free(base);
	return SM_DEBUG_LEAVE_INT(fun_result);
}

/*@client*/
static int
put_delta(rsync_transport_t transport, const char *service, const char *username, const char *local_basedir, const char *remote_basedir, const char *path)
{
	struct qcg_comp_rsync__GetSignatureType gs_request;
	struct qcg_comp_rsync__GetSignatureResponseType gs_response;
	struct qcg_comp_rsync__PutDeltaType pd_request;
	struct qcg_comp_rsync__PutDeltaResponseType pd_response;

	Signature *remote_sig = NULL;
	Delta delta;
	int ret;

	SM_DEBUG_ENTER_STR(service);

	sm_debug("Putting delta for `%s'\n", path);

	TRANSPORT_SETUP("GetSignature");

	SM_TRY_NOT(gs_request.Common = make_params(transport, remote_basedir, username));
	gs_request.Path = (char*) path;

	ret = soap_call___qcg_comp_rsync__GetSignature(RSYNC_GET_SOAP(transport),
			service, NULL, &gs_request, &gs_response);

	if (transport->post_call_check(transport, ret) != SM_OK) {
		sm_error("GetSignature: %s\n", RSYNC_GET_SOAP(transport)->fault->faultstring);
		return SM_DEBUG_LEAVE_INT(0);
	}

	remote_sig = gs_response.Signature;

	if (GetDelta(transport, local_basedir, path, remote_sig, &delta) != RSyncNoError) {
		sm_error("GetDelta failed\n");
		return SM_DEBUG_LEAVE_INT(0);
	}

	soap_set_mime(RSYNC_GET_SOAP(transport), NULL, NULL);
	soap_set_mime_attachment(RSYNC_GET_SOAP(transport), delta.data,
			delta.length, SOAP_MIME_BINARY, "application/octet-stream",
			"delta", NULL, NULL);

	TRANSPORT_SETUP("PutDelta");

	SM_TRY_NOT(pd_request.Common = make_params(transport, remote_basedir, username));
	pd_request.Path = (char*) path;

	ret = soap_call___qcg_comp_rsync__PutDelta(RSYNC_GET_SOAP(transport),
			service, NULL, &pd_request, &pd_response);

	if (transport->post_call_check(transport, ret) != SM_OK)
		sm_error("PutDelta: %s\n", RSYNC_GET_SOAP(transport)->fault->faultstring);

	soap_clr_mime(RSYNC_GET_SOAP(transport));

	return SM_DEBUG_LEAVE_INT(1);

fault:

	sm_error("Internal error in put_delta");

	return SM_DEBUG_LEAVE_INT(0);
}

static SM_BOOL
file_ok(const char *path, const char *filters)
{

	if (path[0] == '/')
		++path;

	if (filters) {
		const char *filter;

		for (filter = filters; *filter; filter += strlen(filter) + 1) {

			sm_debug("Matching '%s' and '%s'", path, filter);

			if (fnmatch(filter, path, 0) == 0) {
				sm_debug("Filtering out '%s'", path);
				return SM_FALSE;
			}
		}
	}

	sm_debug("Accepting '%s'", path);

	return SM_TRUE;
}

/*@client*/
static void
compare_and_match(rsync_transport_t transport, const char *service, const char *username, const char *local_basedir, Files *local, const char *remote_basedir, Files *remote, const char *filter)
{
	int i = 0, j = 0;
	SM_BOOL *local_found, *remote_found, *local_newer;

	SM_DEBUG_ENTER();

	//local file found on remote side
	SM_TRY_NOT(local_found = transport->transport_malloc(transport, local->__sizeFile * sizeof(SM_BOOL)));
	//remote file found on local side
	SM_TRY_NOT(remote_found = transport->transport_malloc(transport, remote->__sizeFile * sizeof(SM_BOOL)));
	//local_newer[i] is valid only if local_found[i] is SM_TRUE
	SM_TRY_NOT(local_newer = transport->transport_malloc(transport, local->__sizeFile * sizeof(SM_BOOL)));

	for (i = 0; i < local->__sizeFile; ++i)
		local_found[i] = local_newer[i] = SM_FALSE;

	for (i = 0; i < remote->__sizeFile; ++i)
		remote_found[i] = SM_FALSE;

	for (i = 0, j = 0; i < local->__sizeFile && j < remote->__sizeFile;) {
		int r = sm_strcmp(local->File[i].Path, remote->File[j].Path);

		if (r < 0)
			++i;

		else if (r == 0) {
			local_found[i] = SM_TRUE;
			remote_found[j] = SM_TRUE;
			local_newer[i] = local->File[i].MTime > remote->File[i].MTime;
			++i;
			++j;
		}

		else
			++j;
	}

	for (i = 0; i < local->__sizeFile; ++i) {

		if (!file_ok(local->File[i].Path, filter))
			continue;

		if (!local_found[i] || local_newer[i])
			put_delta(transport, service, username, local_basedir,
					remote_basedir, local->File[i].Path);
		else
			get_delta(transport, service, username, local_basedir,
					remote_basedir, local->File[i].Path);

	}

	for (i = 0; i < remote->__sizeFile; ++i) {
		if (!file_ok(remote->File[i].Path, filter))
			continue;

		if (!remote_found[i])
			get_delta(transport, service, username, local_basedir,
					remote_basedir, remote->File[i].Path);
	}

	return SM_DEBUG_LEAVE_VOID();

fault:

	sm_error("Internal error");

	return SM_DEBUG_LEAVE_VOID();
}

void
rsync_sync(rsync_transport_t transport, const char *service, const char *username, const char *local_dir, const char *remote_dir, const char *filter)
{
	Files local_files;
	int ret;
	struct qcg_comp_rsync__GetFilesType request;
	struct qcg_comp_rsync__GetFilesResponseType response;

	SM_DEBUG_ENTER();

	GetFiles(transport, local_dir, &local_files); //TODO: obsluga bledow

	TRANSPORT_SETUP("GetFiles");

	SM_TRY_NOT(request.Common = make_params(transport, remote_dir, username));
	ret = soap_call___qcg_comp_rsync__GetFiles(RSYNC_GET_SOAP(transport),
			service, "", &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK)
		compare_and_match(transport, service, username, local_dir,
				&local_files, remote_dir, response.Files, filter);
	else
		sm_error("GetFiles: %s\n",
				RSYNC_GET_SOAP(transport)->fault->faultstring);

	return SM_DEBUG_LEAVE_VOID();

fault:

	sm_error("Internal error in rsync_sync");

	return SM_DEBUG_LEAVE_VOID();
}

void
rsync_local_to_remote(rsync_transport_t transport, const char *service, const char *username, const char *local_basedir, const char *remote_basedir, const char *filter)
{
	Files local_files;
	int i;

	SM_DEBUG_ENTER();

	GetFiles(transport, local_basedir, &local_files); //TODO: obsluga bledow

	for (i = 0; i < local_files.__sizeFile; ++i)
		if (file_ok(local_files.File[i].Path, filter))
			put_delta(transport, service, username, local_basedir,
					remote_basedir, local_files.File[i].Path);

	return SM_DEBUG_LEAVE_VOID();
}

void
rsync_remote_to_local(rsync_transport_t transport, const char *service, const char *username, const char *local_basedir, const char *remote_basedir, const char *filter)
{
	int ret, i;
	struct qcg_comp_rsync__GetFilesType request;
	struct qcg_comp_rsync__GetFilesResponseType response;

	SM_DEBUG_ENTER();

	TRANSPORT_SETUP("GetFiles");

	SM_TRY_NOT(request.Common = make_params(transport, remote_basedir, username));

	TRANSPORT_SETUP("GetFiles");

	ret = soap_call___qcg_comp_rsync__GetFiles(RSYNC_GET_SOAP(transport),
			service, "", &request, &response);

	if (transport->post_call_check(transport, ret) == SM_OK)
		for (i = 0; i < response.Files->__sizeFile; ++i) {
			if (file_ok(response.Files->File[i].Path, filter))
				get_delta(transport, service, username, local_basedir,
						remote_basedir, response.Files->File[i].Path);
		}
	else
		sm_error("GetFiles: %s\n",
				RSYNC_GET_SOAP(transport)->fault->faultstring);

	return SM_DEBUG_LEAVE_VOID();

fault:
	sm_error("Internal error in rsync_remote_to_local");

	return SM_DEBUG_LEAVE_VOID();
}
